unit UScriptEngine;

// Crateur : PhM, Nov 2004
// Description : TScriptEngine : Interprtation d'une commande texte (script ou evt)  appliquer  la simulation

interface

uses Classes, Contnrs, SysUtils, UGlobTypes, Controls, Dialogs, Variants,
     UGlobVar, UGlobConst, UGlobClasses, AGProcUtils, UFileParser, UFaction,
     URegion, UGroup, UUnit, USubUnit, Math;

const _sctNumOfArgs = 128;
      _esvIntVarName = 'esvIntVar';
      _comNone       = 0;
      _comGHQGroup   = 1;
      _comHQGroup    = 2;
      _comCombine    = 3;

type

  // Etat d'un vnement (sauv en GameSim version host uniquement)
  // valeurs de MaxOccurs: 0: evt en sommeil, 999: pas de limites d'occurences
  TEventState = class
    UName : string;      // nom unique pour un scnario donn
    cuOccurs  : integer; // nb d'occurences actuelles de l'evt
    MaxOccurs : integer; // nb d'occurences permises
    function    Serialize  (out SL : TStringList) : boolean;
    function    UnSerialize(SL : TStringList)     : boolean;
  end;

  TEventStates = class(TObjectList)
  protected
    function  Get (Index : Integer): TEventState; overload;
    procedure Put (Index: Integer; Value: TEventState);

  public
    constructor Create;
    function  Add(Item: TEventState): Integer;
    property  Items[Index: Integer] : TEventState read Get write Put; default;

     function  GetByName(aUName : string) : TEventState;
  end;

  // une variable dclare dans le moteur de script et qui subsiste de tour en tour
  TScriptVar = class
    Name : string;
    VarType : TUID; {0: integer; 1: string}
    Value : string;
    function    Serialize  (out SL : TStringList) : boolean;
    function    UnSerialize(SL : TStringList)     : boolean;
  end;

  // Donnes pour crer un message joueur (texte avec flavor)
  TEventMessage = record
    Title       : string;
    Text        : string;
    ImageID     : string;
    DisplayType : TUID;
    RgnUID      : TUID;
    SoundID     : TUID;
    IsBroadCast : boolean;
  end;
                                             
  TScriptEngine = class

  private
    procedure SetCuRegion(Value : TRegion);

  public
    IsSpecialScript : boolean; // global boolean changing slightly the behavior when a Research Script is parsed
    Running : boolean; // True when Parse is under use

    Constructor Create(aOwner : TObject);
    Destructor  Destroy; override;
    function    LoadFromFile(FileName : string) : boolean;
    function    LoadEventFromFile(EventName: string; FileName : string) : boolean;
    procedure   ReplaceInSlParse(Search : string; Replace : string);
    procedure   TriggerEvent(FileName : string; EventName : string; TagList : TStringList; ReplaceList : TStringList; LogFile : string);
    function    Parse(LogFile : string = '') : integer;
    class procedure FinalizeParsing;

  protected
    Owner         : TObject; // GameSim  priori
    slParse       : TStringList; // la SL en cours d'interprtation
    slScriptLog   : TStringList; // dtail des droulement du script
    cuLine        : integer; // la ligne courante traite
    WorkingVars   : TStringList; // sur quel objets on travail
    olBoundVar    : TObjectList; // liste des labels attachs  des slections
    ParsedEvent   : string;  // nom de l'vnement en cours d'analyse
    ConditionsON  : boolean; // on est dans le bloc conditions
    ActionsON     : boolean; // on est dans le bloc actions
    Conditions    : TConditions; // statut des conditions
    BaseOccurs    : integer; // nb d'occurences permises pour l'evt, comme lu dans le fichier d'evts (est modifiable par script)
    cuEventState  : TEventState; // pointeur sur l'tat de l'evt en cours d'analyse
    Args          : array[0.._sctNumOfArgs-1] of string;
    PlayerMessage : TEventMessage; // pour gnrer un message joueur (texte + flavor)
    UsedTurn      : integer; // le tour a considrer pour les tests, soit le tour du GameSim, soit un donn spcifique (pour le debug)

    // Variables d'actions, ne persiste que dans un vnement (puis RZ)
    WorkingGME    : string; // sur quel type de GameEltType le script porte
    GMEParam      : string; // un paramtre pour les sous-types de GME non standard: GHQGroup, HQGroup, ComboUnit, ...
    UseCustom     : boolean; // commutateur pour cration d'unit
    Name          : string;  // Nom de l'elt en interprtation (RaZ sur chaque Apply)
    ShortName     : string;  // shortname (RZ auto)
    Text          : string;  // Text (flavor) (RZ auto)
    FlavorName    : string;  // Name flavor (RZ auto)
    SuFlavorName  : string;  // SU Name flavor (RZ auto)
    Kind          : integer; // sa catgorie (land, nav, etc.)
    ObjType       : integer; // son sous type (= UID du ParentDef, etc.)
    NumCreate     : integer; // nombre d'objets  crer en srie (-1 gal le max possible pour unit, ou 1 pour SU)
    HealthPerc    : integer; // Hits de l'objet (en % du max, ou -1 si le max)
    CohesionPerc  : integer; // idem pour la cohsion
    Level         : integer; // niveau de l'objet
    SocialClass   : TUID;    // pour PopBrick
    EthnicGroup   : TUID;    // pour PopBrick
    Religion      : TUID;    // pour PopBrick
    LoyaltyFac    : integer; // pour PopBrick
    LoyaltyEth    : integer; // pour PopBrick
    AccuGrowth    : integer; // pour PopBrick
    GrowthRate    : integer; // pour PopBrick
    Richness      : integer; // pour PopBrick
    Entrench      : integer; // pour group
    InCS          : boolean; // pour group
    FixType       : integer; // pour l'unit, donn au niveau groupe
    GroupPosture  : TUID;    // pour group
    SpecialCommit : integer; // pour traitement spcial sur un Apply d'objet complexe (unit combine, etc.) _comNone  _comCombine
    fcuRegion     : TRegion;
    MsgNeedRegion : boolean; // commande GenFlavorMsg ignore si pas de rgion slectionne
    LastGHQGroup  : TGroup;  // Un HQStack se dclare  partir du dernier GHQStack cr
    IsBroadCast   : boolean;
    // Curseur
    LstCur        : TCursor;
    DoLegitOwn    : boolean; // will trigger after FinalizeParsing a call to CalcScriptAllLegitOwned
    // *** variables exposes aux scripteurs, cf GetExposedVarValue
    // esv: exposedscriptvariable nom des vars en string: esvIntVar
    // ! elles n'ont aucun reset sur la dure du scriptengine
    esvIntVar        : array[0..99] of integer;
    Registry      : array[0..999] of integer; // another variable system, 2011 version
    ActiveReg     : integer;

    function  IsKnownCommand(Str : string) : boolean;
    function  CommandCanHaveAliasWithSpaceSeparator(Command : string) : boolean;
    function  CommandShouldbeParsed(Command : string) : boolean;
    function  ParseKey(Key, KeyParams : string): integer;
    procedure Log(Msgs : array of variant);
    function  AddBoundVar(VarLabel : string; Selections : array of string) : integer;
    procedure GetSelections(VarLabel : string; var Result : TStringArray);
    function  LabelToUID(aLabel : string) : TUID;
    function  KindLabelToGMEType(aKindLabel : string; var Kind : integer) : integer;
    function  FactionTagToFaction(FactionTag : string) : TFaction;
    procedure Abort;
    function  ExtractParam(ParamStr : string; Params : TStringList) : string;
    procedure ExtractArgs(Param : string);
    function  GetExposedVarIndex(Str : String) : integer;
    function  IsExposedVar    (Str : String) : boolean;
    function  GetExposedVarValue(Str : String) : integer;
    function  StoreToExposedVar (Str : String; Value : integer) : boolean;

    // Operations with Registry
    function  Reg_Activate    (Params : TStringList) : integer;
    function  Reg_DeActivate  (Params : TStringList) : integer;    
    function  Reg_Init        (Params : TStringList) : integer;    
    function  Reg_Sum         (Params : TStringList) : integer;
    function  Reg_Rem         (Params : TStringList) : integer;
    function  Reg_SumNum      (Params : TStringList) : integer;
    function  Reg_Eval        (Params : TStringList) : integer;
    function  Reg_EvalNum     (Params : TStringList) : integer;    

    // Scripts d'vnements: descripteurs, dlimiteurs de blocs, actions formelles
    function  StartEvent    (Params : TStringList) : integer;
    function  EndEvent      (Params : TStringList) : integer;
    function  ToConditions  : integer;
    procedure PreCheckEvent;
    procedure ResetActionsVariables(ResetSelectors : boolean = True);
    function  ToActions     : integer;
    procedure SwitchForcedEventsToTrue;

    // Evaluations
    function  CheckFixedDate    (Params : TStringList) : integer;
    function  CheckMinDate      (Params : TStringList) : integer;
    function  CheckMaxDate      (Params : TStringList) : integer;
    function  Probability       (Params : TStringList) : integer;
    function  EvalUniqueUnit    (Params : TStringList) : integer;
    function  PickUnit          (Params : TStringList) : integer;
    function  EvalEvent         (Params : TStringList) : integer;
    function  CountCtrlAreas    (Params : TStringList) : integer;
    function  CountVP           (Params : TStringList) : integer;
    function  CountUnits        (Params : TStringList) : integer;
    function  CountUnitDef      (Params : TStringList) : integer;
    function  SubUnitCount      (Params : TStringList) : integer;
    function  EvalRgnOwned      (Params : TStringList) : integer;
    function  NotEnemy          (Params : TStringList) : integer;
    function  EvalRgnVP         (Params : TStringList) : integer;
    function  EvalRgnObj        (Params : TStringList) : integer;
    function  EvalBesieged      (Params : TStringList) : integer;
    function  EvalBlockaded     (Params : TStringList) : integer;        
    function  EvalRgnStruc      (Params : TStringList) : integer;
    function  TurnIndex         (Params : TStringList) : integer;
    function  CheckAILevel      (Params : TStringList) : integer;
    function  CheckIsPlayer     (Params : TStringList) : integer;
    function  EvalBlockadePerc  (Params : TStringList) : integer;
    function  EvalLoyalty       (Params : TStringList) : integer;
    function  EvalControl       (Params : TStringList) : integer;
    function  EvalMilitancy     (Params : TStringList) : integer;
    function  EvalContentment   (Params : TStringList) : integer;
    function  EvalEducation     (Params : TStringList) : integer;
    function  EvalTransportLvl  (Params : TStringList) : integer;
    function  EvalMorale        (Params : TStringList) : integer;
    function  EvalMoney         (Params : TStringList) : integer;
    function  EvalEngagementPts (Params : TStringList) : integer;
    function  EvalResource      (Params : TStringList) : integer;
    function  EvalConscript     (Params : TStringList) : integer;
    function  EvalRgnWeather    (Params : TStringList) : integer;
    function  EvalWarSupply     (Params : TStringList) : integer;
    function  EvalVP            (Params : TStringList) : integer;
    function  EvalForeignEntry  (Params : TStringList) : integer;
    function  EvalRegionSel     (Params : TStringList) : integer;
    function  EvalFactionSel    (Params : TStringList) : integer;
    function  EvalGroupSel      (Params : TStringList) : integer;
    function  EvalUnitSel       (Params : TStringList) : integer;
    function  EvalSubUnitSel    (Params : TStringList) : integer;
    function  ActorEnabled      (Params : TStringList) : integer;
    function  ActorDisabled     (Params : TStringList) : integer;
    function  ActorUnderUse     (Params : TStringList) : integer;
    function  CheckOption       (Params : TStringList) : integer;
    function  CanPayOption      (Params : TStringList) : integer;
    function  EvalReplacement   (Params : TStringList) : integer;
    function  EvalBattles       (Params : TStringList) : integer;
    function  EvalSubUnitCount  (Params : TStringList) : integer;
    function  EvalTechKnown     (Params : TStringList) : integer;
    function  FacModifierExists (Params : TStringList) : integer;
    function  EvalFacAttribute  (Params : TStringList) : integer;
    function  EvalIsAtWarWith   (Params : TStringList) : integer;
    function  EvalIsAtPeaceWith (Params : TStringList) : integer;
    function  IsObsoleteModelGen(Params : TStringList) : integer;
    function  IsObsoleteStrucGen(Params : TStringList) : integer;
    function  EvalPrisoners     (Params : TStringList) : integer;
    function  EvalRgnDecisionDef(Params : TStringList) : integer;
    function  EvalRgnDecisionAct(Params : TStringList) : integer;
    function  IsRgnDecisionActValid(Params : TStringList) : integer;
    function  EvalRuler         (Params : TStringList) : integer;
    function  EvalColStatus     (Params : TStringList) : integer;
    function  EvalColStatusArea (Params : TStringList) : integer;
    function  EvalSOI           (Params : TStringList) : integer;
    function  EvalFacRel        (Params : TStringList) : integer;
    function  CheckPowerRatio   (Params : TStringList) : integer;
    function  EvalAreaPwrRatio  (Params : TStringList) : integer;
    function  EvalNatThreats    (Params : TStringList) : integer;
    function  EvalStrucProduction(Params : TStringList) : integer;
    function  EvalStrucCount    (Params : TStringList) : integer;
    function  EvalHasCapital    (Params : TStringList) : integer;
    function  EvalInflation     (Params : TStringList) : integer;
    function  EvalColPen        (Params : TStringList) : integer;
    function  EvalRgnOwnerKind  (Params : TStringList) : integer;

    // end of _eval

    // Actions
    function  CreateObject      (GMEType : string; Params : TStringList) : integer;
    function  SetKind           (Params : TStringList) : integer;
    function  SetSpecialCommit  (Params : TStringList) : integer;
    function  SetGMEParams      (Params : TStringList) : integer;
    function  SetName           (Params : TStringList) : integer;
    function  SetShortName      (Params : TStringList) : integer;
    function  SetFlavorName     (Params : TStringList) : integer;
    function  SetSUFlavorName   (Params : TStringList) : integer;
    function  SetText           (Params : TStringList) : integer;
    function  SetType           (Params : TStringList) : integer;
    function  SetHealth         (Params : TStringList) : integer;
    function  SetCohesion       (Params : TStringList) : integer;
    function  SetLevel          (Params : TStringList) : integer;
    function  SetControl        (Params : TStringList) : integer;
    function  SetLoyaltyFac     (Params : TStringList) : integer;
    function  SetLoyaltyEth     (Params : TStringList) : integer;
    function  SetCivLevel       (Params : TStringList) : integer;
    function  SetRegionStyle    (Params : TStringList) : integer;
    function  SetFacStyle       (Params : TStringList) : integer;
    function  SetTransport      (Params : TStringList) : integer;
    function  SetVP             (Params : TStringList) : integer;
    function  SetBlockState     (Params : TStringList) : integer;
    function  SetColonialPen    (Params : TStringList) : integer;
    function  ChangeColonialPen (Params : TStringList) : integer;    
    function  SetColonialStatus (Params : TStringList) : integer;
    function  SetDevelopment    (Params : TStringList) : integer;
    function  SetPlannedCityName(Params : TStringList) : integer;
    function  SetPopTotalRaw    (Params : TStringList) : integer;
    function  SetAristosPercRaw (Params : TStringList) : integer;
    function  SetUpperPercRaw   (Params : TStringList) : integer;
    function  SetMiddlePercRaw  (Params : TStringList) : integer;
    function  SetWorkersPercRaw (Params : TStringList) : integer;
    function  SetPeasantsPercRaw(Params : TStringList) : integer;
    function  SetServilesPercRaw(Params : TStringList) : integer;
    function  ChgAristosCount   (Params : TStringList) : integer;
    function  ChgUpperCount     (Params : TStringList) : integer;
    function  ChgMiddleCount    (Params : TStringList) : integer;
    function  ChgWorkersCount   (Params : TStringList) : integer;
    function  ChgPeasantsCount  (Params : TStringList) : integer;
    function  ChgServilesCount  (Params : TStringList) : integer;
    function  ChgPopulation     (Params : TStringList) : integer;
    function  ChgPopReligions   (Params : TStringList) : integer;
    function  ChgPopEthnics     (Params : TStringList) : integer;
    function  ChgDevelopment    (Params : TStringList) : integer;
    function  SetPopMilitPerc   (Params : TStringList) : integer;
    function  ChgPopMilitancy   (Params : TStringList) : integer;
    function  ChgPopMilitancyEx (Params : TStringList) : integer;
    function  SetPopEducPerc    (Params : TStringList) : integer;
    function  ChgPopEducation   (Params : TStringList) : integer;
    function  ChgPopEducationEx (Params : TStringList) : integer;
    function  SetPopContentPerc (Params : TStringList) : integer;
    function  ChgPopContent     (Params : TStringList) : integer;
    function  ChgPopContentmentEx(Params : TStringList) : integer;
    function  ConvertSocClass   (Params : TStringList) : integer;
    function  ChgPopDataByEthnic(Params : TStringList) : integer;
    function  ChgPopDataByReligion( Params : TStringList) : integer;
    function  ChgResources_Sites(Params : TStringList) : integer;
    function  DiscoverResources_Sites(Params : TStringList) : integer;    
    function  SetAlliance       (Params : TStringList) : integer;
    function  ChangeRempls      (Params : TStringList) : integer;
    function  SetAutoRaise      (Params : TStringList) : integer;
    function  ChgAutoRaise      (Params : TStringList) : integer;
    function  UpgradeUnit       (Params : TStringList) : integer;
    function  AutoRaiseOn       (Params : TStringList) : integer;
    function  SetNumCreate      (Params : TStringList) : integer;
    function  SetEntranch       (Params : TStringList) : integer;
    function  SetInCS           (Params : TStringList) : integer;
    function  SetFixType        (Params : TStringList) : integer;
    function  CreateFaction     (Params : TStringList) : integer;
    function  SetMapOffsets     (Params : TStringList) : integer;
    function  SetMapBounds      (Params : TStringList) : integer;
    function  SetScenTurns      (Params : TStringList) : integer;
    function  SetScenStart      (Params : TStringList) : integer;
    function  SetScenEnd        (Params : TStringList) : integer;
    function  ChgScenEnd        (Params : TStringList) : integer;
    function  ForceDateTo       (Params : TStringList) : integer;
    function  TurnsPerYear      (Params : TStringList) : integer;
    function  SetScenName       (Params : TStringList) : integer;
    function  SetScnDispName    (Params : TStringList) : integer;
    function  SetPosture        (Params : TStringList) : integer;
    function  GenFlavorMsg      (Params : TStringList) : integer;
    function  GenTextMsg        (Params : TStringList) : integer;
    function  GenMCEvent        (Params : TStringList) : integer;
    function  GenPopUpText      (Params : TStringList) : integer;
    function  GenPopUpAdvText   (Params : TStringList) : integer;
    function  ChgEventOccurs    (Params : TStringList) : integer;
    function  SetEventOccurs    (Params : TStringList) : integer;
    function  SetWinLevel       (Params : TStringList) : integer;
    function  ChgVPCount        (Params : TStringList) : integer;
    function  SetSudDeath       (Params : TStringList) : integer;
    function  SetPettyWar       (Params : TStringList) : integer;
    function  SetFacPassive     (Params : TStringList) : integer;
    function  SetFacUnselect    (Params : TStringList) : integer;
    function  SetIsTutorial     (Params : TStringList) : integer;
    function  SetLoyaltyRule    (Params : TStringList) : integer;
    function  SetObjectives     (Params : TStringList) : integer;
    function  ChgObjective      (Params : TStringList) : integer;
    function  SetWinParams      (Params : TStringList) : integer;
    function  SetWinParamsVP    (Params : TStringList) : integer;
    function  SetLoseParams     (Params : TStringList) : integer;
    function  SetLosePrmsVP     (Params : TStringList) : integer;
    function  SetScenVicCond    (Params : TStringList) : integer;
    function  SetScenText       (Params : TStringList) : integer;
    function  SetSide1          (Params : TStringList) : integer;
    function  SetSide2          (Params : TStringList) : integer;
    function  ScenBGround1      (Params : TStringList) : integer;
    function  ScenBGround2      (Params : TStringList) : integer;
    function  ScenDuration      (Params : TStringList) : integer;
    function  ScenFavored       (Params : TStringList) : integer;
    function  ScenImageBox      (Params : TStringList) : integer;
    function  ChangeGroupFixing (Params : TStringList) : integer;    
    function  SetFixing         (Params : TStringList) : integer;
    function  AddSubFacRgn      (Params : TStringList) : integer;
    function  RemSubFacRgn      (Params : TStringList) : integer;
    function  ChangeLoyaltyFac  (Params : TStringList) : integer;
    function  ChangeLoyalty     (Params : TStringList) : integer;
    function  AddLoyFacCoeff    (Params : TStringList) : integer;
    function  MoveUnit          (Params : TStringList) : integer;
    function  OnAttack          (Params : TStringList) : integer;
    function  ChangeSupply      (Params : TStringList) : integer;
    function  ChangeResStock    (Params : TStringList) : integer;
    function  ChangeAssetsProd  (Params : TStringList) : integer;
    function  ChangeRailPool    (Params : TStringList) : integer;
    function  ChangeRiverPool   (Params : TStringList) : integer;
    function  ChangeMoneyPool   (Params : TStringList) : integer;
    function  ChangeWSUPool     (Params : TStringList) : integer;
    function  ChangeConscripts  (Params : TStringList) : integer;
    function  ChangePriceInc    (Params : TStringList) : integer;
    function  ChangeMorale      (Params : TStringList) : integer;
    function  ChangeEngagementPts (Params : TStringList) : integer;
    function  ChangeUnitPool    (Params : TStringList) : integer;
    function  ChangeUnitType    (Params : TStringList) : integer;
    function  ChangeUnitTypeByGen(Params : TStringList) : integer;    
    function  ObsoleteModelGen  (Params : TStringList) : integer;
    function  ObsoleteStrucGen  (Params : TStringList) : integer;
    function  ChangeStrucPool   (Params : TStringList) : integer;
    function  ChangeStrucType   (Params : TStringList) : integer;
    function  ChangeStrucLevel  (Params : TStringList) : integer;
    function  MoveStruc         (Params : TStringList) : integer;    
    function  ChangeFortBreach  (Params : TStringList) : integer;
    function  SetStrucCustNames (Params : TStringList) : integer;
    function  RemoveStructure   (Params : TStringList) : integer;
    function  SetActorPool      (Params : TStringList) : integer;
    function  PayOption         (Params : TStringList) : integer;    
    function  SetTRSAssetsInfos (Params : TStringList) : integer;
    function  SetRevoltersInfos (Params : TStringList) : integer;
    function  SetUprisingInfos  (Params : TStringList) : integer;
    function  SetScorchEarthInfos(Params : TStringList) : integer;
    function  SetNavalPools     (Params : TStringList) : integer;
    function  SetBlockades      (Params : TStringList) : integer;
    function  SetDistBlockades  (Params : TStringList) : integer;
    function  SetMoraleLevels   (Params : TStringList) : integer;
    function  ChgMoraleLevels   (Params : TStringList) : integer;
    function  AISetPoolRatio    (Params : TStringList) : integer;
    function  AISetLandBuild    (Params : TStringList) : integer;
    function  AISetNavBuild     (Params : TStringList) : integer;
    function  SetUnitPoolInfos  (Params : TStringList) : integer;
    function  ChgUnitPoolInfos  (Params : TStringList) : integer;    
    function  AIChangeAggro     (Params : TStringList) : integer;
    function  AISetAggro        (Params : TStringList) : integer;
    function  AIChangeTheaPrev  (Params : TStringList) : integer;
    function  AISetTheaPrev     (Params : TStringList) : integer;
    function  AIChgeLocalIntrst (Params : TStringList) : integer;
    function  AISetLocalIntrst  (Params : TStringList) : integer;
    function  AIChgeAreaLocalInt(Params : TStringList) : integer;
    function  AISetAreaLocalInt (Params : TStringList) : integer;
    function  AISetParams       (Params : TStringList) : integer;
    function  SetFacMaxEntrench (Params : TStringList) : integer;
    function  ChangeFactionRel  (Params : TStringList) : integer;
    function  SetFactionRel     (Params : TStringList) : integer;
    function  ChangeForeignEntry(Params : TStringList) : integer;
    function  AddHarborNoBlock  (Params : TStringList) : integer;
    function  AddHarborBlockaded(Params : TStringList) : integer;
    function  SetPromoteFlag    (Params : TStringList) : integer;
    function  SetCapitals       (Params : TStringList) : integer;
    function  AddXPLevelToGroup (Params : TStringList) : integer;
    function  AddXPLevelToUnit  (Params : TStringList) : integer;
    function  ChgGroupHealth    (Params : TStringList) : integer;
    function  ChgUnitHealth     (Params : TStringList) : integer;
    function  SetBuildFlag      (Params : TStringList) : integer;
    function  ChgGroupCohesion  (Params : TStringList) : integer;
    function  ChgUnitCohesion   (Params : TStringList) : integer;
    function  AlterCuSubUnit    (Params : TStringList) : integer;
    function  SetLeaderActivity (Params : TStringList) : integer;
    function  SetMsgGeneration  (Params : TStringList) : integer;
    function  SetCombiUnits     (Params : TStringList) : integer;
    function  SetFormedCmdMax   (Params : TStringList) : integer;
    function  ChangeOption      (Params : TStringList) : integer;
    function  AddGarrison       (Params : TStringList) : integer;
    function  ClearGarrisons    (Params : TStringList) : integer;
    function  ChangeOneTurnMod  (Params : TStringList) : integer;
    function  SetUnknownRegions (Params : TStringList) : integer;
    function  SetKnownRegions   (Params : TStringList) : integer;
    function  SetRegionsAllKnown(Params : TStringList) : integer;
    function  SetAllRgnLegitOwn (Params : TStringList) : integer;
    function  SetLegitOwner     (Params : TStringList) : integer;
    function  SetFactionAttribute(Params : TStringList) : integer;
    function  SetTechDiscovery  (Params : TStringList) : integer;
    function  AddNationalRegions(Params : TStringList) : integer;
    function  RemNationalRegions(Params : TStringList) : integer;
    function  AddClaimedRegions (Params : TStringList) : integer;
    function  RemClaimedRegions (Params : TStringList) : integer;
    function  AddDeJureRegions  (Params : TStringList) : integer;
    function  RemDeJureRegions  (Params : TStringList) : integer;
    function  ChangeSellTransac (Params : TStringList) : integer;
    function  ChangeBuyTransac  (Params : TStringList) : integer;
    function  ChangeIntrProd    (Params : TStringList) : integer;
    function  ChangeResDiscovery(Params : TStringList) : integer;
    function  SetSupplyMod      (Params : TStringList) : integer;
    function  CreateCrisis      (Params : TStringList) : integer;
    function  AddFacModifiers   (Params : TStringList) : integer;
    function  RemFacModifiers   (Params : TStringList) : integer;
    function  EnableDiploItem   (Params : TStringList) : integer;
    function  MakeTechKnown     (Params : TStringList) : integer;
    function  AddRulerToFac     (Params : TStringList) : integer;
    function  ChgLeadersPool    (Params : TStringList) : integer;
    function  AddProvLeaders    (Params : TStringList) : integer;
    function  RemProvLeaders    (Params : TStringList) : integer;
    function  SetAllFacTaxes    (Params : TStringList) : integer;
    function  SetTaxes          (Params : TStringList) : integer;
    function  AddDynamicGraphic (Params : TStringList) : integer;
    function  RemDynamicGraphic (Params : TStringList) : integer;
    function  ChangeFacDisplayTag(Params : TStringList) : integer;
    function  SetFacDisplayNames(Params : TStringList) : integer;
    function  SetFacDisplayFrcNames(Params : TStringList) : integer;
    function  SetFacDisplayColors(Params : TStringList) : integer;
    function  ChangeGameRules   (Params : TStringList) : integer;
    function  AddDiploItem      (Params : TStringList) : integer;
    function  RemDiploItem      (Params : TStringList) : integer;
    function  SettleWhitePeace  (Params : TStringList) : integer;
    function  ChangePrisoners   (Params : TStringList) : integer;
    function  ChangeRgnOwner    (Params : TStringList) : integer;
    function  AbsorbFaction     (Params : TStringList) : integer;    
    function  SetSOIs           (Params : TStringList) : integer;
    function  ChangeSOIs        (Params : TStringList) : integer;
    function  SetAdvRevolt      (Params : TStringList) : integer;
    function  SetAdvRevoltFP    (Params : TStringList) : integer;
    function  SetAdvRevoltLoc   (Params : TStringList) : integer;
    function  TriggerAdvRev     (Params : TStringList) : integer;
    function  ChgRgnDecisionFP  (Params : TStringList) : integer;
    function  AddRgnDecisionAct (Params : TStringList) : integer;
    function  RemRgnDecisionAct (Params : TStringList) : integer;
    function  AddRgnDecisionExcl(Params : TStringList) : integer;
    function  RemRgnDecisionExcl(Params : TStringList) : integer;
    function  SetColonialStatusArea(Params : TStringList) : integer;
    function  SetPillage        (Params : TStringList) : integer;
    function  Mobilize          (Params : TStringList) : integer;
    function  Demobilize        (Params : TStringList) : integer;
    function  SetNextElection   (Params : TStringList) : integer;
    function  SetRulerItemActiveState (Params : TStringList) : integer;
    function  SetFAOverride     (Params : TStringList) : integer;
    function  RemFAOverride     (Params : TStringList) : integer;    
    function  AISetDiploRelBias   (Params : TStringList) : integer;
    function  AIChangeDiploRelBias(Params : TStringList) : integer;
    function  AISetDiploItemBias  (Params : TStringList) : integer;
    function  AddMapAlea        (Params : TStringList) : integer;
    function  RemMapAlea        (Params : TStringList) : integer;
    function  AIAddLeaderAgent  (Params : TStringList) : integer;
    function  TakeOverUnits     (Params : TStringList) : integer;

    function  Apply : integer;
    function  ApplyCreateGroup   : integer;
    function  ApplyCreateUnit    : integer;
    function  ApplyCreateSubUnit : integer;
    function  ApplyCreateStruc   : integer;

    public
    cuFaction     : TFaction; // donnes courantes
    cuRegList     : TIntList; // liste d'UID des rgions en slection
    cuGroup       : TGroup;
    cuUnit        : TUnit;
    cuSubUnit     : TSubUnit;
    cuSUList      : TList; // Liste de SU en slection

    property cuRegion : TRegion read fcuRegion write SetCuRegion;

    // Actions (certaines actions sont utilises par le hosting)
    function  SelectFactions      (Params : TStringList) : integer;
    function  SelectRegion        (Params : TStringList) : integer;
    function  SelectRegInArea     (Params : TStringList; ExcludeDisloyal : boolean = False; ExcludeLowControlled : boolean = False) : integer;
    function  SelectUnitInArea    (Params : TStringList) : integer;
    function  SelectGroups        (Params : TStringList) : integer;
    function  SelectUnits         (Params : TStringList) : integer;
    function  SelectSubUnits      (Params : TStringList) : integer; // Peut passer le bloc Condition  False
    function  SelectUniqueUnit    (Params : TStringList) : integer;
    function  RemoveCuUnit        (Params : TStringList) : integer;
    function  SelectUniqueSubUnit (Params : TStringList) : integer; // Peut passer le bloc Condition  False
    function  RemoveUnqUnit       (Params : TStringList) : integer;
    function  RemoveCuSubUnit     (Params : TStringList) : integer;
    function  RemoveSubUnits      (Params : TStringList) : integer;
    function  AddToRegList        (Params : TStringList) : integer;
    function  AddToRegListFromAreas(Params : TStringList) : integer;    
    function  PickFromRegList     (Params : TStringList) : integer;
    function  SelNumerousGroup    (Params : TStringList) : integer;

  end;

  // une tiquette et sa slection attache
  TBindList = class(TStringList)
  public
    VarLabel : string;
  end;


implementation

uses Forms, UGameSim, UProfiler, UGlobProc, URegionDef, ULocalize, UStrucDef,
     UStructures, URegionCommon, UArea, UFactionDef, UUnitDef, UModel, UEconomy,
     UResearch, UDiplomacy, UDiplomacyDef, UResearchDef, URuler, UTerrainDef,
     UScriptDefs, UColonialModule, URegionDecisionsDef, UMap, UConsole,
  UMapAleas, UAIAgent, UAIAgentDef;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// CONSTRUCTOR / DESTRUCTOR
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Constructor TScriptEngine.Create(aOwner : TObject);
begin
  Owner := aOwner;
  if not (Owner is TGameSim) then
    LogMsg(['TScriptEngine.Create', 'Owner is not a TGameSim'], _Critical);
  Inherited Create;
  slParse       := TStringList.Create;
  slScriptLog   := TStringList.Create;
  olBoundVar    := TObjectList.Create;
  olBoundVar.OwnsObjects := True;
  WorkingVars   := TStringList.Create;
  Conditions    := condTrue;
  cuRegList     := TIntList.Create;
  cuSUList      := TList.Create;
  lastGHQGroup  := nil;
  ResetActionsVariables;
end;

Destructor TScriptEngine.Destroy;
begin
  FreeAndNil(slParse);
  FreeAndNil(slScriptLog);
  FreeAndNil(olBoundVar);
  FreeAndNil(WorkingVars);
  FreeAndNil(cuRegList);
  FreeAndNil(cuSUList);

  inherited Destroy;
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// LOAD FROM FILE
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.LoadFromFile(FileName : string) : boolean;
begin
  Result := False;
  if FileExists(FileName) then
  begin
    slParse.LoadFromFile(FileName);
    Result := True;
  end;
end;


// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// LOAD EVENT FROM FILE
// Load into slparse an event from a file
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.LoadEventFromFile(EventName: string; FileName : string) : boolean;
var ScriptSL  : TStringList;
    ParamSL   : TStringList;
    Key       : string;
    KeyParams : string;
    i,j       : integer;

begin

  Result := False;

  try
    ScriptSL := TStringList.Create;
    ParamSL  := TStringList.Create;
    slParse.Clear;

    if FileExists(FileName) then
    begin
      ScriptSL.LoadFromFile(FileName);   // Load ScriptFile
      for i := 0 to ScriptSL.Count -1 do // Check the script for the event to parse
      begin
        Key := Trim(TFileParser.ReadKeyValue(ScriptSL[i]));
        if Uppercase(Key) = Uppercase(__KeyEvent) then     // Search for "StartEvent" command
        begin
          KeyParams := TFileParser.ReadValue(ScriptSL[i]); // Read parameters of the StartEvent command
          RxEW_WordsToStringListEx(KeyParams, ParamSL, ['|', ';']); // Copy the parameters to a StringList
          if ParamSL.Count = 7 then // Check for correct number of parameters
          begin
            if Uppercase(ParamSL.Strings[0]) = Uppercase(EventName) then // if correct event was found
            begin
              j := i;
              repeat // Add strings to slParse till "EndEvent" was found or end of ScriptSL
                slParse.Add(ScriptSL.Strings[j]); // Add all lines of the event to "slParse" so that it can later be parsed
                inc(j);
              until (Uppercase(slParse.Strings[slParse.Count-1]) = Uppercase(__KeyEndEvent)) or (j = ScriptSL.Count); // not Count-1 as inc(j) after the slParse.Add

              if Uppercase(slParse.Strings[slParse.Count-1]) = Uppercase(__KeyEndEvent) then // "EndEvent" was found --> Everything OK!
              begin
                Result := True;
                Break; // Exit the for-loop, goes to finally and exit the function
              end
              else  // No "EndEvent" keyword found
              begin
                slParse.Clear; // Delete the Parse SL, as it contains an invalid event
                LogMsg(['TScriptEngine.LoadEventFromFile: "StartEvent" was found but no "EndEvent"', 'File:', FileName, 'Event:', EventName], _Warning);
                Break; // Exit the for-loop, goes to finally and exit the function. Result = false
              end;
            end; // End: Copy strings to slParse
          end    // End: Start Event has the correct number of parameters
          else
          begin
            LogMsg(['TScriptEngine.LoadEventFromFile: Start event command with erroneous number of params, found:', 'File:', FileName, 'Line:',i], _Warning);
            Continue;
          end;
        end; // End: Keyword "StartEvent" found
      end;   // End: Search in script file for the correcte event

    end; // End: If file exists


  finally
    FreeAndNil(ScriptSL);
    FreeAndNil(ParamSL);
  end;
end;



// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// REPLACE IN SLPARSE
// ex: replace all 'CMN' by 'FRA'. Case sensitive to avoid trouble like replacing
// 'France' by 'CMNnce' ...
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
procedure   TScriptEngine.ReplaceInSlParse(Search : string; Replace : string);
var i       : integer;
    tempStr : string;
    strPos  : integer;
begin

  if Search = '' then exit;

  for i := 0 to slParse.Count -1 do
  begin
    while Pos(Search, slParse.Strings[i]) > 0 do // while loop needed if more than one string to replace in a line
    begin
      strPos  := Pos(Search, slParse.Strings[i]);
      tempStr := Copy(slParse.Strings[i], 1, strPos -1);
      tempStr := tempStr + Replace;
      tempStr := tempStr + Copy(slParse.Strings[i], strPos + length(Search), length(slParse.Strings[i]));
      slParse.Strings[i] := tempStr;
    end;
  end;
end;





// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// TRIGGER EVENT
// Execute the event code from the Template file given in the Tech/DiploItem/decision/other subsystem
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
procedure TScriptEngine.TriggerEvent(FileName : string; EventName : string; TagList : TStringList; ReplaceList : TStringList; LogFile : string);
var ScriptName    : string;
    DebugStep     : integer;
    i             : integer;
    j             : integer;
    strPos        : integer;
    NativeTag     : string;
    tempStr       : string;
begin

  DebugStep := 0;

  try

    // Some initial checks
    DebugStep := 10;
    if TagList.Count <> ReplaceList.Count then
    begin
      LogMsg(['TScriptEngine.TriggerEvent', 'The TagList has a different length than the replace list!', FileName], _Critical);
    end;

    // Get the Event file
    DebugStep := 20;
    if (FileName <> '') and (EventName <> '') then
    begin
      ScriptName := EventsPath_ + '\' + FileNameOnly(FileName) + '.' + EventsExt_;
      if not FileExists(ScriptName) then
        ScriptName := EventsPath_ + '\' + FileName + '.' + EventsExt_;
      if not FileExists(ScriptName) then
      begin
        LogMsg(['TScriptEngine.TriggerEvent', 'Cannot find event file:', FileName], _Reporting);
        Exit;
      end;

      // Get the event from the event file and parse it.
      if LoadEventFromFile(Trim(EventName), ScriptName) = True then
      begin
        for i := 0 to TagList.Count -1 do
        begin
          ReplaceInSlParse(TagList[i], ReplaceList[i]); // change the tags
        end;

        // Hack to have all native tags included in the event, by duplicating the command.
        // example if ChangeUnitPool = $uni_AUS_Tra1;5 is encountered with AUS has current faction
        // then a new line will be added dynamically in memory, ChangeUnitPool = $uni_AHU_Tra1;5
        // same for all nativetags/subfactions...
        for i := 0 to slParse.Count-1 do
        begin
          if (Pos('_' + cuFaction.FactionTag + '_', slParse[i]) > 0)
          and ( (Pos('CHANGEUNITPOOL', UpperCase(slParse[i])) > 0) or
                (Pos('CHANGEUNITTYPE', UpperCase(slParse[i])) > 0) )
          then
          begin
            for j := 1 to RxEW_GetWordCount(cuFaction.NativeTag) do
            begin
              NativeTag := ExString(cuFaction.NativeTag, j);
              if NativeTag = cuFaction.FactionTag then Continue;

              strPos  := Pos('_' + cuFaction.FactionTag + '_', slParse[i]);
              tempStr := Copy(slParse[i], 1, strPos -1);
              tempStr := tempStr + '_' + NativeTag + '_';
              tempStr := tempStr + Copy(slParse[i], strPos + length('_' + NativeTag + '_'), length(slParse[i]));
              slParse.Insert(i+1, tempStr);
            end;
          end;
        end;


        Parse(LogFile + '.txt'); // execute (1 event stored in slParse if all went well)
      end
      else
        LogMsg(['TScriptEngine.TriggerEvent', 'This event not found while triggering a specific script', ScriptName, EventName], _Reporting);
    end; // End: if (FileName <> '') and (EventName <> '')

  except
    on E : Exception do
      LogMsg(['TScriptEngine.TriggerEvent','Exception caught:', E.Message, 'DebugStep:', DebugStep], _Critical);
  end;

end;


// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// IS KNOWN COMMAND
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.IsKnownCommand(Str : string) : boolean;
var i : integer;
begin
  Result := False;
  for i := 1 to High(__seCommands) do
    if Trim(UpperCase(Str)) = UpperCase(__seCommands[i]) then
    begin
      Result := True;
      Exit;
    end;

end;

function TScriptEngine.CommandCanHaveAliasWithSpaceSeparator(Command : string) : boolean;
begin
  Result := True;
//  if Pos(__KeySelectSubUnits, Command) > 0 then Result := False;
//  if Pos(__KeySelectSubUnit,  Command) > 0 then Result := False;
end;
function TScriptEngine.CommandShouldbeParsed(Command : string) : boolean;
begin
  Result := True;
  if Pos(__KeySelectSubUnits, Command) > 0 then Result := False;
  if Pos(__KeySelectSubUnit,  Command) > 0 then Result := False;
  if Pos(__KeyAIAddLeaderAgent, Command) > 0 then Result := False;
end;
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// PARSE
// Rsultat: nb d'erreurs du script, voir slScriptLog
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.Parse(LogFile : string = '') : integer;
var i : integer;
    Key, Str : string;
    cuResult : integer;
    slReport : TStringList;
    WarnCount : integer;
    //slBAKScriptLog : TStringList;
    TriggerScript : boolean;
begin
  Result := 0;
  TriggerScript := IsSpecialScript; // specify here if we are running a script coming from a tech, attribute, etc.
                                     // i.e not the 'main' script engine but one created for a single operation

  Running := True;
  try
    try
      (*
      // if a research script, keep aside the master script log
      if IsResearchScript then
      begin
        slBAKScriptLog := TStringList.Create;
        slBAKScriptLog.Assign(slScriptLog);
      end;
      *)

      slScriptLog.Clear;
      if slParse.Count = 0 then Exit;

      // Reporting in a Pre Parse file
      slReport := TStringList.Create; // outside of the test below to remove a warning...
      if not TriggerScript then
      begin
        slReport.Assign(slParse);
        for i := 0 to slReport.Count-1 do slReport[i] := PadL(ITS(i), 6, '0') + ' ' + slReport[i];
        slReport.SaveToFile(ScriptPreParseFile_);
      end;

      // .inc dans le rpertoire include OU
      // .sct dans le rpertoire d'evt ainsi que sous dir
      repeat
        i := TFileParser.ParseForIncludes(slParse, EventsPath_, True, '.' + EventsExt_);
        i := i + TFileParser.ParseForIncludes(slParse, IncludesPath_);
      until i = 0; // inc de inc/sct

      // Reporting in a Post Parse file
      if (not TriggerScript) then
      begin
        slReport.Assign(slParse);
        for i := 0 to slReport.Count-1 do slReport[i] := PadL(ITS(i), 6, '0') + ' ' + slReport[i];
        slReport.SaveToFile(ScriptPostParseFile_);
      end;
      slReport.Free;

      // Init the turn to use (can be updated in SetScenStart)
      if sysScriptForceTurn_ = '' then
        UsedTurn := TGameSim(Owner).cuTurn
      else
        UsedTurn := DateToTurn(sysScriptForceTurn_);

      // ALIAS
      // les entres spares par | ou ; peuvent avoir des alias  et deviendront un param par entre quand on appel ParseKey (un param pouvant avoir des arguments)
      // le sparateur entre 2 alias est |ou ; mais ne peut tre espace (certains alias ont des espaces comme les rgions)
      // On ne peut donc mettre sauf si prvu via CommandCanHaveAliasWithSpaceSeparator
      // d'alias dans les arguments d'un paramtre ($famLeader 5 comme paramtre ne sera pas pars pour l'alias)
      for i := 0 to slParse.Count -1 do
      begin
        Str := slParse[i];
        AliasWithSpaceSeparatorIsPossible_ := CommandCanHaveAliasWithSpaceSeparator(Str); // always True now (they can all have that)

        // Some commmands are not parsed here but in their function, typically when we have Regions $Reg1 $Reg2 (a parameter with aliases as args)
        if CommandShouldbeParsed(Str) then
          TFileParser.ParseForAliases(Str, '$');
        AliasWithSpaceSeparatorIsPossible_ := True; // default
        slParse[i] := Str;
      end;

      // le script reste en latin, on enlve tout accents
      if locCurrentLanguage_ in [locRussian_, locPolish_] then
        for i := 0 to slParse.Count-1 do
          slParse[i] := ReplaceAccents(slParse[i]);

      LastGHQGroup := nil;
      ResetActionsVariables(not TriggerScript); // don't reset action variables if this is a triggered script (research, RGD), as we
      for i := 0 to slParse.Count -1 do         // initialized on purpose before 
      begin
        cuLine := i;

        if i /1000 = i div 1000 then GameSim.APM;

        Key := Trim(TFileParser.ReadKeyValue(slParse[i]));
        if Key = '' then Key := Trim(slParse[i]);
          if IsKnownCommand(Key) then
          begin
            cuResult := ParseKey(Key, TFileParser.ReadStrKey(Key, slParse[i]));
            Inc(Result, cuResult);
            if FullDebug_ and (cuResult <> 0) then
              LogMsg(['FullDebug: Script Engine Error(s), while parsing line', i, slParse[i]], _Error, False, True);
          end
          else
          if (Length(Key) > 0)and(not(Key[1] in ['/', '{', '(', '#']))and(Key <> 'NULL')
          and(UpperCase(Key) <> 'SCRIPTNAME')     // include dans script de scnario (utilis uniquement par CSVSplitter)
          then begin
            slScriptLog.Add('=> Unknown keyword at line ' + ITS(i) + ' ' + Key);
            LogMsg(['FullDebug: Script Engine, while parsing line', i, 'Unknown Keyword: ', Key], _Reporting, False, True);
          end;
      end;

      // Slow: only for the main script engine, trigger scripts should not
      // call that each time except at the very end of all processes
      if not TriggerScript then
        FinalizeParsing;

      if DoLegitOwn then
      begin
        GameSim.RegionCmns.CalcScriptAllLegitOwned;
        DoLegitOwn := False;
      end;

      if Result > 0 then
        LogMsg(['Errors found in script:', Result, ' Check', LogFile], _Error);

    except
      on E: exception do
        LogMsg(['TScriptEngine.Parse','Exception caught:', E.Message, 'while parsing line', cuLine, ':', Key], _Critical);
    end;
  finally
    slScriptLog.Insert(0, '--------------START OF SCRIPT----------------');
    if Result = 0 then
      slScriptLog.Insert(0, 'No error in this script')
    else
      slScriptLog.Insert(0, 'Error(s) found: ' + ITS(Result));

    slScriptLog.Add('------------------------------');
    slScriptLog.Add('Error(s) found: ' + ITS(Result));
    WarnCount := 0;
    for i := 0 to slScriptLog.Count-1 do
      if Pos('[WARNING]', UpperCase(slScriptLog[i])) > 0 then
        Inc(WarnCount);
    slScriptLog.Add('Warning(s) found: ' + ITS(WarnCount));
    slScriptLog.Add('--------------END OF SCRIPT----------------');
    if LogFile <> '' then
    begin
      if not DirectoryExists(ExtractFilePath(LogFile)) then
         ForceDirectories(ExtractFilePath(LogFile)); // for research
      if LogFile = ScriptReportFile_ then
        slScriptLog.SaveToFile(LogFile + ITS(GameSim.cuTurn) + '.txt')
      else
        slScriptLog.SaveToFile(LogFile);
    end;

    Running := False;
  end;

end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// FINALIZE PARSING
// To be called when a batch of scripts has been run, not on each trigger e.g
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
class procedure TScriptEngine.FinalizeParsing;
var i,j : integer;
    DebugStep : string;
begin

  try
    for i := 0 to GameSim.Factions.Count -1 do
    begin
      DebugStep := 'Faction ' + ITS(i);
      // Ajout de possessions => il faut tendre la liste des rgions  l'alentour
      GameSim.Factions[i].Regions.ExtendNoEmpty;
      // Calcul blocus si besoin
      GameSim.Factions[i].CalcBlockadePerc;
      GameSim.Factions[i].GetStrucNbOnMap(0, True, True); // Force recalc, necessary for SynchColStatus
      // Update on all new groups, the ones created did not receive any update at all
      for j := 0 to GameSim.Factions[i].Count-1 do
        if GameSim.Factions[i][j].NeedUpdate then
          GameSim.Factions[i][j].Update;
    end;

    DebugStep := 'Colonial Update';
    GameSim.ColModule.SynchColonialStatus();

    // MJ Region Cmns
    DebugStep := 'Regions Update';    
    GameSim.CalcRegionCommons([ukAll]);

  except
    on E : Exception do
      LogMsg(['TScriptEngine.FinalizeParsing','Exception caught:', E.Message, 'DebugStep:', DebugStep], _Error);
  end;

end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// PARSE KEY
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.ParseKey(Key, KeyParams : string): integer;
var SL : TStringList;
begin
  Result := 0;
  Key := UpperCase(Key);
  KeyParams := Trim(KeyParams);
  SL := TStringList.Create;

  try
    RxEW_WordsToStringListEx(KeyParams, SL, ['|', ';']);

    // Commandes fonctionnant galement hors vnements ou bornes d'vnements

    if Key = UpperCase(__KeyScenName) then
    begin
      Result := SetScenName(SL);
      Exit;
    end
    else
    if Key = UpperCase(__KeyScenDispName) then
    begin
      Result := SetScnDispName(SL);
      Exit;
    end
    else
    if Key = UpperCase(__KeyEvent) then
    begin
      Result := StartEvent(SL);
      Exit;
    end
    else
    if Key = UpperCase(__KeyEndEvent) then
    begin
      Result := EndEvent(SL);
      Exit;
    end
    else
    if Key = UpperCase(__KeySelectFaction) then
    begin
      Result := SelectFactions(SL);
      Exit;
    end
    else
    if Key = UpperCase(__KeyCountCtrlAreas) then
    begin
      CountCtrlAreas(SL);
      Exit;
    end
    else
    if Key = UpperCase(__KeyCountVP) then
    begin
      Result := CountVP(SL);
      Exit;
    end
    else
    if Key = UpperCase(__KeyPickUnit) then
    begin
      Result := PickUnit(SL);
      Exit;
    end
    else
    if Key = UpperCase(__KeyMsgNeedRegion) then
    begin
      Result := SetMsgGeneration(SL);
      Exit;
    end
    else
    if Key = UpperCase(__KeySubUnitCount) then
    begin
      Result := SubUnitCount(SL);
      Exit;
    end

    // Registry manipulation
    else
    if Key = UpperCase(__KeyRegActivate) then
    begin
      Result := Reg_Activate(SL);
      Exit;
    end
    else
    if Key = UpperCase(__KeyRegDeActivate) then
    begin
      Result := Reg_DeActivate(SL);
      Exit;
    end
    else
    if Key = UpperCase(__KeyRegInit) then
    begin
      Result := Reg_Init(SL);
      Exit;
    end
    else
    if Key = UpperCase(__KeyRegSum) then
    begin
      Result := Reg_Sum(SL);
      Exit;
    end
    else
    if Key = UpperCase(__KeyRegRem) then
    begin
      Result := Reg_Rem(SL);
      Exit;
    end
    else
    if Key = UpperCase(__KeyRegSumNum) then
    begin
      Result := Reg_SumNum(SL);
      Exit;
    end
    else
    if Key = UpperCase(__KeyRegRem) then
    begin
      Result := Reg_Rem(SL);
      Exit;
    end
    else
    if Key = UpperCase(__KeyRegEval) then
    begin
      Result := Reg_Eval(SL);
      Exit;
    end
    else
    if Key = UpperCase(__KeyRegEvalNum) then
    begin
      Result := Reg_EvalNum(SL);
      Exit;
    end;

    // Les commandes qui suivent ne s'appliquent que dans le cadre d'un vnement
    if ParsedEvent = '' then Exit;

    // => on est dans un vnement

    // Commandes inconditionnelles (mais ne fct que dans le bloc evt)
    if Key = UpperCase(__KeyResetSelUnit) then
    begin
      cuUnit := nil;
      Exit;
    end;

    // Entre dans le bloc Conditions
    if Key = UpperCase(__KeyConditions) then
    begin
      Result := ToConditions;
      Exit;
    end;

    // on est dans le bloc Conditions et l'valuation continue:
    // on autorise les Selections et Evaluations
    if ConditionsON and (Conditions <> condFalse) then
    begin
      if Key = UpperCase(__KeyFixedDate) then
        Result := CheckFixedDate(SL)
      else
      if Key = UpperCase(__KeyMinDate) then
        Result := CheckMinDate(SL)
      else
      if Key = UpperCase(__KeyMaxDate) then
        Result := CheckMaxDate(SL)
      else
      if Key = UpperCase(__KeyProbability) then
        Result := Probability(SL)
      else
      if Key = UpperCase(__KeyAICheck) then
        Result := CheckAILevel(SL)
      else
      if Key = UpperCase(__KeyCheckIsPlayer) then
        Result := CheckIsPlayer(SL)
      else
      if Key = UpperCase(__KeyEvalEvent) then
        Result := EvalEvent(SL)
      else
      if Key = UpperCase(__KeyCountUnits) then
        Result := CountUnits(SL)
      else
      if Key = UpperCase(__KeyCountUnitDef) then
        Result := CountUnitDef(SL)
      else
      if Key = UpperCase(__KeyEvalUnqUnit) then
        Result := EvalUniqueUnit(SL)
      else
      if Key = UpperCase(__KeyEvalRgnOwned) then
        Result := EvalRgnOwned(SL)
      else
      if Key = UpperCase(__KeyNotEnemy) then
        Result := NotEnemy(SL)
      else
      if Key = UpperCase(__KeyEvalRgnVP) then
        Result := EvalRgnVP(SL)
      else
      if Key = UpperCase(__KeyEvalRgnObj) then
        Result := EvalRgnObj(SL)
      else
      if Key = UpperCase(__KeyEvalRgnStruc) then
        Result := EvalRgnStruc(SL)
      else
      if Key = UpperCase(__KeyTurnIndex) then
        Result := TurnIndex(SL)
      else
      if Key = UpperCase(__KeyEvalBlockade) then
        Result := EvalBlockadePerc(SL)
      else
      if Key = UpperCase(__KeyEvalLoyalty) then
        Result := EvalLoyalty(SL)
      else
      if Key = UpperCase(__KeyEvalControl) then
        Result := EvalControl(SL)
      else
      if Key = UpperCase(__KeyEvalTrptLvl) then
        Result := EvalTransportLvl(SL)
      else
      if Key = UpperCase(__KeyEvalMorale) then
        Result := EvalMorale(SL)
      else
      if Key = UpperCase(__KeyEvalVP) then
        Result := EvalVP(SL)
      else
      if Key = UpperCase(__KeyEvalConscript) then
        Result := EvalConscript(SL)
      else
      if Key = UpperCase(__KeyEvalRgnWeather) then
        Result := EvalRgnWeather(SL)
      else
      if Key = UpperCase(__KeyEvalSubUnitCt) then
        Result := EvalSubUnitCount(SL)
      else
      if Key = UpperCase(__KeyEvalMoney) then
        Result := EvalMoney(SL)
      else
      if Key = UpperCase(__KeyEvalWSU) then
        Result := EvalWarSupply(SL)
      else
      if Key = UpperCase(__KeyEvalForEntry) then
        Result := EvalForeignEntry(SL)
      else
      if Key = UpperCase(__KeyEvalFactionSel) then
        Result := EvalFactionSel(SL)
      else
      if Key = UpperCase(__KeyEvalGroupSel) then
        Result := EvalGroupSel(SL)
      else
      if Key = UpperCase(__KeyEvalUnitSel) then
        Result := EvalUnitSel(SL)
      else
      if Key = UpperCase(__KeyEvalSubUnitSel) then
        Result := EvalSubUnitSel(SL)
      else
      if Key = UpperCase(__KeyEvalRegionSel) then
        Result := EvalRegionSel(SL)
      else
      if Key = UpperCase(__KeyActorEnabled) then
        Result := ActorEnabled(SL)
      else
      if Key = UpperCase(__KeyActorDisabled) then
        Result := ActorDisabled(SL)
      else
      if Key = UpperCase(__KeyActorUnderUse) then
        Result := ActorUnderUse(SL)
      else
      if Key = UpperCase(__KeyCheckOption) then
        Result := CheckOption(SL)
      else
      if Key = UpperCase(__KeyCanPayOption) then
        Result := CanPayOption(SL)
      else
      if Key = UpperCase(__KeyEvalReplacement) then
        Result := EvalReplacement(SL)
      else
      if Key = UpperCase(__KeyEvalEngagementPts) then
        Result := EvalEngagementPts(SL)
      else
      if Key = UpperCase(__KeyEvalBattles) then
        Result := EvalBattles(SL)
      else
      if Key = UpperCase(__KeyEvalResource) then
        Result := EvalResource(SL)
      else
      if Key = UpperCase(__KeyEvalTechKnown) then
        Result := EvalTechKnown(SL)
      else
      if Key = UpperCase(__KeyFacModExists) then
        Result := FacModifierExists(SL)
      else
//      if Key = UpperCase(__KeyEvalDIAttribute) then
//        Result := EvalDIAttribute(SL)
//      else
      if Key = UpperCase(__KeyEvalFacAttribute) then
        Result := EvalFacAttribute(SL)
      else
//      if Key = UpperCase(__KeyEvalDiploItem) then
//        Result := EvalDiploItem(SL)
//      else
      if Key = UpperCase(__KeyEvalIsAtWarWith) then
        Result := EvalIsAtWarWith(SL)
      else
      if Key = UpperCase(__KeyEvalIsAtPeaceWith) then
        Result := EvalIsAtPeaceWith(SL)
      else
      if Key = UpperCase(__KeyIsObsoleteMdlGen) then
        Result := IsObsoleteModelGen(SL)
      else
      if Key = UpperCase(__KeyIsObsoleteStrGen) then
        Result := IsObsoleteStrucGen(SL)
      else
      if Key = UpperCase(__KeyEvalRuler) then
        Result := EvalRuler(SL)
      else
      if Key = UpperCase(__KeyEvalColStatusArea) then
        Result := EvalColStatusArea(SL)
      else
      if Key = UpperCase(__KeyEvalColStatus) then
        Result := EvalColStatus(SL)
      else
      if Key = UpperCase(__KeyEvalSOI) then
        Result := EvalSOI(SL)
      else
      if Key = UpperCase(__KeyEvalFacRel) then
        Result := EvalFacRel(SL)
      else
      if Key = UpperCase(__KeyCheckPowerRatio) then
        Result := CheckPowerRatio(SL)
      else
      if Key = UpperCase(__KeyEvalAreaPwrRatio) then
        Result := EvalAreaPwrRatio(SL)
      else
      if Key = UpperCase(__KeyEvalStrucProd) then
        Result := EvalStrucProduction(SL)
      else
      if Key = UpperCase(__KeyEvalStrucCount) then
        Result := EvalStrucCount(SL)
      else
      if Key = UpperCase(__KeyEvalHasCapital) then
        Result := EvalHasCapital(SL)
      else

      // New Eval VGN+
      if Key = UpperCase(__KeyEvalMilitancy) then
        Result := EvalMilitancy(SL)
      else
      if Key = UpperCase(__KeyEvalContentment) then
        Result := EvalContentment(SL)
      else
      if Key = UpperCase(__KeyEvalEducation) then
        Result := EvalEducation(SL)
      else
      if Key = UpperCase(__KeyEvalColPen) then
        Result := EvalColPen(SL)
      else
      if Key = UpperCase(__KeyEvalInflation) then
        Result := EvalInflation(SL)
      else
      if Key = UpperCase(__KeyEvalRgnOwnerKind) then
        Result := EvalRgnOwnerKind(SL)
      else
      // Selectors also added in Action block
      if Key = UpperCase(__KeySelectRegion) then
        Result := SelectRegion(SL)
      else
      if (Key = UpperCase(__KeySelectRegInArea))or(Key = UpperCase(__KeySelRegFromArea)) then
        Result := SelectRegInArea(SL)
      else
      if Key = UpperCase(__KeyAddToRegList) then
        Result := AddToRegList(SL)
      else
      if Key = UpperCase(__KeyAddToRegListFromAreas) then
        Result := AddToRegListFromAreas(SL)
      else
      if Key = UpperCase(__KeyPickFromRegList) then
        Result := PickFromRegList(SL)
      else
      if Key = UpperCase(__KeySelectGroup) then
        Result := SelectGroups(SL)
      else
      if Key = UpperCase(__KeySelMostNumGroup) then
        Result := SelNumerousGroup(SL)
      else
      if Key = UpperCase(__KeySelectUnit) then
        Result := SelectUnits(SL)
      else
      if (Key = UpperCase(__KeySelUnqUnit)) or (Key = UpperCase(__KeySelectUnqUnit)) then
        Result := SelectUniqueUnit(SL)
      else
      if Key = UpperCase(__KeySelUnqSubUnit) then
        Result := SelectUniqueSubUnit(SL)
      else
      if Key = UpperCase(__KeySelUnitsInArea) then
        Result := SelectUnitInArea(SL)
      else
      if (Key = UpperCase(__KeySelectSubUnits)) or (Key = UpperCase(__KeySelectSubUnit)) then
        Result := SelectSubUnits(SL)
      else  
      if Key = UpperCase(__KeyEvalPrisoners) then
        Result := EvalPrisoners(SL)
      else
      if Key = UpperCase(__KeyEvalRgnDecisionDef) then
        Result := EvalRgnDecisionDef(SL)
      else
      if Key = UpperCase(__KeyEvalRgnDecisionAct) then
        Result := EvalRgnDecisionAct(SL)
      else
      if Key = UpperCase(__KeyIsRgnDecActValid) then
        Result := IsRgnDecisionActValid(SL)
      else
      if Key = UpperCase(__KeyEvalNatThreats) then
        Result := EvalNatThreats(SL)
      else
      if Key = UpperCase(__KeyEvalBesieged) then
        Result := EvalBesieged(SL)
      else
      if Key = UpperCase(__KeyEvalBlockaded) then
        Result := EvalBlockaded(SL);

      // switch condition to true for forced events
      SwitchForcedEventsToTrue;

    end;

    // Entre dans le bloc Actions
    if (Conditions <> condFalse )and (Key = UpperCase(__KeyActions)) then
    begin
      Result := ToActions; {et on passe ici en condTrue}
      Exit;
    end;

    // Les autres commandes ne s'appliquent que si on est dans les Actions et que
    // les autres Conditions sont runies

    if not ActionsON then Exit;
    if not (Conditions = condTrue) then Exit;

    // Selecteurs
    if Key = UpperCase(__KeySelectRegion) then
      Result := SelectRegion(SL)
    else
    if (Key = UpperCase(__KeySelectRegInArea))or(Key = UpperCase(__KeySelRegFromArea)) then
      Result := SelectRegInArea(SL)
    else
    if Key = UpperCase(__KeySelectGroup) then
      Result := SelectGroups(SL)
    else
    if Key = UpperCase(__KeySelMostNumGroup) then
      Result := SelNumerousGroup(SL)
    else
    if Key = UpperCase(__KeySelectUnit) then
      Result := SelectUnits(SL)
    else
    if (Key = UpperCase(__KeySelUnqUnit)) or (Key = UpperCase(__KeySelectUnqUnit)) then
      Result := SelectUniqueUnit(SL)
    else
    if Key = UpperCase(__KeySelUnqSubUnit) then
      Result := SelectUniqueSubUnit(SL)
    else
    if Key = UpperCase(__KeySelUnitsInArea) then
      Result := SelectUnitInArea(SL)
    else
    if (Key = UpperCase(__KeySelectSubUnit)) or (Key = UpperCase(__KeySelectSubUnits)) then
      Result := SelectSubUnits(SL)
    else
    if Key = UpperCase(__KeyRemoveCuUnit) then
      Result := RemoveCuUnit(SL)
    else
    if Key = UpperCase(__KeyRemoveCuSubUnit) then
      Result := RemoveCuSubUnit(SL)
    else
    if (Key = UpperCase(__KeyRemoveUnqUnit)) or (Key = UpperCase(__KeyRemUnqUnit)) then
      Result := RemoveUnqUnit(SL)
    else
    if Key = UpperCase(__KeyRemoveSubUnits) then
      Result := RemoveSubUnits(SL)
    else
    if Key = UpperCase(__KeyGenMessage) then
      Result := GenFlavorMsg(SL)
    else
    if Key = UpperCase(__KeyGenTextMsg) then
      Result := GenTextMsg(SL)
    else
    if Key = UpperCase(__KeyGenMCEvent) then
      Result := GenMCEvent(SL)
    else
    if Key = UpperCase(__KeyGenPopUpText) then
      Result := GenPopUpText(SL)
    else
    if Key = UpperCase(__KeyGenPopUpAdvTxt) then   // Ajout Yoann 'Advanced Text'
      Result := GenPopUpAdvText(SL)
    else
    if Key = UpperCase(__KeyPickFromRegList) then
      Result := PickFromRegList(SL)
    else
    if Key = UpperCase(__KeyCreateFaction) then
      Result := CreateFaction(SL)
    else
    if Key = UpperCase(__KeyCreateGroup) then
      Result := CreateObject(_geGroup, SL)
    else
    if Key = UpperCase(__KeyCreateUnit) then
    begin
      UseCustom := False;
      Result := CreateObject(_geUnit, SL);
    end
    else
    if Key = UpperCase(__KeyCustomUnit) then
    begin
      UseCustom := True;
      Result := CreateObject(_geUnit, SL);
    end
    else
    if Key = UpperCase(__KeyUpgradeUnit) then
    begin
      Result := UpgradeUnit(SL);
    end
    else
    if Key = UpperCase(__KeyCreateSubUnit) then
      Result := CreateObject(_geSubU, SL)
    else
    if Key = UpperCase(__KeyCreateStruc) then
      Result := CreateObject(_geStruc, SL)
    else
    if Key = UpperCase(__KeyCreatePop) then
      Result := CreateObject(_gePop, SL)
    else
    if Key = UpperCase(__KeyApplyCommand) then
      Result := Apply
    else
    if Key = UpperCase(__KeySetName) then
      Result := SetName(SL)
    else
    if Key = UpperCase(__KeyShortName) then
      Result := SetName(SL)
    else
    if Key = UpperCase(__KeyFlavorName) then
      Result := SetFlavorName(SL)
    else
    if Key = UpperCase(__KeySUFlavorName) then
      Result := SetSUFlavorName(SL)
    else
    if Key = UpperCase(__KeySetPromFlag) then
      Result := SetPromoteFlag(SL)
    else
    if (Key = UpperCase(__KeySetCapitals))or(Key = UpperCase(__KeySetCapital)) then
      Result := SetCapitals(SL)
    else
    if Key = UpperCase(__KeyChgForeignEntry) then
      Result := ChangeForeignEntry(SL)
    else
    if Key = UpperCase(__KeyText) then
      Result := SetName(SL)
    else
    if Key = UpperCase(__KeySetKind) then
      Result := SetKind(SL)
    else
    if Key = UpperCase(__KeySetCommit) then
      Result := SetSpecialCommit(SL)
    else
    if Key = UpperCase(__KeyGMEParam) then
      Result := SetGMEParams(SL)
    else
    if Key = UpperCase(__KeySetType) then
      Result := SetType(SL)
    else
    if Key = UpperCase(__KeySetHealth) then
      Result := SetHealth(SL)
    else
    if Key = UpperCase(__KeySetCohesion) then
      Result := SetCohesion(SL)
    else
    if Key = UpperCase(__KeySetLevel) then
      Result := SetLevel(SL)
    else
    if (Key = UpperCase(__KeySetLoyalty))or(Key = UpperCase(__KeyLoyaltyFac))or(Key = UpperCase(__KeyInsurgent)) then
      Result := SetLoyaltyFac(SL)
    else
    if Key = UpperCase(__KeyLoyaltyEth) then
      Result := SetLoyaltyEth(SL)
    else
    if Key = UpperCase(__KeySetControl) then
      Result := SetControl(SL)
    else
    if Key = UpperCase(__KeyAlliance) then
      Result := SetAlliance(SL)
    else
    if (Key = UpperCase(__KeyRemplacement)) or (Key = UpperCase(__KeyReplacement)) then
      Result := ChangeRempls(SL)
    else
    if (Key = UpperCase(__KeySetAutoRaise)) or (Key = UpperCase(__KeyAutoRaise)) then
      Result := SetAutoRaise(SL)
    else
    if Key = UpperCase(__KeyChgAutoRaise) then
      Result := ChgAutoRaise(SL)
    else
    if Key = UpperCase(__KeyCivLevel) then
      Result := SetCivLevel(SL)
    else
    if (Key = UpperCase(__KeyRegionStyle)) or (Key = UpperCase(__KeyCityStyle)) then
      Result := SetRegionStyle(SL)
    else
    if Key = UpperCase(__KeySetFacStyle) then
      Result := SetFacStyle(SL)
    else
    if (Key = UpperCase(__KeyTransport))or(Key = UpperCase(__KeySetTransport)) then
      Result := SetTransport(SL)
    else
    if Key = UpperCase(__KeyVP) then
      Result := SetVP(SL)
    else
    if Key = UpperCase(__KeyBlockState) then
      Result := SetBlockState(SL)
    else
    if (Key = UpperCase(__KeyColonialPen)) or (Key = UpperCase(__KeySetColonialPen)) then
      Result := SetColonialPen(SL)
    else
    if Key = UpperCase(__KeyChgColonialPen) then
      Result := ChangeColonialPen(SL)
    else
    if Key = UpperCase(__KeySetColStatusArea) then
      Result := SetColonialStatusArea(SL)
    else
    if Key = UpperCase(__KeyDevelopment) then
      Result := SetDevelopment(SL)
    else
    if Key = UpperCase(__KeyPlanCityName) then
      Result := SetPlannedCityName(SL)
    else
    if Key = UpperCase(__KeySetPopTotalRaw) then
      Result := SetPopTotalRaw(SL)
    else
    if Key = UpperCase(__KeyPopMilitancy) then
      Result := SetPopMilitPerc(SL)
    else
    if Key = UpperCase(__KeyChgPopMilitancy) then
      Result := ChgPopMilitancy(SL)
    else
    if Key = UpperCase(__KeyChgPopMilitancyEx) then
      Result := ChgPopMilitancyEx(SL)
    else
    if Key = UpperCase(__KeyPopContentPerc) then
      Result := SetPopContentPerc(SL)
    else
    if Key = UpperCase(__KeyChgPopContent) then
      Result := ChgPopContent(SL)
    else
    if (Key = UpperCase(__KeyChgPopContentEx)) or (Key = UpperCase(__KeyChgPopContentEx2)) then 
      Result := ChgPopContentmentEx(SL)
    else
    if Key = UpperCase(__KeyConvertSocClass) then
      Result := ConvertSocClass(SL)
    else
    if Key = UpperCase(__KeyChgPopDataByEthnic) then
      Result := ChgPopDataByEthnic(SL)
    else
    if Key = UpperCase(__KeyChgPopDataByReligion) then
      Result := ChgPopDataByReligion(SL)
    else
    if Key = UpperCase(__KeySetAristosPercRaw) then
      Result := SetAristosPercRaw(SL)
    else
    if Key = UpperCase(__KeyChgAristosCount) then
      Result := ChgAristosCount(SL)
    else
    if Key = UpperCase(__KeySetUpperPercRaw) then
      Result := SetUpperPercRaw(SL)
    else
    if Key = UpperCase(__KeyChgUpperCount) then
      Result := ChgUpperCount(SL)
    else
    if Key = UpperCase(__KeySetMiddlePercRaw) then
      Result := SetMiddlePercRaw(SL)
    else
    if Key = UpperCase(__KeyChgMiddleCount) then
      Result := ChgMiddleCount(SL)
    else
    if Key = UpperCase(__KeySetWorkersPercRaw) then
      Result := SetWorkersPercRaw(SL)
    else
    if Key = UpperCase(__KeyChgWorkersCount) then
      Result := ChgWorkersCount(SL)
    else
    if Key = UpperCase(__KeySetPeasantsPercRaw) then
      Result := SetPeasantsPercRaw(SL)
    else
    if Key = UpperCase(__KeyChgPeasantsCount) then
      Result := ChgPeasantsCount(SL)
    else
    if Key = UpperCase(__KeySetServilesPercRaw) then
      Result := SetServilesPercRaw(SL)
    else
    if Key = UpperCase(__KeyChgServilesCount) then
      Result := ChgServilesCount(SL)
    else
    if Key = UpperCase(__KeyPopEducation) then
      Result := SetPopEducPerc(SL)
    else
    if Key = UpperCase(__KeyChgPopEducation) then
      Result := ChgPopEducation(SL)
    else
    if Key = UpperCase(__KeyChgPopEducationEx) then
      Result := ChgPopEducationEx(SL)
    else
    if Key = UpperCase(__KeyChgPopulation) then
      Result := ChgPopulation(SL)
    else
    if Key = UpperCase(__KeyChgReligions) then
      Result := ChgPopReligions(SL)
    else
    if Key = UpperCase(__KeyChgEthnics) then
      Result := ChgPopEthnics(SL)
    else
    if Key = UpperCase(__KeyChgResSites) then
      Result := ChgResources_Sites(SL)
    else
    if Key = UpperCase(__KeyDiscoverResSites) then
      Result := DiscoverResources_Sites(SL)
    else
    if Key = UpperCase(__KeyUnknownRegions) then
      Result := SetUnknownRegions(SL)
    else
    if Key = UpperCase(__KeyKnownRegions) then
      Result := SetKnownRegions(SL)
    else
    if Key = UpperCase(__KeyRegionsAllKnown) then
      Result := SetRegionsAllKnown(SL)
    else
    if Key = UpperCase(__KeySetAllRgnLegitOwn) then
      Result := SetAllRgnLegitOwn(SL)
    else
    if Key = UpperCase(__KeySetLegitOwner) then
      Result := SetLegitOwner(SL)
    else
    if Key = UpperCase(__KeyAutoRaiseOn) then
      Result := AutoRaiseOn(SL)
    else
    if Key = UpperCase(__KeyNumCreate) then
      Result := SetNumCreate(SL)
    else
    if (Key = UpperCase(__KeySetEntrench)) or (Key = UpperCase(__KeyEntranch)) then
      Result := SetEntranch(SL)
    else
    if Key = UpperCase(__KeyInCS) then
      Result := SetInCS(SL)
    else
    if Key = UpperCase(__KeyFixType) then
      Result := SetFixType(SL)
    else
    if Key = UpperCase(__KeyPosture) then
      Result := SetPosture(SL)
    else
    if Key = UpperCase(__KeyMapOffSets) then
      Result := SetMapOffSets(SL)
    else
    if Key = UpperCase(__KeyMapBounds) then
      Result := SetMapBounds(SL)
    else
    if Key = UpperCase(__KeyScenTurns) then
      Result := SetScenTurns(SL)
    else
    if Key = UpperCase(__KeyScenStart) then
      Result := SetScenStart(SL)
    else
    if Key = UpperCase(__KeyScenEnd) then
      Result := SetScenEnd(SL)
    else
    if Key = UpperCase(__KeyChgScenEnd) then
      Result := ChgScenEnd(SL)
    else
    if Key = UpperCase(__KeyForceDateTo) then
      Result := ForceDateTo(SL)
    else
    if Key = UpperCase(__KeyTurnsPerYear) then
      Result := TurnsPerYear(SL)
    else
    if Key = UpperCase(__KeyAddToRegList) then
      Result := AddToRegList(SL)
    else
    if Key = UpperCase(__KeyAddToRegListFromAreas) then
      Result := AddToRegListFromAreas(SL)
    else
    if Key = UpperCase(__KeyChgEvtOccurs) then
      Result := ChgEventOccurs(SL)
    else
    if Key = UpperCase(__KeySetEvtOccurs) then
      Result := SetEventOccurs(SL)
    else
    if Key = UpperCase(__KeyChgGroupFixing) then
      Result := ChangeGroupFixing(SL)
    else
    if (Key = UpperCase(__KeySetFixing)) or (Key = UpperCase(__KeyChangeFixing)) then  // change obsolte mais support
      Result :=SetFixing(SL)
    else
    if Key = UpperCase(__KeyMoveUnit) then
      Result := MoveUnit(SL)
    else
    if Key = UpperCase(__KeyOnAttack) then
      Result := OnAttack(SL)
    else
    if Key = UpperCase(__KeySetWinLevel) then
      Result := SetWinLevel(SL)
    else
    if (Key = UpperCase(__KeyChgVPCount)) or (Key = UpperCase(__KeyChgVPCount2)) then
      Result := ChgVPCount(SL)
    else
    if Key = UpperCase(__KeySetSudDeath) then
      Result := SetSudDeath(SL)
    else
    if Key = UpperCase(__KeySetPettyWar) then
      Result := SetPettyWar(SL)
    else
    if Key = UpperCase(__KeySetFacPassive) then
      Result := SetFacPassive(SL)
    else
    if Key = UpperCase(__KeySetFacUnSelect) then
      Result := SetFacUnSelect(SL)
    else
    if Key = UpperCase(__KeyIsTutorialScen) then
      Result := SetIsTutorial(SL)
    else
    if Key = UpperCase(__KeyLoyaltyRule) then
      Result := SetLoyaltyRule(SL)
    else
    if (Key = UpperCase(__KeyObjectives)) or (Key = UpperCase(__KeySetObjectives)) then
      Result := SetObjectives(SL)
    else
    if Key = UpperCase(__KeyChgObjective) then
      Result := ChgObjective(SL)
    else
    if Key = UpperCase(__KeyWinParams) then
      Result := SetWinParams(SL)
    else
    if Key = UpperCase(__KeyWinParamsVP) then
      Result := SetWinParamsVP(SL)
    else
    if (Key = UpperCase(__KeyLoseParams)) or (Key = UpperCase(__KeyLooseParams)) then
      Result := SetLoseParams(SL)
    else
    if (Key = UpperCase(__KeyLoseParamsVP)) or (Key = UpperCase(__KeyLooseParamsVP)) then
      Result := SetLosePrmsVP(SL)
    else
    if Key = UpperCase(__KeyScenVicCond) then
      Result := SetScenVicCond(SL)
    else
    if Key = UpperCase(__KeyScenText) then
      Result := SetScenText(SL)
    else
    if Key = UpperCase(__KeySide1) then
      Result := SetSide1(SL)
    else
    if Key = UpperCase(__KeySide2) then
      Result := SetSide2(SL)
    else
    if Key = UpperCase(__KeyScenBGrnd1) then
      Result := ScenBGround1(SL)
    else
    if Key = UpperCase(__KeyScenBGrnd2) then
      Result := ScenBGround2(SL)
    else
    if Key = UpperCase(__KeyScenDuration) then
      Result := ScenDuration(SL)
    else
    if Key = UpperCase(__KeyScenFavored) then
      Result := ScenFavored(SL)
    else
    if Key = UpperCase(__KeyScenImageBox) then
      Result := ScenImageBox(SL)
    else
    if Key = UpperCase(__KeyAddSubFacRgn) then
      Result := AddSubFacRgn(SL)
    else
    if Key = UpperCase(__KeyRemSubFacRgn) then
      Result := RemSubFacRgn(SL)
    else
    if Key = UpperCase(__KeyLoyFacCoeff) then
      Result := AddLoyFacCoeff(SL)
    else
    if Key = UpperCase(__KeyChgeLoyaltyFac) then
      Result := ChangeLoyaltyFac(SL)
    else
    if Key = UpperCase(__KeyChgeLoyalty) then
      Result := ChangeLoyalty(SL)
    else
    if Key = UpperCase(__KeyChangeSupply) then
      Result := ChangeSupply(SL)
    else
    if Key = UpperCase(__KeyChgeAseProd) then
      Result := ChangeAssetsProd(SL)
    else
    if Key = UpperCase(__KeyChgeResStock) then
      Result := ChangeResStock(SL)
    else
    if Key = UpperCase(__KeyChgeIntrProd) then
      Result := ChangeIntrProd(SL)
    else
    if Key = UpperCase(__KeyChgeResDiscov) then
      Result := ChangeResDiscovery(SL)
    else
    if Key = UpperCase(__KeyChangeRailPool) then
      Result := ChangeRailPool(SL)
    else
    if Key = UpperCase(__KeyChangeRiverPool) then
      Result := ChangeRiverPool(SL)
    else
    if Key = UpperCase(__KeyChangeUnitPool) then
      Result := ChangeUnitPool(SL)
    else
    if Key = UpperCase(__KeyChangeUnitType) then
      Result := ChangeUnitType(SL)
    else
    if Key = UpperCase(__KeyChangeUnitTypeByGen) then
      Result := ChangeUnitTypeByGen(SL)
    else
    if Key = UpperCase(__KeyObsoleteMdlGen) then
      Result := ObsoleteModelGen(SL)
    else
    if Key = UpperCase(__KeyObsoleteStrGen) then
      Result := ObsoleteStrucGen(SL)
    else
    if Key = UpperCase(__KeyChangeStrucPool) then
      Result := ChangeStrucPool(SL)
    else
    if Key = UpperCase(__KeyChangeStrucType) then
      Result := ChangeStrucType(SL)
    else
    if Key = UpperCase(__KeyChgStrucLevel) then
      Result := ChangeStrucLevel(SL)
    else
    if Key = UpperCase(__KeyRemoveStructure) then
      Result := RemoveStructure(SL)
    else
    if Key = UpperCase(__KeySetStrucCustomNames) then
      Result := SetStrucCustNames(SL)
    else
    if (Key = UpperCase(__KeyChgeActorPool)) or(Key = UpperCase(__KeySetActorPool)) then
      Result := SetActorPool(SL)
    else
    if Key = UpperCase(__KeyChgConscript) then
      Result := ChangeConscripts(SL)
    else
    if Key = UpperCase(__KeyChgFacRelationships) then
      Result := ChangeFactionRel(SL)
    else
    if Key = UpperCase(__KeySetFacRelationships) then
      Result := SetFactionRel(SL)
    else
    if Key = UpperCase(__KeyChgWSUPool) then
      Result := ChangeWSUPool(SL)
    else
    if (Key = UpperCase(__KeyChgMoneyPool)) or (Key = UpperCase(__KeyChgMoneyPool2)) then
      Result := ChangeMoneyPool(SL)
    else
    if Key = UpperCase(__KeyChgFacMorale) then
      Result := ChangeMorale(SL)
    else
    if Key = UpperCase(__KeyChgFacEngagementPts) then
      Result := ChangeEngagementPts(SL)
    else
    if Key = UpperCase(__KeyChgPriceInc) then
      Result := ChangePriceInc(SL)
    else
    if Key = UpperCase(__KeySetTrsAssets) then
      Result := SetTRSAssetsInfos(SL)
    else
    if Key = UpperCase(__KeySetRevolters) then
      Result := SetRevoltersInfos(SL)
    else
    if Key = UpperCase(__KeySetPopUprising) then
      Result := SetUprisingInfos(SL)
    else
    if Key = UpperCase(__KeySetScorchEarth) then
      Result := SetScorchEarthInfos(SL)
    else
    if Key = UpperCase(__KeySetNavalPool) then
      Result := SetNavalPools(SL)
    else
    if Key = UpperCase(__KeySetBlockades) then
      Result := SetBlockades(SL)
    else
    if Key = UpperCase(__KeySetDistantBlockades) then
      Result := SetDistBlockades(SL)
    else
    if Key = UpperCase(__KeyHarborNoBlock) then
      Result := AddHarborNoBlock(SL)
    else
    if Key = UpperCase(__KeyHarborBlockade) then
      Result := AddHarborBlockaded(SL)
    else
    if Key = UpperCase(__KeyMoraleLevels) then
      Result := SetMoraleLevels(SL)
    else
    if Key = UpperCase(__KeyChgMoraleLevels) then
      Result := ChgMoraleLevels(SL)
    else
    if (Key = UpperCase(__KeyAISetPoolRatio)) or (Key = UpperCase(__KeySetAIPoolRatios)) then
      Result := AISetPoolRatio(SL)
    else
    if Key = UpperCase(__KeySetUnitPoolInfos) then
      Result := SetUnitPoolInfos(SL)
    else
    if Key = UpperCase(__KeyChgUnitPoolInfos) then
      Result := ChgUnitPoolInfos(SL)
    else
    if Key = UpperCase(__KeyAISetAggro)  then
      Result := AISetAggro(SL)
    else
    if Key = UpperCase(__KeyAIChgAggro) then
      Result := AIChangeAggro(SL)
    else
    if Key = UpperCase(__KeyAISetTheaPrev)  then
      Result := AISetTheaPrev(SL)
    else
    if Key = UpperCase(__KeyAIChgTheaPrev) then
      Result := AIChangeTheaPrev(SL)
    else
    if Key = UpperCase(__KeyAISetParams) then
      Result := AISetParams(SL)
    else
    if Key = UpperCase(__KeyAISetDiploRelBias) then
      Result := AISetDiploRelBias(SL)
    else
    if Key = UpperCase(__KeyAIChgDiploRelBias) then
      Result := AIChangeDiploRelBias(SL)
    else
    if Key = UpperCase(__KeyAISetDiploItemBias) then
      Result := AISetDiploItemBias(SL)
    else
    if Key = UpperCase(__KeyAIChgLocalIntrst) then
      Result := AIChgeLocalIntrst(SL)
    else
    if Key = UpperCase(__KeyAISetLocalIntrst) then
      Result := AISetLocalIntrst(SL)
    else
    if Key = UpperCase(__KeyAIChgAreaLocalIntrst) then
      Result := AIChgeAreaLocalInt(SL)
    else
    if Key = UpperCase(__KeyAISetAreaLocalIntrst) then
      Result := AISetAreaLocalInt(SL)
    else
    if Key = UpperCase(__KeyAISetLandBuild) then
      Result := AISetLandBuild(SL)
    else
    if Key = UpperCase(__KeyAISetNavBuild) then
      Result := AISetNavBuild(SL)
    else
    if Key = UpperCase(__KeySetFacMaxEntrench) then
      Result := SetFacMaxEntrench(SL)
    else
    if Key = UpperCase(__KeyAddXpGroup) then
      Result := AddXPLevelToGroup(SL)
    else
    if Key = UpperCase(__KeyAddXpUnit) then
      Result := AddXPLevelToUnit(SL)
    else
    if Key = UpperCase(__KeyChgGroupHealth) then
      Result := ChgGroupHealth(SL)
    else
    if Key = UpperCase(__KeyChgUnitHealth) then
      Result := ChgUnitHealth(SL)
    else
    if Key = UpperCase(__KeySetBuildFlag) then
      Result := SetBuildFlag(SL)
    else
    if Key = UpperCase(__KeyChgGroupCoh) then
      Result := ChgGroupCohesion(SL)
    else
    if Key = UpperCase(__KeyChgUnitCoh) then
      Result := ChgUnitCohesion(SL)
    else
    if (Key = UpperCase(__KeyAltercuSubUnit)) or (Key = UpperCase(__KeyAltercuSubUnits)) then
      Result := AlterCuSubUnit(SL)
    else
    if Key = UpperCase(__KeySetLdrActivity) then
      Result := SetLeaderActivity(SL)
    else
    if Key = UpperCase(__KeySetCombiUnits) then
      Result := SetCombiUnits(SL)
    else
    if Key = UpperCase(__KeySetFormedCmdMax) then
      Result := SetFormedCmdMax(SL)
    else
    if (Key = UpperCase(__KeyChgOption)) or (Key = UpperCase(__KeyChangeOption)) then
      Result := ChangeOption(SL)
    else
    if Key = UpperCase(__KeyAddGarrison) then
      Result := AddGarrison(SL)
    else
    if Key = UpperCase(__KeyClearGarrisons) then
      Result := ClearGarrisons(SL)
    else
    if Key = UpperCase(__KeyChgOneTurnMod) then
      Result := ChangeOneTurnMod(SL)
    else
    if Key = UpperCase(__KeySetFactionAtt) then
      Result := SetFactionAttribute(SL)
    else
    if Key = UpperCase(__KeySetTechDiscovery) then
      Result := SetTechDiscovery(SL)
    else
    if Key = UpperCase(__KeyAddNationalRgns) then
      Result := AddNationalRegions(SL)
    else
    if Key = UpperCase(__KeyRemNationalRgns) then
      Result := RemNationalRegions(SL)
    else
    if Key = UpperCase(__KeyAddClaimedRgns) then
      Result := AddClaimedRegions(SL)
    else
    if Key = UpperCase(__KeyRemClaimedRgns) then
      Result := RemClaimedRegions(SL)
    else
    if (Key = UpperCase(__KeyAddDeJureRgns)) or (Key = UpperCase(__KeySetDeJure)) then
      Result := AddDeJureRegions(SL)
    else
    if Key = UpperCase(__KeyRemDeJureRgns) then
      Result := RemDeJureRegions(SL)
    else
    if Key = UpperCase(__KeySetSupplyMod) then
      Result := SetSupplyMod(SL)
    else
    if Key = UpperCase(__KeyCreateCrisis) then
      Result := CreateCrisis(SL)
    else
    if Key = UpperCase(__KeyChgSellTransac) then
      Result := ChangeSellTransac(SL)
    else
    if Key = UpperCase(__KeyChgBuyTransac) then
      Result := ChangeBuyTransac(SL)
    else
    if Key = UpperCase(__KeyAddFacModifiers) then
      Result := AddFacModifiers(SL)
    else
    if Key = UpperCase(__KeyRemFacModifiers) then
      Result := RemFacModifiers(SL)
    else
    if Key = UpperCase(__KeyEnableDiploItem) then
      Result := EnableDiploItem(SL)
    else
    if Key = UpperCase(__KeyMakeTechKnown) then
      Result := MakeTechKnown(SL)
    else
    if Key = UpperCase(__KeyAddRulerToFac) then
      Result := AddRulerToFac(SL)
    else
    if Key = UpperCase(__KeyChgLeadersPool) then
      Result := ChgLeadersPool(SL)
    else
    if Key = UpperCase(__KeyAddProvLeaders) then
      Result := AddProvLeaders(SL)
    else
    if Key = UpperCase(__KeyRemProvLeaders) then
      Result := RemProvLeaders(SL)
    else
    if Key = UpperCase(__KeySetAllFacTaxes) then
      Result := SetAllFacTaxes(SL)
    else
    if Key = UpperCase(__KeySetTaxes) then
      Result := SetTaxes(SL)
    else
    if Key = UpperCase(__KeyAddDynGraphic) then
      Result := AddDynamicGraphic(SL)
    else
    if Key = UpperCase(__KeyRemDynGraphic) then
      Result := RemDynamicGraphic(SL)
    else
    if Key = UpperCase(__KeyChgFacDisplayTag) then
      Result := ChangeFacDisplayTag(SL)
    else
    if Key = UpperCase(__KeyChgGameRules) then
      Result := ChangeGameRules(SL)
    else
//    if Key = UpperCase(__KeyAddDIToAllFac) then
//      Result := AddDIToAllFac(SL)
//    else
    if Key = UpperCase(__KeyAddDiploItem) then
      Result := AddDiploItem(SL)
    else
    if Key = UpperCase(__KeyRemDiploItem) then
      Result := RemDiploItem(SL)
    else
    if Key = UpperCase(__KeySettleWhitePeace) then
      Result := SettleWhitePeace(SL)
    else
    if Key = UpperCase(__KeyChangePrisoners) then
      Result := ChangePrisoners(SL)
    else
    if Key = UpperCase(__KeyChangeRgnOwner) then
      Result := ChangeRgnOwner(SL)
    else
    if Key = UpperCase(__KeyAbsorbFaction) then
      Result := AbsorbFaction(SL)
    else
    if Key = UpperCase(__KeyChgSOIVals) then
      Result := ChangeSOIs(SL)
    else
    if Key = UpperCase(__KeySetSOIVals) then
      Result := SetSOIs(SL)
    else
    if Key = UpperCase(__KeySetAdvRevolt) then
      Result := SetAdvRevolt(SL)
    else
    if Key = UpperCase(__KeySetAdvRevoltFP) then
      Result := SetAdvRevoltFP(SL)
    else
    if Key = UpperCase(__KeySetAdvRevoltLoc) then
      Result := SetAdvRevoltLoc(SL)
    else
    if Key = UpperCase(__KeyChgRgnDecisionFP) then
      Result := ChgRgnDecisionFP(SL)
    else
    if Key = UpperCase(__KeyAddRgnDecisionAct) then
      Result := AddRgnDecisionAct(SL)
    else
    if Key = UpperCase(__KeyRemRgnDecisionAct) then
      Result := RemRgnDecisionAct(SL)
    else
    if Key = UpperCase(__KeyAddRgnDecisionExcl) then
      Result := AddRgnDecisionExcl(SL)
    else
    if Key = UpperCase(__KeyRemRgnDecisionExcl) then
      Result := RemRgnDecisionExcl(SL)
    else
    if (Key = UpperCase(__KeyColStatus)) or (Key = UpperCase(__KeySetColStatus)) then
      Result := SetColonialStatus(SL)
    else
    if Key = UpperCase(__KeyTriggerTribalRev) then
      Result := TriggerAdvRev(SL)
    else
    if Key = UpperCase(__KeySetFacDisplayNames) then
      Result := SetFacDisplayNames(SL)
    else
    if Key = UpperCase(__KeySetFacDisplayFrcNames) then
      Result := SetFacDisplayFrcNames(SL)
    else
    if Key = UpperCase(__KeySetFacDisplayColors) then
      Result := SetFacDisplayColors(SL)
    else
    if Key = UpperCase(__KeyMobilize) then
      Result := Mobilize(SL)
    else
    if Key = UpperCase(__KeyDemobilize) then
      Result := Demobilize(SL)
    else
    if Key = UpperCase(__KeySetPillage) then
      Result := SetPillage(SL)
    else
    if Key = UpperCase(__KeySetNextElection) then
      Result := SetNextElection(SL)
    else
    if Key = UpperCase(__KeySetRulerItemActiveState) then
      Result := SetRulerItemActiveState(SL)
    else
    if Key = UpperCase(__KeySetFAOverride) then
      Result := SetFAOverride(SL)
    else
    if Key = UpperCase(__KeyRemFAOverride) then
      Result := RemFAOverride(SL)
    else
    if Key = UpperCase(__KeyChangeFortBreach) then
      Result := ChangeFortBreach(SL)
    else
    if Key = UpperCase(__KeyAddMapAlea) then
      Result := AddMapAlea(SL)
    else
    if Key = UpperCase(__KeyRemMapAlea) then
      Result := RemMapAlea(SL)
    else
    if Key = UpperCase(__KeyAIAddLeaderAgent) then
      Result := AIAddLeaderAgent(SL)
    else
    if Key = UpperCase(__KeyPayOption) then
      Result := PayOption(SL)
    else
    if Key = UpperCase(__KeyTakeOverUnits) then
      Result := TakeOverUnits(SL)
    else
    if Key = UpperCase(__KeyMoveStruc) then
      Result := MoveStruc(SL)

    // These Evals can also manipulate registry, so we must add them in the Actions block
    else
    if Key = UpperCase(__KeyEvalStrucProd) then
      Result := EvalStrucProduction(SL)
    else
    if Key = UpperCase(__KeyEvalStrucCount) then
      Result := EvalStrucCount(SL)

    // Registry manipulation
    else
    if Key = UpperCase(__KeyRegActivate) then
      Result := Reg_Activate(SL)
    else
    if Key = UpperCase(__KeyRegDeActivate) then
      Result := Reg_DeActivate(SL)
    else
    if Key = UpperCase(__KeyRegInit) then
      Result := Reg_Init(SL)
    else
    if Key = UpperCase(__KeyRegSum) then
      Result := Reg_Sum(SL)
    else
    if Key = UpperCase(__KeyRegSumNum) then
      Result := Reg_SumNum(SL)
    else
    if Key = UpperCase(__KeyRegEval) then
      Result := Reg_Eval(SL)
    else
    if Key = UpperCase(__KeyRegEvalNum) then
      Result := Reg_EvalNum(SL)

  finally
    SL.Free;
  end;
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ADD ERROR
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
procedure TScriptEngine.Log(Msgs : array of variant);
var Str : string;
    i : integer;
begin
  for i := 0 to High(Msgs) do
    Str := Str + ' ' + VarToStr(Msgs[i]);
  slScriptLog.Add('Line ' + ITS(cuLine) + ': ' + Str);
  if Pos('=>', Str) > 0 then // immediate writing if error (temporary, we will recreate the file at the end)
    slScriptLog.SaveToFile(ScriptReportFile_ + ITS(GameSim.cuTurn) + '.txt');
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ABORT
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
procedure TScriptEngine.Abort;
begin
  ResetActionsVariables;
  WorkingVars.Clear;
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// EXTRACT PARAM
// Si le paramtre ParamStr existe dans la SL, le renvoi avec ses arguments
// StratTown <= 6 >= 2 est un paramtre avec 4 arguments
// Nb: les arguments ne peuvent avoir d'alias (=> sparateurs espace) SAUF si
// rfrenc dans CommandCanHaveAliasWithSpaceSeparator
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function  TScriptEngine.ExtractParam(ParamStr : string; Params : TStringList) : string;
var i : integer;
begin
  Result := '';
  for i := 0 to Params.Count -1 do
  begin
    if Pos(UpperCase(ParamStr), UpperCase(Trim(Params[i]))) > 0 then
    begin
      Result := Trim(Params[i]);
      Exit;
    end;
  end;
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// EXTRACT ARGS
// Extrait les arguments d'un paramtre dans le tableau Args
// StratTown <= 6 >= 2 place 4 args dans le tableau: <=, 6, >= et 2
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
procedure  TScriptEngine.ExtractArgs(Param : string);
var Count, i : integer;
begin
  for i := 0 to _sctNumOfArgs-1 do Args[i] := '';
  Count := RxEW_GetWordCount(Param, ' ');
  for i := 2 to Min(_sctNumOfArgs+1, Count) do
    Args[i-2] := ExString(Param, i, ' ');
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// GET EXPOSED VAR INDEX
// -1: n'est pas une esv
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.GetExposedVarIndex(Str : String) : integer;
var Idx : integer;
begin
  Result := -1;
  Idx := ExtractPosNumeral(Str);
  if (Idx = -1)or(Idx > High(esvIntVar)) then Exit;
  if Pos(_esvIntVarName, Str) <> 1 then Exit;
  Result := Idx;
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// IS EXPOSED VAR
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.IsExposedVar(Str : String) : boolean;
begin
  Result := GetExposedVarIndex(Str) <> -1;
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// GET EXPOSED VAR VALUE
// si Str est une esv (exposed script variable), sa valeur est renvoye
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.GetExposedVarValue(Str : String) : integer;
var Idx : integer;
begin
  Result := Low(integer);
  Idx := GetExposedVarIndex(Str);
  if Idx = -1 then Exit;
  Result := esvIntVar[Idx];
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// STORE TO VAR
// si Str est une esv (exposed script variable), elle est valorise
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.StoreToExposedVar(Str : String; Value : integer) : boolean;
var Idx : integer;
begin
  Result := False;
  Idx := GetExposedVarIndex(Str);
  if Idx = -1 then Exit;
  esvIntVar[Idx] := Value;
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// REGISTRY ACTIVATE
// This does not reset the value of the registry, it only specify the active registry
// RegActivate = RegIdx
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.Reg_Activate(Params : TStringList) : integer;
begin
  Result := 0;
  if not (Params.Count in [1]) then
  begin
    Inc(Result);
    Log(['=> RegActivate command, need to have the registry number]']);
    Exit;
  end;

  if STI(Params[0]) > High(Registry) then
  begin
    Inc(Result);
    Log(['=> RegActivate command, Too high registry number, maximum is', High(Registry)]);
    Exit;
  end;
  ActiveReg := STI(Params[0]);

  Log(['RegActivate', ActiveReg, 'Value', Registry[ActiveReg]]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// REGISTRY DE ACTIVATE
// Deactivate the active registry, allowing some commands to return to their
// usage without registry
// RegDeActivate = RegIdx
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.Reg_DeActivate(Params : TStringList) : integer;
begin
  Result := 0;
  if not (Params.Count <> 0) then
  begin
    Inc(Result);
    Log(['=> RegDeActivate command, No parameter needed.']);
    Exit;
  end;
  ActiveReg := _noUID;

  Log(['RegDeactivate']);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// REGISTRY INIT
// This reset the value of the registry to 0 and activate it (becomes the active registry)
// RegInit = RegIdx
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.Reg_Init(Params : TStringList) : integer;
begin
  Result := 0;
  if not (Params.Count in [1]) then
  begin
    Inc(Result);
    Log(['=> RegInit command, need to have the registry number]']);
    Exit;
  end;

  if STI(Params[0]) > High(Registry) then
  begin
    Inc(Result);
    Log(['=> RegInit command, Too high registry number, maximum is', High(Registry)]);
    Exit;
  end;
  ActiveReg := STI(Params[0]);
  Registry[ActiveReg] := 0;

  Log(['RegInit', ActiveReg]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// REGISTRY SUM
// Will sum all specified registry and add this value into the current value of the
// active registry
// Syntax RegSum = Registry Idx1 Idx2 Idx3 ..[;Best NumBest][;Worst NumWorse]
// RegSum = Registry 1 2 3 :Will sum registry 1, 2 and 3 into the active registry
// RegSum = Registry 1 2 4 Best 2 :Will take the 2 highest values of registries 1, 2 & 4  and add these into the active registry
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.Reg_Sum(Params : TStringList) : integer;
var i : integer;
    Param : string;
    Best, Worst, All : boolean;
    NumAdd : integer;
    Registries : string;
    RegIdx : array[0.._sctNumOfArgs-1] of integer;
    IL : TIntList;
begin
  Result := 0;
  NumAdd := 0;

  if ActiveReg <= 0 then // registry 0 is never used even if it exists
  begin
    Inc(Result);
    Log(['=> RegSum, No Active registry to hold the result.']);
    Exit;
  end;

  Param :=  ExtractParam('BEST', Params);
  Best := Param <> '';
  if Best then
  begin
    ExtractArgs(Param);
    if Args[0] = '' then
    begin
      Inc(Result);
      Log(['=> RegSum, Best specified but now how many.']);
      Exit;
    end;
    NumAdd := STI(Args[0]);
  end;

  Param :=  ExtractParam('WORST', Params);
  Worst := Param <> '';
  if Worst then
  begin
    ExtractArgs(Param);
    if Args[0] = '' then
    begin
      Inc(Result);
      Log(['=> RegSum, Worst specified but now how many.']);
      Exit;
    end;
    NumAdd := STI(Args[0]);
  end;

  All := (not Best) and (not Worst);

  for i := 0 to High(RegIdx) do
    RegIdx[i] := 0;
  Registries := ExtractParam('REGISTRY', Params);
  ExtractArgs(Registries);

  // get the registries to use
  for i := 0 to High(Args) do
  begin
    if Args[i] = '' then Continue;
    RegIdx[i] := STI(Args[i]);
  end;

  IL := TIntList.Create;

  for i := 0 to High(RegIdx) do
  begin
    if RegIdx[i] = 0 then Break; // registry 0 is never used
    if RegIdx[i] = ActiveReg then
    begin
      Inc(Result);
      Log(['=> RegSum, Can''t sum Active registry with self.']);
      Exit;
    end;

    // Add the value of each selected registry
    IL.Add(Registry[RegIdx[i]]);
  end;

  IL.Sort();

  // operate
  if All then
  begin
    for i := 0 to IL.Count-1 do
      Inc(Registry[ActiveReg], IL[i]);
  end
  else
  if Best then
  begin
    for i := 0 to Min(NumAdd-1, IL.Count-1) do
      Inc(Registry[ActiveReg], IL[i]);
  end
  else
  if Worst then
  begin
    for i := IL.Count-1 downto Max(IL.Count-NumAdd, 0) do
      Inc(Registry[ActiveReg], IL[i]);
  end;

  Log(['RegSum', IIF(All, 'All', ''), IIF(Best, 'Best ' + ITS(NumAdd), ''), IIF(Worst, 'Worst ' + ITS(NumAdd), ''),
  'Into Registry', ActiveReg, 'end value', Registry[ActiveReg]]);
  
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// REGISTRY REMOVE

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.Reg_Rem(Params : TStringList) : integer;
var Val, RegIdx1, RegIdx2 : integer;
begin
  Result := 0;

  if (not (Params.Count in [1,3])) then
  begin
    Inc(Result);
    Log(['=> Registry Remove Command, syntax is: RegRem = <RegistryIdx1> ; <Minus> <RegistryIdx2> or RegRem = <RegistryIdx>']);
  end;
  if Result > 0 then Exit;

  if Params.Count = 1 then
  begin
    Val := STI(Params[0]);
    Dec(Registry[ActiveReg], Val);
    Log(['RegRem', 'Removing from Registry', ActiveReg, 'the value', Val, 'end value', Registry[ActiveReg]]);
  end
  else
  begin
    RegIdx1 := STI(Params[0]);
    RegIdx2 := STI(Params[2]);
    Dec(Registry[RegIdx1], Registry[RegIdx2]);
    Log(['RegRem', 'Removing from Registry', RegIdx1, 'the value', Registry[RegIdx2], 'end value', Registry[RegIdx1]]);
  end;

end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// REGISTRY SUM NUM
// Add a numeric into a registry
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.Reg_SumNum(Params : TStringList) : integer;
var Reg, Val : integer;
begin
  Result := 0;

  if (Params.Count <> 2) then
  begin
    Inc(Result);
    Log(['=> Registry Sum Numeric Command, syntax is: RegSumNum = <Registry> ; <Value>']);
  end;
  if Result > 0 then Exit;

  Reg := STI(Params[0]);
  Val := STI(Params[1]);

  Inc(Registry[Reg], Val);

  Log(['RegSumNum', 'Into Registry', Reg, 'adding', Val, 'to it, end value', Registry[Reg]]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// REGISTRY EVALUATE
// Evaluate 2 registry against each other
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.Reg_Eval(Params : TStringList) : integer;
var Operator : string;
    Reg1, Reg2 : integer;
begin
  Result := 0;

  if (Params.Count <> 3) then
  begin
    Inc(Result);
    Log(['=> Registry Evaluate Command, syntax is: RegEval = <Registry1> ; <Operator (one of: < <= = >= >) >; <Registry2>']);
  end;
  if Result > 0 then Exit;

  Reg1 := STI(Params[0]);
  Reg2 := STI(Params[2]);
  Operator := Params[1];

  if Operator = '<' then
  begin
    if (Registry[Reg1] < Registry[Reg2]) = False then Conditions := condFalse;
  end
  else
  if Operator = '<=' then
  begin
    if (Registry[Reg1] <= Registry[Reg2]) = False then Conditions := condFalse;
  end
  else
  if Operator = '>' then
  begin
    if (Registry[Reg1] > Registry[Reg2]) = False then Conditions := condFalse;
  end
  else
  if Operator = '>=' then
  begin
    if (Registry[Reg1] >= Registry[Reg2]) = False then Conditions := condFalse;
  end
  else
  if Operator = '=' then
  begin
    if (Registry[Reg1] = Registry[Reg2]) = False then Conditions := condFalse;
  end;

  Log(['RegEval:', 'Registry', Reg1, 'Value:', Registry[Reg1],
  Operator, 'Registry', Reg2, 'Value:', Registry[Reg2], ':',
  IIF(Conditions = condEvaluating, 'True', 'False')]);

end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// REGISTRY EVALUATE NUMERIC
// Evaluate 1 registry against a numeric
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.Reg_EvalNum(Params : TStringList) : integer;
var Value : integer;
    Operator : string;
    Reg1 : integer;
begin
  Result := 0;

  if (Params.Count <> 3) then
  begin
    Inc(Result);
    Log(['=> Registry Evaluate Numeric Command, syntax is: RegEvalNum = <Registry1 Idx> ; <Operator (one of: < <= = >= >) >; <Value>']);
  end;
  if Result > 0 then Exit;

  Reg1 := STI(Params[0]);
  Value := STI(Params[2]);
  Operator := Params[1];

  if Operator = '<' then
  begin
    if (Registry[Reg1] < Value) = False then Conditions := condFalse;
  end
  else
  if Operator = '<=' then
  begin
    if (Registry[Reg1] <= Value) = False then Conditions := condFalse;
  end
  else
  if Operator = '>' then
  begin
    if (Registry[Reg1] > Value) = False then Conditions := condFalse;
  end
  else
  if Operator = '>=' then
  begin
    if (Registry[Reg1] >= Value) = False then Conditions := condFalse;
  end
  else
  if Operator = '=' then
  begin
    if (Registry[Reg1] = Value) = False then Conditions := condFalse;
  end;

  Log(['RegEval:', 'Registry', Reg1, 'Value:', Registry[Reg1],
  Operator, 'Value:', Value, ':',
  IIF(Conditions = condEvaluating, 'True', 'False')]);

end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ADD BOUND VAR
// Ajoute un label/variable, et y attache une slection d'lments
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.AddBoundVar(VarLabel : string; Selections : array of string) : integer;
var BL : TBindList;
    i  : integer;
begin
  BL := TBindList.Create;
  BL.VarLabel := VarLabel;
  for i := 0 to High(Selections) do
    BL.Add(Selections[i]);
  olBoundVar.Add(BL);
  Result := olBoundVar.Count -1;
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// GET SELECTIONS
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
procedure TScriptEngine.GetSelections(VarLabel : string; var Result : TStringArray);
var i,j  : integer;
    BL   : TBindList;
    Tmp  : TStringArray;
begin
  // ne pas toucher au Result entrant
  for i := 0 to olBoundVar.Count -1 do
  begin
    BL := TBindList(olBoundVar[i]);
    if BL.VarLabel = VarLabel then
    begin
      // Sortir chaque slection 'primaire'
      SetLength(Result, BL.Count);
      for j := 0 to BL.Count -1 do
      begin
        Result[j] := BL[j];
      end;
      Tmp := nil;
      // parcours chaque slection, et voir si c'est un UID o elle mme une slection
      for j := 0 to Length(Result) -1 do
      begin
        if IsNumeric(Result[j]) then Continue; // un UID
        // une variable,  dcoder
        GetSelections(Result[j], Tmp);
      end;
      // slections dcodes,  ajouter
      for j := 0 to Length(Tmp) -1 do
      begin
        SetLength(Result, Length(Result) +1);
        Result[Length(Result)] := Tmp[j];
      end;
      Exit;
    end;
  end;
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// LABEL TO UID
// Cherche le label aLabel dans olBoundVar, et retourne l'UID correspondant (1er terme de la slection)
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.LabelToUID(aLabel : string) : TUID;
var SA : TStringArray;
begin
  GetSelections(aLabel, SA);
  if Length(SA) > 0 then
    Result := STI(SA[0])
  else
    Result := _noUID;
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// KIND LABEL TO GME TYPE
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.KindLabelToGMEType(aKindLabel : string; var Kind : integer) : integer;
begin
  Kind := _gkUnknown;
  Result := 0;
  if IsNumeric(aKindLabel) then
    Kind := STI(aKindLabel)
  else
  // ancienne implmentation avant les alias
  begin
    if UpperCase(aKindLabel) = UpperCase(__prmLand)  then Kind := _gkLand
    else
    if UpperCase(aKindLabel) = UpperCase(__prmNaval) then Kind := _gkNav;
    if Kind =  _gkUnknown then
    begin
      Log(['=> SetKind command, failed to recognize this kind: ',aKindLabel]);
      Result := 1;
    end;
  end;

  if (Kind < _gkLand)  or (Kind > _gkNav) then
  begin
    Log(['=> SetKind command, failed to recognize this kind: ',aKindLabel]);
    Result := 1;
  end;
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// FACTION TAG TO FACTION
// allow not finding the faction
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.FactionTagToFaction(FactionTag : string) : TFaction;
var fac : integer;
    GS  : TGameSim;
begin
  Result := nil;
  GS := TGameSim(Owner);
  for fac := 0 to GS.Factions.Count -1 do
  begin
    if UpperCase(GS.Factions[fac].FactionDef.FactionTag) = Trim(UpperCase(FactionTag)) then
    begin
      Result := GS.Factions[fac];
      Exit;
    end;
  end;
end;

// *** DEBUT DESCRIPTEURS DE BLOCS ***

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// START EVENT
// Indique que l'on entre dans une structure vnement
// Trace les occurences initiales permises
// StartEvent GUIDTitle|Occurs|DisplayType|FlavorText|ImageID|RegionFocus|SoundID
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.StartEvent(Params : TStringList) : integer;
var ES   : TEventState;
    DisplayType : TUID;
    FlavorText : string;
    RgnUID : TUID;
    ImageID : string;
    SoundID : TUID;
    Str : string;
begin
  Result := 0;
  if not (Params.Count in [7]) then
  begin
    Inc(Result);
    Log(['=> Start event command, erroneous number of params, found:', Params.Count, 'need: 7. Format is StartEvent GUIDTitle|Occurs|DisplayType|FlavorText|ImageID|RegionFocus|SoundID. Use NULL for FlavorText, ImageID, RegionFocus, SoundID as null values']);
  end;
  if ParsedEvent <> '' then
  begin
    Inc(Result);
    Log(['=> Start event command, there is already an event being parsed:', ParsedEvent, 'continuing on it.']);
  end;
  if Result <> 0 then Exit;

  ParsedEvent := Params[0];
  Conditions := condTrue; // s'il n'y a pas de bloc conditions, vrai par dfaut, si un bloc est dclar: repassse  condEvaluating

  // on ne localise pas le titre et le texte flavor, sera fait au display suivant la langue du "viewer"
  Str := Params[4]; TFileParser.ParseForAliases(Str); Params[4] := Str;
  Str := Params[5]; TFileParser.ParseForAliases(Str); Params[5] := Str;
  Str := Params[6]; TFileParser.ParseForAliases(Str); Params[6] := Str;

  // Ce nb d'occurences de base n'est utilis que pour tre sauve dans le EventState,
  // et seulement si l'evt n'est pas encore rfrenc
  BaseOccurs  := STI(Params[1]);

  DisplayType := STI(Params[2]);
  if Params[3] = 'NULL' then FlavorText := '' else
    FlavorText  := Params[3];
  if Params[4] = 'NULL' then ImageID := '' else
    ImageID  := Params[4];
  if Params[5] = 'NULL' then RgnUID := _noUID else
    RgnUID      := STI(Params[5]);
  if Params[6] = 'NULL' then SoundID := _noUID else
    SoundID     := STI(Params[6]);

  // les autres slecteurs simples ne sont pas RAZ
  cuRegList.Clear;
  cuSUList.Clear;

  if (cuFaction <> nil)and(RgnUID <> _noUID) then
  begin
    cuRegion := cuFaction.Regions.Fetch(RgnUID);
    cuRegList.Add(cuRegion.UID);
  end
  else
  begin
    if (not IsSpecialScript) then
      cuRegion := nil;
  end;    

  if (FlavorText = '')and(DisplayType = _evtFullDisplay) then
    DisplayType := _evtLineDisplay;

  // prparer le message flavor qu'il faudra crer, si les conditions deviennent vraies
  // sera fait automatiquement dans le bloc action, si RgnUID <> _noUID autrement
  //  appeler explicitement dans les Actions
  if (DisplayType <> _evtNoDisplay)and(cuFaction <> nil) then
  begin
    PlayerMessage.Title   := ParsedEvent;
    PlayerMessage.Text    := FlavorText;
    PlayerMessage.ImageID := ImageID;
    PlayerMessage.DisplayType := DisplayType;
    PlayerMessage.RgnUID  := RgnUID;
    PlayerMessage.SoundID := SoundID;
    PlayerMessage.IsBroadCast := False;
  end
  else
  if (DisplayType <> _evtNoDisplay)and((cuFaction = nil)or IsBroadcast) then // Broadcast
  begin
    PlayerMessage.Title   := ParsedEvent;
    PlayerMessage.Text    := FlavorText;
    PlayerMessage.ImageID := ImageID;
    PlayerMessage.DisplayType := DisplayType;
    PlayerMessage.RgnUID  := RgnUID;
    PlayerMessage.SoundID := SoundID;
    PlayerMessage.IsBroadCast := True;
  end
  else // RZ du record
  begin
    PlayerMessage.Title   := '';
    PlayerMessage.Text    := '';
    PlayerMessage.ImageID := '';
    PlayerMessage.DisplayType := _evtNoDisplay;
    PlayerMessage.RgnUID  := _noUID;
    PlayerMessage.SoundID := _noUID;
    PlayerMessage.IsBroadCast := False;
  end;

  Log(['Started processing event:', ParsedEvent]);
  ES := TGameSim(Owner).EventStates.GetByName(ParsedEvent);
  if ES = nil then
    Log(['This event is not yet referenced, base number of allowed occurences', BaseOccurs])
  else
    Log(['Event already referenced, current occurences', ES.cuOccurs, 'Max allowed:', ES.MaxOccurs, IIF(ES.cuOccurs >= ES.MaxOccurs, '  >>Sleeping<<', '  <<Active>>')])
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// END EVENT
// Indique que l'on sort d'une structure vnement
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.EndEvent(Params : TStringList) : integer;
begin
  Result := 0;
  if not (Params.Count in [0]) then
  begin
    Log(['=> Notice: End event command, you don''t need to name the ended event.']);
  end;
  if ParsedEvent = '' then
  begin
    Inc(Result);
    Log(['=> End event command, no event was being parsed.']);
  end;
  if Result <> 0 then Exit;

  // l'vnement a t dclench, au augmente son compteur d'occurences
  if (Conditions = condTrue)
  or(Conditions = condEvaluating) then // quand l'evt n'a pas de bloc action (sert pour incrmenter l'occurence en tant que timer/flag)
  begin
    if (cuEventState.MaxOccurs <> _evtUnlimited)or(cuEventState.cuOccurs < _evtUnlimited){999} then
      Inc(cuEventState.cuOccurs);
  end;

  Log(['Finished processing event:', ParsedEvent]);
  Log(['------------------------------']);
  ParsedEvent := '';
  cuEventState := nil;
  ConditionsON := False;
  Conditions   := condTrue;
  ActionsON    := False;
  BaseOccurs   := 1;
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// TO CONDITIONS
// Pour entrer dans l'analyse du bloc Conditions
// On analyse de suite si le nb max d'occurence n'est pas dpass
// note: evt en sommeil = evt avec MaxOccurs  0
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.ToConditions  : integer;
begin
  Result := 0;
  if ParsedEvent = '' then
  begin
    Inc(Result);
    Log(['=> Conditions command, no event was being parsed.']);
  end;
  if ActionsON then
  begin
    Inc(Result);
    Log(['=> Conditions command, the Actions block is already being parsed.']);
  end;

  if Result <> 0 then Exit;

  ConditionsON := True;
  Conditions   := condEvaluating; // il faut tester les conditions

  // Rfrencement et test sur occurences:
  PreCheckEvent;
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// PRECHECK EVENT
// Rfrence dans le GameSim de l'evt si inconnu
// Test sur les occurences permises et passage  valutation False si non autoris
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
procedure TScriptEngine.PreCheckEvent;
begin

  GameSim := TGameSim(Owner);

  cuEventState := GameSim.EventStates.GetByName(ParsedEvent);
  // l'evt n'est pas rfrenc, on l'ajoute
  if cuEventState = nil then
  begin
    cuEventState := TEventState.Create;
    cuEventState.UName := ParsedEvent;
    cuEventState.cuOccurs := 0;
    cuEventState.MaxOccurs := BaseOccurs;
    GameSim.EventStates.Add(cuEventState);
  end;

  // on regarde maintenant si l'vnement n'a pas dpass le nb d'occurences permises
  if cuEventState.cuOccurs >= cuEventState.MaxOccurs then
  begin
    Conditions   := condFalse;
    Log(['Too many occurences for event', ParsedEvent, 'further processing aborted.']);
  end;
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// RESET ACTIONS VARIABLES
// RZ de toute variable d'action (autrement elles persistes sauf chgt implicite, le long d'un evt)
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
procedure TScriptEngine.ResetActionsVariables(ResetSelectors : boolean = True);
begin
  WorkingGME    := '';
  UseCustom     := False;
//  cuFaction     := nil; pas de reset entre 2 evts
  if ResetSelectors then {pas de RZ sur slcteurs quand on passe  la partie Action d'un evt}
  begin
    cuRegion      := nil;   // il n'y a jamais RAZ des selecteurs entre 2 evts, mais normalement tt commande de slection les RAZ avant d'tre parcourue...
    cuGroup       := nil;
    cuUnit        := nil;
    cuSubUnit     := nil;
    LastGHQGroup  := nil;
    cuRegList.Clear;
    cuSUList.Clear;
  end;
  Name          := _genName;
  ShortName     := '';
  FlavorName    := '';
  SUFlavorName  := '';
  Text          := '';
  Kind          := _gkLand;
  ObjType       := _noUID;
  NumCreate     := -1;
  HealthPerc    := 100;
  CohesionPerc  := 100;
  Level         := 1;
  SocialClass   := _noUID; // converti en valeur alatoire
  EthnicGroup   := _noUID; // converti en l'UID de la faction controlant la rgion
  Religion      := 0;
  LoyaltyFac    := popMaxLoyalty_;
  LoyaltyEth    := popMaxLoyalty_;
  AccuGrowth    := 0;
  GrowthRate    := 0;
  Richness      := -1; // valeur alatoire fct de la classe sociale
  Entrench      := 0;
  InCS          := False;
  FixType       := _fixUnFixed;
  GroupPosture  := _psAgressive;
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// TO ACTIONS
// Pour entrer dans l'analyse du bloc Actions
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.ToActions     : integer;
var i : integer;
begin
  Result := 0;
  if ParsedEvent = '' then
  begin
    Inc(Result);
    Log(['=> Actions command, no event was being parsed.']);
  end;

  if Conditions = condFalse then
  begin
    Inc(Result);
    Log(['=> Actions command, Evaluating actions while Conditions are not met!']);
  end;

  if Result <> 0 then Exit;

  // Possible s'il n'y a pas de bloc de conditions (les occurences peuvent passer les conditions  Faux)
  if cuEventState = nil then
    PreCheckEvent;

  // Reset variables d'actions
  ResetActionsVariables(False);

  // on ne fait rien de plus
  if Conditions = condFalse then Exit;

  // Bloc action ok, on le commence:
  ConditionsON := False; // on peut d'ailleurs autoriser aucune condition
  ActionsON    := True;
  Conditions   := condTrue; // les conditions sont runies

  // Gnrer un message joueur si  faire et qu'une rgion est dj renseigne (indication en dur dans l'evt d'une rgion de focus)
  if (PlayerMessage.DisplayType <> _evtNoDisplay) and (PlayerMessage.RgnUID <> _noUID) then
  begin
    if (cuFaction <> nil) and (not PlayerMessage.IsBroadCast) then
    begin
      with PlayerMessage do
        cuFaction.Messages.MakeFlavorMsg(Title, Text, ImageID, DisplayType, RgnUID, SoundID);
    end
    else
    if PlayerMessage.IsBroadCast then
    begin
      for i := 0 to GameSim.Factions.Count-1 do
        with PlayerMessage do
          GameSim.Factions[i].Messages.MakeFlavorMsg(Title, Text, ImageID, DisplayType, RgnUID, SoundID);
    end;
  end;

  Log(['Entering triggered actions for event', ParsedEvent]);
end;


// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SWITCH FORCED EVENTS TO TRUE
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
procedure TScriptEngine.SwitchForcedEventsToTrue;
var i : integer;
begin
  if Conditions <> condFalse then Exit;

  for i := Low(ForcedEvents) to High(ForcedEvents) do
  begin
    if UpperCase(ParsedEvent) = UpperCase(ForcedEvents[i]) then
    begin
      Conditions := condEvaluating;
      Log(['Switched Forced Event To True:', ParsedEvent]);
      Exit;
    end;
  end;
end;

// *** FIN DESCRIPTEURS DE BLOCS ***

// *** DEBUT CONDITIONS ***

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// CHECK FIXED DATE
// Test sur une date fixe
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.CheckFixedDate(Params : TStringList) : integer;
begin
  Result := 0;
  if not (Params.Count in [1]) then
  begin
    Inc(Result);
    Log(['=> CheckFixedDate command, need to have a date as argument']);
  end;
  if (Length(Params[0]) <> 10) then
  begin
    Inc(Result);
    Log(['=> CheckFixedDate command, Erroneous format for date, the format must have 10 chars.']);
  end;
  if Result <> 0 then Exit;

  Params[0] := NormalizeDate(Params[0]);

  if UsedTurn <> DateToTurn(Params[0]) then
    Conditions := condFalse;

  Log(['Fixed date evaluated:', Params[0], 'converted to turn ', DateToTurn(Params[0]), 'current turn', TGameSim(Owner).cuTurn, IIF(Conditions = condEvaluating, 'True', 'False')]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// CHECK MIN DATE
// le tour doit tre au moins gal au tour de la date passe
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.CheckMinDate(Params : TStringList) : integer;
var Turn : integer;
    OffSet : integer;
begin
  Result := 0;
  OffSet := 0;
  if not (Params.Count in [1, 2]) then
  begin
    Inc(Result);
    Log(['=> CheckMinDate command, need to have a date as argument']);
  end;
  if (Length(Params[0]) <> 10) then
  begin
    Inc(Result);
    Log(['=> CheckMinDate command, Erroneous format for date, the format must have 10 chars.']);
  end;
  // il y a une frquence  tester
  if Params.Count = 2 then
  begin
    if Pos('STEP', Uppercase(Params[1])) = 0 then
    begin
    Inc(Result);
    Log(['=> CheckMinDate command, The second parameter must be STEP x, where x is an offset in days']);
    end;
  end;

  if Result <> 0 then Exit;

  Params[0] := NormalizeDate(Params[0]);
  Turn := DateToTurn(Params[0]);

  if UsedTurn < Turn then
  begin
    Conditions := condFalse;
  end;

  // Test supplmentaire sur le step
  if Params.Count = 2 then
  begin
    Params[1] := Trim(Copy(Params[1], Pos('STEP', Params[1]) + 5, 255));
    OffSet := STI(Trim(Params[1])) div DaysPerTurn_;
    if (TGameSim(Owner).cuTurn - Turn) MOD OffSet <> 0 then
      Conditions := condFalse;
  end;

  Log(['Min date evaluated:', Params[0], 'converted to turn ',
  Turn, 'current turn', TGameSim(Owner).cuTurn,
  IIF(OffSet <> 0, 'OffSet : ' + ITS(OffSet), ''),
  IIF(Conditions = condEvaluating, 'True', 'False')]);

end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// CHECK MAX DATE
// le tour doit tre au plus gal au tour de la date passe
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.CheckMaxDate(Params : TStringList) : integer;
begin
  Result := 0;
  if not (Params.Count in [1]) then
  begin
    Inc(Result);
    Log(['=> CheckMaxDate command, need to have a date as argument']);
  end;
  if (Length(Params[0]) <> 10) then
  begin
    Inc(Result);
    Log(['=> CheckMaxDate command, Erroneous format for date, need yyyy.mm.dd or dd/mm/yyyy']);
  end;
  if Result <> 0 then Exit;

  Params[0] := NormalizeDate(Params[0]);

  if UsedTurn > DateToTurn(Params[0])  then
    Conditions := condFalse;

  Log(['Max date evaluated:', Params[0], 'converted to turn ', DateToTurn(Params[0]), 'current turn', TGameSim(Owner).cuTurn, IIF(Conditions = condEvaluating, 'True', 'False')]);

end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// PROBABILITY
// Un D100 <= au paramtre pass
// on peut ajouter le nom d'une variable expose (prcalcule!)
// et un coefficient associ  cette variable
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.Probability(Params : TStringList) : integer;
var Proba, Roll : integer;
    VarName : string;
    ModValue, VarCoeff : integer;
    VarValue : integer;
begin
  Result := 0;
  if not (Params.Count in [1, 3]) then
  begin
    Inc(Result);
    Log(['=> Probability command, need to have a % as argument: Probability [;esvIntVar(x)] [;VarCoeff]']);
  end;

  VarName := ''; VarCoeff := 0;
  if Params.Count = 3 then
  begin
    VarName  := Params[1];
    if not IsExposedVar(VarName) then
    begin
      Inc(Result);
      Log(['=> Probability command, unknown exposed var', VarName]);
      Exit;
    end;
    VarCoeff := STI(Params[2]);
  end;

  if IsExposedVar(VarName) then
    VarValue := GetExposedVarValue(VarName)
  else
    VarValue := 0;
      
  ModValue := VarValue * VarCoeff;

  Roll  := STI(Params[0]);
  Inc(Roll, ModValue);
  Proba := Dice(100);
  if Proba > Roll  then
    Conditions := condFalse;

  Log(['Probability evaluated:', 'Probability', Roll, 'rolled', Proba, IIF(Conditions = condEvaluating, 'True', 'False')]);

end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// EVALUATE UNIQUE UNITS
// On value pour savoir si l'unit est sur carte ou pas, en multi-factions
// Agit comme slecteur galement
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.EvalUniqueUnit(Params : TStringList) : integer;
var fac, unt : integer;
    Faction : TFaction;
    aUnit : TUnit;
    FixType : integer;
    SU : TSubUnit;
    ReverseResult : boolean;
begin
  Result := 0;
  if not (Params.Count in [1..3]) then
  begin
    Inc(Result);
    Log(['=> Evaluate Unique Unit command, need to have 1 to 3 parameters: EvalUnqUnit = <UnitUniqueName> [;FixationType][;NOT]']);
  end;

  ReverseResult := ExtractParam('NOT', Params) <> '';

  if Params.Count >= 2 then
  begin
    if UpperCase(Params[1]) = 'FIXED' then
      FixType := High(integer)
    else
      FixType := STI(Params[1], True, low(integer)); // Param[1] might be also the 'NOT'
  end
  else
    FixType := Low(integer);

  Params[0] := Trim(Params[0]);

  for fac := 0 to GameSim.Factions.Count -1 do
  begin
    Faction := GameSim.Factions[fac];
    Faction.CalcUnitList;
    for unt := 0 to Faction.UnitList.Count -1 do
    begin
      aUnit := TUnit(Faction.UnitList[unt]);
      if Trim(aUnit.InternalName) = Params[0] then
      begin
        Log(['Unique unit:', Params[0], 'found']);
        if (FixType = Low(integer))
        or(aUnit.FixType = FixType)
        or( (aUnit.FixType <> _fixUnfixed)and(FixType = High(integer))) then
        begin
          cuUnit    := aUnit;
          if cuUnit.Count > 0 then
            cuSubUnit := cuUnit[0];
          cuFaction := cuUnit.pFaction;
          cuGroup   := cuUnit.pGroup;
          cuRegion  := cuFaction.Regions.Fetch(cuUnit.StartSubSpace);
          if ReverseResult then Conditions := condFalse;
          Exit;
        end
        else Log(['... but Fixtype differs']);
      end;
    end;
  end;

  // si pas trouv au niveau unit, on regarde toutes les Sous-units
  for fac := 0 to GameSim.Factions.Count -1 do
  begin
    Faction := GameSim.Factions[fac];
    SU := Faction.GetSubUnitByName(Params[0], True);
    if SU = nil then Continue;
    aUnit := TUnit(SU.pUnit);
    if (FixType = Low(integer))
    or(aUnit.FixType = FixType)
    or( (aUnit.FixType <> _fixUnfixed)and(FixType = High(integer))) then
    begin
      cuSubUnit := SU;
      cuUnit    := TUnit(SU.pUnit);
      cuGroup   := cuUnit.pGroup;
      cuFaction := cuGroup.pFaction;
      cuRegion  := cuFaction.Regions.Fetch(cuGroup.StartSubSpace);
      Log(['Select Unique unit:', 'found as an unit which have a sub unit of name', Params[0]]);
      if ReverseResult then Conditions := condFalse;
      Exit;
    end;
  end;

  // pas trouv, si on a mis NOT c'est ce que l'on voulait, donc on laisse Conditions  True,
  // autrement on valorise
  if not ReverseResult then Conditions := condFalse;
  Log(['Unique unit:', Params[0], 'NOT found / FixType differs']);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// PICK UNIT
// Slection d'une unit (et chgt des slecteurs cuGroup et cuRegion si trouv) dans une zone
// ** ou une rgion **
// PickUnit = AreaOrRegionUID; UIDType; SubTypeUID
// Valeurs spciales:
// UIDType: 0 une rgion, 1: une area
// si SubType = _noUID (-1), on prend tout les types
// !!! Les leaders purs sont tjrs exclus
// SELUNITINAREA fait pareil !
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.PickUnit(Params : TStringList) : integer;
var Area      : TArea;
    SubType   : TUID;
    i         : integer;
    Region    : TRegion;
    RgnUID    : TUID;
    Bag       : TRandomBag;
    GpKind    : TUID;
begin
  Result := 0;
  if not (Params.Count in [3]) then
  begin
    Inc(Result);
    Log(['=> PickUnit command, need to have 3 parameters, syntax is: PickUnit = AreaOrRegionUID; UIDType; SubTypeUID']);
  end;

  Area := nil;
  RgnUID := _noUID;

  if cuFaction = nil then
  begin
    Log(['=> PickUnit command, no faction selected']);
    Inc(Result);
    Exit;
  end;

  if Params[1] = '1' then
  begin
    Area := Areas[STI(Params[0])];
    if Area = nil then
    begin
      Inc(Result);
      Log(['=> PickUnit command, the Area', STI(Params[0]), 'is unknown.']);
    end;
  end
  else
    RgnUID := STI(Params[0]);

  SubType  := STI(Params[2]);

  if SubType > _stmNb-1 then
  begin
    Inc(Result);
    Log(['=> PickUnit command, SubType too high.', SubType]);
  end;

  if Result > 0 then Exit;

  cuGroup  := nil;
  cuUnit   := nil;
  cuRegion := nil;
  cuRegList.Clear;

  // Test sur une rgion
  if RgnUID <> _noUID then
  begin
    Region := cuFaction.Regions.Fetch(RgnUID);
    if SubType = _noUID then
      GpKind := IIF(RegionDefs.IsLandRegion(RgnUID), _gkLand, _gkNav)
    else
    if SubType = _stmShip then
      GpKind := _gkNav
    else
      GpKind := _gkLand;

    if GameSim.RegionCmns[RgnUID].BlockState = _regUnBlocked then
      cuUnit := Region.PickUnit(GpKind, SubType, True);

    if cuUnit <> nil then
    begin
      cuGroup  := cuUnit.pGroup;
      cuRegion := cuGroup.pStartRegion;
      Log(['PickUnit command, picked this unit', cuUnit.Name, 'in group', cuGroup.Name, 'in region', cuRegion.UID, cuRegion.Name]);
      Exit;
    end;
  end
  else
  // test sur une zone
  begin
    Bag := TRandomBag.Create;
    try
      for i := 0 to Area.Count -1 do
      begin
        if GameSim.RegionCmns[Area[i]].BlockState = _regUnBlocked then
          Bag.Add(Area[i]);
      end;    

      while Bag.Count > 0 do
      begin

        Region := cuFaction.Regions.Fetch(Bag.PickInt);
        if Region = nil then Continue;

        if SubType = _noUID then
          GpKind := IIF(RegionDefs.IsLandRegion(RgnUID), _gkLand, _gkNav)
        else
        if SubType = _stmShip then
          GpKind := _gkNav
        else
          GpKind := _gkLand;

        cuUnit := Region.PickUnit(GpKind, SubType, True);
        if cuUnit <> nil then
        begin
          cuGroup  := cuUnit.pGroup;
          cuRegion := cuGroup.pStartRegion;
          cuRegList.Add(cuRegion.UID);
          Log(['PickUnit command (Area), picked this unit', cuUnit.Name, 'in group', cuGroup.Name, 'in region', cuRegion.UID, cuRegion.Name]);
          Exit;
        end;
      end;

      Log(['PickUnit command, failed to Pick a unit with these parameters (Blocked regions discarded)', Params[0], Params[1], Params[2], 'Group, Unit, Region selectors set to nil now.']);
      
    finally
      Bag.Free;
    end;
  end;

end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// EVAL EVENT
// Test le nombre d'occurences (dclenches) d'un vnement
// EvalEvent  = NameOfEvent;Operator;NumOccurs
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.EvalEvent (Params : TStringList) : integer;
var GS : TGameSim;
    ES : TEventState;
    NbOccurs : integer;
begin
  Result := 0;
  if not (Params.Count in [3]) then
  begin
    Inc(Result);
    Log(['=> EvalEvent command, need to have 3 parameters: Name of Event, Operator (=, >=, <=) and number of occurences']);
  end;

  GS := TGameSim(Owner);
  NbOccurs := STI(Params[2]);

  ES := GS.EventStates.GetByName(Params[0]);
  if ES = nil then
  begin
    Conditions := condFalse;
    Log(['EvalEvent evaluated:', Params[0], 'is unknown (or not referenced yet), Conditions set to False']);
    Exit;
  end;

  Params[1] := Trim(Params[1]);

  // Evt dj rfrenc, on test
  if Params[1] = '=' then
  begin
    if not (ES.cuOccurs = NbOccurs) then Conditions := condFalse;
    Log(['EvalEvent evaluated:', 'Checking', ES.UName, ES.cuOccurs, 'equal to', NbOccurs, IIF(Conditions = condEvaluating, 'True', 'False')]);
    Exit;
  end
  else
  if Params[1] = '<=' then
  begin
    if not (ES.cuOccurs <= NbOccurs) then Conditions := condFalse;
    Log(['EvalEvent evaluated:', 'Checking', ES.UName, ES.cuOccurs, 'at most equal to', NbOccurs, IIF(Conditions = condEvaluating, 'True', 'False')]);
    Exit;
  end
  else
  if Params[1] = '>=' then
  begin
    if not (ES.cuOccurs >= NbOccurs) then Conditions := condFalse;
    Log(['EvalEvent evaluated:', 'Checking', ES.UName, ES.UName, ES.cuOccurs, 'at least equal to', NbOccurs, IIF(Conditions = condEvaluating, 'True', 'False')]);
    Exit;
  end
  else
  if Params[1] = '<' then
  begin
    if not (ES.cuOccurs < NbOccurs) then Conditions := condFalse;
    Log(['EvalEvent evaluated:', 'Checking', ES.UName, ES.cuOccurs, 'strictly inferior to', NbOccurs, IIF(Conditions = condEvaluating, 'True', 'False')]);
    Exit;
  end
  else
  if Params[1] = '>' then
  begin
    if not (ES.cuOccurs > NbOccurs) then Conditions := condFalse;
    Log(['EvalEvent evaluated:', 'Checking', ES.UName, ES.UName, ES.cuOccurs, 'strictly superior to', NbOccurs, IIF(Conditions = condEvaluating, 'True', 'False')]);
    Exit;
  end
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// COUNT CONTROLED AREAS
// Compte le nombre de zones controles entirement (toutes les rgions strats/VP controles par la faction)
// et range le rsultat dans la variable passe en premier paramtre
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.CountCtrlAreas (Params : TStringList) : integer;
var AreaUID : TUID;
    i : integer;
    OwnedVPs, TotalVPs : integer;
    CtrlArea : integer;
begin
  Result := 0;
  CtrlArea := 0;

  if (cuFaction = nil) then
  begin
    Inc(Result);
    Log(['=> Count Controlled Areas: No faction selected']);
  end;
  if (Params.Count < 2) then
  begin
    Inc(Result);
    Log(['=> Count Controlled Areas: Need the esvIntVar where to store the result, and one argument for each area evaluated']);
    Exit;
  end;
  if not IsExposedVar(Params[0]) then
  begin
    Inc(Result);
    Log(['=> Count Controlled Areas: not recognized as an exposed var', Params[0]]);
  end;
  if Result <> 0 then Exit;

  for i := 1 to Params.Count -1 do
  begin
    AreaUID := STI(Params[i]);
    if Areas[AreaUID] = nil then
    begin
      Inc(Result);
      Log(['=> Count Controlled Areas: Area not found', Params[i]]);
    end;
    OwnedVPs := GameSim.RegionCmns.GetOwnedVPsInArea(AreaUID, cuFaction);
    TotalVPs := GameSim.RegionCmns.GetTotalVPsInArea(AreaUID);
    if TotalVPs = OwnedVPs then
      Inc(CtrlArea);
  end;

  StoreToExposedVar(Params[0], CtrlArea);
  Log(['Count Controlled Areas:', cuFaction.Name, 'controls entirely', CtrlArea, 'area(s), value stored in', Params[0]]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// COUNT VP
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.CountVP(Params : TStringList) : integer;
var Value : integer;
begin
  Result := 0;

  if (cuFaction = nil) then
  begin
    Inc(Result);
    Log(['=> Count VP: No faction selected']);
  end;
  if (Params.Count <> 1) then
  begin
    Inc(Result);
    Log(['=> Count VP: Need the esvIntVar where to store the result']);
    Exit;
  end;
  if not IsExposedVar(Params[0]) then
  begin
    Inc(Result);
    Log(['=> Count VP: not recognized as an exposed var', Params[0]]);
  end;
  if Result <> 0 then Exit;

  Value := cuFaction.GetHeldVPCount;

  StoreToExposedVar(Params[0], Value);
  Log(['Count VP:', cuFaction.Name, 'possess: ', Value, ' VP(s), value stored in', Params[0]]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// COUNT UNITS
// Test dans une zone ou une province un nombre de sous-types d'units versus une valeur (et un oprateur)
// CountUnits = $Deep_South; 1; $ENG; $Regular; >=; 5
// CountUnits = $Philadelphia; 0; $ENG; $Regular; >=; 5
// Syntaxe : CountUnits = AreaOrRegionUID; UIDType; FacUID; SubTypeUID; Operator; Value
// Valeurs spciales:
// UIDType: 0 une rgion, 1: une area
// si SubType = _noUID (-1), on prend tout les types
// DICE accept
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.CountUnits(Params : TStringList) : integer;
var GS : TGameSim;
    Area     : TArea;
    Faction  : TFaction;
    SubType  : TUID;
    Operator : string;
    Value    : integer;
    NbFound  : integer;
    i        : integer;
    Region   : TRegion;
    RgnUID   : TUID;
    Roll     : integer;
begin
  Result := 0;
  if not (Params.Count in [6]) then
  begin
    Inc(Result);
    Log(['=> CountUnits command, need to have 6 parameters, syntax is: EvalUnits = AreaOrRegionUID; UIDType; FacUID; SubTypeUID; Operator; Value']);
  end;

  GS := TGameSim(Owner);
  Area := nil;
  RgnUID := _noUID;

  if Params[1] = '1' then
  begin
    Area := Areas[STI(Params[0])];
    if Area = nil then
    begin
      Inc(Result);
      Log(['=> CountUnits command, the Area', STI(Params[0]), 'is unknown.']);
    end;
  end
  else
    RgnUID := STI(Params[0]);

  if IsNumeric(Params[2]) then
    Faction := GS.Factions.GetFactionByUID(STI(Params[2]), False)
  else
    Faction := FactionTagToFaction(Params[2]);

  if Faction = nil then
  begin
    Inc(Result);
    Log(['=> CountUnits command, the Faction', STI(Params[2]), 'is unknown.']);
  end;

  SubType  := STI(Params[3]);

  if SubType > _stmNb-1 then
  begin
    Inc(Result);
    Log(['=> CountUnits command, SubType too high.', SubType]);
  end;

  if Result > 0 then Exit;

  Operator := Trim(Params[4]);
  Value    := STI(Params[5]);
  NbFound := 0;

  // Test sur une rgion
  if RgnUID <> _noUID then
  begin
    Region := Faction.Regions.Fetch(RgnUID);
    NbFound := Region.GetNbUnitsTheseSubTypes([SubType]);
  end
  else
  // test sur une zone
  for i := 0 to Area.Count -1 do
  begin
    Region := Faction.Regions.Fetch(Area[i], False);
    if Region = nil then Continue;
    Inc(NbFound, Region.GetNbUnitsTheseSubTypes([SubType]));
  end;

  if Operator = '=' then
  begin
    if not (Value = NbFound) then Conditions := condFalse;
    Log(['CountUnits evaluated:', 'Checking', Value, 'equal to', NbFound, IIF(Conditions = condEvaluating, 'True', 'False')]);
    Exit;
  end
  else
  if Operator = '<=' then
  begin
    if not (NbFound <= Value) then Conditions := condFalse;
    Log(['CountUnits evaluated:', 'Checking', NbFound, 'at most equal to', Value, IIF(Conditions = condEvaluating, 'True', 'False')]);
    Exit;
  end
  else
  if Operator = '>=' then
  begin
    if not (NbFound >= Value) then Conditions := condFalse;
    Log(['CountUnits evaluated:', 'Checking', NbFound, 'at least equal to', Value, IIF(Conditions = condEvaluating, 'True', 'False')]);
    Exit;
  end
  else
  if Operator = 'DICE' then
  begin
    Roll := Dice(Value);
    if not (Roll > NbFound) then Conditions := condFalse;
    Log(['CountUnits evaluated:', 'Checking', NbFound, 'vs Dice Roll of ', Roll, IIF(Conditions = condEvaluating, 'True', 'False')]);
    Exit;
  end;
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// COUNT UNIT DEF
// Count a number of UnitDef in a region, area or on the entire map
// ECN 1.00
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.CountUnitDef(Params : TStringList) : integer;
var
    AreaUID  : TUID;
    UnitUID  : TUID;
    UD       : TUnitDef;
    Operator : string;
    Value    : integer;
    NbFound  : integer;
    i        : integer;
    Region   : TRegion;
    Roll     : integer;
    ilRgnList: TIntList;
begin
  Result := 0;
  if not (Params.Count in [4]) then
  begin
    Inc(Result);
    Log(['=> CountUnitDef command, need to have 4 parameters, syntax is: CountUnitDef = MAPorREGorAreaUID; UnitDefType; Operator; Value']);
  end;

  // Faction checking
  if cuFaction = nil then
  begin
    Inc(Result);
    Log(['=> CountUnitDef command, no faction selected.']);
  end;

  // Area checking
  if Params[0] = 'MAP' then
    AreaUID := -1
  else
  if Params[0] = 'REG' then
    AreaUID := -2
  else
    AreaUID := STI(Params[1], True, -3);
  if (AreaUID = -3)or(AreaUID > Areas.Count-1) then
  begin
    Inc(Result);
    Log(['=> CountUnitDef command, unknown Area Alias ', Params[1]]);
  end;

  // Region Checking
  if (AreaUID = -2)and(cuRegion = nil) then
  begin
    Inc(Result);
    Log(['=> CountUnitDef command, no region selected.']);
  end;

  // UnitDef checking
  UnitUID := STI(Params[1], True, -1);
  if (UnitUID = -1)or(UnitUID > UnitDefs.Count-1) then
  begin
    Inc(Result);
    Log(['=> CountUnitDef command, unknown UnitDef Alias ', Params[2]]);
  end;
  UD  := UnitDefs[UnitUID];

  if Result > 0 then Exit;

  ilRgnList := TIntList.Create;

  try
    Operator := Trim(Params[2]);
    Value    := STI(Params[3]);
    NbFound := 0;

    if AreaUID = -1 then
    begin
      for i := 0 to RegionDefs.Count-1 do
        ilRgnList.Add(i);
    end
    else
    if AreaUID = -2 then
      ilRgnList.Add(cuRegion.UID)
    else
      ilRgnList.AssignFrom(Areas[AreaUID]);

    // Counting UD
    for i := 0 to ilRgnList.Count -1 do
    begin
      Region := cuFaction.Regions.Fetch(ilRgnList[i], False);
      if Region = nil then Continue;
      Inc(NbFound, Region.GetNbUnitDef(UD, True));
    end;

    // Checking against operator
    if Operator = '=' then
    begin
      if not (Value = NbFound) then Conditions := condFalse;
      Log(['CountUnitDef evaluated:', 'Checking', Value, 'equal to', NbFound, IIF(Conditions = condEvaluating, 'True', 'False')]);
      Exit;
    end
    else
    if Operator = '<=' then
    begin
      if not (NbFound <= Value) then Conditions := condFalse;
      Log(['CountUnitDef evaluated:', 'Checking', NbFound, 'at most equal to', Value, IIF(Conditions = condEvaluating, 'True', 'False')]);
      Exit;
    end
    else
    if Operator = '>=' then
    begin
      if not (NbFound >= Value) then Conditions := condFalse;
      Log(['CountUnitDef evaluated:', 'Checking', NbFound, 'at least equal to', Value, IIF(Conditions = condEvaluating, 'True', 'False')]);
      Exit;
    end
    else
    if Operator = '<' then
    begin
      if not (NbFound < Value) then Conditions := condFalse;
      Log(['CountUnitDef evaluated:', 'Checking', NbFound, 'lower than', Value, IIF(Conditions = condEvaluating, 'True', 'False')]);
      Exit;
    end
    else
    if Operator = '>' then
    begin
      if not (NbFound > Value) then Conditions := condFalse;
      Log(['CountUnitDef evaluated:', 'Checking', NbFound, 'greater than', Value, IIF(Conditions = condEvaluating, 'True', 'False')]);
      Exit;
    end
    else

    if Operator = 'DICE' then
    begin
      Roll := Dice(Value);
      if not (Roll > NbFound) then Conditions := condFalse;
      Log(['CountUnitDef evaluated:', 'Checking', NbFound, 'vs Dice Roll of ', Roll, IIF(Conditions = condEvaluating, 'True', 'False')]);
      Exit;
    end;

  finally
    ilRgnList.Free;
  end;
end;


// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// EVAL REGION OWNED
// Condition value  Vrai si la faction courante est le possesseur de la rgion
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.EvalRgnOwned(Params : TStringList) : integer;
var RgnUID : TUID;
    ReverseResult : boolean;
    FriendToo : boolean;
    Found : boolean;
begin
  Result := 0;
  if Params.Count > 4 then
  begin
    Inc(Result);
    Log(['=> EvalRgnOwned command, syntax is: EvalRgnOwned = RegionUID  [;NOT] [;FriendToo] [;UpdateSel]']);
  end;
  if (cuFaction = nil) then
  begin
    Inc(Result);
    Log(['=> EvalRgnOwned: No faction selected']);
  end;

  if IsNumeric(Params[0]) then
    RgnUID := STI(Params[0])
  else
    RgnUID := _noUID;
  if (RgnUID = _noUID) then
  begin
    Inc(Result);
    Log(['=> EvalRgnOwned: Param0 is not a known region (check if the alias is valid):', Params[0]]);
  end;
  if Result <> 0 then Exit;


  ReverseResult := ExtractParam('NOT', Params) <> '';
  FriendToo :=     ExtractParam('FRIENDTOO', Params) <> '';

  if FriendToo then
    Found := GameSim.RegionCmns.FetchOwner(RgnUID).IsFriendOf(cuFaction)
  else
    Found := GameSim.RegionCmns.FetchOwner(RgnUID) = cuFaction;

  if  (ExtractParam('UPDATESEL', Params) <> '') and Found then
  begin
    cuFaction := GameSim.RegionCmns.FetchOwner(RgnUID);
    cuRegion := cuFaction.Regions.Fetch(RgnUID);
  end;

  if ReverseResult then
  begin
    if Found then Conditions := condFalse;
    Log(['EvalRgnOwned:', 'Checking', RgnUID, GameSim.RegionCmns[RgnUID].Name, 'is not owned', IIF(Conditions = condEvaluating, 'True', 'False')]);
  end
  else
  begin
    if not Found then Conditions := condFalse;
    Log(['EvalRgnOwned:', 'Checking', RgnUID, GameSim.RegionCmns[RgnUID].Name, 'is owned', IIF(Conditions = condEvaluating, 'True', 'False')]);
  end;
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// NOT ENEMY
// ! critre Not Enemy du script exclue une rgion peu loyale (conside ennemie dans ce cas)
//  la diffrence de TRegion.IsEnemyControled qui n'en tient pas compte
// !! un ennemi dans les murs ne controle pas la rgion, utiliser EvalRgnOwned pour cela
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.NotEnemy (Params : TStringList) : integer;
var RgnUID : TUID;
begin
  Result := 0;
  if not (Params.Count in [1]) then
  begin
    Inc(Result);
    Log(['=> NotEnemy command, syntax is: NotEnemy = RegionUID']);
  end;
  if (cuFaction = nil) then
  begin
    Inc(Result);
    Log(['=> NotEnemy: No faction selected']);
  end;

  if IsNumeric(Params[0]) then
    RgnUID := STI(Params[0])
  else
    RgnUID := _noUID;
  if (RgnUID = _noUID) then
  begin
    Inc(Result);
    Log(['=> NotEnemy: Param0 is not a known region (check if the alias is valid):', Params[0]]);
  end;
  if Result <> 0 then Exit;

  if cuFaction.Regions.Fetch(RgnUID).IsEnemyControled(True) then Conditions := condFalse;

  Log(['NotEnemy:', 'Checking', RgnUID, GameSim.RegionCmns[RgnUID].Name, 'Not Enemy to ', cuFaction.Name + ': result is ',
  IIF(Conditions = condEvaluating, 'True', 'False')]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// EVAL REGION VICTORY POINT
// Condition value  Vrai si la faction courante  les VP (ventuels) de la rgion
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.EvalRgnVP(Params : TStringList) : integer;
var RgnUID : TUID;
    ReverseResult : boolean;
    Found : boolean;
begin
  Result := 0;
  if Params.Count > 2 then
  begin
    Inc(Result);
    Log(['=> EvalRgnVP command, syntax is: EvalRgnVP = RegionUID [;NOT]']);
  end;
  if (cuFaction = nil) then
  begin
    Inc(Result);
    Log(['=> EvalRgnVP: No faction selected']);
  end;

  if IsNumeric(Params[0]) then
    RgnUID := STI(Params[0])
  else
    RgnUID := _noUID;
  if (RgnUID = _noUID) then
  begin
    Inc(Result);
    Log(['=> EvalRgnVP: Param0 is not a known region (check if the alias is valid):', Params[0]]);
  end;
  if Result <> 0 then Exit;

  ReverseResult := ExtractParam('NOT', Params) <> '';

  Found := GameSim.RegionCmns.GetVPCountFromRegion(cuFaction, RgnUID) <> 0;

  if ReverseResult then
  begin
    if Found then Conditions := condFalse;
    Log(['EvalRgnVP:', 'Checking', RgnUID, GameSim.RegionCmns[RgnUID].Name, 'for the absence of VP', IIF(Conditions = condEvaluating, 'True', 'False')]);
  end
  else
  begin
    if not Found then Conditions := condFalse;
    Log(['EvalRgnVP:', 'Checking', RgnUID, GameSim.RegionCmns[RgnUID].Name, 'for the presence of VP', IIF(Conditions = condEvaluating, 'True', 'False')]);
  end;
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// EVAL REGION OBJECTIVE
// Condition value  Vrai si la faction courante possde l'objectif (ventuel) de la rgion
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.EvalRgnObj(Params : TStringList) : integer;
var RgnUID : TUID;
    ReverseResult : boolean;
    Found : boolean;
begin
  Result := 0;
  if Params.Count > 2 then
  begin
    Inc(Result);
    Log(['=> EvalRgnObj command, syntax is: EvalRgnObj = RegionUID [;NOT]']);
  end;
  if (cuFaction = nil) then
  begin
    Inc(Result);
    Log(['=> EvalRgnObj: No faction selected']);
  end;

  if IsNumeric(Params[0]) then
    RgnUID := STI(Params[0])
  else
    RgnUID := _noUID;
  if (RgnUID = _noUID) then
  begin
    Inc(Result);
    Log(['=> EvalRgnObj: Param0 is not a known region (check if the alias is valid):', Params[0]]);
  end;
  if Result <> 0 then Exit;
  ReverseResult := ExtractParam('NOT', Params) <> '';

  Found := GameSim.RegionCmns.HasObjectiveInRegion(cuFaction, RgnUID);

  if ReverseResult then
  begin
    if Found then Conditions := condFalse;
    Log(['EvalRgnObj:', 'Checking', RgnUID, GameSim.RegionCmns[RgnUID].Name, 'for the absence of objective', IIF(Conditions = condEvaluating, 'True', 'False')]);
  end
  else
  begin
    if not Found then Conditions := condFalse;
    Log(['EvalRgnObj:', 'Checking', RgnUID, GameSim.RegionCmns[RgnUID].Name, 'for the presence of objective', IIF(Conditions = condEvaluating, 'True', 'False')]);
  end;
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// EVAL BESIEGED
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.EvalBesieged(Params : TStringList) : integer;
var RgnUID : TUID;
    ReverseResult : boolean;
    Found : boolean;
    All : boolean;
    pStruc : Tstructure;
    Faction : TFaction;
    Region : TRegion;
begin
  Result := 0;            
  if Params.Count > 2 then
  begin
    Inc(Result);
    Log(['=> EvalBesieged command, syntax is: EvalBesieged = [ALL] [;NOT]']);
  end;
  if (cuFaction = nil) then
  begin
    Inc(Result);
    Log(['=> EvalBesieged: No faction selected']);
  end;
  if (cuRegion = nil) then
  begin
    Inc(Result);
    Log(['=> EvalBesieged: No region selected']);
  end;

  if Result <> 0 then Exit;
  ReverseResult := ExtractParam('NOT', Params) <> '';
  All           := ExtractParam('ALL', Params) <> '';

  Faction := cuFaction;
  RgnUID := cuRegion.UID;

  if not All then
    Found := cuRegion.Besieged
  else
  begin
    Faction := GameSim.RegionCmns.GetCityInfos(RgnUID, pStruc);
    if Faction = nil then Found := False else
    begin
      Region := Faction.Regions.Fetch(RgnUID);
      Found := Region.Besieged;
    end;

  end;

  if ReverseResult then
  begin
    if Found then Conditions := condFalse;
    if Faction = nil then
      Log(['EvalBesieged:', 'Checking', RgnUID, GameSim.RegionCmns[RgnUID].Name, 'no faction owns a structure here', IIF(Conditions = condEvaluating, 'True', 'False')])
    else
      Log(['EvalBesieged:', 'Checking', RgnUID, GameSim.RegionCmns[RgnUID].Name, 'is this faction is under siege:', Faction.UID, Faction.Name, IIF(Conditions = condEvaluating, 'True', 'False')]);
  end
  else
  begin
    if not Found then Conditions := condFalse;
    if Faction = nil then
      Log(['EvalBesieged:', 'Checking', RgnUID, GameSim.RegionCmns[RgnUID].Name, 'no faction owns a structure here', IIF(Conditions = condEvaluating, 'True', 'False')])
    else
      Log(['EvalBesieged:', 'Checking', RgnUID, GameSim.RegionCmns[RgnUID].Name, 'is this faction is under siege:', Faction.UID, Faction.Name, IIF(Conditions = condEvaluating, 'True', 'False')]);
  end;
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// EVAL BLOCKADED
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.EvalBlockaded(Params : TStringList) : integer;
var RgnUID : TUID;
    ReverseResult : boolean;
    Found : boolean;
    All : boolean;
    Faction : TFaction;
    Region : TRegion;
    pStruc : Tstructure;
begin
  Result := 0;
  if Params.Count > 2 then
  begin
    Inc(Result);
    Log(['=> EvalBlockaded command, syntax is: EvalBlockaded = ALL [;NOT]']);
  end;
  if (cuFaction = nil) then
  begin
    Inc(Result);
    Log(['=> EvalBlockaded: No faction selected']);
  end;
  if (cuRegion = nil) then
  begin
    Inc(Result);
    Log(['=> EvalBlockaded: No region selected']);
  end;

  if Result <> 0 then Exit;
  ReverseResult := ExtractParam('NOT', Params) <> '';
  All           := ExtractParam('ALL', Params) <> '';

  Faction := cuFaction;
  RgnUID := cuRegion.UID;

  if not All then
    Found := cuRegion.IsBlockaded
  else
  begin
    Faction := GameSim.RegionCmns.GetCityInfos(RgnUID, pStruc);
    if Faction = nil then Found := False else
    begin
      Region := Faction.Regions.Fetch(RgnUID);
      Found := Region.IsBlockaded;
    end;

  end;

  if ReverseResult then
  begin
    if Found then Conditions := condFalse;
    Log(['EvalBlockaded:', 'Checking', RgnUID, GameSim.RegionCmns[RgnUID].Name, 'if this faction is under blockade', Faction.UID, Faction.Name, IIF(Conditions = condEvaluating, 'True', 'False')]);
  end
  else
  begin
    if not Found then Conditions := condFalse;
    Log(['EvalBlockaded:', 'Checking', RgnUID, GameSim.RegionCmns[RgnUID].Name, 'if this faction is under blockade', Faction.UID, Faction.Name, IIF(Conditions = condEvaluating, 'True', 'False')]);
  end;
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// EVAL REGION STRUCTURE
// True if the faction (or any faction with FACALL) has at least one
// structure, based on a particular criteria
// ex: EvalRgnStruc = $Paris|FACALL|STRUCATR $attIsSewer
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.EvalRgnStruc(Params : TStringList) : integer;
type
  TStructureSearchType = (sstAny, sstKind, sstSubKind, sstFamily, sstAttribute, sstDef);
var RgnUID    : TUID;
    Param     : string;
    i         : integer;
    SST       : TStructureSearchType;
    Struc     : TStructure;
    Found     : boolean;
    ReverseResult : boolean;
    FacAll    : boolean;
begin
  Result := 0;
  if not (Params.Count in [3, 4]) then
  begin
    Inc(Result);
    Log(['=> EvalRgnStruc command, syntax is: EvalRgnStruc = RegionUID; <FactionValue>; <StructureSearchType (one of STRUCKIND, STRUCDEF, STRUCSUB, STRUCFAM, STRUCATR) value >[;NOT]']);
  end;

  RgnUID := _noUID;
  if IsNumeric(Params[0]) then
    RgnUID := STI(Params[0])
  else
  begin
    Inc(Result);
    Log(['=> EvalRgnStruc: Param0 is not a known region (check if the alias is valid):', Params[0]]);
  end;
  if Result <> 0 then Exit;

  if (RgnUID = _noUID) and (cuRegion <> nil) then
    RgnUID := cuRegion.UID;

  if RgnUID = _noUID then
  begin
    Log(['[Warning] EvalRgnStruc', 'The region is still at -1, exiting silently.']);
    Exit;
  end;
  
  FacAll := Params[1] = 'FACALL';                   // Note: extractparams can not discriminate the ALL of FACALL and STRUCALL, so I use STRUCANY for the 3rd param

  if not FacAll then
    cuFaction := FactionTagToFaction(Params[1]);

  Found := False;
  ReverseResult := ExtractParam('NOT', Params) <> '';

  //  STRUCANY, STRUCKIND, STRUCDEF
  SST := sstAny;
  Param := ExtractParam('STRUCKIND', Params);
  if Param <> '' then
  begin
    SST := sstKind;
    ExtractArgs(Param);
  end;
  Param := ExtractParam('STRUCDEF', Params);
  if Param <> '' then
  begin
    SST := sstDef;
    ExtractArgs(Param);
  end;
  Param := ExtractParam('STRUCSUB', Params);
  if Param <> '' then
  begin
    SST := sstSubKind;
    ExtractArgs(Param);
  end;
  Param := ExtractParam('STRUCFAM', Params);
  if Param <> '' then
  begin
    SST := sstFamily;
    ExtractArgs(Param);
  end;
  Param := ExtractParam('STRUCATR', Params);
  if Param <> '' then
  begin
    SST := sstAttribute;
    ExtractArgs(Param);
  end;
  if FacAll then
  begin
    for i := 0 to GameSim.RegionCmns[RgnUID].pListStruc.Count-1 do
    begin
      Struc := GameSim.RegionCmns[RgnUID].pListStruc[i];
      if SST = sstAny then Found := True;
      if (SST = sstDef)  and (Struc.DefUID  = STI(Args[0])) then Found := True;
      if (SST = sstKind) and (Struc.GetKind = STI(Args[0])) then Found := True;
      if (SST = sstSubKind) and (Struc.StrucDef.SubKind = STI(Args[0])) then Found := True;
      if (SST = sstFamily) and (Struc.StrucDef.Family = STI(Args[0])) then Found := True;
      if (SST = sstAttribute) and (Struc.StrucDef.HasThisAttribute(Args[0])) then Found := True;

      if Found then
      begin
        cuFaction := Struc.pFaction;
        cuRegion  := cuFaction.Regions.Fetch(RgnUID, False);
      end;
    end;
  end
  else
  begin
    cuRegion := cuFaction.Regions.Fetch(RgnUID, False);
    if cuRegion <> nil then
    begin
      for i := 0 to cuRegion.Structures.Count-1 do
      begin
        Struc := cuRegion.Structures[i];
        if SST = sstAny then Found := True;
        if (SST = sstDef)  and (Struc.DefUID  = STI(Args[0])) then Found := True;
        if (SST = sstKind) and (Struc.GetKind = STI(Args[0])) then Found := True;
        if (SST = sstSubKind) and (Struc.StrucDef.SubKind = STI(Args[0])) then Found := True;
        if (SST = sstFamily) and (Struc.StrucDef.Family = STI(Args[0])) then Found := True;
        if (SST = sstAttribute) and (Struc.StrucDef.HasThisAttribute(Args[0])) then Found := True;
      end;
    end;
  end;

  if ReverseResult then
  begin
    if Found then Conditions := condFalse;
    Log(['EvalRgnStruc:', 'Checking', RgnUID, GameSim.RegionCmns[RgnUID].Name, 'for the absence of a particular structure', IIF(Conditions = condEvaluating, 'True', 'False')]);
  end
  else
  begin
    if not Found then Conditions := condFalse;
    Log(['EvalRgnStruc:', 'Checking', RgnUID, GameSim.RegionCmns[RgnUID].Name, 'for the presence of a particular structure', IIF(Conditions = condEvaluating, 'True', 'False')]);
  end;

end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// TURN INDEX
// Test l'index du tour dans l'anne, permet de faire des tests trimestriels
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.TurnIndex(Params : TStringList) : integer;
var i : integer;
    Str : string;
begin
  Result := 0;
  Str := '';
  for i := 0 to Params.Count-1 do
  begin
    if STI(Params[i]) = GetTurnIndexInYear(GameSim.cuTurn) then
    begin
      Str := Params[i] + ' ';
      Log(['TurnIndex:', 'Testing', Str, 'vs current turn index:',
      GetTurnIndexInYear(GameSim.cuTurn), 'True']);
      Exit;
    end;
  end;
  Log(['TurnIndex:', 'Testing', Str, 'vs current turn index:',
      GetTurnIndexInYear(GameSim.cuTurn), 'False']);
  Conditions := condFalse;
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// CHECK AI LEVEL
// Test si le niveau d'IA de la faction est >= au niveau demand
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.CheckAILevel(Params : TStringList) : integer;
var AILevel : TUID;
begin
  Result := 0;
  if not (Params.Count in [1]) then
  begin
    Inc(Result);
    Log(['=> CheckAILevel command, syntax is: CheckAILevel = AILevel']);
  end;
  if (cuFaction = nil) then
  begin
    Inc(Result);
    Log(['=> CheckAILevel: No faction selected']);
  end;

  if IsNumeric(Params[0]) then
    AILevel := STI(Params[0])
  else
    AILevel := _noUID;
  if (AILevel = _noUID) then
  begin
    Inc(Result);
    Log(['=> CheckAILevel: Param0 is not a known AILevel (check if the alias is valid):', Params[0]]);
  end;
  if Result <> 0 then Exit;

  if not aiActivated_ then
    Conditions := condFalse
  else
    if cuFaction.AILevel < AILevel then
      Conditions := condFalse;

  Log(['CheckAILevel:', 'Checking AI Level of faction ', cuFaction.Name, 'at least equal to', AILevel,
  IIF(Conditions = condEvaluating, 'True', 'False')]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// IS PLAYER (human)
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.CheckIsPlayer(Params : TStringList) : integer;
var Reverse : boolean;
begin
  Result := 0;
  if not (Params.Count in [0, 1]) then
  begin
    Inc(Result);
    Log(['=> CheckIsPlayer command, syntax is: CheckIsPlayer = [;NOT]']);
  end;
  if (cuFaction = nil) then
  begin
    Inc(Result);
    Log(['=> CheckIsPlayer: No faction selected']);
  end;

  if Result <> 0 then Exit;

  Reverse := ExtractParam('NOT', Params) <> '';

  if (cuFaction.AILevel <> _aiHuman) or (Reverse and (cuFaction.AILevel = _aiHuman)) then
      Conditions := condFalse;

  Log(['CheckIsPlayer:', 'Checking if faction is under human player control', cuFaction.UID, cuFaction.Name, IIF(Conditions = condEvaluating, 'True', 'False')]);
end;


// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// EVAL BLOCKADE PERCENTAGE
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.EvalBlockadePerc  (Params : TStringList) : integer;
var Value : integer;
begin
  Result := 0;
  if Params.Count <> 2 then
  begin
    Inc(Result);
    Log(['=> Evaluate Blockade Percentage command, syntax is: EvalBlockadePerc = <Operator>;<Value>']);
  end;
  if (cuFaction = nil) then
  begin
    Inc(Result);
    Log(['=> EvalBlockadePerc: No faction selected']);
  end;
  if Result > 0 then Exit;

  Value := STI(Params[1]);

  if Params[0] = '<' then
  begin
    if not (cuFaction.BlockadePerc < Value) then Conditions := condFalse;
  end
  else
  if Params[0] = '<=' then
  begin
    if not (cuFaction.BlockadePerc <= Value) then Conditions := condFalse;
  end
  else
  if Params[0] = '>' then
  begin
    if not (cuFaction.BlockadePerc > Value) then Conditions := condFalse;
  end
  else
  if Params[0] = '>=' then
  begin
    if not (cuFaction.BlockadePerc >= Value) then Conditions := condFalse;
  end
  else
  if Params[0] = '=' then
  begin
    if not (cuFaction.BlockadePerc = Value) then Conditions := condFalse;
  end
  else
  if UpperCase(Params[0]) = 'DICE' then
  begin
    if (Dice(Value) > cuFaction.BlockadePerc) then Conditions := condFalse;
  end;

  Log(['BlockadePerc:', cuFaction.Name, 'blockade level', cuFaction.BlockadePerc, Params[0], Params[1], ':',
  IIF(Conditions = condEvaluating, 'True', 'False')]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// EVAL LOYALTY
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.EvalLoyalty(Params : TStringList) : integer;
var Value : integer;
begin
  Result := 0;
  if Params.Count <> 2 then
  begin
    Inc(Result);
    Log(['=> Evaluate Loyalty command, syntax is: EvalLoyalty = <Operator>;<Value>']);
  end;
  if (cuFaction = nil) then
  begin
    Inc(Result);
    Log(['=> EvalLoyalty: No faction selected']);
  end;
  if (cuRegion = nil) then
  begin
    Inc(Result);
    Log(['=> EvalLoyalty: No region selected']);
  end;
  if Result > 0 then Exit;

  Value := STI(Params[1]);

  if Params[0] = '<' then
  begin
    if not (cuRegion.Loyalty < Value) then Conditions := condFalse;
  end
  else
  if Params[0] = '<=' then
  begin
    if not (cuRegion.Loyalty <= Value) then Conditions := condFalse;
  end
  else
  if Params[0] = '>' then
  begin
    if not (cuRegion.Loyalty > Value) then Conditions := condFalse;
  end
  else
  if Params[0] = '>=' then
  begin
    if not (cuRegion.Loyalty >= Value) then Conditions := condFalse;
  end
  else
  if Params[0] = '=' then
  begin
    if not (cuRegion.Loyalty = Value) then Conditions := condFalse;
  end
  else
  if UpperCase(Params[0]) = 'DICE' then
  begin
    if (Dice(Value) > cuRegion.Loyalty) then Conditions := condFalse;
  end;

  Log(['EvalLoyalty:', cuRegion.Name, 'for', cuFaction.Name, 'Test done:', cuRegion.Loyalty, Params[0], Params[1], ':',
  IIF(Conditions = condEvaluating, 'True', 'False')]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// EVAL CONTROL
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.EvalControl(Params : TStringList) : integer;
var Value : integer;
begin
  Result := 0;
  if Params.Count <> 2 then
  begin
    Inc(Result);
    Log(['=> Evaluate Control command, syntax is: EvalControl = <Operator>;<Value>']);
  end;
  if (cuFaction = nil) then
  begin
    Inc(Result);
    Log(['=> EvalControl: No faction selected']);
  end;
  if (cuRegion = nil) then
  begin
    Inc(Result);
    Log(['=> EvalControl: No region selected']);
  end;
  if Result > 0 then Exit;

  Value := STI(Params[1]);

  if Params[0] = '<' then
  begin
    if not (cuRegion.GetControlWithAllies < Value) then Conditions := condFalse;
  end
  else
  if Params[0] = '<=' then
  begin
    if not (cuRegion.GetControlWithAllies <= Value) then Conditions := condFalse;
  end
  else
  if Params[0] = '>' then
  begin
    if not (cuRegion.GetControlWithAllies > Value) then Conditions := condFalse;
  end
  else
  if Params[0] = '>=' then
  begin
    if not (cuRegion.GetControlWithAllies >= Value) then Conditions := condFalse;
  end
  else
  if Params[0] = '=' then
  begin
    if not (cuRegion.GetControlWithAllies = Value) then Conditions := condFalse;
  end
  else
  if UpperCase(Params[0]) = 'DICE' then
  begin
    if (Dice(Value) > cuRegion.GetControlWithAllies) then Conditions := condFalse;
  end;

  Log(['EvalControl (with allies):', cuRegion.Name, 'for', cuFaction.Name, 'Test done:', cuRegion.GetControlWithAllies, Params[0], Params[1], ':',
  IIF(Conditions = condEvaluating, 'True', 'False')]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// EVAL MILITANCY
// with ALL, all national regions are evaluated (average), otherwise a single region
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.EvalMilitancy(Params : TStringList) : integer;
var Value, cuValue : integer;
    All : boolean;
    ReverseResult : boolean; // only useful for DICE...
begin
  Result := 0;
  if Params.Count < 2 then
  begin
    Inc(Result);
    Log(['=> Evaluate Militancy command, syntax is: EvalMilitancy = <Operator><;Value>[;NOT][;ALL]']);
  end;
  if (cuFaction = nil) then
  begin
    Inc(Result);
    Log(['=> EvalMilitancy: No faction selected']);
  end;

  Value := STI(Params[1]);
  All := ExtractParam('ALL', Params) <> '';
  ReverseResult := ExtractParam('NOT', Params) <> '';

  if not All then
  begin
    if (cuRegion = nil) then
    begin
      Inc(Result);
      Log(['=> EvalMilitancy: No region selected']);
    end;
  end;
  if Result > 0 then Exit;

  if not All then
    cuValue := GameSim.RegionCmns[cuRegion.UID].Pop_MilitancyPerc
  else
    cuValue := cuFaction.GetAvgMilitancyPerc(True);

  if Params[0] = '<' then
  begin
    if not (cuValue < Value) then Conditions := condFalse;
  end
  else
  if Params[0] = '<=' then
  begin
    if not (cuValue <= Value) then Conditions := condFalse;
  end
  else
  if Params[0] = '>' then
  begin
    if not (cuValue > Value) then Conditions := condFalse;
  end
  else
  if Params[0] = '>=' then
  begin
    if not (cuValue >= Value) then Conditions := condFalse;
  end
  else
  if Params[0] = '=' then
  begin
    if not (cuValue = Value) then Conditions := condFalse;
  end
  else
  if UpperCase(Params[0]) = 'DICE' then
  begin
    if (Dice(Value) > cuValue) then Conditions := condFalse;
  end;

  if ReverseResult then
  begin
    if Conditions = condFalse then
      Conditions := condEvaluating
    else
      Conditions := condFalse;
  end;

  Log(['EvalMilitancy:', cuFaction.Name, 'Test done in ', IIF(All, 'All Nat. region', cuRegion.Name), ': ',
  cuValue, Params[0], Params[1], ':', IIF(Conditions = condEvaluating, 'True', 'False'), IIF(ReverseResult, '(Result was Reversed)', '')]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// EVAL CONTENTMENT
// with ALL, all national regions are evaluated (average), otherwise a single region
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.EvalContentment(Params : TStringList) : integer;
var Value, cuValue : integer;
    All : boolean;
    ReverseResult : boolean; // only useful for DICE...
begin
  Result := 0;
  if Params.Count < 2 then
  begin
    Inc(Result);
    Log(['=> Evaluate Contentment command, syntax is: EvalContentment = <Operator><;Value>[;NOT][;ALL]']);
  end;
  if (cuFaction = nil) then
  begin
    Inc(Result);
    Log(['=> EvalContentment: No faction selected']);
  end;

  Value := STI(Params[1]);
  All := ExtractParam('ALL', Params) <> '';
  ReverseResult := ExtractParam('NOT', Params) <> '';

  if not All then
  begin
    if (cuRegion = nil) then
    begin
      Inc(Result);
      Log(['=> EvalContentment: No region selected']);
    end;
  end;
  if Result > 0 then Exit;

  if not All then
    cuValue := GameSim.RegionCmns[cuRegion.UID].Pop_ContentmentPerc
  else
    cuValue := cuFaction.GetAvgContentmentPerc(True);

  if Params[0] = '<' then
  begin
    if not (cuValue < Value) then Conditions := condFalse;
  end
  else
  if Params[0] = '<=' then
  begin
    if not (cuValue <= Value) then Conditions := condFalse;
  end
  else
  if Params[0] = '>' then
  begin
    if not (cuValue > Value) then Conditions := condFalse;
  end
  else
  if Params[0] = '>=' then
  begin
    if not (cuValue >= Value) then Conditions := condFalse;
  end
  else
  if Params[0] = '=' then
  begin
    if not (cuValue = Value) then Conditions := condFalse;
  end
  else
  if UpperCase(Params[0]) = 'DICE' then
  begin
    if (Dice(Value) > cuValue) then Conditions := condFalse;
  end;

  if ReverseResult then
  begin
    if Conditions = condFalse then
      Conditions := condEvaluating
    else
      Conditions := condFalse;
  end;

  Log(['EvalContentment:', cuFaction.Name, 'Test done in ', IIF(All, 'All Nat. region', cuRegion.Name), ': ',
  cuValue, Params[0], Params[1], ':', IIF(Conditions = condEvaluating, 'True', 'False'), IIF(ReverseResult, '(Result was Reversed)', '')]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// EVAL EDUCATION
// with ALL, all national regions are evaluated (average), otherwise a single region
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.EvalEducation(Params : TStringList) : integer;
var Value, cuValue : integer;
    All : boolean;
    ReverseResult : boolean; // only useful for DICE...
begin
  Result := 0;
  if Params.Count < 2 then
  begin
    Inc(Result);
    Log(['=> Evaluate Education command, syntax is: EvalEducation = <Operator><;Value>[;NOT][;ALL]']);
  end;
  if (cuFaction = nil) then
  begin
    Inc(Result);
    Log(['=> EvalEducation: No faction selected']);
  end;

  Value := STI(Params[1]);
  All := ExtractParam('ALL', Params) <> '';
  ReverseResult := ExtractParam('NOT', Params) <> '';

  if not All then
  begin
    if (cuRegion = nil) then
    begin
      Inc(Result);
      Log(['=> EvalEducation: No region selected']);
    end;
  end;
  if Result > 0 then Exit;

  if not All then
    cuValue := GameSim.RegionCmns[cuRegion.UID].Pop_EducationPerc
  else
    cuValue := cuFaction.GetAvgEducationPerc(True);

  if Params[0] = '<' then
  begin
    if not (cuValue < Value) then Conditions := condFalse;
  end
  else
  if Params[0] = '<=' then
  begin
    if not (cuValue <= Value) then Conditions := condFalse;
  end
  else
  if Params[0] = '>' then
  begin
    if not (cuValue > Value) then Conditions := condFalse;
  end
  else
  if Params[0] = '>=' then
  begin
    if not (cuValue >= Value) then Conditions := condFalse;
  end
  else
  if Params[0] = '=' then
  begin
    if not (cuValue = Value) then Conditions := condFalse;
  end
  else
  if UpperCase(Params[0]) = 'DICE' then
  begin
    if (Dice(Value) > cuValue) then Conditions := condFalse;
  end;

  if ReverseResult then
  begin
    if Conditions = condFalse then
      Conditions := condEvaluating
    else
      Conditions := condFalse;
  end;

  Log(['EvalEducation:', cuFaction.Name, 'Test done in ', IIF(All, 'All Nat. region', cuRegion.Name), ': ',
  cuValue, Params[0], Params[1], ':', IIF(Conditions = condEvaluating, 'True', 'False'), IIF(ReverseResult, '(Result was Reversed)', '')]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// EVAL TRANSPORT LEVEL
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.EvalTransportLvl(Params : TStringList) : integer;
var Value : integer;
    RgnCmn : TRegionCmn;
    OnlyFunc, OnlyDest : boolean;
    Level : TUID;
begin
  Result := 0;
  if Params.Count <> 3 then
  begin
    Inc(Result);
    Log(['=> Evaluate Transport Level command, syntax is: EvalTransportLvl = <Operator>;<Value>;<Both / OnlyFuncRR / OnlyDestRR>']);
  end;
  if (cuFaction = nil) then
  begin
    Inc(Result);
    Log(['=> EvalControl: No faction selected']);
  end;
  if (cuRegion = nil) then
  begin
    Inc(Result);
    Log(['=> EvalControl: No region selected']);
  end;
  if Result > 0 then Exit;

  RgnCmn := GameSim.RegionCmns[cuRegion.UID];

  Value := STI(Params[1]);

  OnlyFunc := ExtractParam('OnlyFuncRR', Params) <> '';
  OnlyDest := ExtractParam('OnlyDestRR', Params) <> '';

  Level := RgnCmn.TransportLvl;
  if Level = _trsRailDestroy then
    Level := _trsRailRoad
  else
  if Level = _trsDRailDestroy then
    Level := _trsDRailRoad;

  // Rail as structure in construction  
  if (Level < _trsRailRoad) and GameSim.RegionCmns.StrucsHaveThisAttribute(cuRegion.UID, _attRailroad1) then
    Level := _trsRailRoad;
  if (Level < _trsRailRoad) and GameSim.RegionCmns.StrucsHaveThisAttribute(cuRegion.UID, _attRailroad2) then
    Level := _trsDRailRoad;

  if OnlyFunc and ((RgnCmn.TransportLvl in [_trsRailDestroy, _trsRailDestroy])
  or (GameSim.RegionCmns.StrucsHaveThisAttribute(cuRegion.UID, _attRailroad1))
  or (GameSim.RegionCmns.StrucsHaveThisAttribute(cuRegion.UID, _attRailroad2))) then
    Conditions := condFalse
  else
  if OnlyDest and (RgnCmn.TransportLvl < _trsRailDestroy) then
    Conditions := condFalse
  else
  if Params[0] = '<' then
  begin
    if not (Level < Value) then Conditions := condFalse;
  end
  else
  if Params[0] = '<=' then
  begin
    if not (Level <= Value) then Conditions := condFalse;
  end
  else
  if Params[0] = '>' then
  begin
    if not (Level > Value) then Conditions := condFalse;
  end
  else
  if Params[0] = '>=' then
  begin
    if not (Level >= Value) then Conditions := condFalse;
  end
  else
  if Params[0] = '=' then
  begin
    if not (Level = Value) then Conditions := condFalse;
  end
  else
  if UpperCase(Params[0]) = 'DICE' then
  begin
    if (Dice(Value) > Level) then Conditions := condFalse;
  end;

  Log(['EvalTransportLvl:', cuRegion.Name, 'for', cuFaction.Name, 'Test done:', Level, Params[0], Params[1], ':',
  IIF(Conditions = condEvaluating, 'True', 'False')]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// EVAL MORALE
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.EvalMorale  (Params : TStringList) : integer;
var Value : integer;
    OtherFac : TFaction;
begin
  Result := 0;
  if (Params.Count <> 2) then
  begin
    Inc(Result);
    Log(['=> Evaluate Morale command, syntax is: EvalMorale = <Operator>;<Value_or_Faction_tag>']);
  end;
  if (cuFaction = nil) then
  begin
    Inc(Result);
    Log(['=> EvalMorale: No faction selected']);
  end;
  if Result > 0 then Exit;

  OtherFac := nil;
  if (Length(Params[1]) >= 3) and (Params[1][1] in ['A'..'Z']) then
  begin
    OtherFac := FactionTagToFaction(Params[1]);
    if OtherFac <> nil then Log(['EvalMorale: Evaluating against', OtherFac.Name, 'morale.']);
  end;

  if OtherFac = nil then
    Value := STI(Params[1])
  else
    Value := OtherFac.Morale;

  if Params[0] = '<' then
  begin
    if not (cuFaction.Morale < Value) then Conditions := condFalse;
  end
  else
  if Params[0] = '<=' then
  begin
    if not (cuFaction.Morale <= Value) then Conditions := condFalse;
  end
  else
  if Params[0] = '>' then
  begin
    if not (cuFaction.Morale > Value) then Conditions := condFalse;
  end
  else
  if Params[0] = '>=' then
  begin
    if not (cuFaction.Morale >= Value) then Conditions := condFalse;
  end
  else
  if Params[0] = '=' then
  begin
    if not (cuFaction.Morale = Value) then Conditions := condFalse;
  end
  else
  if UpperCase(Params[0]) = 'DICE' then
  begin
    if (Dice(Value) > cuFaction.Morale) then Conditions := condFalse;
  end;

  Log(['EvalMorale:', cuFaction.Name, 'Morale:', cuFaction.Morale, 'against', Value, ' - Operator:', Params[0], 'Result:',
  IIF(Conditions = condEvaluating, 'True', 'False')]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// EVAL MONEY
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.EvalMoney  (Params : TStringList) : integer;
var Value : integer;
    OtherFac : TFaction;
begin
  Result := 0;
  if (Params.Count <> 2) then
  begin
    Inc(Result);
    Log(['=> Evaluate Money command, syntax is: EvalMoney = <Operator>;<Value_or_Faction_tag>']);
  end;
  if (cuFaction = nil) then
  begin
    Inc(Result);
    Log(['=> EvalMoney: No faction selected']);
  end;
  if Result > 0 then Exit;

  OtherFac := nil;
  if (Length(Params[1]) >= 3) and (Params[1][1] in ['A'..'Z']) then
  begin
    OtherFac := FactionTagToFaction(Params[1]);
    if OtherFac <> nil then Log(['EvalMoney: Evaluating against', OtherFac.Name, 'Money.']);
  end;

  if OtherFac = nil then
    Value := STI(Params[1])
  else
    Value := OtherFac.Money;

  if Params[0] = '<' then
  begin
    if not (cuFaction.Money < Value) then Conditions := condFalse;
  end
  else
  if Params[0] = '<=' then
  begin
    if not (cuFaction.Money <= Value) then Conditions := condFalse;
  end
  else
  if Params[0] = '>' then
  begin
    if not (cuFaction.Money > Value) then Conditions := condFalse;
  end
  else
  if Params[0] = '>=' then
  begin
    if not (cuFaction.Money >= Value) then Conditions := condFalse;
  end
  else
  if Params[0] = '=' then
  begin
    if not (cuFaction.Money = Value) then Conditions := condFalse;
  end
  else
  if UpperCase(Params[0]) = 'DICE' then
  begin
    if (Dice(Value) > cuFaction.Money) then Conditions := condFalse;
  end;

  Log(['EvalMoney:', cuFaction.Name, 'Money:', cuFaction.Money, 'against', Value, ' - Operator:', Params[0], 'Result:',
  IIF(Conditions = condEvaluating, 'True', 'False')]);
end;


// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// EVAL BATTLES
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.EvalBattles(Params : TStringList) : integer;
var NoOfBattles : integer;
    OperatorBtl : string;
    OperatorNM  : string;
    ValueNM     : integer;
    i           : integer;
    BtlCounter  : integer;

begin
  Result := 0;
  if (Params.Count <> 4) then
  begin
    Inc(Result);
    Log(['=> Evaluate Battles command, syntax is: EvalBattles = <Operator>;<No. of Battles>;<Operator>;<Value>']);
  end;
  if (cuFaction = nil) then
  begin
    Inc(Result);
    Log(['=> EvalBattles: No faction selected']);
  end;
  if Result > 0 then Exit;

  OperatorBtl := Params[0];
  NoOfBattles := STI(Params[1]);
  OperatorNM  := Params[2];
  ValueNM     := STI(Params[3]);
  BtlCounter  := 0;

  if OperatorNM = '<' then
  begin
    for i := 0 to cuFaction.BattlesNMChg.Count -1 do
      if cuFaction.BattlesNMChg.Items[i] - ValueNM < 0 then inc(BtlCounter);
  end
  else
  if OperatorNM = '<=' then
  begin
    for i := 0 to cuFaction.BattlesNMChg.Count -1 do
      if cuFaction.BattlesNMChg.Items[i] - ValueNM <= 0 then inc(BtlCounter);
  end
  else
  if OperatorNM = '=' then
  begin
    for i := 0 to cuFaction.BattlesNMChg.Count -1 do
      if cuFaction.BattlesNMChg.Items[i] - ValueNM = 0 then inc(BtlCounter);
  end
  else
    if OperatorNM = '>=' then
  begin
    for i := 0 to cuFaction.BattlesNMChg.Count -1 do
      if cuFaction.BattlesNMChg.Items[i] - ValueNM >= 0 then inc(BtlCounter);
  end
  else
    if OperatorNM = '>' then
  begin
    for i := 0 to cuFaction.BattlesNMChg.Count -1 do
      if cuFaction.BattlesNMChg.Items[i] - ValueNM > 0 then inc(BtlCounter);
  end;

  if OperatorBtl = '<' then
  begin
    if not (BtlCounter - NoOfBattles < 0)  then Conditions := condFalse;
  end
  else
  if OperatorBtl = '<=' then
  begin
    if not (BtlCounter - NoOfBattles <= 0)  then Conditions := condFalse;
  end
  else
    if OperatorBtl = '=' then
  begin
    if not (BtlCounter - NoOfBattles = 0)  then Conditions := condFalse;
  end
  else
    if OperatorBtl = '>=' then
  begin
    if not (BtlCounter - NoOfBattles >= 0)  then Conditions := condFalse;
  end
  else
    if OperatorBtl = '>' then
  begin
    if not (BtlCounter - NoOfBattles > 0)  then Conditions := condFalse;
  end;

  Log(['EvalBattles:', cuFaction.Name , OperatorBtl, NoOfBattles, 'Battles with an NM change of ',OperatorNM, ValueNM, 'Result:', IIF(Conditions = condEvaluating, 'True', 'False')]);
end;


// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// EVAL SUB UNIT COUNT
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.EvalSubUnitCount(Params : TStringList) : integer;
var Value : integer;
begin
  Result := 0;
  if (Params.Count > 2) then
  begin
    Inc(Result);
    Log(['=> Evaluate SubUnit Count command, syntax is: EvalSubUnitCount = <Operator>;<Value>']);
  end;
  if Result > 0 then Exit;

  if Params.Count > 1 then
    Value := STI(Params[1])
  else
    Value := 100; // DICE or DICE_NOT

  if Params[0] = '<' then
  begin
    if not (cuSUList.Count < Value) then Conditions := condFalse;
  end
  else
  if Params[0] = '<=' then
  begin
    if not (cuSUList.Count <= Value) then Conditions := condFalse;
  end
  else
  if Params[0] = '>' then
  begin
    if not (cuSUList.Count > Value) then Conditions := condFalse;
  end
  else
  if Params[0] = '>=' then
  begin
    if not (cuSUList.Count >= Value) then Conditions := condFalse;
  end
  else
  if Params[0] = '=' then
  begin
    if not (cuSUList.Count = Value) then Conditions := condFalse;
  end
  else
  if UpperCase(Params[0]) = 'DICE' then
  begin
    if (Dice(Value) > cuSUList.Count) then Conditions := condFalse;
  end;
  if UpperCase(Params[0]) = 'DICE_NOT' then
  begin
    if (Dice(Value) < cuSUList.Count) then Conditions := condFalse;
  end;
  Log(['EvalSubUnitCount:', 'Selected SUs:', cuSUList.Count, 'against', Value, ' - Operator:', Params[0], 'Result:',
  IIF(Conditions = condEvaluating, 'True', 'False')]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// EVAL ENGAGEMENT PTS
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.EvalEngagementPts(Params : TStringList) : integer;
var Value : integer;
    OtherFac : TFaction;
begin
  Result := 0;
  if rulEngagementPts_ = _eptNoEngagementPts then
  begin
    Inc(Result);
    Log(['=> Evaluate EngagementPts command, game rule not activated, command ignored']);
  end;
  if (Params.Count <> 2) then
  begin
    Inc(Result);
    Log(['=> Evaluate EngagementPts command, syntax is: EvalEngagementPts = <Operator>;<Value_or_Faction_tag>']);
  end;
  if (cuFaction = nil) then
  begin
    Inc(Result);
    Log(['=> EvalEngagementPts: No faction selected']);
  end;
  if Result > 0 then Exit;

  OtherFac := nil;
  if (Length(Params[1]) >= 3) and (Params[1][1] in ['A'..'Z']) then
  begin
    OtherFac := FactionTagToFaction(Params[1]);
    if OtherFac <> nil then Log(['EvalEngagementPts: Evaluating against', OtherFac.Name, 'Engagement Points.']);
  end;

  if OtherFac = nil then
    Value := STI(Params[1])
  else
    Value := OtherFac.EngagementPts;

  if Params[0] = '<' then
  begin
    if not (cuFaction.EngagementPts < Value) then Conditions := condFalse;
  end
  else
  if Params[0] = '<=' then
  begin
    if not (cuFaction.EngagementPts <= Value) then Conditions := condFalse;
  end
  else
  if Params[0] = '>' then
  begin
    if not (cuFaction.EngagementPts > Value) then Conditions := condFalse;
  end
  else
  if Params[0] = '>=' then
  begin
    if not (cuFaction.EngagementPts >= Value) then Conditions := condFalse;
  end
  else
  if Params[0] = '=' then
  begin
    if not (cuFaction.EngagementPts = Value) then Conditions := condFalse;
  end
  else
  if UpperCase(Params[0]) = 'DICE' then
  begin
    if (Dice(Value) > cuFaction.EngagementPts) then Conditions := condFalse;
  end;

  Log(['EvalEngagementPts:', cuFaction.Name, 'EngagementPts:', cuFaction.EngagementPts, 'against', Value, ' - Operator:', Params[0], 'Result:',
  IIF(Conditions = condEvaluating, 'True', 'False')]);
end;


// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// EVAL RESOURCE
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.EvalResource(Params : TStringList) : integer;
var Value : integer;
    OtherFac : TFaction;
    ResUID : TUID;
begin
  Result := 0;
  if (Params.Count <> 3) then
  begin
    Inc(Result);
    Log(['=> Evaluate Resource command, syntax is: EvalResource = <ResUID>;<Operator>;<Value_or_Faction_tag>']);
  end;
  if (cuFaction = nil) then
  begin
    Inc(Result);
    Log(['=> Evaluate Resource command: No faction selected']);
  end;
  if Result > 0 then Exit;

  ResUID := STI(Params[0]);

  OtherFac := nil;
  if (Length(Params[2]) >= 3) and (Params[2][1] in ['A'..'Z']) then
  begin
    OtherFac := FactionTagToFaction(Params[2]);
    if OtherFac <> nil then Log(['EvalResource: Evaluating against', OtherFac.Name, 'Resource.']);
  end;

  if OtherFac = nil then
    Value := STI(Params[2])
  else
    Value := OtherFac.GetResource(ResUID);

  if Params[1] = '=' then
  begin
    if not (cuFaction.GetResource(ResUID) <> Value) then Conditions := condFalse;
  end
  else
  if Params[1] = '<' then
  begin
    if not (cuFaction.GetResource(ResUID) < Value) then Conditions := condFalse;
  end
  else
  if Params[1] = '<=' then
  begin
    if not (cuFaction.GetResource(ResUID) <= Value) then Conditions := condFalse;
  end
  else
  if Params[1] = '>' then
  begin
    if not (cuFaction.GetResource(ResUID) > Value) then Conditions := condFalse;
  end
  else
  if Params[1] = '>=' then
  begin
    if not (cuFaction.GetResource(ResUID) >= Value) then Conditions := condFalse;
  end
  else
  if UpperCase(Params[1]) = 'DICE' then
  begin
    if (Dice(Value) > cuFaction.GetResource(ResUID)) then Conditions := condFalse;
  end;

  Log(['EvalResource:', cuFaction.Name, 'Resources:', cuFaction.GetResource(ResUID), 'against', Value, ' - Operator:', Params[0], 'Result:',
  IIF(Conditions = condEvaluating, 'True', 'False')]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// EVAL CONSCRIPT
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.EvalConscript  (Params : TStringList) : integer;
var Value : integer;
    OtherFac : TFaction;
begin
  Result := 0;
  if (Params.Count <> 2) then
  begin
    Inc(Result);
    Log(['=> Evaluate Conscript command, syntax is: EvalConscript = <Operator>;<Value_or_Faction_tag>']);
  end;
  if (cuFaction = nil) then
  begin
    Inc(Result);
    Log(['=> EvalConscript: No faction selected']);
  end;
  if Result > 0 then Exit;

  OtherFac := nil;
  if (Length(Params[1]) >= 3) and (Params[1][1] in ['A'..'Z']) then
  begin
    OtherFac := FactionTagToFaction(Params[1]);
    if OtherFac <> nil then Log(['EvalConscript: Evaluating against', OtherFac.Name, 'Conscript.']);
  end;

  if OtherFac = nil then
    Value := STI(Params[1])
  else
    Value := OtherFac.Conscript;

  if Params[0] = '<' then
  begin
    if not (cuFaction.Conscript < Value) then Conditions := condFalse;
  end
  else
  if Params[0] = '<=' then
  begin
    if not (cuFaction.Conscript <= Value) then Conditions := condFalse;
  end
  else
  if Params[0] = '>' then
  begin
    if not (cuFaction.Conscript > Value) then Conditions := condFalse;
  end
  else
  if Params[0] = '>=' then
  begin
    if not (cuFaction.Conscript >= Value) then Conditions := condFalse;
  end
  else
  if Params[0] = '=' then
  begin
    if not (cuFaction.Conscript = Value) then Conditions := condFalse;
  end
  else
  if UpperCase(Params[0]) = 'DICE' then
  begin
    if (Dice(Value) > cuFaction.Conscript) then Conditions := condFalse;
  end;

  Log(['EvalConscript:', cuFaction.Name, 'Conscript:', cuFaction.Conscript, 'against', Value, ' - Operator:', Params[0], 'Result:',
  IIF(Conditions = condEvaluating, 'True', 'False')]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// EVAL REGION WEATHER
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.EvalRgnWeather(Params : TStringList) : integer;
var Value, RgnUID : integer;
    WeatherUID : TUID;
    Operator : string;
begin
  Result := 0;
  if (Params.Count <> 3) then
  begin
    Inc(Result);
    Log(['=> Evaluate Region Weahter command, syntax is: EvalRgnWeather = <RegionUID>;<Operator>;<Value>']);
  end;
  if Result > 0 then Exit;

  RgnUID := STI(Params[0]);
  WeatherUID := GameSim.Weather.GetWeather(RgnUID);
  Operator := Params[1];
  Value := STI(Params[2]);

  if Operator = '<' then
  begin
    if not (WeatherUID < Value) then Conditions := condFalse;
  end
  else
  if Operator = '<=' then
  begin
    if not (WeatherUID <= Value) then Conditions := condFalse;
  end
  else
  if Operator = '>' then
  begin
    if not (WeatherUID > Value) then Conditions := condFalse;
  end
  else
  if Operator = '>=' then
  begin
    if not (WeatherUID >= Value) then Conditions := condFalse;
  end
  else
  if Operator = '=' then
  begin
    if not (WeatherUID = Value) then Conditions := condFalse;
  end
  else
  if UpperCase(Operator) = 'DICE' then
  begin
    if (Dice(Value) > WeatherUID) then Conditions := condFalse;
  end;

  Log(['EvalRgnWeather:', 'Weather in region', RgnUID, RegionDefs[RgnUID].Name, 'against', Value, ' - Operator:', Params[1], 'Result:',
  IIF(Conditions = condEvaluating, 'True', 'False')]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// EVAL WAR SUPPLY
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.EvalWarSupply(Params : TStringList) : integer;
var Value : integer;
    OtherFac : TFaction;
begin
  Result := 0;
  if (Params.Count <> 2) then
  begin
    Inc(Result);
    Log(['=> Evaluate War Supply command, syntax is: EvalWSU = <Operator>;<Value_or_Faction_tag>']);
  end;
  if (cuFaction = nil) then
  begin
    Inc(Result);
    Log(['=> EvalWSU: No faction selected']);
  end;
  if Result > 0 then Exit;

  OtherFac := nil;
  if (Length(Params[1]) >= 3) and (Params[1][1] in ['A'..'Z']) then
  begin
    OtherFac := FactionTagToFaction(Params[1]);
    if OtherFac <> nil then Log(['EvalConscript: Evaluating against', OtherFac.Name, 'WSU.']);
  end;

  if OtherFac = nil then
    Value := STI(Params[1])
  else
    Value := OtherFac.WarSupply;

  if Params[0] = '<' then
  begin
    if not (cuFaction.WarSupply < Value) then Conditions := condFalse;
  end
  else
  if Params[0] = '<=' then
  begin
    if not (cuFaction.WarSupply <= Value) then Conditions := condFalse;
  end
  else
  if Params[0] = '>' then
  begin
    if not (cuFaction.WarSupply > Value) then Conditions := condFalse;
  end
  else
  if Params[0] = '>=' then
  begin
    if not (cuFaction.WarSupply >= Value) then Conditions := condFalse;
  end
  else
  if Params[0] = '=' then
  begin
    if not (cuFaction.WarSupply = Value) then Conditions := condFalse;
  end
  else
  if UpperCase(Params[0]) = 'DICE' then
  begin
    if (Dice(Value) > cuFaction.WarSupply) then Conditions := condFalse;
  end;

  Log(['EvalWSU:', cuFaction.Name, 'WSU:', cuFaction.WarSupply, 'against', Value, ' - Operator:', Params[0], 'Result:',
  IIF(Conditions = condEvaluating, 'True', 'False')]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// EVAL VP
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.EvalVP(Params : TStringList) : integer;
var Value : integer;
    OtherFac : TFaction;
begin
  Result := 0;
  if (Params.Count <> 2) then
  begin
    Inc(Result);
    Log(['=> Evaluate Victory Points command, syntax is: EvalVP = <Operator>;<Value_or_Faction_tag>']);
  end;
  if (cuFaction = nil) then
  begin
    Inc(Result);
    Log(['=> EvalVP: No faction selected']);
  end;
  if Result > 0 then Exit;

  OtherFac := nil;
  if (Length(Params[1]) >= 3) and (Params[1][1] in ['A'..'Z']) then
  begin
    OtherFac := FactionTagToFaction(Params[1]);
    if OtherFac <> nil then Log(['EvalVP: Evaluating against', OtherFac.Name, 'VPs.']);
  end;

  if OtherFac = nil then
    Value := STI(Params[1])
  else
    Value := OtherFac.VPTotal;

  if Params[0] = '<' then
  begin
    if not (cuFaction.VPTotal < Value) then Conditions := condFalse;
  end
  else
  if Params[0] = '<=' then
  begin
    if not (cuFaction.VPTotal <= Value) then Conditions := condFalse;
  end
  else
  if Params[0] = '>' then
  begin
    if not (cuFaction.VPTotal > Value) then Conditions := condFalse;
  end
  else
  if Params[0] = '>=' then
  begin
    if not (cuFaction.VPTotal >= Value) then Conditions := condFalse;
  end
  else
  if Params[0] = '=' then
  begin
    if not (cuFaction.VPTotal = Value) then Conditions := condFalse;
  end
  else
  if UpperCase(Params[0]) = 'DICE' then
  begin
    if (Dice(Value) > cuFaction.VPTotal) then Conditions := condFalse;
  end;

  Log(['EvalVP:', cuFaction.Name, 'VP:', cuFaction.WarSupply, 'against', Value, ' - Operator:', Params[0], 'Result:',
  IIF(Conditions = condEvaluating, 'True', 'False')]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// EVAL FOREIGN ENTRY
// TODO : "EvalIntValue" on pourrait valuer n'importe quelle valeur numrique de faction assez facilement
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.EvalForeignEntry (Params : TStringList) : integer;
var cuValue, OthValue : integer;
    OtherFac, NmyFac : TFaction;
begin
  Result := 0;
  if Params.Count <> 2 then
  begin
    Inc(Result);
    Log(['=> Evaluate Foreign Entry command, syntax is: EvalForeignEntry = <Operator>;<Value_or_Faction_tag>']);
  end;
  if (cuFaction = nil) then
  begin
    Inc(Result);
    Log(['=> EvalForeignEntry: No faction selected']);
  end;
  if Result > 0 then Exit;
  if (cuFaction.FactionDef.ForeignEntryKind = _foeDisallowed) then
  begin
    Inc(Result);
    Log(['=> EvalForeignEntry: The faction do not use Foreign Entry']);
  end;
  if Result > 0 then Exit;

  if (Length(Params[1]) = 3) and (UpperCase(Params[1]) = 'MAX') then
  begin
    OthValue := cuFaction.GetForeignEntryUpperBound;
  end
  else
  if (Length(Params[1]) >= 3) and (Params[1][1] in ['A'..'Z']) then
  begin
    OtherFac := FactionTagToFaction(Params[1]);
    if OtherFac <> nil then Log(['EvalForeignEntry: Evaluating against', OtherFac.Name, 'ForeignEntry.']);
    if OtherFac.FactionDef.ForeignEntryKind <> _foeAllowed then
    begin
      Inc(Result);
      Log(['=> EvalForeignEntry: You are evaluating against a faction which either do not use FE, or is already a Counterer.']);
      Exit;
    end;
    OthValue := OtherFac.ForeignEntry;
  end
  else
    OthValue := STI(Params[1]);

  if cuFaction.FactionDef.ForeignEntryKind = _foeCounterer then
  begin
    NmyFac := cuFaction.pStrongEnemyFac;
    if NmyFac = nil then
    begin
      Log(['[WARNING] EvalForeignEntry: (not an error) You evaluate a Counterer faction which has no current enemy, exiting evaluation.']);
      Exit;
    end;
    cuValue := NmyFac.ForeignEntry;
  end
  else
    cuValue := cuFaction.ForeignEntry;

  if Params[0] = '<' then
  begin
    if not (cuValue < OthValue) then Conditions := condFalse;
  end
  else
  if Params[0] = '<=' then
  begin
    if not (cuValue <= OthValue) then Conditions := condFalse;
  end
  else
  if Params[0] = '>' then
  begin
    if not (cuValue > OthValue) then Conditions := condFalse;
  end
  else
  if Params[0] = '>=' then
  begin
    if not (cuValue >= OthValue) then Conditions := condFalse;
  end
  else
  if Params[0] = '=' then
  begin
    if not (cuValue = OthValue) then Conditions := condFalse;
  end
  else
  if UpperCase(Params[0]) = 'DICE' then
  begin
    if (Dice(OthValue) > cuValue) then Conditions := condFalse;
  end;

  Log(['EvalForeignEntry:', cuFaction.Name, 'ForeignEntry:', cuValue, 'against', OthValue, ' - Operator:', Params[0], 'Result:',
  IIF(Conditions = condEvaluating, 'True', 'False')]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// EVAL REGION / FACTION / GROUP / UNIT SELECTED
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.EvalRegionSel  (Params : TStringList) : integer;
var ReverseResult : boolean;
begin
  Result := 0;
  ReverseResult := ExtractParam('NOT', Params) <> '';
  if (not ReverseResult) and (cuRegion = nil) then Conditions := condFalse;
  if (ReverseResult) and (cuRegion <> nil) then Conditions := condFalse;
  Log(['Evaluate Region Selected :', IIF(Conditions = condEvaluating, 'True', 'False')]);
end;

function TScriptEngine.EvalFactionSel (Params : TStringList) : integer;
var ReverseResult : boolean;
begin
  Result := 0;
  ReverseResult := ExtractParam('NOT', Params) <> '';
  if (not ReverseResult) and (cuFaction = nil) then Conditions := condFalse;
  if (ReverseResult) and (cuFaction <> nil) then Conditions := condFalse;
  Log(['Evaluate Faction Selected :', IIF(Conditions = condEvaluating, 'True', 'False')]);
end;

function TScriptEngine.EvalGroupSel   (Params : TStringList) : integer;
var ReverseResult : boolean;
begin
  Result := 0;
  ReverseResult := ExtractParam('NOT', Params) <> '';
  if (not ReverseResult) and (cuGroup = nil) then Conditions := condFalse;
  if (ReverseResult) and (cuGroup <> nil) then Conditions := condFalse;
  Log(['Evaluate Froup Selected :', IIF(Conditions = condEvaluating, 'True', 'False')]);
end;

function TScriptEngine.EvalUnitSel    (Params : TStringList) : integer;
var ReverseResult : boolean;
begin
  Result := 0;
  ReverseResult := ExtractParam('NOT', Params) <> '';
  if (not ReverseResult) and (cuUnit = nil) then Conditions := condFalse;
  if (ReverseResult) and (cuUnit <> nil) then Conditions := condFalse;
  Log(['Evaluate Unit Selected :', IIF(Conditions = condEvaluating, 'True', 'False')]);
end;

function TScriptEngine.EvalSubUnitSel    (Params : TStringList) : integer;
var ReverseResult : boolean;
begin
  Result := 0;
  ReverseResult := ExtractParam('NOT', Params) <> '';
  if (not ReverseResult) and (cuSubUnit = nil) then Conditions := condFalse;
  if (ReverseResult) and (cuSubUnit <> nil) then Conditions := condFalse;
  Log(['Evaluate Sub Unit Selected :', IIF(Conditions = condEvaluating, 'True', 'False')]);
end;


// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// EVAL TECH KNOWN 
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.EvalTechKnown(Params : TStringList) : integer;
var Value    : integer;
    OtherFac : TFaction;
    DiscPerc : integer;
    RI_UID   : integer;

begin
  Result := 0;

  {Check for errors}
  if rulResearches_ = False then
  begin
    Inc(Result);
    Log(['=> EvalTechKnown: Game rule not activated, command ignored']);
    Exit;
  end;
  if (Params.Count <> 3) then
  begin
    Inc(Result);
    Log(['=> EvalTechKnown: Syntax is: EvalTechKnown = <RI_UID>;<Operator>;<Value_or_Faction_tag>']);
  end;
  if (cuFaction = nil) then
  begin
    Inc(Result);
    Log(['=> EvalTechKnown: No faction selected']);
  end;
  if not IsNumeric(Params[0]) then
  begin
    Log(['=> EvalTechKnown: Parameter 1 is not a number. Check if correct Alias used.']);
    Inc(Result);
  end;
  if Result > 0 then Exit;

  RI_UID := STI(Params[0]);

  {Get own Discovery Percentage}
  if cuFaction.olResearch.IsMastered(RI_UID) then
    DiscPerc := 100
  else
  if cuFaction.olResearch.IsKnown(RI_UID) then
    DiscPerc := cuFaction.olResearch.GetByUID(RI_UID).Discovery
  else
    DiscPerc := -999; {Tech not known}

  {Check if faction vs. faction comparism}
  OtherFac := nil;
  if (Length(Params[2]) >= 3) and (Params[2][1] in ['A'..'Z']) then
  begin
    OtherFac := FactionTagToFaction(Params[2]);
    if OtherFac <> nil then Log(['EvalTechKnown: Evaluating against', OtherFac.Name]);
  end;

  {Get Discovery Percentage of other Faction if a faction vs. faction comparism is done}
  if OtherFac = nil then
    Value := STI(Params[2])
  else
  begin
    if OtherFac.olResearch.IsMastered(RI_UID) then
      Value := 100
    else
    if OtherFac.olResearch.IsKnown(RI_UID) then
      Value := OtherFac.olResearch.GetByUID(RI_UID).Discovery
    else
      Value := -999; {Tech not known}
  end;

  {Compare}
  if Params[1] = '<' then
  begin
    if not (DiscPerc < Value) then Conditions := condFalse;
  end
  else
  if Params[1] = '<=' then
  begin
    if not (DiscPerc <= Value) then Conditions := condFalse;
  end
  else
  if Params[1] = '>' then
  begin
    if not (DiscPerc > Value) then Conditions := condFalse;
  end
  else
  if Params[1] = '>=' then
  begin
    if not (DiscPerc >= Value) then Conditions := condFalse;
  end
  else
  if Params[1] = '=' then
  begin
    if not (DiscPerc = Value) then Conditions := condFalse;
  end
  else
  if UpperCase(Params[1]) = 'DICE' then
  begin
    if (Dice(Value) > DiscPerc) then Conditions := condFalse;
  end;

  Log(['EvalTechKnown:', cuFaction.Name, 'RI UID', RI_UID, 'Discovery Percentage:', DiscPerc, 'against', Value, ' - Operator:', Params[0], 'Result:', IIF(Conditions = condEvaluating, 'True', 'False')]);
end;


// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ACTOR ENABLED
// Test sur la dfinition de l'option dans la faction
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.ActorEnabled(Params : TStringList) : integer;
var ActorUID  : TUID;
    SubType   : string;
begin
  Result := 0;
  if not (Params.Count in [1..2]) then
  begin
    Inc(Result);
    Log(['=> Actor Enabled, syntax is: ActorEnabled = GameActorUID [;SubType]']);
  end;
  if (cuFaction = nil) then
  begin
    Inc(Result);
    Log(['=> ActorEnabled: No faction selected']);
  end;
  ActorUID := STI(Params[0]);

  if Params.Count = 2 then
    SubType := Params[1]
  else
    SubType  := _gmaNULL;
  if (ActorUID < _gmaMinUID)or(ActorUID > _gmaMaxUID) then
  begin
    Inc(Result);
    Log(['=> ActorEnabled: Erroneous value for the Actor UID, the correct range is', _gmaMinUID, 'to', _gmaMaxUID, 'and the parameter was:', ActorUID, 'if you get a 0, then the alias can be incorrect.']);
  end;
  if cuFaction.GetActorPool(ActorUID, SubType).MaxUse < 1 then Conditions := condFalse;
  Log(['ActorEnabled :', IIF(Conditions = condEvaluating, 'True', 'False')]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ACTOR DISABLED
// Test sur la dfinition de l'option dans la faction
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.ActorDisabled(Params : TStringList) : integer;
var ActorUID  : TUID;
    SubType   : string;
begin
  Result := 0;
  if not (Params.Count in [1..2]) then
  begin
    Inc(Result);
    Log(['=> Actor Disabled, syntax is: ActorDisabled = GameActorUID [;SubType]']);
  end;
  if (cuFaction = nil) then
  begin
    Inc(Result);
    Log(['=> ActorDisabled: No faction selected']);
  end;
  ActorUID := STI(Params[0]);
  if Params.Count = 2 then
    SubType := Params[1]
  else
    SubType  := _gmaNULL;
  if (ActorUID < _gmaMinUID)or(ActorUID > _gmaMaxUID) then
  begin
    Inc(Result);
    Log(['=> ActorDisabled: Erroneous value for the Actor UID, the correct range is', _gmaMinUID, 'to', _gmaMaxUID, 'and the parameter was:', ActorUID, 'if you get a 0, then the alias can be incorrect.']);
  end;

  if cuFaction.GetActorPool(ActorUID, SubType).MaxUse > 0 then Conditions := condFalse;
  Log(['ActorDisabled :', IIF(Conditions = condEvaluating, 'True', 'False')]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ACTOR UNDER USE
// Test sur l'usage de l'acteur par le joueur
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.ActorUnderUse(Params : TStringList) : integer;
var ActorUID, Req, Proc : TUID;
begin
  Result := 0;
  if not (Params.Count in [1..2]) then
  begin
    Inc(Result);
    Log(['=> Actor Under Use, syntax is: ActorUnderUse = GameActorUID [;NOT]']);
  end;
  if (cuFaction = nil) then
  begin
    Inc(Result);
    Log(['=> ActorUnderUse: No faction selected']);
  end;
  ActorUID := STI(Params[0]);
  if (ActorUID < _gmaMinUID)or(ActorUID > _gmaMaxUID) then
  begin
    Inc(Result);
    Log(['=> ActorUnderUse: Erroneous value for the Actor UID, the correct range is', _gmaMinUID, 'to', _gmaMaxUID, 'and the parameter was:', ActorUID, 'if you get a 0, then the alias can be incorrect.']);
  end;

  cuFaction.GetActorsGroupInfos([ActorUID], Req, Proc);

  if (Proc =  _noUID)and(ExtractParam('NOT', Params) =  '') then Conditions := condFalse;
  if (Proc <> _noUID)and(ExtractParam('NOT', Params) <> '') then Conditions := condFalse;

  Log(['ActorUnderUse :', IIF(Conditions = condEvaluating, 'True', 'False')]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// CHECK OPTION
// option: the 'instance' or 'usage' of a complex actor by a faction
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.CheckOption(Params : TStringList) : integer;
var ActorUID, Value : TUID;
    OptionUID, Operator : string;
    CA : TComplexActor;

begin
  Result := 0;
  if (Params.Count <> 4) then
  begin
    Inc(Result);
    Log(['=> Check Option, syntax is: CheckOption = GameActorUID ; OptionUID ; Operator (=, >=, <=, >, >) ; Value']);
  end;
  if (cuFaction = nil) then
  begin
    Inc(Result);
    Log(['=> CheckOption: No faction selected']);
  end;
  ActorUID  :=  STI(Params[0]);
  OptionUID := Params[1];
  Operator  := Params[2];
  Value     := STI(Params[3]);

  if (ActorUID < _gmaMinUID)or(ActorUID > _gmaMaxUID) then
  begin
    Inc(Result);
    Log(['=> CheckOption: Erroneous value for the Actor UID, the correct range is', _gmaMinUID, 'to', _gmaMaxUID, 'and the parameter was:', ActorUID, 'if you get a 0, then the alias can be incorrect.']);
  end;

  if Result > 0 then Exit;

  CA := cuFaction.GetComplexActorBuilt(ActorUID, OptionUID, _gmaNULL);
  if Operator = '=' then
  begin
    if (Value <> CA.Requested) then Conditions := condFalse;
    Log(['CheckOption evaluated:', OptionUID, 'Checking if', Value, 'equal to (request by player or AI)', CA.Requested, IIF(Conditions = condEvaluating, 'True', 'False')]);
    Exit;
  end
  else
  if Operator = '<=' then
  begin
    if (CA.Requested > Value) then Conditions := condFalse;
    Log(['CheckOption evaluated:', OptionUID, 'Checking if', Value, 'inferior or equal to (request by player or AI)', CA.Requested, IIF(Conditions = condEvaluating, 'True', 'False')]);
    Exit;
  end
  else
  if Operator = '>=' then
  begin
    if (CA.Requested < Value) then Conditions := condFalse;
    Log(['CheckOption evaluated:', OptionUID, 'Checking if', Value, 'superior or equal to (request by player or AI)', CA.Requested, IIF(Conditions = condEvaluating, 'True', 'False')]);
    Exit;
  end
  else
  if Operator = '<' then
  begin
    if (CA.Requested >= Value) then Conditions := condFalse;
    Log(['CheckOption evaluated:', OptionUID, 'Checking if', Value, 'inferior or equal to (request by player or AI)', CA.Requested, IIF(Conditions = condEvaluating, 'True', 'False')]);
    Exit;
  end
  else
  if Operator = '>' then
  begin
    if (CA.Requested <= Value) then Conditions := condFalse;
    Log(['CheckOption evaluated:', OptionUID, 'Checking if', Value, 'superior or equal to (request by player or AI)', CA.Requested, IIF(Conditions = condEvaluating, 'True', 'False')]);
    Exit;
  end;
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// CAN PAY OPTION
// Check if a faction can pay an option
// Coeff is used to see if you could pay twice, thrice the cost (to have some leftover margin for the AI e.g)
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.CanPayOption(Params : TStringList) : integer;
var ActorUID  : TUID;
    SubType   : string;
    APR : TActorPoolRec;
    Coeff : integer;
begin
  Result := 0;
  if not (Params.Count in [1..3]) then
  begin
    Inc(Result);
    Log(['=> CanPayOption, syntax is: CanPayOption = <GameActorUID> [;<SubType>] [;<Coeff>]']);
  end;
  if (cuFaction = nil) then
  begin
    Inc(Result);
    Log(['=> CanPayOption: No faction selected']);
  end;
  ActorUID := STI(Params[0]);

  if Params.Count >= 2 then
    SubType := Params[1]
  else
    SubType  := _gmaNULL;

  if Params.Count >= 3 then
    Coeff := STI(Params[2])
  else
    Coeff := 100;

  if (ActorUID < _gmaMinUID)or(ActorUID > _gmaMaxUID) then
  begin
    Inc(Result);
    Log(['=> CanPayOption: Erroneous value for the Actor UID, the correct range is', _gmaMinUID, 'to', _gmaMaxUID, 'and the parameter was:', ActorUID, 'if you get a 0, then the alias can be incorrect.']);
  end;

  APR := cuFaction.GetActorPool(ActorUID, SubType);

  if IsZeroAssetCost(APR.Cost) then
  begin
    Log(['[WARNING] CanPayOption: Actor Pool not found or is a zero cost actor, so suspicious.']);
    Exit;
  end;

  MultiplyAssets(APR.Cost, Coeff/100, False);

  if cuFaction.CanPayAssets(APR.Cost) then
  begin

    Log(['CanPayOption: ', APR.Title, 'can be paid by faction', cuFaction.UID, cuFaction.Name,
    'Faction assets are:', AssetCostToString(ResourcesListToAssetCost(cuFaction.dlResourcesStock)),
    'Option costs are:', AssetCostToString(APR.Cost), 'Pay Coeff', Coeff]);
  end
  else
  begin
    Conditions := condFalse;
    Log(['CanPayOption: ', APR.Title, 'can''t be paid by faction', cuFaction.UID, cuFaction.Name,
    'Faction assets are:', AssetCostToString(ResourcesListToAssetCost(cuFaction.dlResourcesStock)),
    'Option costs are:', AssetCostToString(APR.Cost), 'Pay Coeff', Coeff]);
  end;


end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// CHANGE OPTION
// Change the usage of an option (the instance of a complex actor), i.e yes/no or MC choice
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.ChangeOption(Params : TStringList) : integer;
var ActorUID, Value : TUID;
    OptionUID : string;
    CA : TComplexActor;

begin
  Result := 0;
  if (Params.Count <> 3) then
  begin
    Inc(Result);
    Log(['=> ChangeOption, syntax is: ChangeOption / ChgOption = GameActorUID ; OptionUID ; ValueChange']);
  end;
  if (cuFaction = nil) then
  begin
    Inc(Result);
    Log(['=> ChangeOption: No faction selected']);
  end;
  if Result > 0 then Exit;

  ActorUID  :=  STI(Params[0]);
  OptionUID := Params[1];
  Value     := STI(Params[2]);

  if (ActorUID < _gmaMinUID)or(ActorUID > _gmaMaxUID) then
  begin
    Inc(Result);
    Log(['=> ChangeOption: Erroneous value for the Actor UID, the correct range is', _gmaMinUID, 'to', _gmaMaxUID, 'and the parameter was:', ActorUID, 'if you get a 0, then the alias can be incorrect.']);
  end;

  CA := InitComplexActor;
  CA.Requested := Value;
  CA.KeyField1 := OptionUID;
  cuFaction.ChangeComplexActorBuilt(ActorUID, CA);
  Log(['ChangeOption', 'ActorUID', ActorUID, 'of subtype', Params[1], 'Changed by', Value]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// EVAL REPLACEMENT
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.EvalReplacement(Params : TStringList) : integer;
var cuNumber  : integer;
    FactionTag: string;
    Value     : integer;
    Operator  : string;
    Family    : TUID;
begin
  Result := 0;
  if (Params.Count <> 4) then
  begin
    Inc(Result);
    Log(['=> Evaluate Replacements, syntax is: EvalReplacement = FactionTag ; FamilyUID ; Operator (one of: =, <, >, <=, >=, DICE) ; Value']);
  end;
  if (cuFaction = nil) then
  begin
    Inc(Result);
    Log(['=> EvalReplacement: No faction selected']);
  end;
  if Result > 0 then Exit;

  FactionTag := Params[0];
  Family     := STI(Params[1]);
  Operator   := Params[2];
  Value      := STI(Params[3]);
  cuNumber   := cuFaction.GetRemplNumber(FactionTag, Family, True);

  if Operator = '=' then
  begin
    if not (cuNumber = Value) then Conditions := condFalse;
  end
  else
  if Operator = '<' then
  begin
    if not (cuNumber < Value) then Conditions := condFalse;
  end
  else
  if Operator = '<=' then
  begin
    if not (cuNumber <= Value) then Conditions := condFalse;
  end
  else
  if Operator = '>' then
  begin
    if not (cuNumber > Value) then Conditions := condFalse;
  end
  else
  if Operator = '>=' then
  begin
    if not (cuNumber >= Value) then Conditions := condFalse;
  end
  else
  if UpperCase(Operator) = 'DICE' then
  begin
    if (Dice(Value) > cuNumber) then Conditions := condFalse;
  end;

  Log(['EvalReplacement:', cuFaction.Name, 'RPL:', factionTag, FamilyToString(Family), 'Current Number', cuNumber, 'against', Value, ' - Operator:', Operator, 'Result:',
  IIF(Conditions = condEvaluating, 'True', 'False')]);

end;


// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//  ADD GARRISON
// Add a garrison to the faction. StrucType is the kind in Simple Garrison mode and
// the struc family in advanced mode. In both case, -1 is allowed and means 'every type'
// ChancePerCheck is ChancePerRPL in simple mode and ChancePerLVL in advanced mode (max 100% chance in the end, no multiple gen here)
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.AddGarrison(Params : TStringList) : integer;
var
  AreaUID   : TUID;
  StrucType: TUID;
  UnitUID   : TUID;
  Chance : integer;
begin
  Result := 0;
  if not (Params.Count in [4,8]) then // 4 = without the placeholder, 8: normal amount
  begin
    Inc(Result);
    Log(['=> Add Garrison , syntax is: AddGarrison = AreaUID or -1 ; StrucTypeUID(Kind_or_Fam) ; UnitDefUID ; ChancePerCheck; PlaceHolder1; PlaceHolder2; PlaceHolder3;PlaceHolder4']);
  end;
  if (cuFaction = nil) then
  begin
    Inc(Result);
    Log(['=> AddGarrison: No faction selected']);
  end;
  AreaUID   := STI(Params[0]);
  StrucType := STI(Params[1]);
  UnitUID   := STI(Params[2]);
  Chance    := STI(Params[3]);

  if (AreaUID > Areas.Count-1) then  // -1 = all map
  begin
    Inc(Result);
    Log(['=> Add Garrison , Erroneous AreaUID', AreaUID]);
  end;
  if (rulAutoGarrison_ = _gmdSimple) and // no test in advanced as the struc families are user defined
   ((StrucType < _noUID) or (StrucType > _stuMax)) then
  begin
    Inc(Result);
    Log(['=> Add Garrison , Erroneous StrucKind', StrucType]);
  end;
  if (UnitUID < 0) or (UnitUID > UnitDefs.Count-1) then
  begin
    Inc(Result);
    Log(['=> Add Garrison , Erroneous UnitUID', UnitUID]);
  end;
  if (rulAutoGarrison_ = _gmdSimple) and (Chance <= 0) then
  begin
    Inc(Result);
    Log(['=> Add Garrison , Erroneous Chance value', Chance]);
  end;
  if (rulAutoGarrison_ = _gmdAdvanced) and (Chance <= 0) then
    Chance := 100;

  if Result > 0 then Exit;

  cuFaction.AddGarrisonsInfos(AreaUID, StrucType, UnitUID, Chance);
  Log(['Add Garrison , Added this garrison:', 'AreaUID:', AreaUID, 'StrucType (Kind or Family):', StrucType, 'UnitDefUID:', UnitUID, 'Chance PerRPL or LVL:', Chance]);

end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//  CLEAR GARRISONS
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.ClearGarrisons(Params : TStringList) : integer;
begin
  Result := 0;
  if (Params.Count > 0) then
  begin
    Log(['=> [Warning] Clear Garrisons Garrison , syntax is: ClearGarrisons, no parameter are needed.']);
  end;
  if (cuFaction = nil) then
  begin
    Inc(Result);
    Log(['=> Clear Garrisons: No faction selected']);
  end;
  if Result > 0 then Exit;

  cuFaction.ClearGarrisonsInfos;
  Log(['Clear Garrisons , All Garrisons cleared for faction', cuFaction.UID, cuFaction.Name]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// CHANGE ONE TURN MODIFIER
// Faction-wide modifiers which are cleared every turn (active during one turn + hosting phase)
// _otmForcedMarch is the Forced March Modifier
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.ChangeOneTurnMod  (Params : TStringList) : integer;
begin
  Result := 0;
  if (Params.Count <> 4) then
  begin
    Log(['=> Change One Turn Modifier, Need 4 parameters, syntax is: ChgeOneTurnMod, <ModifierID> ; <FactionTags>, <AreaUID or -1> ; <Amount or -9999 for reset>']);
    Inc(Result);
  end;
  if (cuFaction = nil) then
  begin
    Inc(Result);
    Log(['=> Change One Turn Modifier: No faction selected']);
  end;
  Params[0] := LowerCase(Params[0]);

  if (LowerCase(Params[0]) <> _otmForcedMarch) then
  begin
    Inc(Result);
    Log(['=> Change One Turn Modifier: Unknown Modifier, found:', LowerCase(Params[0])]);
  end;
  if Result > 0 then Exit;

  cuFaction.OneTurn_ChangeModifier(Params[0], Params[1], STI(Params[2]), STI(Params[3]));
  if Params[0] = _otmForcedMarch then
    Log(['Change One Turn Modifier, Forced March,', Params[0], 'Faction Tags', Params[1], 'AreaUID:', Params[2], 'value is:', cuFaction.OneTurn_FetchForcedMarchModifier(Params[1], STI(Params[2]))]);

end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET UNKNOWN REGIONS
// Set the regions which are unknown to the faction(s). All the rest wil be known
// when this command is called
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetUnknownRegions (Params : TStringList) : integer;
var i,j     : integer;
    Faction : TFaction;
begin
  Result :=0;
  if (rulExploration_ < _rulSimple)  then
  begin
    Inc(Result);
    Log(['=> SetUnknownRegions: This command can''t be used when Exploration Rule is None.']);
  end;
  if (cuFaction = nil)and(rulExploration_ > _rulSimple)  then
  begin
    Inc(Result);
    Log(['=> SetUnknownRegions: This command apply to one only faction when Exploration Rule is Advanced or above.']);
  end;
  if Result > 0 then Exit;

  if rulExploration_ = _rulSimple then
  begin
    for i := 0 to GameSim.Factions.Count-1 do
    begin
      Faction := GameSim.Factions[i];
      for j := 0 to Params.Count-1 do
        Faction.SetRegionUnknown(STI(Params[j]));
    end;
  end
  else
  begin
    for j := 0 to Params.Count-1 do
      cuFaction.SetRegionUnknown(STI(Params[j]));
  end;
  for j := 0 to Params.Count-1 do
    Log(['SetUnknownRegions: ', STI(Params[j]), RegionDefs[STI(Params[j])].Name, 'is now unknown.']);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET KNOWN REGIONS
// Set new regions which are now known to the faction(s).
// SetKnownRegions = RgnUID1|RgnUID2 etc.
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetKnownRegions (Params : TStringList) : integer;
var i,j     : integer;
    Faction : TFaction;
begin
  Result :=0;
  if (rulExploration_ < _rulSimple)  then
  begin
    Inc(Result);
    Log(['=> SetKnownRegions: This command can''t be used when Exploration Rule is None.']);
  end;
  if (cuFaction = nil)and(rulExploration_ > _rulSimple)  then
  begin
    Inc(Result);
    Log(['=> SetKnownRegions: This command apply to one only faction when Exploration Rule is Advanced or above.']);
  end;
  if Result > 0 then Exit;

  if rulExploration_ = _rulSimple then
  begin
    for i := 0 to GameSim.Factions.Count-1 do
    begin
      Faction := GameSim.Factions[i];
      for j := 0 to Params.Count-1 do
        Faction.SetRegionKnown(STI(Params[j]));
    end;
  end
  else
  begin
    for j := 0 to Params.Count-1 do
      cuFaction.SetRegionKnown(STI(Params[j]));
  end;
  for j := 0 to Params.Count-1 do
    Log(['SetKnownRegions: ', STI(Params[j]), RegionDefs[STI(Params[j])].Name, 'is now known.']);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET REGIONS ALL KNOWN
// This command is added by CSVSplitter itself if there is a cities.inc
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetRegionsAllKnown(Params : TStringList) : integer;
var i       : integer;
    Faction : TFaction;
begin
  Result :=0;
  if (rulExploration_ < _rulSimple)  then
  begin
    Inc(Result);
    Log(['=> SetRegionsAllKnown: This command can''t be used when Exploration Rule is None.']);
  end;
  if (cuFaction = nil)and(rulExploration_ > _rulSimple)  then
  begin
    Inc(Result);
    Log(['=> SetRegionsAllKnown: This command apply to one only faction when Exploration Rule is Advanced or above.']);
  end;
  if Result > 0 then Exit;

  if rulExploration_ = _rulSimple then
  begin
    for i := 0 to GameSim.Factions.Count-1 do
    begin
      Faction := GameSim.Factions[i];
      Faction.SetRegionsAllKnown;
    end;
  end
  else
  begin
    cuFaction.SetRegionsAllKnown;
  end;
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET ALL REGIONS LEGIT OWNERS
// This command is added by CSVSplitter itself if there is a cities.inc
// will give to each faction each region where they have 51%+ MC in their Legit Owner list
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetAllRgnLegitOwn (Params : TStringList) : integer;
begin
  Result := 0;
  if (rulDiplomacyMode_ < _rulAdvanced)  then
  begin
    Exit;
    Log(['=> SetAllRgnLegitOwn: This command can''t be used when Diplomacy mode is not advanced.']);
  end;

  DoLegitOwn := True; // will be done after FinalizeParsing, once RgnCmn Update done

end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET LEGIT OWNER
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetLegitOwner (Params : TStringList) : integer;
begin
  Result := 0;
  if (rulDiplomacyMode_ < _rulAdvanced)  then
  begin
    Exit;
    Log(['=> SetLegitOwner: This command can''t be used when Diplomacy mode is not advanced.']);
  end;
  if cuFaction = nil then
  begin
    Inc(Result);
    Log(['=> SetLegitOwner, at least one faction must be selected']);
  end;
  if (cuRegion = nil) then
  begin
    Inc(Result);
    Log(['=> EvalLoyalty: No region selected']);
  end;
  if Result > 0 then Exit;

  GameSim.RegionCmns[cuRegion.UID].LegitOwner := cuFaction.UID;
  GameSim.RegionCmns[cuRegion.UID].pLegitOwner := cuFaction;
  GameSim.RegionCmns.UpdateFactionsLegitOwned;

  // this command don't reverse a ChangeRgnOwner, which has more profound effects

end;


// *** FIN CONDITIONS & EVALUATIONS ***

// *** DEBUT ACTIONS ***
// Toutes les variables d'actions commencent rinitialises ici

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SELECT FACTIONS
// implmentation basique:
// on ne slectionne qu'une faction, qui est l'UID pass en param
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SelectFactions(Params : TStringList) : integer;
var i   : integer;
    UID : TUID;
    NewFaction : TFaction;
begin
  Result := 0;

  for i := 0 to Params.Count -1 do
  begin
    if IsNumeric(Params[i]) then
    begin
      SetNumeric(Params[i], UID);
      NewFaction := TGameSim(Owner).Factions.GetFactionByUID(UID, False);
      if NewFaction = nil then
      begin
        Inc(Result);
        Log(['=> SelectFaction, no faction selected, criteria: ', UID]);
        Exit;
      end;  
    end
    else
      NewFaction := FactionTagToFaction(Params[i]);

     IsBroadCast := UpperCase(Params[i]) = 'CMN';

    // Reset selectors
    if cuFaction <> NewFaction then
    begin
      if IsSpecialScript and (not IsBroadCast) then // if a special script, maintain the region in selection (i.e a RGD will create rebels against the one playing the RGD)
        cuRegion := NewFaction.Regions.Fetch(cuRegion.UID)
      else
        cuRegion  := nil;
      cuGroup   := nil;
      cuUnit    := nil;
      LastGHQGroup := nil;
    end;
    cuFaction := NewFaction;
  end;

  // pas forcment une erreur de script, la faction peut ne pas encore exister
  if (cuFaction = nil)and(not IsBroadCast) then
  begin
    Log(['[WARNING] SelectFaction, no faction selected, criteria: ', UID])
  end
  else
  if cuFaction <> nil then
    Log(['SelectFaction, selected:', cuFaction.Name]);

  if Params.Count = 0 then
  begin
    Inc(Result);
    Log(['=> SelectFaction, no valid parameter']);
  end;
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SELECT REGION
// Syntaxe: SelectRegion RgnUID[;NotEnemy][;CondSkip][;NoAdj]
// si pas de paramtres, slectionner la rgion, mme si non controle.
// NotEnemy: si la rgion n'est pas controle, en prendre une adjacente (controle)
// au hasard
// NoAdj:  utiliser avec NotEnemy, on ne prend pas adj si ennemi
// SelectRegion prend la rgion indique, revient  faire un AddToRegList de la rgion puis PickFromRegList
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SelectRegion (Params : TStringList) : integer;
var i   : integer;
    TempReg : TRegion;
    AdjReg  : TRegion;
    UID : TUID;
    Param : string;
    Bag : TRandomBag;
begin
  Result := 0;

  //  ce n'est plus une erreur, pour les CivLevel et VP, on a besoin de SelectRegion, mais sans appel  une faction
  if cuFaction = nil then
  begin
    Log(['SelectRegion, at least one faction must be selected, selecting first faction']);
    cuFaction := GameSim.Factions[0];
  end;

  if not(Params.Count in [1..4]) then
  begin
    Inc(Result);
    Log(['=> SelectRegion, erroneous number of params', Params.Count, 'Syntax: SelectRegion RgnUID[;NotEnemy][;NotAdj][;CondSkip]']);
    Exit;
  end;

  if (not(IsNumeric(Params[0])))or(STI(Params[0]) <= _noUID) then
  begin
    Inc(Result);
    Log(['=> SelectRegion, First parameter must be a region UID (check if the alias was translated correctly):', Params[0]]);
    Exit;
  end;

  // dj une rgion slectionne
  if (ExtractParam('CondSkip', Params) <> '') and (cuRegList.Count > 0) then
    Exit;

  UID := STI(Params[0]);
  TempReg := cuFaction.Regions.Fetch(UID);
  if TempReg = nil then // ne doit pas arriver sauf si l'UID ne correspond  rien
  begin
    Inc(Result);
    Log(['=> SelectRegion, Internal Error, no region found, criteria: ', UID]);
    Exit;
  end;

  // Gestion d'une slection sur rgion adjacente si la rgion choisie doit tre amie
  Param := ExtractParam('NotEnemy', Params);

  // gestion sans NotEnemy, prendre la rgion passe en param, sans regarder le contrle et on sort
  if Param = '' then
  begin
    cuRegion := TempReg;
    cuRegList.Add(cuRegion.UID);
    Log(['SelectRegion,', cuRegion.UID, cuRegion.Name, 'selected (simple mode)']);
    Exit;
  end;

  // gestion avec NotEnemy, mais rgion amie
  if not TempReg.IsEnemyControled(True) then
  begin
    cuRegion := TempReg;
    cuRegList.Add(cuRegion.UID);
    Log(['SelectRegion,', cuRegion.UID, cuRegion.Name, 'selected (NotEnemy mode but region is friendly)']);
    Exit;
  end;

  Param := ExtractParam('NotAdj', Params);
  if Param <> '' then
  begin
    cuRegion := nil; // ajout 19/03/07  
    Log(['SelectRegion, no valid selection, Selection is now empty.']);  
    Exit; // cuRegion est dsormais  nil
  end;

  // Adj autoris et NotEnemy  grer
  Bag := TRandomBag.Create;
  try
  begin
    // il faut trouver une rgion terrestre adjacente & amie
    TempReg.CalcLinks(True);
    for i := 0 to TempReg.pListLinks.Count -1 do
    begin
      AdjReg := TRegion(TempReg.pListLinks[i]);
      if not RegionDefs.IsLandRegion(AdjReg.UID) then Continue;
      if AdjReg.IsEnemyControled(True) then Continue;
      // on l'ajoute
      Bag.Add(AdjReg.UID);
    end;
    // on prend une rgion au hasard et on sort
    if Bag.Count > 0 then
    begin
      cuRegion := cuFaction.Regions.Fetch(Bag.PickInt);
      cuRegList.Add(cuRegion.UID);
      Log(['SelectRegion,', cuRegion.Name, 'selected']);
      Exit;
    end;
  end;
  finally
    Bag.Free;
  end;

end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SELECT REGION IN AREA
// Syntaxe: SelectRegInArea AreaUID[;CondSkip][;NotEnemy][;OpenHarbor][;OpenHarborWSU][;OwnedVPCities][;OwnedCities][;Squared][;WildTerrain][;OwnedObjCities][;WarSupplies]
// Squared peut s'utiliser avec les autres options
// Utilis en hosting par THost.CheckAutoRaise et THost.CheckGameActorsTransaction,
// rpercuter si modif de syntaxe existente (ok si que ajout de nouvelles options)
// on peut faire plusieurs appels  cette fonction ainsi qu'a SelectRegion, pour ajouter dans la cuRegList
// Usages: a) choisir cuRegion  partir de l'area, et b) ajouter des elts dans cuRegList
// TODO  passer dans RegionCmn?
//
// Note sur l'exclusivit des paramtres: ExtractParams ne peut pas diffrencier entre 2 paramtres si l'un est une sous chaine de l'autre
// (et on ne doit pas faire une comparaison exacte, car il peut y avoir des arguments), donc OpenHarbor se retrouve dans OpenHarborWSU par exemple
// C'est pour cela que l'on a WarSupplies et non pas WSU comme paramtre possible
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SelectRegInArea(Params : TStringList; ExcludeDisloyal : boolean = False; ExcludeLowControlled : boolean = False) : integer;
var i   : integer;
    TempReg : TRegion;
    UID : TUID;
    Param : string;
    dlRegions : TDoubleList;
    dlRegionsTemp : TDoubleList;
    Bag : TRandomBag;
    Area : TArea;
    GS : TGameSim;
    pStruc : TStructure;
    IsLandArea : boolean;
    Struc : TStructure;
    Squared : boolean;
    Region : TRegion;
    WSUProd : integer;
begin
  Result := 0;
  GS := TGameSim(Owner);

  // SelRegInArea ou SelectRegInArea est ok pour le Splitter

  if cuFaction = nil then
  begin
    Inc(Result);
    Log(['=> SelectRegInArea, at least one faction must be selected']);
    Exit;
  end;

  if Params.Count = 0 then
  begin
    Inc(Result);
    Log(['=> SelectRegInArea, erroneous number of params', Params.Count, 'Syntax: SelectRegInArea AreaUID[;CondSkip][;NotEnemy][;OpenHarbor][;OwnedVPCities][;OwnedObjCities][;OwnedCities][;OwnedCitiesEmptyFirst][;Squared][;WildTerrains]']);
    Exit;
  end;

  if (not(IsNumeric(Params[0])))or(STI(Params[0]) <= _noUID) then
  begin
    Inc(Result);
    Log(['=> SelectRegInArea, First parameter must be an area UID']);
    Exit;
  end;

//  cuRegion := nil; supprim 19/03/07

  // dj des rgions slectionnes, on sort si CondSkip
  if (ExtractParam('CondSkip', Params) <> '') and (cuRegList.Count > 0) then
  begin
    Log(['SelectRegInArea, CondSkip and List is not empty, exiting, List count:', cuRegList.Count]);
    Exit;
  end;

  UID := STI(Params[0]);

  dlRegions     := TDoubleList.Create;
  dlRegionsTemp := TDoubleList.Create;
  Bag := TRandomBag.Create;
  Squared := ExtractParam('Squared', Params) <> '';
  try

    // candidats initiaux: toutes les rgions de la zone
    // on considre implicitement que l'on veut les rgions terrestres dans une area terrestre
    // et inversement pour une area maritime
    Area := Areas[UID];
    IsLandArea := Area.IsLandArea;
    if Area = nil then
    begin
      Inc(Result);
      Log(['=> SelectRegInArea, Unknown Area, UID:', UID]);
      Exit;
    end;
    for i := 0 to Area.Count -1 do
    begin
      if GS.RegionCmns[Area[i]].BlockState <> _regUnBlocked then Continue;

      if RegionDefs.IsLandRegion(Area[i]) and IsLandArea then
        dlRegions.Add(Area[i], 1)
      else
      begin
        if (not RegionDefs.IsLandRegion(Area[i])) and (not IsLandArea) then
          dlRegions.Add(Area[i], 1)
      end;
    end;

    // Exclusion suivant les critres passs en param

    Param := ExtractParam('NotEnemy', Params);
    if Param <> '' then
    begin
      for i := dlRegions.Count-1 downto 0 do
      begin
        UID := dlRegions.Keys[i];
        TempReg := cuFaction.Regions.Fetch(UID);
        if TempReg.IsEnemyControled(True) then
          dlRegions.Delete(i)
      end;
    end;


    // OpenHarborWSU is tested first because OpenHarbor text can be found within OpenHarborWSU
    Param := ExtractParam('OpenHarborWSU', Params);
    if Param <> '' then
    begin
      for i := dlRegions.Count-1 downto 0 do
      begin
        UID := dlRegions.Keys[i];
        TempReg := cuFaction.Regions.Fetch(UID);

        WSUProd := GS.RegionCmns[UID].WSUProd; // simple eco
        if TempReg.Structures.HasThisKind(_stuProdSite) then Inc(WSUProd, 10); // adv eco

        if (not TempReg.OperatingHarbor)or(TempReg.GetCity = nil) then // il faut supprimer les anchorages de la liste, graphiquement cela ne colle pas then
          dlRegions.Delete(i)
        else
        begin
          dlRegions.Vals[i] := Max(1, IIF(Squared, Sqr(WSUProd +1), WSUProd +1));
          if TempReg.GetNbSubUnitsTheseFamilies([_famNavalEngineer]) > 0 then
            dlRegions.Vals[i] := dlRegions.Vals[i] *2;
        end;
      end;
    end
    else
    begin
      Param := ExtractParam('OpenHarbor', Params);
      if Param <> '' then
      begin
        for i := dlRegions.Count-1 downto 0 do
        begin
          UID := dlRegions.Keys[i];
          TempReg := cuFaction.Regions.Fetch(UID);
          if (not TempReg.OperatingHarbor)or(TempReg.GetCity = nil) then // il faut supprimer les anchorages de la liste, graphiquement cela ne colle pas
            dlRegions.Delete(i)
          else
          begin
            Struc := GS.RegionCmns.GetDockableHarbor(cuFaction, UID);
            if Struc <> nil then
            begin
              dlRegions.Vals[i] := IIF(Squared, Sqr(Struc.Level), Struc.Level);
              if TempReg.GetNbSubUnitsTheseFamilies([_famNavalEngineer]) > 0 then
                dlRegions.Vals[i] := dlRegions.Vals[i] *2;
            end;
          end;
        end;
      end;
    end;

    Param := ExtractParam('WarSupplies', Params);
    if Param <> '' then
    begin
      for i := dlRegions.Count-1 downto 0 do
      begin
        UID := dlRegions.Keys[i];
        TempReg := cuFaction.Regions.Fetch(UID);
        WSUProd := GS.RegionCmns[UID].WSUProd; // simple eco
        if TempReg.Structures.HasThisKind(_stuProdSite) then Inc(WSUProd, 10); // adv eco

        if WSUProd = 0 then
        begin
          UID := dlRegions.Keys[i];
          if GS.RegionCmns.GetCityInfos(UID, pStruc) <> cuFaction then
          begin
            dlRegions.Delete(i);
            Continue;
          end
          else
            dlRegions.Vals[i] := 1
        end
        else
          dlRegions.Vals[i] := IIF(Squared, Sqr(WSUProd +1), WSUProd +1);
      end;
    end;

    Param := ExtractParam('OwnedVPCities', Params);
    if Param <> '' then
    begin
      for i := dlRegions.Count-1 downto 0 do
      begin
        UID := dlRegions.Keys[i];
        if GS.RegionCmns.GetVPCountFromRegion(cuFaction, UID) = 0 then
          dlRegions.Delete(i)
        else
        begin
          GS.RegionCmns.GetCityInfos(UID, Struc);
          if Struc <> nil then
            dlRegions.Vals[i] := IIF(Squared, Sqr(Struc.Level), Struc.Level);
        end;
      end;
    end;

    Param := ExtractParam('OwnedFort', Params);
    if Param <> '' then
    begin
      for i := dlRegions.Count-1 downto 0 do
      begin
        TempReg := cuFaction.Regions.Fetch(UID);
        Struc := TempReg.GetFort;
        if (TempReg.GetFort = nil) then
          dlRegions.Delete(i)
        else
        begin
          dlRegions.Vals[i] := IIF(Squared, Sqr(Struc.Level), Struc.Level);
        end;
      end;
    end;

    Param := ExtractParam('HasRR', Params);
    if Param <> '' then
    begin
      for i := dlRegions.Count-1 downto 0 do
      begin
        UID := dlRegions.Keys[i];
        if GameSim.RegionCmns[UID].TransportLvl in [_trsRailRoad, _trsDRailRoad] then
          DoNothing
        else
          dlRegions.Delete(i)
      end;
    end;

    Param := ExtractParam('OwnedObjCities', Params);
    if Param <> '' then
    begin
      for i := dlRegions.Count-1 downto 0 do
      begin
        UID := dlRegions.Keys[i];
        if not GS.RegionCmns.HasObjectiveInRegion(cuFaction, UID) then
          dlRegions.Delete(i)
        else
        begin
          dlRegions.Vals[i] := IIF(Squared, Sqr(cuFaction.GetObjValue(UID)), cuFaction.GetObjValue(UID));
        end;
      end;
    end;

    Param := ExtractParam('OwnedCitiesEmptyFirst', Params);
    if Param <> '' then
    begin
      // on enlve toute ville que l'on occupe ou qui n'est pas  nous
      dlRegionsTemp.Assign(dlRegions);
      for i := dlRegions.Count-1 downto 0 do
      begin
        UID := dlRegions.Keys[i];
        if GS.RegionCmns.GetCityInfos(UID, pStruc) <> cuFaction then
        begin
          dlRegions.Delete(i);
          Continue;
        end;
        Region := cuFaction.Regions.Fetch(UID);
        if Region.HasCSDefender then
        begin
          dlRegions.Delete(i);
          Continue;
        end;
        // we must had there the two tests below, which are done normally at the very end,
        // as this procedure is in 2 steps (empty first) and it can lead to a bug
        if ExcludeLowControlled and (Region.GetControlWithAllies < bldMinControlBuildUnit_) then
        begin
          dlRegions.Delete(i);
          Continue;
        end;
        if ExcludeDisloyal and (Region.GetLoyaltyWithAllies < bldMinLoyaltyBuildUnit_) then
        begin
          dlRegions.Delete(i);
          Continue;
        end;

        if pStruc <> nil then
        begin
          dlRegions.Vals[i] := IIF(Squared, Sqr(pStruc.Level), pStruc.Level);
          if Region.GetDepot <> nil then dlRegions.Vals[i] := dlRegions.Vals[i] *10;
          if Region.GetFort  <> nil then dlRegions.Vals[i] := dlRegions.Vals[i] *10;
          if GS.RegionCmns.CanGetVPOrObjFromRegion(cuFaction, UID) then dlRegions.Vals[i] := dlRegions.Vals[i] *10;
        end
        else
          dlRegions.Delete(i);
      end;
      // plus aucune rgion, on fait comme OwnedCities
      if dlRegions.Count = 0 then
      begin
        dlRegions.Assign(dlRegionsTemp);
        for i := dlRegions.Count-1 downto 0 do
        begin
          UID := dlRegions.Keys[i];
          if GS.RegionCmns.GetCityInfos(UID, pStruc) <> cuFaction then
            dlRegions.Delete(i)
          else
          begin
            GS.RegionCmns.GetCityInfos(UID, Struc);
            if Struc <> nil then
              dlRegions.Vals[i] := IIF(Squared, Sqr(Struc.Level), Struc.Level);
          end;
        end;
      end;
    end
    else
    // OwnedCitiesEmptyFirst et OwnedCities doivent etre exclusifs entre eux, cf note en tte  de fonction
    begin
      Param := ExtractParam('OwnedCities', Params);
      if Param <> '' then
      begin
        for i := dlRegions.Count-1 downto 0 do
        begin
          UID := dlRegions.Keys[i];
          if GS.RegionCmns.GetCityInfos(UID, pStruc) <> cuFaction then
            dlRegions.Delete(i)
          else
          begin
            GS.RegionCmns.GetCityInfos(UID, Struc);
            if Struc <> nil then
              dlRegions.Vals[i] := IIF(Squared, Sqr(Struc.Level), Struc.Level);
          end;
        end;
      end;
    end;

    Param := ExtractParam('WildTerrains', Params);
    if Param <> '' then
    begin
      for i := dlRegions.Count-1 downto 0 do
      begin
        UID := dlRegions.Keys[i];
        if (GS.RegionCmns.GetCity(UID) <> nil) or (not TerrainDefs[RegionDefs[UID].TerrainType].IsWildTerrain) then
          dlRegions.Delete(i);
      end;
    end;

    // Ajout dans cuRegList + MJ cuRegion (rgion valide dans les rgions slectionnes)
    if dlRegions.Count > 0 then
    begin
      dlRegions.Vals.NormalizeVals();
      for i := 0 to dlRegions.Count -1 do
      begin
        TempReg := cuFaction.Regions.Fetch(dlRegions.Keys[i]);
        if ExcludeLowControlled and (TempReg.GetControlWithAllies < bldMinControlBuildUnit_) then Continue;
        if ExcludeDisloyal and (TempReg.GetLoyaltyWithAllies < bldMinLoyaltyBuildUnit_) then Continue;

        Bag.Add(dlRegions.Keys[i], dlRegions.Vals[i]);
        cuRegList.Add(dlRegions.Keys[i]);
      end;

      // MJ cuRegion
      UID := Bag.PickInt;
      if UID > _noUID then
      begin
        cuRegion := cuFaction.Regions.Fetch(UID);
        Log(['SelectRegInArea,', cuRegion.Name, 'selected & Regions List has now ', cuRegList.Count, 'regions.']);
      end
      else
        Log(['SelectRegInArea, no region selected']); 
    end
    else
      Log(['SelectRegInArea, no region selected']);
  finally
    dlRegions  .Free;
    dlRegionsTemp.Free;
    Bag.Free;
  end;
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SELECT UNIT IN AREA    *** OBSOLETE, ne passe plus en FullDebug ***
// Syntaxe : SelUnitsInArea AreaUID ; SubTypeUID ; HomeAreaUID
// Permet de slectionner une unit d'une faction dans une zone donne
// paramtres supplmentaires: le sous-type, la zone AreaBound de l'unit (-1 si non utilis)
// PICKUNIT fait pareil !
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SelectUnitInArea(Params : TStringList) : integer;
var i   : integer;
    UID : TUID;
    IL : TIntList;
    Bag : TRandomBag;
    Area : TArea;
    Region : TRegion;
    SubType : TUID;
//    HomeArea : TUID;
    aUnit : TUnit;
begin
  Result := 0;
  if cuFaction = nil then
  begin
    Inc(Result);
    Log(['=> SelectUnitInArea, at least one faction must be selected']);
    Exit;
  end;

  if not (Params.Count in [3]) then
  begin
    Inc(Result);
    Log(['=> SelectUnitInArea, erroneous number of params', Params.Count, 'Syntax: SelUnitsInArea AreaUID [; SubTypeUID][; HomeAreaUID]']);
    Exit;
  end;

  if (not(IsNumeric(Params[0])))or(STI(Params[0]) <= _noUID) then
  begin
    Inc(Result);
    Log(['=> SelectUnitInArea, First parameter must be an area UID']);
    Exit;
  end;

  if FullDebug_ then
  begin
    Inc(Result);
    Log(['=> SelectUnitInArea, This command is deprecated, use PickUnit']);
    Exit;
  end;

  UID      := STI(Params[0]);
  SubType  := STI(Params[1]);
//  HomeArea := STI(Params[2]);

  IL := TIntList.Create;
  Bag := TRandomBag.Create;

  try
    Area := Areas[UID];
    if Area = nil then
    begin
      Inc(Result);
      Log(['=> SelectUnitInArea, Unknown Area, UID:', UID]);
      Exit;
    end;

    // candidats initiaux: toutes les units de la zone
    for i := 0 to Area.Count -1 do
    begin
      Region := cuFaction.Regions.Fetch(Area[i], False);
      if Region = nil then Continue;
      Region.GetUnitList(IL, False);
    end;

    // Exclusion suivant les critres passs en param
    if SubType <> _noUID then
    begin
      for i := IL.Count-1 downto 0 do
      begin
        aUnit := cuFaction.GetUnitByUID(IL[i]);
        if aUnit.SubType <> SubType then
          IL.Delete(i);
      end;
    end;
    (*
    if HomeArea <> _noUID then
    begin
      for i := IL.Count-1 downto 0 do
      begin
        aUnit := cuFaction.GetUnitByUID(IL[i]);
        if aUnit.AreaBoundUID <> HomeArea then
          IL.Delete(i);
      end;
    end;
    *)
    // Prendre une unit valide dans les units candidates
    if IL.Count > 0 then
    begin
      for i := 0 to IL.Count -1 do
        Bag.Add(IL[i]);
      UID := Bag.PickInt;
      cuUnit := cuFaction.GetUnitByUID(UID, False);
      Log(['SelectUnitInArea,', cuUnit.Name, 'selected']);
      Exit;
    end;


  finally
    IL.Free;
    Bag.Free;
  end;


end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SELECT GROUPS
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SelectGroups (Params : TStringList) : integer;
var i   : integer;
    UID : TUID;
begin
  Result := 0;
  if cuFaction = nil then
  begin
    Inc(Result);
    Log(['=> SelectGroups, at least one faction must be selected']);
    Exit;
  end;

  for i := 0 to Params.Count -1 do
  begin
    if IsNumeric(Params[i]) then
    begin
      SetNumeric(Params[i], UID);
    end
    else
      UID := LabelToUID(Params[i]);
    cuGroup := cuFaction.GetGroupByUID(UID, True);
  end;

  if cuGroup = nil then
  begin
    Inc(Result);
    Log(['=> SelectGroups, no groups selected, criteria: ', UID])
  end
  else
    Log(['SelectGroups, 1 Group selected:', cuGroup.Name]);

  if Params.Count = 0 then
  begin
    Inc(Result);
    Log(['=> SelectGroups, no valid parameter']);
  end;
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SELECT UNITS
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SelectUnits   (Params : TStringList) : integer;
var i   : integer;
    UID : TUID;
begin
  Result := 0;
  if cuGroup = nil then
  begin
    Inc(Result);
    Log(['=> SelectUnits, at least one group must be selected']);
    Exit;
  end;

  for i := 0 to Params.Count -1 do
  begin
    if IsNumeric(Params[i]) then
      SetNumeric(Params[i], UID)
    else
      UID := LabelToUID(Params[i]);
    cuUnit := cuGroup.GetUnitByUID(UID);
  end;

  if cuUnit = nil then
  begin
    Inc(Result);
    Log(['=> SelectUnits, no units selected, criteria: ', UID])
  end
  else
    Log(['SelectUnits, 1 Unit selected:', cuUnit.Name]);

  if Params.Count = 0 then
  begin
    Inc(Result);
    Log(['=> SelectUnits, no valid parameter']);
  end;
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SELECT SUBUNITS / SELECT SUBUNIT
// SelectSubUnits Region <RgnUID>|Area <AreaUID>|Families <Fam1> <Fam2> ..|Models <Mdl1> <Mdl2> ..|FactionTags <Tag1> <Tag2> ..
// |Domains <_domLand> <_domNav> <_domAir>
// |Attributes <Attr1> <Attr2> | Generations <Gen1> <Gen2> ..
// |OnlyFixed|OnlyNotFixed|OnlyPermFixed|OnlyNPermFixed|Enemy|FriendOnly|FriendAndSelf|UniqueName <Name>
// |OnlyBesieged|OnlyBesieging
// |OnlyGHQCommand|OnlyHQCommand
// |HealthCohPerc <MinHealth> <MaxHealth> <MinCoh> <MaxCoh>
// |PreserveCMN: if true, a SubUnit with tag CMN will be seen as CMN, otherwise code will mask it and use parent faction tag
// Si area est omis, c'est sur toute la carte
// ex SelectSubUnits Area $South_West|Models $mdl_USA_Cav1 $mdl_USA_Cav2
// _SELECTSUBUNITS
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SelectSubUnits(Params : TStringList) : integer;
var i,j, sub      : integer;
    grp           : integer;
    UID           : TUID;
    Param         : string;
    iArg          : integer;
    AllTypes      : boolean;
    UIDList       : TIntList;
    Group         : TGroup;
    SU            : TSubUnit;
    DoModels      : boolean;
    DoFamilies    : boolean;
    slFactionTags : TStringList;
    slAttributes  : TStringList;
    slGenerations : TStringList;    
    ilDomains     : TIntList;
    OnlyFixed     : boolean;
    OnlyPermFixed : boolean;
    OnlyNPermFixed: boolean;
    OnlyNotFixed  : boolean;
    OnlyBesieged  : boolean;
    OnlyBesieging : boolean;
    OnlyGHQCommand: boolean;
    OnlyHQCommand : boolean;
    PreserveCMN   : boolean;
    bHealthMax    : string;
    HealthMax     : integer;
    bHealthMin    : string;
    HealthMin     : integer;
    bCohMax       : string;
    CohMax        : integer;
    bCohMin       : string;
    CohMin        : integer;
    Enemy         : boolean;
    EnemyNeutral  : boolean;
    FriendOnly    : boolean;
    FriendAndSelf : boolean;
    SelfOnly      : boolean;
    OpDone        : boolean;
    RgnUID        : TUID;
    UniqueName    : string;
begin
  Result := 0;
  if cuFaction = nil then
  begin
    Inc(Result);
    Log(['=> SelectSubUnits, no faction selected']);
    Exit;
  end;

  UIDList  := TIntList.Create;
  slFactionTags := TStringList.Create;
  slAttributes  := TStringList.Create;
  slGenerations := TStringList.Create;
  ilDomains := TIntList.Create;
  cuSUList.Clear;
  AllTypes := False; DoModels := False; DoFamilies := False;

  try
    cuRegList.Clear;

    // IMPORTANT: troubles when parsing aliases, for example
    // region $Kent, Va;Models $Mdl1 $Mdl2
    // so we translate manually the aliases, one by one
    // this command do all alias PARSING here, not in the main TScriptEngine.Parse


    // Region, Zone ou toute la carte
    Param :=  ExtractParam('Area', Params);
    if Param <> '' then
    begin
      ExtractArgs(Param);
      if Args[0] = '' then
      begin
        Inc(Result);
        Log(['=> SelectSubUnits, Area specified but no Area UID']);
        Exit;
      end;
      if Length(Args[0]) > 1 then
        TFileParser.ParseForAliases(Args[0]);
      UID   := STI(Args[0]);
      for i := 0 to Areas[UID].Count-1 do
      begin
        cuRegList.Add(Areas[UID][i]);
      end;
    end
    else
    begin
      Param :=  ExtractParam('Region', Params);
      if Param = '' then
        ExtractParam('Regions', Params);
      if Param <> '' then
      begin
        //ExtractArgs(Param);
        Args[0] := Copy(Param, 7, 255); // manual extraction, args should not have space within them like $Kent, Va
        if (Args[0] = '') and (cuRegion <> nil) then
          UID := cuRegion.UID
        else
        begin
          if Length(Args[0]) > 1 then
            TFileParser.ParseForAliases(Args[0]);
          UID   := STI(Args[0]);
        end;  
        cuRegList.Add(UID);
      end
    end;

    // Liste par famille
    Param :=  ExtractParam('Families', Params);
    if Param = '' then
      Param :=  ExtractParam('Family', Params);
    if Param <> '' then
    begin
      DoFamilies := True;
      ExtractArgs(Param);
      i := 0;
      while (i <= High(Args))and(Args[i] <> '') do
      begin
        TFileParser.ParseForAliases(Args[i]);
        iArg := STI(Args[i]);
        if iArg = -1 then
        begin
          AllTypes := True;
          Break;
        end
        else
          UIDList.Add(iArg);
        Inc(i);
      end;
    end;

    // Liste par ModelUID
    Param :=  ExtractParam('Models', Params);
    if Param = '' then
      Param :=  ExtractParam('Model', Params);
    if Param <> '' then
    begin
      DoModels := True;
      ExtractArgs(Param);
      i := 0;
      while (i <= High(Args))and(Args[i] <> '') do
      begin
        TFileParser.ParseForAliases(Args[i]);
        iArg := STI(Args[i]);
        if iArg = -1 then
        begin
          AllTypes := True;
          Break;
        end
        else
          UIDList.Add(iArg);
        Inc(i);
      end;
    end;

    if DoModels and DoFamilies then
    begin
      Inc(Result);
      Log(['=> SelectSubUnits, You must select either by Families or by Models.']);
      Exit;
    end
    else
    if (not DoModels) and (not DoFamilies) then
      AllTypes := True;


    // Restriction sur les FactionTags (par dfaut, tous)
    Param :=  ExtractParam('FactionTags', Params);
    if Param = '' then
      Param :=  ExtractParam('FactionTag', Params);
    if Param <> '' then
    begin
      ExtractArgs(Param);
      i := 0;
      while (i <= High(Args))and(Args[i] <> '') do
      begin
        slFactionTags.Add(Args[i]);
        Inc(i);
      end;
    end;

    // Restriction sur les Attributes (par dfaut, tous)
    Param :=  ExtractParam('Attributes', Params);
    if Param = '' then
      Param :=  ExtractParam('Attribute', Params);
    if Param <> '' then
    begin
      ExtractArgs(Param);
      i := 0;
      while (i <= High(Args))and(Args[i] <> '') do
      begin
        slAttributes.Add(Args[i]);
        Inc(i);
      end;
    end;

    // Restriction sur les Gnrations (par dfaut, tous)
    Param :=  ExtractParam('Generations', Params);
    if Param = '' then
      Param :=  ExtractParam('Generation', Params);
    if Param <> '' then
    begin
      ExtractArgs(Param);
      i := 0;
      while (i <= High(Args))and(Args[i] <> '') do
      begin
        slGenerations.Add(Args[i]);
        Inc(i);
      end;
    end;

    // Restriction sur les Domaines (par dfaut, tous)
    Param :=  ExtractParam('Domains', Params);
    if Param = '' then
      Param :=  ExtractParam('Domain', Params);
    if Param <> '' then
    begin
      ExtractArgs(Param);
      i := 0;
      while (i <= High(Args))and(Args[i] <> '') do
      begin
        TFileParser.ParseForAliases(Args[i]);
        iArg := STI(Args[i]);
        if iArg in [_domLand .. _domNav] then
          ilDomains.Add(iArg);
        Inc(i);
      end;
    end;

    // Various Restrictions
    OnlyFixed     := ExtractParam('OnlyFixed', Params) <> '';
    OnlyNotFixed  := ExtractParam('OnlyNotFixed', Params) <> '';
    OnlyPermFixed := ExtractParam('OnlyPermFixed', Params) <> '';
    OnlyNPermFixed:= ExtractParam('OnlyNPermFixed', Params) <> '';
    Enemy         := ExtractParam('Enemy', Params) <> '';
    EnemyNeutral  := ExtractParam('EnemyNeutral', Params) <> '';    
    FriendOnly    := ExtractParam('FriendOnly', Params) <> '';
    FriendAndSelf := ExtractParam('FriendAndSelf', Params) <> '';
    UniqueName    := ExtractParam('UniqueName', Params);
    OnlyBesieged  := ExtractParam('OnlyBesieged', Params) <> '';
    OnlyBesieging := ExtractParam('OnlyBesieging', Params) <> '';
    OnlyGHQCommand:= ExtractParam('OnlyGHQCommand', Params) <> '';
    OnlyHQCommand := ExtractParam('OnlyHQCommand', Params) <> '';
    PreserveCMN   := ExtractParam('PreserveCMN', Params) <> '';
    // Health & Coh
    Param := ExtractParam('HealthCohPerc', Params);
    if Param <> '' then
    begin
      ExtractArgs(Param);
      HealthMin := STI(Args[0]);
      HealthMax := STI(Args[1]);
      CohMin    := STI(Args[2]);
      CohMax    := STI(Args[3]);
    end
    else
    begin
      HealthMin := 0;
      HealthMax := 100;
      CohMin    := 0;
      CohMax    := 100;
    end;

    if UniqueName <> '' then
    begin
      ExtractArgs(UniqueName);
      UniqueName := '';
      for j := 0 to High(Args) do
      UniqueName := Trim(UniqueName + ' ' + Args[j]); // Name can have space, but each space make an argument!
    end;

    if OnlyFixed and OnlyNotFixed then
    begin
      Inc(Result);
      Log(['=> SelectSubUnits, OnlyFixed/OnlyNotFixed: Please select something remotely possible.']);
      Exit;
    end;

    if (Enemy and FriendOnly) or (Enemy and FriendAndSelf) or (FriendOnly and FriendAndSelf) then
    begin
      Inc(Result);
      Log(['=> SelectSubUnits, Enemy/FriendOnly/FriendAndSelf: Please select something not absurd.']);
      Exit;
    end;

    if cuRegList.Count = 0 then // all map
      for i := 0 to GameSim.RegionCmns.Count-1 do
        cuRegList.Add(i);

    SelfOnly := (not Enemy) and (not FriendOnly) and (not FriendAndSelf);

    // Checking all list
    for i := 0 to cuRegList.Count-1 do
    begin
      RgnUID := cuRegList[i];
      for grp := 0 to GameSim.RegionCmns[RgnUID].pListGroup.Count-1 do
      begin
        Group := GameSim.RegionCmns[RgnUID].pListGroup[grp];
        if SelfOnly and (cuFaction.UID <> Group.Faction) then Continue;
        if Enemy and (not cuFaction.IsEnemyOf(Group.Faction)) then Continue;
        if EnemyNeutral and (cuFaction.IsFriendOf(Group.Faction)) then Continue;
        if FriendOnly and (not cuFaction.IsFriendOf(Group.Faction) or (cuFaction.UID = Group.Faction)) then Continue;
        if FriendAndSelf and (not cuFaction.IsFriendOf(Group.Faction)) then Continue;

        Group.MakeSubUList;
        for sub := 0 to Group.SubUList.Count-1 do
        begin
          SU := TSubUnit(Group.SubUList[sub]);

          // Faction Tags
          if (slFactionTags.Count > 0) and (slFactionTags.IndexOf(SU.FactionTag(PreserveCMN)) = -1) then Continue;

          // Attributes
          if (slAttributes.Count > 0) then
          begin
            OpDone := False;
            for j := 0 to slAttributes.Count-1 do
            begin
              if SU.HasThisAttribute(slAttributes[j]) then
              begin
                OpDone := True;
                Break;
              end;
            end;
            if not OpDone then Continue;
          end;

          // Generations
          if (slGenerations.Count > 0) then
          begin
            OpDone := False;
            for j := 0 to slGenerations.Count-1 do
            begin
              if SU.Model.Generation = slGenerations[j] then
              begin
                OpDone := True;
                Break;
              end;
            end;
            if not OpDone then Continue;
          end;

          // Domains
          if (ilDomains.Count > 0) and (not ilDomains.InList(SU.Domain)) then Continue;

          // Fix types
          if (TUnit(SU.pUnit).FixType =  0) and OnlyFixed then Continue;
          if (TUnit(SU.pUnit).FixType <> 0) and OnlyNotFixed then Continue;
          if ((TUnit(SU.pUnit).FixType = -999) or (TUnit(SU.pUnit).FixType = 999)) and OnlyNPermFixed then Continue;
          if ((TUnit(SU.pUnit).FixType <> -999) and (TUnit(SU.pUnit).FixType <> 999)) and OnlyPermFixed then Continue;

          // Name
          if (UniqueName <> '') and (UpperCase(SU.Name) <> UpperCase(UniqueName)) then Continue;

          // health or cohesion
          if (HealthMin <> 0) or (HealthMax <> 100) or (CohMin <> 0) or (CohMax <> 100) then
          begin
            if SU.GetHealthPerc < HealthMin then Continue;
            if SU.GetHealthPerc > HealthMax then Continue;
            if SU.GetCohesionPerc < CohMin then Continue;
            if SU.GetCohesionPerc > CohMax then Continue;
          end;

          // besieger/besieging
          if OnlyBesieged  and (not Group.IsBesieged) then Continue;
          if OnlyBesieging and (not Group.IsBesieging) then Continue;

          // Command
          if OnlyGHQCommand and (not SU.IsGHQStackLeader) then Continue;
          if OnlyHQCommand  and (not SU.IsHQStackLeader)  then Continue;

          // All types or by models or families
          if AllTypes then
          begin
            cuSUList.Add(SU);
            Log(['SelectSubUnits', SU.UID, SU.Name, 'added, List Count now at:', cuSUList.Count]);
          end
          else
          if DoFamilies and (AllTypes or UIDList.InList(SU.Model.Family)) then
          begin
            cuSUList.Add(SU);
            Log(['SelectSubUnits', SU.UID, SU.Name, 'added, List Count now at:', cuSUList.Count]);
          end
          else
          if DoModels and (AllTypes or UIDList.InList(SU.Model.UID)) then
          begin
            cuSUList.Add(SU);
            Log(['SelectSubUnits', SU.UID, SU.Name, 'added, List Count now at:', cuSUList.Count]);
          end;

        end; // SU
      end; // grp
    end; // rgn

  finally
    // Sept09 Addition: latest select update all selectors
    if cuSUList.Count > 0 then
    begin
      cuSubUnit := cuSUList[cuSUList.Count-1];
      cuUnit    := TUnit(cuSubUnit.pUnit);
      cuGroup   := TGroup(cuSubUnit.pGroup);
      cuRegion  := cuGroup.pStartRegion;
      cuFaction := TFaction(cuSubUnit.pFaction);
    end
    else
    begin
      cuSubUnit := nil;
      cuUnit    := nil;
      cuGroup   := nil;
    end;

    // Nov 2009: update cuRegList
    cuRegList.Clear;
    for i := 0 to cuSUList.Count-1 do
      if cuRegList.IndexOf(ITS(TSubUnit(cuSUList[i]).StartSubSpace)) = -1 then
        cuRegList.Add(TSubUnit(cuSUList[i]).StartSubSpace);

    // Jan 2011: Registry
    if ActiveReg > 0 then
    begin
      Inc(Registry[ActiveReg], cuSUList.Count); // was cleared at start of function
      Log(['SelectSubUnits', 'Registry', ActiveReg, 'used, adding',  cuSUList.Count, 'to it, for an end value of', Registry[ActiveReg]]);
    end;

    Log(['SelectSubUnits finished', 'Regions Selected:', cuRegList.Count, 'SubUnits Selected:', cuSUList.Count]);
            
    UIDList.Free;
    ilDomains.Free;
    slFactionTags.Free;
  end;
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SELECT UNIQUE UNIT
// Slectionne une unit d'internal Name prcis (qqs la faction). Si elle n'existe pas, tous les selecteurs
// passent  nil, autrement il sont tous valoriss, ! peut entrainer un changement de faction intempestif
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SelectUniqueUnit(Params : TStringList) : integer;
var fac, unt : integer;
    Faction : TFaction;
    aUnit : TUnit;
    SearchMode : string;
    SU : TSubUnit;
    OnlyUnfixed : boolean;
    OnlyFixed : boolean;
    RestrictFac : boolean;
begin
  Result := 0;
  if not (Params.Count in [1 .. 4]) then
  begin
    Inc(Result);
    Log(['=> Select Unique Unit command, need to have a unique name as an argument & Optionnaly a search mode. Syntax: SelUnqUnit = Name [;CondSkip][;OnlyUnfixed][;OnlyFixed] [;RestrictFac]']);
    Exit;
  end;

  SearchMode := ExtractParam('NoSelGroup', Params);
  if SearchMode = '' then SearchMode := ExtractParam('CondSkip', Params);
  OnlyUnfixed := ExtractParam('OnlyUnfixed', Params) <> '';
  OnlyFixed := ExtractParam('OnlyFixed', Params) <> '';
  RestrictFac := ExtractParam('RestrictFac', Params) <> '';

  if (SearchMode <> '')and(cuGroup <> nil) then
  begin
    Log(['[Warning] Select Unique unit: Evaluation skipped, NoSelGroup/CondSkip parameter & a group is already selected']);
    Exit;
  end;
                                                   // REM PHM 03/2010, otherwise RestrictFac can't work
  cuSubUnit := nil; cuUnit := nil; cuGroup := nil; {cuFaction := nil;} cuRegion := nil; cuRegList.Clear;

  Params[0] := Trim(Params[0]);

  for fac := 0 to GameSim.Factions.Count -1 do
  begin
    Faction := GameSim.Factions[fac];
    if RestrictFac and (Faction <> cuFaction) then Continue;
    Faction.CalcUnitList;
    for unt := Faction.UnitList.Count -1 downto 0 do
    begin
      aUnit := TUnit(Faction.UnitList[unt]);
      if OnlyFixed and (aUnit.FixType = 0) then Continue;
      if OnlyUnfixed and (aUnit.FixType <> 0) then Continue;
      if Trim(aUnit.InternalName) = Params[0] then
      begin
        cuUnit    := aUnit;
        cuGroup   := cuUnit.pGroup;
        cuFaction := cuGroup.pFaction;
        cuRegion  := cuFaction.Regions.Fetch(cuGroup.StartSubSpace);
        cuRegList.Add(cuRegion.UID);
        Log(['Select Unique Unit, found:', cuFaction.Name, cuGroup.Name, cuUnit.Name, 'in', cuRegion.Name]);
        Exit;
      end;
    end;
  end;
  // si pas trouv au niveau unit, on regarde toutes les Sous-units
  for fac := 0 to GameSim.Factions.Count -1 do
  begin
    Faction := GameSim.Factions[fac];
    if RestrictFac and (Faction <> cuFaction) then Continue;
    SU := Faction.GetSubUnitByName(Params[0], True);
    if SU = nil then Continue;
    if OnlyFixed and (TUnit(SU.pUnit).FixType = 0) then Continue;
    if OnlyUnfixed and (TUnit(SU.pUnit).FixType <> 0) then Continue;

    cuSubUnit := SU;
    cuUnit    := TUnit(SU.pUnit);
    cuGroup   := cuUnit.pGroup;
    cuFaction := cuGroup.pFaction;
    cuRegion  := cuFaction.Regions.Fetch(cuGroup.StartSubSpace);
    cuRegList.Add(cuRegion.UID);
    Log(['Select Unique unit:', 'found into a unit as a sub unit of name:', Params[0], cuFaction.Name, cuGroup.Name, cuUnit.Name, 'in', cuRegion.Name]);
    Exit;
  end;

  Log(['Select Unique unit:', Params[0], ' not found']);
  // Peut servir  une valuation dans le bloc Conditions
  if ConditionsON then
  begin
    Conditions := condFalse;
    Log(['We were in Conditions block, Conditions are now False.']);
  end;
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// REMOVE UNIQUE UNIT
// Dtruit cuUnit (si non nil), passe cuGroup au Groupe de l'unit enleve s'il reste
// une unit, passe cuRegion  la rgion de l'unit (tt le temps)
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.RemoveCuUnit(Params : TStringList) : integer;
var Group : TGroup;
    UnitName : string;
begin
  Result := 0;
  if not (Params.Count in [0]) then
  begin
    Log(['[Warning] Remove Current Unit: There is no argument in this command.']);
  end;
  if cuUnit = nil then // pas une erreur
  begin
    Log(['[Warning] Remove Current Unit: No Selected unit.']);
    Exit;
  end;

  UnitName := cuUnit.Name + ' (InternalName: ' + cuUnit.InternalName + ')';
  Group := cuUnit.pGroup;
  Group.Remove(cuUnit);
  cuRegion  := Group.pStartRegion;
  cuFaction := Group.pFaction;
  if Group.Count = 0 then
  begin
    TFaction(Group.pFaction).KillGroup(Group);
    cuGroup := nil;
  end
  else
  begin
    Group.Update;
    cuGroup := Group;
  end;
  
  GameSim.RegionCmns.Update(cuRegion.UID);

  Log(['Unit ', UnitName, 'removed from play.']);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SELECT UNIQUE SUB UNIT
// Slectionne une SU du Name prcis (qqs la faction). Si elle n'existe pas, tous les selecteurs
// passent  nil, autrement il sont tous valoriss, ! peut entrainer un changement de faction intempestif
// Pas de notion d'Internal Name au niveau SU, pas changeable par le joueur
// Region RgnUID pour faire la recherche sur une region donne
// Area   AreaUID pour faire la recherche sur une zone donne
// _SELUNQSUBUNIT
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SelectUniqueSubUnit(Params : TStringList) : integer;
var fac,i   : integer;
    Faction : TFaction;
    SU      : TSubUnit;
    Param   : string;
    SUName  : string;
    UID     : TUID;
    RegList : TIntList;
begin
  Result := 0;
  if not (Params.Count in [1..3]) then
  begin
    Inc(Result);
    Log(['=> Select Unique Sub Unit command, Syntax is : SelUnqSubUnit = Name [;Region RgnUID] [;Area AreaUID] [;CondSkip or NoSelGroup]']);
    Exit;
  end;

  Param := ExtractParam('NoSelGroup', Params);
  if Param = '' then Param := ExtractParam('CondSkip', Params);

  if (Param <> '')and(cuGroup <> nil) then
  begin
    Log(['[Warning] Select Unique Sub unit: Evaluation skipped, NoSelGroup/CondSkip parameter & a group is already selected']);
    Exit;
  end;

  cuSubUnit := nil; cuUnit := nil; cuGroup := nil; cuFaction := nil; cuRegion := nil; cuRegList.Clear;

  SUName := Trim(Params[0]);
  RegList := TIntList.Create;

  try
    // Region, Zone ou toute la carte (RegList.Count reste  0 en ce cas)
    Param :=  ExtractParam('Area', Params);
    if Param <> '' then
    begin
      ExtractArgs(Param);
      if Args[0] = '' then
      begin
        Inc(Result);
        Log(['=> SelUnqSubUnit, Area specified but no Area UID']);
        Exit;
      end;
      UID   := STI(Args[0]);
      for i := 0 to Areas[UID].Count-1 do
      begin
        RegList.Add(Areas[UID][i]);
      end;
    end
    else
    begin
      Param :=  ExtractParam('Region', Params);
      if Param <> '' then
      begin
        ExtractArgs(Param);
        if Args[0] = '' then
        begin
          Inc(Result);
          Log(['=> SelectSubUnits, Region specified but no Region UID']);
          Exit;
        end;
        UID   := STI(Args[0]);
        RegList.Add(UID);
      end
    end;

    for fac := 0 to GameSim.Factions.Count -1 do
    begin
      Faction := GameSim.Factions[fac];
      SU := Faction.GetSubUnitByName(SUName, True);
      if  (SU <> nil)
      and ( (RegList.Count = 0) or (RegList.InList(TUnit(SU.pUnit).StartSubSpace)) ) then
      begin
        cuSubUnit := SU;
        cuUnit    := TUnit(SU.pUnit);
        cuGroup   := cuUnit.pGroup;
        cuFaction := cuGroup.pFaction;
        cuRegion  := cuFaction.Regions.Fetch(cuGroup.StartSubSpace);

        if RegList.Count > 0 then
          Log(['Note: Search was limited to', RegList.Count, IIF(RegList.Count = 1, 'region.', 'regions.')]);

        Log(['Select Unique sub unit, found:', cuFaction.Name, cuGroup.Name, cuUnit.Name, SU.Name, 'in', cuRegion.Name]);

        // we reset the reglist and the cuSUList to the region of the unique SU (if this is a problem, keeps the cuSUList at least) + 20120124
        cuRegList.Clear;
        cuSUList.Clear;
        cuRegList.Add(cuRegion.UID);
        cuSUList.Add(cuSubUnit);

        Exit;
      end;                               
    end;

    // if here, unique SU was not found
    if RegList.Count > 0 then
      Log(['Note: Search was limited to', RegList.Count, IIF(RegList.Count = 1, 'region.', 'regions.')]);

    Log(['Select Unique SubUnit:', SUName, ' not found', IIF(RegList.Count = 0, 'on map', 'in the region(s) list.')]);

    // and we can be in conditions, so invalidate it
    if ConditionsON then
    begin
      Conditions := condFalse;
      Log(['We were in Conditions block, Conditions are now False.']);
    end;

  finally
    RegList.Free;
  end;
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// REMOVE UNIQUE UNIT
// Selectionne et dtruit en une seule passe une unit de la faction suivant son nom
// Peut chercher les units combines et ne dtruire que les SU s'y rapportant
// (d'o la ncessit de faire en une passe, les units combines n'tant plus des units)
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.RemoveUnqUnit(Params : TStringList) : integer;
var Group : TGroup;
    grp, unt : integer;
    UnitName : string;
    OpDone : boolean;
    PrevCount : integer;
begin
  Result := 0;
  if Params.Count <> 1 then
  begin
    Inc(Result);
    Log(['[Warning] Remove Unique Unit: Need exactly one argument in this command. Syntax is RemoveUnqUnit = UnitName or RemUnqUnit = UnitName']);
  end;
  if cuFaction = nil then
  begin
    Inc(Result);
    Log(['[Warning] Remove Unique Unit: No Selected Faction.']);
    Exit;
  end;
  UnitName := Params[0];
  OpDone   := False;

  try

    for grp := 0 to cuFaction.Count -1 do
    begin
      Group := cuFaction[grp];
      for unt := 0 to Group.Count -1 do
      begin
        // Unit 'simple'
        if Group[unt].Name = UnitName then
        begin
          OpDone := True;
          cuGroup := Group;
          cuRegion := Group.pStartRegion;
          Group.Remove(Group[unt]);
          Log(['RemoveUnqUnit, operation done on ', UnitName, 'in region', Group.StartRegionName, 'Removed entirely the unit.']);
          if Group.Count = 0 then
          begin
            cuFaction.KillGroup(Group);
            cuGroup := nil;
          end
          else
            Group.Update;
          GameSim.RegionCmns.Update(cuRegion.UID);

          Break;
        end;

        // Regarder par unit combine
        PrevCount := Group[unt].Count;
        if Group[unt].RemoveCombinedUnitByName(UnitName) then
        begin
          OpDone := True;
          Log(['RemoveUnqUnit, operation done on ', UnitName, 'in region', Group.StartRegionName, 'Removed', PrevCount - Group[unt].Count, 'SUs within the unit.']);
          cuUnit  := Group[unt];
          cuGroup := Group;
          cuRegion := Group.pStartRegion;

          if Group[unt].Count = 0 then
          begin
            Group.Remove(Group[unt]);
            cuUnit := nil;
          end;
          if Group.Count = 0 then
          begin
            cuFaction.KillGroup(Group);
            cuGroup := nil;
          end
          else
            Group.Update;

          GameSim.RegionCmns.Update(cuRegion.UID);
          Break;
        end;
      end;
      if OpDone then Break;
    end;

  except
    on E : Exception do
      LogMsg(['RemoveUnqUnit', E.Message], IIF(FullDebug_, _Error, _Reporting));
  end;
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// REMOVE UNIQUE SUB UNIT
// Dtruit cuSubUnit (si non nil), passe cuGroup au Groupe de l'unit enleve s'il reste
// une unit, passe cuRegion  la rgion de l'unit (tt le temps)
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.RemoveCuSubUnit(Params : TStringList) : integer;
var SUInfos : string;
begin
  Result := 0;
  if (Params.Count > 0) and (Params[0] <> 'NULL') then
  begin
    Log(['[Warning] Remove Current Sub Unit: There is no need for an argument in this command. Syntax is RemoveCuSubUnit']);
  end;
  if cuSubUnit = nil then
  begin
    Log(['[Warning] RemoveCuSubUnit: No Selected SubUnit, skipping.']);
    Exit;
  end;

  try
    cuUnit    := TUnit(cuSubUnit.pUnit);
    cuGroup   := cuUnit.pGroup;
    cuFaction := cuGroup.pFaction;
    cuRegion  := cuFaction.Regions.Fetch(cuGroup.StartSubSpace);

    SUInfos := cuSubUnit.Name + ' Unit: ' + cuUnit.Name + ' Group: ' + cuGroup.Name + ' Faction: ' + cuFaction.Name + ' in region: ' + cuRegion.Name;

    // cf TFaction.RemoveSubUnit now
    cuFaction.RemoveSubUnit(cuSubUnit);
{
    cuUnit.Remove(cuSubUnit);
    cuSubUnit := nil;
    if cuUnit.Count = 0 then
    begin
      cuGroup.Remove(cuUnit);
      cuUnit := nil;
    end;
    if cuGroup.Count = 0 then
    begin
      cuFaction.KillGroup(cuGroup);
      cuGroup := nil;
    end
    else
      cuGroup.Update;

    GameSim.RegionCmns.Update(cuRegion.UID);
}
    Log(['RemoveCuSubUnit, operation done on ' + SUInfos]);
  except
    on E : Exception do
      LogMsg(['RemoveCuSubUnit', E.Message], IIF(FullDebug_, _Error, _Reporting));
  end;
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// REMOVE SUB UNITS
// Supprime physiquement toutes les SU de la liste de slection (et les units/groupes parents si besoin)
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.RemoveSubUnits(Params : TStringList) : integer;
var i : integer;
begin
  Result := 0;
  if (Params.Count >0) and (Pos('NULL', Params[0]) = 0) then
  begin
    Log(['[Warning] Remove Sub Units: There is no need for an argument in this command, the list of currently selected SUs will be used. Syntax is RemoveSubUnits']);
  end;
  if cuSUList.Count = 0 then
  begin
    Log(['[Warning] Remove Sub Units: Empty List.']);
    Exit;
  end;

  try
    for i := 0 to cuSUList.Count-1 do
    begin
      cuSubUnit := cuSUList[i];
      RemoveCuSubUnit(Params);
    end;
    cuSUList.Clear;
    cuUnit    := nil;
    cuGroup   := nil;
    cuRegion  := nil;
  except
    on E : Exception do
      LogMsg(['RemoveCuSubUnit', E.Message], IIF(FullDebug_, _Error, _Reporting));
  end;
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ADD TO REGION LIST
// Ajout de rgions  la liste des rgions slectionnables (cuRegList)
// nb: la liste est vide entre chaque vnement
// nb: on peut ajouter des rgions  la liste via SelectRegInArea en mettant des critres, etc.
// on peut ajouter +ieurs fois une rgion pour en modifier la proba de tirage
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.AddToRegList(Params : TStringList) : integer;
var i : integer;
begin
  Result := 0;
  if Params.Count = 0 then
  begin
    Inc(Result);
    Log(['=> Add to Region List: need to have a list of regions UID as an argument (alias allowed): AddToRegList UID1;UID2;UID3']);
    Exit;
  end;
  //nb: les args sont spars par ;, ils sont donc devenus une entre de la SL par UID
  for i := 0 to Params.Count -1 do
  begin
    if IsNumeric(Params[i]) then
      cuRegList.Add(STI(Params[i]));
  end;
    Log(['Add to Region List: Added', Params.Count, 'in the current regions list, for a total of', cuRegList.Count]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// PICK FROM REGION LIST
// Tirage d'une rgion  partir de la liste des rgions slectionnables (cuRegList)
// PickFromRegList [;NotEnemy] [;NoLowMC] [NotAdj] [;NoBasicBias] [;FavorLocalRatio] [;FavorLackLocalRatio]
// [;FavorCityLevel] [;FavorHarborLevel]  [;FavorFortLevel] [;FavorLoyalty] [;FavorLackLoyalty] [;FavorVP]
// [;FavorOBJ] [;FavorColPen] [;FavorLackColPen] [;AllowBlocked]
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.PickFromRegList(Params : TStringList) : integer;
var i,adj  : integer;
    Bag : TRandomBag;
    UID : TUID;
    TempReg, AdjReg : TRegion;
    NotEnemy, NotAdj, NoBasicBias, FavorLocalRatio, FavorLackLocalRatio, NoLowMC : boolean;
    FavorCityLevel, FavorHarborLevel, FavorFortLevel, FavorLoyalty, FavorLackLoyalty, FavorVP, FavorOwnOBJ : boolean;
    FavorColPen, FavorLackColPen, AllowBlocked : boolean;
    Coeff : integer;
    Ratio : double;
    RgnUID : TUID;
    dlRgnCoeffs : TDoubleList;
begin
  Result := 0;
  if cuFaction = nil then
  begin
    Log(['[Warning]  Pick From Region List: No faction selected']);
    Exit;
  end;
  cuRegion := nil;
  if  cuRegList.Count = 0 then
  begin
    Log(['[Warning] Pick From Region List: No region in the pick list.']);
    Exit;
  end;

  NotEnemy               := ExtractParam('NotEnemy', Params) <> '';
  NoLowMC                := ExtractParam('NoLowMC', Params) <> '';
  NotAdj                 := ExtractParam('NotAdj', Params) <> '';
  NoBasicBias            := ExtractParam('NoBasicBias', Params) <> '';
  FavorLocalRatio        := ExtractParam('FavorLocalRatio', Params) <> '';
  FavorLackLocalRatio    := ExtractParam('FavorLackLocalRatio', Params) <> '';
  FavorCityLevel         := ExtractParam('FavorCityLevel', Params) <> '';
  FavorHarborLevel       := ExtractParam('FavorHarborLevel', Params) <> '';
  FavorFortLevel         := ExtractParam('FavorFortLevel', Params) <> '';
  FavorLoyalty           := ExtractParam('FavorLoyalty', Params) <> '';
  FavorLackLoyalty       := ExtractParam('FavorLackLoyalty', Params) <> '';
  FavorVP                := ExtractParam('FavorVP', Params) <> '';
  FavorOwnOBJ            := ExtractParam('FavorOwnOBJ', Params) <> '';
  FavorColPen            := ExtractParam('FavorColPen', Params) <> '';
  FavorLackColPen        := ExtractParam('FavorLackColPen', Params) <> '';
  AllowBlocked           := ExtractParam('AllowBlocked', Params) <> '';

  Bag := TRandomBag.Create;
  dlRgnCoeffs := TDoubleList.Create;

  try
    for i := 0 to cuRegList.Count-1 do
    begin
      RgnUID := cuRegList[i];
      TempReg := cuFaction.Regions.Fetch(RgnUID);

      // Add all regions one by one, NotEnemy filter completely, otherwise DiscardNotMet will filter completely too
      if NotEnemy and TempReg.IsEnemyControled(True) then Continue;

      if NoLowMC and (TempReg.Control < 50) then Continue;


      if (GameSim.RegionCmns[RgnUID].BlockState <> _regUnblocked) and (not AllowBlocked) then
        Continue;

      // If there is the basic bias (retrocompatibility), then add a ... basic bias!
      if NoBasicBias = False then
        Coeff := Min( (cuRegList.Count-i) *2, 100) 
      else
        Coeff := 100;

      if FavorLocalRatio then
      begin
        Ratio := GameSim.RegionCmns.GetLocalRatio(cuFaction, RgnUID);
        Coeff := Round(Coeff * Ratio); // between 0.2 and 10
      end;
      if FavorLackLocalRatio then
      begin
        Ratio := GameSim.RegionCmns.GetLocalRatio(cuFaction, RgnUID);
        Coeff := Round(Coeff / Ratio); // between 0.2 and 10
      end;
      if FavorCityLevel then
      begin
        Coeff := Coeff * GameSim.RegionCmns.GetCityLevel(RgnUID);
      end;
      if FavorHarborLevel then
      begin
        Coeff := Coeff * GameSim.RegionCmns.GetHarborLevel(RgnUID);
      end;
      if FavorFortLevel then
      begin
        Coeff := Coeff * GameSim.RegionCmns.GetFortLevel(RgnUID);
      end;
      if FavorLoyalty then
      begin
        Coeff := Coeff * TempReg.Loyalty;
      end;
      if FavorLackLoyalty then
      begin
        Coeff := Coeff * (100-TempReg.Loyalty);
      end;
      if FavorVP then
      begin
        Coeff := Coeff * GameSim.RegionCmns[RgnUID].VictoryPts;
      end;
      if FavorOwnOBJ then
      begin
        Coeff := Coeff * cuFaction.GetObjValue(RgnUID);
      end;
      if FavorColPen then
      begin
        Coeff := Coeff * TempReg.ColonialPen;
      end;
      if FavorLackColPen then
      begin
        Coeff := Coeff * (100 - TempReg.ColonialPen);
      end;

      dlRgnCoeffs.Add(cuRegList[i], Coeff);
    end;

    // normalize values to not overload the bag
    dlRgnCoeffs.Vals.NormalizeVals;
    for i := 0 to dlRgnCoeffs.Count-1 do
      Bag.Add(dlRgnCoeffs.Keys[i], dlRgnCoeffs.Vals[i]);


    if (Bag.Count = 0)and (NotAdj = True) then
    begin
      Log(['PickFromRegList, no valid selection, Region Selector is now nil and Selection List is now empty..']);
      cuRegList.Clear; // cuRegion est dj  nil
      Exit;
    end;

    // NotAdj not found: exclude others criterias implicitely...
    // aucune rgion est amie, on passe aux adjacentes (NotAdj pas trouv)
    if (Bag.Count = 0)and(NotAdj = False) then
    begin
      for i := 0 to cuRegList.Count-1 do
      begin
        TempReg := cuFaction.Regions.Fetch(cuRegList[i]);
        // il faut trouver une rgion terrestre amie adjacente
        TempReg.CalcLinks(True);
        for adj := 0 to TempReg.pListLinks.Count -1 do
        begin
          AdjReg := TRegion(TempReg.pListLinks[adj]);
          if not RegionDefs.IsLandRegion(AdjReg.UID) then Continue;
          if AdjReg.IsEnemyControled(True) then Continue;
          // on l'ajoute
          Coeff := Min( (cuRegList.Count-i) *2, 100); // les premiers elts sont favoriss
          Bag.Add(AdjReg.UID, Coeff);
        end;
      end;
    end;

    // Pick a region
    if Bag.Count <> 0 then
    begin
      UID := Bag.PickInt;
      cuRegion := cuFaction.Regions.Fetch(UID);
      Log(['Pick From Region List:', cuRegion.Name, 'selected']);
    end
    else
    begin
      Log(['Pick From Region List: no qualified region, Region Selector is now nil and Selection List is now empty.']);
    end;

  finally
    Bag.Free;
    dlRgnCoeffs.Free;
  end;
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ADD TO REGION LIST FROM AREAS
// Add the regions of the given areas
// <AreaCapitalOnly> : only the calculated capital areas regions will be added (one per area)
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.AddToRegListFromAreas(Params : TStringList) : integer;
var i,j : integer;
    Param : string;
    AreaUID : TUID;
    Area : TArea;
begin
  Result := 0;
  if Params.Count = 0 then
  begin
    Inc(Result);
    Log(['=> AddToRegListFromAreas: need to have a list of Area UID as an argument (alias allowed): AddToRegListFromAreas = AreaUID1;AreaUID2;AreaUID3[;<AreaCapitalOnly>]']);
    Exit;
  end;

  // cuRegList is emptied at the start of each event and when a PickUnit, SelectSubUnits, etc. is done
  
  Param := ExtractParam('AreaCapitalOnly', Params);
  if Param <> '' then
  begin
    for i := 0 to Params.Count -2 do
    begin
      AreaUID := STI(Params[i], False, -1);

      if (AreaUID <= -1) or (AreaUID > Areas.Count-1) then
      begin
        Log(['=> AddToRegListFromAreas:', 'Unknown area alias', Params[i]]);
        Inc(Result);
        Exit;
      end;
      Area := Areas[AreaUID];
      cuRegList.Add(Area.GetCalcCapital);
    end;
  end
  else
  begin
    for i := 0 to Params.Count -1 do
    begin
      AreaUID := STI(Params[i], False, -1);
      if (AreaUID <= -1) or (AreaUID > Areas.Count-1) then
      begin
        Log(['=> AddToRegListFromAreas:', 'Unknown area alias', Params[i]]);
        Inc(Result);
        Exit;
      end;
      Area := Areas[AreaUID];
      for j := 0 to Area.Count-1 do
        cuRegList.Add(Area[j]);
    end;
  end;

  Log(['Add to Region List: Added', Params.Count, 'in the current regions list, for a total of', cuRegList.Count]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SELECT (most) NUMEROUS GROUP
// Slectionne le groupe le plus nombreux de la faction en cours de slection
// la rgion du groupe devient la rgion en cours
// un paramtre optionnel: n'excuter la commande que si cuGroup est nil
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SelNumerousGroup(Params : TStringList) : integer;
var grp, MaxNb  : integer;
    Group       : TGroup;
    SearchMode  : string;
    Param       : string;
    AreaUID     : TUID;
    CPCount     : boolean;
    EvenFixed   : boolean;
    i           : integer;
    bRestrictToThisTag : boolean;
    FactionTag  : string;
    bRestrictToThisKind : boolean;
    GroupKind : TUID;
begin
  Result := 0;
  if cuFaction = nil then
  begin
    Inc(Result);
    Log(['=> Select Most Numerous Group: No faction selected']);
    Exit;
  end;

  if not Params.Count in [0..5] then
  begin
    Inc(Result);
    Log(['=> Select Most Numerous Group: Wrong number of parameters. Syntax is SelMostNumGroup = [;NoSelGroup] [;Area AreaUID] [;CPCount] [;EvenFixed] [;RestrictToThisTag FactionTag] [;RestrictToThisKind KindAlias]' ]);
    Exit;
  end;

  MaxNb := -1;

  SearchMode := ExtractParam('NoSelGroup', Params);

  if (SearchMode <> '')and(cuGroup <> nil) then
  begin
    Log(['[Warning] Select Most Numerous Group: Evaluation skipped, NoSelGroup parameter & a group is already selected']);
    Exit;
  end;

  cuGroup := nil; cuRegion := nil;

  Param := ExtractParam('Area', Params);
  if Param <> '' then
  begin
    ExtractArgs(Param);
    if not IsNumeric(Args[0]) then
    begin
      Log(['=> Select Most Numerous Group: Area parameter but invalid Area UID (check alias).']);
      Inc(Result);
      Exit;
    end;
    AreaUID := STI(Args[0]);
  end
  else
    AreaUID := _noUID;

  CPCount   := ExtractParam('CPCount', Params) <> '';
  EvenFixed := ExtractParam('EvenFixed', Params) <> '';
  Param := ExtractParam('RestrictToThisTag', Params);
  bRestrictToThisTag := Param <> '';
  if bRestrictToThisTag then
  begin
    ExtractArgs(Param);
    FactionTag := Args[0];
  end;
  Param := ExtractParam('RestrictToThisKind', Params);
  bRestrictToThisKind := Param <> '';
  if bRestrictToThisKind then
  begin
    ExtractArgs(Param);
    GroupKind := STI(Args[0]);
  end
  else GroupKind := _gkLand;


  // parcourir tous les groupes de la faction

  for grp := 0 to cuFaction.Count -1 do
  begin
    Group := cuFaction[grp];
    if (AreaUID <> _noUID)and(not Areas[AreaUID].InList(Group.StartSubSpace)) then Continue;
    if Group.IsFixed and (not EvenFixed) then Continue;
    if bRestrictToThisTag and (Group.CountSUThisTag(FactionTag) < Group.GetSUCount div 2) then Continue;
    if bRestrictToThisKind and (Group.Kind = GroupKind) then Continue;
    if IIF(CPCount, Group.GetUsedCommand, Group.GetSUCount) > MaxNb then //fixed: + AACW 20/03/07
    begin
      cuGroup := Group;
      MaxNb   := IIF(CPCount, Group.GetUsedCommand, Group.GetSUCount);
    end;
  end;

  // we now set cuSUList also (+PhM 28 may 08) - allow to check nb of SUs with EvalSubUnitCount
  if cuGroup <> nil then
  begin
    cuRegion := cuFaction.Regions.Fetch(cuGroup.StartSubSpace);
    cuRegList.Add(cuRegion.UID);
    cuGroup.MakeSubUList;
    cuSUList.Clear;
    for i := 0 to cuGroup.SubUList.Count-1 do
      cuSUList.Add(cuGroup.SubUList);
    Log(['Select Most Numerous Group:', cuGroup.UID, cuGroup.Name, 'in', cuRegion.UID, cuRegion.Name]);
  end;
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// CREATE OBJECT
// implmentation basique: un param, qui est le label donn  l'objet
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.CreateObject(GMEType : string; Params : TStringList) : integer;
var BoundVar : TStringArray;
begin
  Result := 0;
  if WorkingGME <> '' then
    Log(['=> CreateObject, was working on a', WorkingGME, 'switching to another object']);
  WorkingGME := GMEType;

  // il y a un param, c'est le label de travail de l'objet, et l'objet reoit de suite un UID
  if (Params.Count > 0)and(WorkingGME <> _geSubU) then
  begin
    SetLength(BoundVar, 1);
    try
      BoundVar[0] := ITS(MakeGmUID(_genUID));
      // affecter au label donn par le script l'UID gnre
      AddBoundVar(Params[0], BoundVar[0]);
      WorkingVars.Add(Params[0]);
    finally
      BoundVar := nil;
    end;
  end;

  // nb : les SubUnits ne peuvent tre rfrences ainsi mais on peut en crer plusieurs,
  // params[0] sert au nb  crer
  if (Params.Count > 0)and(WorkingGME = _geSubU) then
    SetNumeric(Params[0], NumCreate, 1);

end;


// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET KIND
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetKind(Params : TStringList) : integer;
begin
  if Params.Count <> 1 then
  begin
    Result := 1;
    Log(['=> SetKind command, erroneous number of params', Params.Count]);
    Exit;
  end;
  Result := KindLabelToGMEType(Params[0], Kind);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET SPECIAL COMMIT
// Sous paramtre qui est test en Apply pour savoir s'il y a un traitement particulier
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetSpecialCommit(Params : TStringList) : integer;
begin
  Result := 0;
  if Params.Count <> 1 then
  begin
    Result := 1;
    Log(['=> SetSpecialCommit command, erroneous number of params', Params.Count]);
    Exit;
  end;
  SpecialCommit := STI(Params[0]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET GME PARAMS
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetGMEParams(Params : TStringList) : integer;
begin
  Result := 0;
  if Params.Count <> 1 then
  begin
    Result := 1;
    Log(['=> SetGMEParams command, erroneous number of params', Params.Count]);
    Exit;
  end;
  GMEParam := Params[0];
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET NAME
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetName(Params : TStringList) : integer;
begin
  Result := 0;
  if Params.Count <> 1 then
  begin
    Result := 1;
    Log(['=> SetName command, erroneous number of params', Params.Count]);
    Exit;
  end;
  Name := Params[0];
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET SHORT NAME
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetShortName(Params : TStringList) : integer;
begin
  Result := 0;
  if Params.Count <> 1 then
  begin
    Result := 1;
    Log(['=> SetShortName command, erroneous number of params', Params.Count]);
    Exit;
  end;
  ShortName := Params[0];
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET FLAVOR NAME
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetFlavorName(Params : TStringList) : integer;
begin
  Result := 0;
  if Params.Count <> 1 then
  begin
    Result := 1;
    Log(['=> SetFlavorName command, erroneous number of params', Params.Count]);
    Exit;
  end;
  FlavorName := Params[0];
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET SU FLAVOR NAME
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetSUFlavorName(Params : TStringList) : integer;
var i : integer;
begin
  Result := 0;
  SUFlavorName := '';
  for i := 0 to Params.Count -1 do
    SUFlavorName := SUFlavorName + Params[i] + '|';
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET TEXT
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetText(Params : TStringList) : integer;
begin
  Result := 0;
  if Params.Count <> 1 then
  begin
    Result := 1;
    Log(['=> SetText command, erroneous number of params', Params.Count]);
    Exit;
  end;
  Text := Params[0];
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET TYPE
// Cas normal: 1 paramtre numrique, l'UID du TypeDef de l'objet  crer
// autre cas: 1 paramtre chaine, Nom unique du TypeDef
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetType(Params : TStringList) : integer;
var i : integer;
begin
  Result := 0;
  if Params.Count <> 1 then
  begin
    Result := 1;
    Log(['=> SetType command, erroneous number of params', Params.Count]);
    Exit;
  end;
  if IsNumeric(Params[0]) then
    ObjType := STI(Params[0])
  else
  begin
    // conversion name > UID TypeDef
    if WorkingGME = _geStruc then
    begin
      for i := 0 to StrucDefs.Count -1 do
      begin
        if UpperCase(StrucDefs[i].Name) = UpperCase(Params[0]) then
        begin
          ObjType := StrucDefs[i].UID;
          Exit;
        end;
      end;
      Inc(Result);
      Log(['=> SetType command, unable to convert the string', Params[0], 'into a valid structure identifier']);
      Exit;
    end
    else
    begin
      Inc(Result);
      Log(['=> SetType command', 'Unknown identifier (check if Alias is OK):', Params[0], 'This kind of object', WorkingGME, 'must have his identifier specified as a numeric']);
      Exit;
    end;
  end;
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET HEALTH
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetHealth(Params : TStringList) : integer;
begin
  Result := 0;
  if Params.Count <> 1 then
  begin
    Result := 1;
    Log(['=> SetHealth command, erroneous number of params', Params.Count]);
    Exit;
  end;
  HealthPerc := STI(Params[0]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET COHESION
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetCohesion(Params : TStringList) : integer;
begin
  Result := 0;
  if Params.Count <> 1 then
  begin
    Result := 1;
    Log(['=> SetCohesion command, erroneous number of params', Params.Count]);
    Exit;
  end;
  CohesionPerc := STI(Params[0]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET LEVEL
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetLevel(Params : TStringList) : integer;
begin
  Result := 0;
  if Params.Count <> 1 then
  begin
    Result := 1;
    Log(['=> SetLevel command, erroneous number of params', Params.Count]);
    Exit;
  end;
  Level := STI(Params[0]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET CONTROL
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetControl(Params : TStringList) : integer;
begin
  Result := 0;
  if not (Params.Count in [1, 2]) then
  begin
    Inc(Result);
    Log(['=> SetControl command, erroneous number of params', Params.Count]);
  end;
  if cuFaction = nil then
  begin
    Inc(Result);
    Log(['=> SetControl command, no faction selected']);
  end;
  if (cuRegion = nil) and (Params.Count = 1) then
  begin
    Log(['SetControl command, no region selected']);
    Exit;
  end;
  if Result <> 0 then Exit;

  if Params.Count = 1 then
    cuRegion.Control := STI(Params[0])
  else
  begin
    cuRegion := cuFaction.Regions.Fetch(STI(Params[0]));
    cuRegion.Control := STI(Params[1])
  end;
  Log(['SetControl', cuRegion.Name, 'set to', cuRegion.Control]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET LOYALTY (FACTION)
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetLoyaltyFac(Params : TStringList) : integer;
begin
  Result := 0;
  // on veut fixer  qui va la loyaut d'une rgion (= UID de la faction)
  if not (Params.Count in [1, 2]) then
  begin
    Inc(Result);
    Log(['=> SetLoyalty command, erroneous number of params', Params.Count]);
  end;
  if cuFaction = nil then
  begin
    Inc(Result);
    Log(['=> SetLoyalty command, no faction selected']);
  end;
  if (cuRegion = nil) and (Params.Count = 1) then
  begin
    Inc(Result);
    Log(['=> SetLoyalty command, no region selected']);
  end;
  if Result <> 0 then Exit;

  if rulSimpleLoyalty_ then
  begin
    if Params.Count = 1 then
      GameSim.RegionCmns[cuRegion.UID].SplLoyalty := STI(Params[0])
    else
    begin
      cuRegion := cuFaction.Regions.Fetch(STI(Params[0]));
      GameSim.RegionCmns[cuRegion.UID].SplLoyalty := STI(Params[1])
    end;
    Log(['SetLoyalty', cuRegion.Name, 'set to', GameSim.RegionCmns[cuRegion.UID].SplLoyalty]);
  end
  else
  begin
    if Params.Count = 1 then
      cuRegion.Loyalty := STI(Params[0])
    else
    begin
      cuRegion := cuFaction.Regions.Fetch(STI(Params[0]));
      cuRegion.Loyalty := STI(Params[1])
    end;
    Log(['SetLoyalty', cuRegion.Name, 'set to', cuRegion.Loyalty, 'for', cuFaction.Name]);
  end;
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET LOYALTY ETHNIE
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetLoyaltyEth(Params : TStringList) : integer;
begin
  Result := 0;
  if Params.Count <> 1 then
  begin
    Result := 1;
    Log(['=> SetLoyaltyEth command, erroneous number of params', Params.Count]);
    Exit;
  end;
  LoyaltyEth := STI(Params[0]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET CIV LEVEL  SetTransportLevel
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetCivLevel(Params : TStringList) : integer;
var AllRegions : boolean;
    i : integer;
    CivLevel : integer;
begin
  Result := 0;
  if not (Params.Count in [1, 2]) then
  begin
    Inc(Result);
    Log(['=> SetCivLevel command, erroneous number of params, Syntax is: CivLevel = [ALL;] Value', Params.Count]);
  end;
  AllRegions := ExtractParam('ALL', Params) <> '';

  if (cuRegion = nil) and (not AllRegions) then
  begin
    Inc(Result);
    Log(['=> SetCivLevel command, no region selected (and no ALL flag specified)']);
  end;
  if Params.Count = 1 then SetNumeric(Params[0], CivLevel)
  else
    SetNumeric(Params[1], CivLevel);
  if (CivLevel = Low(integer)) then
  begin
    Inc(Result);
    Log(['=> SetCivLevel command', 'CivLevel is not an integer']);
  end;
  if Result <> 0 then Exit;

  if Params.Count = 1 then
  begin
    if RegionDefs.IsLandRegion(cuRegion.UID) then
    begin
      GameSim.RegionCmns[cuRegion.UID].CivLevel := CivLevel;
      Log(['SetCivLevel', cuRegion.Name, 'set to', CivLevel]);
    end
    else
      Log(['SetCivLevel ignored:', cuRegion.Name, 'can''t set the civ level of a non land region']);
  end
  else
  begin
    for i := 0 to GameSim.RegionCmns.Count-1 do
    begin
      if RegionDefs.IsLandRegion(i) then
        GameSim.RegionCmns[i].CivLevel := CivLevel;
    end;
    Log(['SetCivLevel', 'all regions set to', CivLevel, 'CivLevel']);
  end;
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET REGION STYLE
// CityStyle obsolete but supported
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetRegionStyle  (Params : TStringList) : integer;
begin
  Result := 0;
  {$IF (_gameset = 'ACW')}
  Exit; // colonnes non valorises
  {$IFEND}
  if not (Params.Count in [1]) then
  begin
    Inc(Result);
    Log(['=> SetCityStyle/SetRegionStyle command, erroneous number of params, Syntax is: CityStyle = Value']);
  end;
  if cuRegion = nil then
  begin
    Inc(Result);
    Log(['=> SetRegionStyle (SetCityStyle) command, no region selected']);
  end;
  if Result > 0 then Exit;
  GameSim.RegionCmns[cuRegion.UID].RegionStyle := STI(Params[0]);
  Log(['SetRegionStyle', cuRegion.UID, cuRegion.Name, 'set to', GameSim.RegionCmns[cuRegion.UID].RegionStyle]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET FACTION STYLE VARIANT
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetFacStyle(Params : TStringList) : integer;
begin
  Result := 0;
  if not (Params.Count in [1]) then
  begin
    Inc(Result);
    Log(['=> SetFacStyle command, erroneous number of params, Syntax is: SetFacStyle = Value']);
  end;
  if cuFaction = nil then
  begin
    Inc(Result);
    Log(['=> SetFacStyle command, No Faction selected']);
  end;
  if Result > 0 then Exit;
  cuFaction.StyleVariant := STI(Params[0]);
  Log(['SetFacStyle', cuFaction.Name, 'set to', cuFaction.StyleVariant]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET TRANSPORT
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetTransport  (Params : TStringList) : integer;
var AllRegions : boolean;
    i : integer;
    TransportLvl : integer;
begin
  Result := 0;
  if not (Params.Count in [1, 2]) then
  begin
    Inc(Result);
    Log(['=> Transport command, erroneous number of params, Syntax is: Transport = [ALL;] Value', Params.Count]);
  end;
  AllRegions := ExtractParam('ALL', Params) <> '';

  if (cuRegion = nil) and (not AllRegions) then
  begin
    Inc(Result);
    Log(['=> Transport command, no region selected (and no ALL flag specified)']);
  end;
  if Params.Count = 1 then SetNumeric(Params[0], TransportLvl)
  else
    SetNumeric(Params[1], TransportLvl);
  if (TransportLvl = Low(integer)) then
  begin
    Inc(Result);
    Log(['=> Transport command', 'TransportLvl is not an integer']);
  end;
  if Result <> 0 then Exit;

  if Params.Count = 1 then
  begin
    if RegionDefs.IsLandRegion(cuRegion.UID) then
    begin // don't reduce transport level, June 2011 - if needed, add an optional third parameter
      GameSim.RegionCmns[cuRegion.UID].TransportLvl := Max(GameSim.RegionCmns[cuRegion.UID].TransportLvl, TransportLvl);
      Log(['Transport Level', cuRegion.Name, 'set to', TransportLvl]);
    end
    else
      Log(['Transport Level ignored:', cuRegion.Name, 'can''t set the transport level of a non land region']);
  end
  else
  begin
    for i := 0 to GameSim.RegionCmns.Count-1 do
    begin
      if RegionDefs.IsLandRegion(i) then
        GameSim.RegionCmns[i].TransportLvl := TransportLvl;
    end;
    Log(['Transport Level', 'all regions set to', TransportLvl, 'Transport level']);
  end;
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET VICTORY POINTS
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetVP(Params : TStringList) : integer;
var VP : integer;
begin
  Result := 0;
  if not (Params.Count in [1]) then
  begin
    Inc(Result);
    Log(['=> SetVP command, erroneous number of params, Syntax is: VP = Value', Params.Count]);
  end;

  if (cuRegion = nil) then
  begin
    Inc(Result);
    Log(['=> SetVP command, no region selected']);
  end;
  if Params.Count = 1 then SetNumeric(Params[0], VP);

  if (VP = Low(integer)) then
  begin
    Inc(Result);
    Log(['=> SetVP command', 'VP is not an integer']);
  end;
  if Result <> 0 then Exit;

  GameSim.RegionCmns[cuRegion.UID].VictoryPts := VP;
  Log(['SetVP', cuRegion.Name, 'set to', VP, 'VP(s)']);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET BLOCK STATE
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetBlockState(Params : TStringList) : integer;
var Value : integer;
    AreaUID  : TUID;
    RgnUID : TUID;
    i : integer;
    Area : TArea;
begin
  Result := 0;
  if not (Params.Count in [1, 2]) then
  begin
    Inc(Result);
    Log(['=> BlockState command, erroneous number of params, Syntax is: BlockState = <Value> [;<AreaAlias>]', Params.Count]);
  end;

  if (cuRegion = nil) and (Params.Count = 1) then
  begin
    Inc(Result);
    Log(['=> BlockState command, no region selected and no area specified.']);
  end;
  if Params.Count >= 1 then SetNumeric(Params[0], Value);
  if Params.Count >= 2 then SetNumeric(Params[1], AreaUID)
  else AreaUID := _noUID;

  if (Value = Low(integer)) then
  begin
    Inc(Result);
    Log(['=> BlockState command', 'BlockState is not an integer']);
  end;
  if Result <> 0 then Exit;

  if AreaUID = _noUID then
  begin
    GameSim.RegionCmns[cuRegion.UID].BlockState := Value;
    Log(['BlockState', cuRegion.UID, cuRegion.Name, 'set to', Value]);
  end
  else
  begin
    Area := Areas[AreaUID];

    for i := 0 to Area.Count-1 do
    begin
      RgnUID := Area[i];
      GameSim.RegionCmns[RgnUID].BlockState := Value;
      Log(['BlockState', RgnUID, GameSim.RegionCmns[RgnUID].Name, 'set to', Value]);
    end;
  end;


end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET COLONIAL PENETRATION
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetColonialPen(Params : TStringList) : integer;
begin
  Result := 0;
  if not (Params.Count in [1, 2]) then
  begin
    Inc(Result);
    Log(['=> SetColonialPen command, erroneous number of params', Params.Count]);
  end;
  if cuFaction = nil then
  begin
    Inc(Result);
    Log(['=> SetColonialPen command, no faction selected']);
  end;
  if (cuRegion = nil) and (Params.Count = 1) then
  begin
    Log(['SetColonialPen command, no region selected']);
    Exit;
  end;
  if (cuRegion <> nil)and (not RegionDefs.IsLandRegion(cuRegion.UID)) then
  begin
    Log(['SetColonialPen command, not a land region.']);
  end;
  if Result <> 0 then Exit;

  if Params.Count = 1 then
    cuRegion.ColonialPen := STI(Params[0])
  else
  begin
    cuRegion := cuFaction.Regions.Fetch(STI(Params[0]));
    cuRegion.ColonialPen := STI(Params[1])
  end;
  Log(['SetColonialPen', cuRegion.Name, 'set to', cuRegion.ColonialPen]);
end;
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// CHANGE COLONIAL PENETRATION
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.ChangeColonialPen(Params : TStringList) : integer;
begin
  Result := 0;
  if not (Params.Count in [1, 2]) then
  begin
    Inc(Result);
    Log(['=> ChgColonialPen command, erroneous number of params', Params.Count]);
  end;
  if cuFaction = nil then
  begin
    Inc(Result);
    Log(['=> ChgColonialPen command, no faction selected']);
  end;
  if (cuRegion = nil) and (Params.Count = 1) then
  begin
    Log(['ChgColonialPen command, no region selected']);
    Exit;
  end;
  if (cuRegion <> nil)and (not RegionDefs.IsLandRegion(cuRegion.UID)) then
  begin
    Log(['ChgColonialPen command, not a land region.']);
  end;
  if Result <> 0 then Exit;

  if cuRegion.ColonialPen = 101 then
  begin
    Log(['[Warning]', 'ChgColonialPen', cuRegion.Name, 'no change will be done, as already at 101. Change status instead']);
    Exit;
  end;

  if Params.Count = 1 then
    cuRegion.ColonialPen := Min(100, cuRegion.ColonialPen + STI(Params[0]))
  else
  begin
    cuRegion := cuFaction.Regions.Fetch(STI(Params[0]));
    cuRegion.ColonialPen := Min(100, cuRegion.ColonialPen +  + STI(Params[1]));
  end;
  
  Log(['ChgColonialPen', cuRegion.Name, 'changed to', cuRegion.ColonialPen]);
end;
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET COLONIAL STATUS
// SetColStatus = Value
// SetColorStatus = RgnUID Value
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetColonialStatus(Params : TStringList) : integer;
begin
  Result := 0;
  if not (Params.Count in [1, 2]) then
  begin
    Inc(Result);
    Log(['=> SetColStatus command, erroneous number of params', Params.Count]);
  end;
  if cuFaction = nil then
  begin
    Inc(Result);
    Log(['=> SetColStatus command, no faction selected']);
  end;
  if (cuRegion = nil) and (Params.Count = 1) then
  begin
    Inc(Result);
    Log(['=> SetColStatus command, no region selected']);
    Exit;
  end;
  if (cuRegion <> nil)and (not RegionDefs.IsLandRegion(cuRegion.UID)) then
  begin
    Inc(Result);
    Log(['=> SetColStatus command, not a land region.']);
  end;
  if Result <> 0 then Exit;

  if Params.Count = 1 then
    cuRegion.ColonialStatus := STI(Params[0])
  else
  begin
    cuRegion := cuFaction.Regions.Fetch(STI(Params[0]));
    cuRegion.ColonialStatus := STI(Params[1])
  end;

  // ***** Automatically done in SetColonialStatus: *****
  // If the region is in a colonial area and is the capital region, change also CP to bare minimum
  // if the status was set to OtherNonNat or Nat, then CP was set to 101 automatically

  if cuRegion.UID = GameSim.ColModule.GetColonialAreaCapitalFromRegion(cuRegion.UID) then
    GameSim.ColModule.SynchColonialStatus(GameSim.ColModule.GetColAreaFromRegion(cuRegion.UID).UID);

  Log(['SetColStatus', cuRegion.Name, 'set to', cuRegion.ColonialStatus]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET DEVELOPMENT
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetDevelopment(Params : TStringList) : integer;
var Value : integer;
begin
  Result := 0;
  if not (Params.Count in [1]) then
  begin
    Inc(Result);
    Log(['=> SetDevelopment command, erroneous number of params, Syntax is: Development = Value', Params.Count]);
  end;

  if (cuRegion = nil) then
  begin
    Inc(Result);
    Log(['=> SetDevelopment command, no region selected']);
  end;
  if Params.Count = 1 then SetNumeric(Params[0], Value);

  if (Value = Low(integer)) then
  begin
    Inc(Result);
    Log(['=> SetDevelopment command', 'Development is not an integer']);
  end;
  if (cuRegion <> nil)and (not RegionDefs.IsLandRegion(cuRegion.UID)) then
  begin
    Log(['SetDevelopment command, not a land region.']);
  end;
  if Result <> 0 then Exit;

  GameSim.RegionCmns[cuRegion.UID].Development := Value;
  Log(['SetDevelopment', cuRegion.Name, 'set to', Value]);
end;
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// CHANGE DEVELOPMENT
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.ChgDevelopment(Params : TStringList) : integer;
var Value     : integer;
begin
  Result := 0;
  if not (Params.Count in [1]) then
  begin
    Inc(Result);
    Log(['=> ChgDevelopment command, erroneous number of params, Syntax is: ChgDevelopment = Value', Params.Count]);
  end;
  if (cuRegion = nil) then
  begin
    Inc(Result);
    Log(['=> ChgDevelopment command, no region selected']);
  end;
  if (cuRegion <> nil) and (not RegionDefs.IsLandRegion(cuRegion.UID)) then
  begin
    Log(['ChgDevelopment command, not a land region.']);
  end;
  if Params.Count = 1 then SetNumeric(Params[0], Value);
  if (Value = Low(integer)) then
  begin
    Inc(Result);
    Log(['=> ChgDevelopment command', 'Development is not an integer']);
  end;  
  if Result <> 0 then Exit;
  GameSim.RegionCmns[cuRegion.UID].Development := GameSim.RegionCmns[cuRegion.UID].Development + STI(Params[0]);
  Log(['ChgDevelopment:', cuRegion.Name, 'new value: ', GameSim.RegionCmns[cuRegion.UID].Development]);  
end; 
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET PLANNED CITY NAME
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetPlannedCityName(Params : TStringList) : integer;
begin
  Result := 0;
  if not (Params.Count in [1]) then
  begin
    Inc(Result);
    Log(['=> SetPlannedCityName command, erroneous number of params, Syntax is: PlanCityName = Name', Params.Count]);
  end;

  if (cuRegion = nil) then
  begin
    Inc(Result);
    Log(['=> SetPlannedCityName command, no region selected']);
  end;
  if (cuRegion <> nil)and (not RegionDefs.IsLandRegion(cuRegion.UID)) then
  begin
    Log(['SetPlannedCityName command, not a land region.']);
  end;
  if Result <> 0 then Exit;

  GameSim.RegionCmns[cuRegion.UID].CityName := Params[0];
  Log(['SetPlannedCityName', cuRegion.Name, 'set to', Params[0]]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET POPULATION TOTAL RAW
// to be used only if the % of social classes are not set yet
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetPopTotalRaw(Params : TStringList) : integer;
var Value : integer;
begin
  Result := 0;
  if not (Params.Count in [1]) then
  begin
    Inc(Result);
    Log(['=> SetPopTotalRaw command, erroneous number of params, Syntax is: SetPopTotalRaw = Value', Params.Count]);
  end;

  if (cuRegion = nil) then
  begin
    Inc(Result);
    Log(['=> SetPopTotalRaw command, no region selected']);
  end;
  if Params.Count = 1 then SetNumeric(Params[0], Value);

  if (Value = Low(integer)) then
  begin
    Inc(Result);
    Log(['=> SetPopTotalRaw command', 'SetPopTotal is not an integer']);
  end;
  if (cuRegion <> nil)and (not RegionDefs.IsLandRegion(cuRegion.UID)) then
  begin
    Log(['SetPopTotalRaw command, not a land region.']);
  end;
  if Result <> 0 then Exit;

  GameSim.RegionCmns[cuRegion.UID].SetPopTotalRaw(Value);
  Log(['SetPopTotalRaw', cuRegion.Name, 'Set to', Value]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET POP MILITANCY
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetPopMilitPerc(Params : TStringList) : integer;
var Value : integer;
begin
  Result := 0;
  if not (Params.Count in [1]) then
  begin
    Inc(Result);
    Log(['=> SetPopMilitPerc command, erroneous number of params, Syntax is: MilitancyPerc = Value', Params.Count]);
  end;

  if (cuRegion = nil) then
  begin
    Inc(Result);
    Log(['=> SetPopMilitPerc command, no region selected']);
  end;
  if Params.Count = 1 then SetNumeric(Params[0], Value);

  if (Value = Low(integer)) then
  begin
    Inc(Result);
    Log(['=> SetPopMilitPerc command', 'Militancy is not an integer']);
  end;
  if (cuRegion <> nil)and (not RegionDefs.IsLandRegion(cuRegion.UID)) then
  begin
    Log(['SetPopMilitPerc command, not a land region.']);
  end;
  if Result <> 0 then Exit;

  GameSim.RegionCmns[cuRegion.UID].Pop_MilitancyPerc := Value;
  Log(['SetPopMilitPerc', cuRegion.Name, 'set to', Value]);
end;
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// CHANGE POPULATION - MILITANCY
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.ChgPopMilitancy(Params : TStringList) : integer;
var Value     : integer;
begin
  Result := 0;
  if not (Params.Count in [1]) then
  begin
    Inc(Result);
    Log(['=> ChgPopMilitancy command, erroneous number of params, Syntax is: ChgPopMilitancy = Value', Params.Count]);
  end;
  if (cuRegion = nil) then
  begin
    Inc(Result);
    Log(['=> ChgPopMilitancy command, no region selected']);
  end;
  if (cuRegion <> nil) and (not RegionDefs.IsLandRegion(cuRegion.UID)) then
  begin
    Log(['ChgPopMilitancy command, not a land region.']);
  end;
  if Params.Count = 1 then SetNumeric(Params[0], Value);
  if (Value = Low(integer)) then
  begin
    Inc(Result);
    Log(['=> ChgPopMilitancy command', 'PopMilitancy is not an integer']);
  end;
  if Result <> 0 then Exit;
  GameSim.RegionCmns[cuRegion.UID].Pop_MilitancyPerc := GameSim.RegionCmns[cuRegion.UID].Pop_MilitancyPerc + STI(Params[0]);
  Log(['ChgPopMilitancy:', cuRegion.Name, 'new value: ', GameSim.RegionCmns[cuRegion.UID].Pop_MilitancyPerc]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// CHANGE POP MILITANCY EX
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.ChgPopMilitancyEx (Params : TStringList) : integer;
var Value1, Value2, Value3, Value4, Value5, Value6     : integer;
    All : boolean;
    i : integer;
begin
  Result := 0;
  if not (Params.Count in [6, 7]) then
  begin
    Inc(Result);
    Log(['=> ChgPopMilitancyEx command, erroneous number of params, Syntax is: ChgPopMilitancyEx = ServileValue; PeasantValue; WorkerValue; MiddleValue; UpperValue; AristoValue [;ALL]', Params.Count]);
  end;

  All := ExtractParam('ALL', Params) <> '';

  if not All then
  begin
    if (cuRegion = nil) then
    begin
      Inc(Result);
      Log(['=> ChgPopMilitancyEx command, no region selected']);
    end;
    if (cuRegion <> nil) and (not RegionDefs.IsLandRegion(cuRegion.UID)) then
    begin
      Log(['ChgPopMilitancyEx command, not a land region.']);
    end;
  end;
  if (cuFaction = nil) then
  begin
    Inc(Result);
    Log(['=> ChgPopMilitancyEx command, no faction selected']);
  end;

  SetNumeric(Params[0], Value1);
  SetNumeric(Params[1], Value2);
  SetNumeric(Params[2], Value3);
  SetNumeric(Params[3], Value4);
  SetNumeric(Params[4], Value5);
  SetNumeric(Params[5], Value6);
  if (Value1 = Low(integer))or(Value2 = Low(integer))or(Value3 = Low(integer))
  or(Value4 = Low(integer))or(Value5 = Low(integer))or(Value6 = Low(integer)) then
  begin
    Inc(Result);
    Log(['=> ChgPopMilitancyEx command', 'One of the value is not an integer']);
  end;

  if Result <> 0 then Exit;

  if not All then
  begin
    GameSim.RegionCmns[cuRegion.UID].ChangeMilitancyPercEx(Value1, Value2, Value3, Value4, Value5, Value6);
    Log(['ChgPopMilitancyEx:', cuRegion.Name, 'new (agregated) value: ', GameSim.RegionCmns[cuRegion.UID].Pop_MilitancyPerc]);
  end
  else
  begin
    for i := 0 to cuFaction.ilNationalRgns.Count-1 do
    begin
      cuRegion := cuFaction.Regions.Fetch(cuFaction.ilNationalRgns[i]);
      if GameSim.RegionCmns.FetchOwner(cuRegion.UID) = cuFaction then
      begin
        GameSim.RegionCmns[cuRegion.UID].ChangeMilitancyPercEx(Value1, Value2, Value3, Value4, Value5, Value6);
        Log(['ChgPopMilitancyEx:', cuRegion.Name, 'new (agregated) value: ', GameSim.RegionCmns[cuRegion.UID].Pop_MilitancyPerc]);
      end;
    end;
  end;
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET POPULATION - ARISTOS - RAW
// in %
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetAristosPercRaw(Params : TStringList) : integer;
var Value : integer;
begin
  Result := 0;
  if not (Params.Count in [1]) then
  begin
    Inc(Result);
    Log(['=> SetPopAristosPerc command, erroneous number of params, Syntax is: PopAristosPerc = Value', Params.Count]);
  end;

  if (cuRegion = nil) then
  begin
    Inc(Result);
    Log(['=> SetPopAristosPerc command, no region selected']);
  end;
  if Params.Count = 1 then SetNumeric(Params[0], Value);

  if (Value = Low(integer)) then
  begin
    Inc(Result);
    Log(['=> SetPopAristosPerc command', 'PopAristos is not an integer']);
  end;
  if (cuRegion <> nil)and (not RegionDefs.IsLandRegion(cuRegion.UID)) then
  begin
    Inc(Result);
    Log(['=> SetPopAristosPerc command, not a land region.']);
  end;
  if (GameSim.RegionCmns[cuRegion.UID].GetPopTotalRaw = 0) then
  begin
    Inc(Result);
    Log(['=> SetPopAristosPerc command, Population total is not set (still at 0).']);
  end;


  if Result <> 0 then Exit;

  GameSim.RegionCmns[cuRegion.UID].Pop_AristoCount := Round(Value/100 * GameSim.RegionCmns[cuRegion.UID].GetPopTotalRaw);
  Log(['SetPopAristosPerc', cuRegion.Name, 'count set to', GameSim.RegionCmns[cuRegion.UID].Pop_AristoCount]);
end;
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// CHANGE POPULATION - ARISTOS
// in number
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.ChgAristosCount(Params : TStringList) : integer;
var Value     : integer;
begin
  Result := 0;
  if not (Params.Count in [1]) then
  begin
    Inc(Result);
    Log(['=> ChgPopAristos command, erroneous number of params, Syntax is: ChgPopAristos = Value', Params.Count]);
  end;
  if (cuRegion = nil) then
  begin
    Inc(Result);
    Log(['=> ChgPopAristos command, no region selected']);
  end;
  if (cuRegion <> nil) and (not RegionDefs.IsLandRegion(cuRegion.UID)) then
  begin
    Log(['ChgPopAristos command, not a land region.']);
  end;
  if Params.Count = 1 then SetNumeric(Params[0], Value);
  if (Value = Low(integer)) then
  begin
    Inc(Result);
    Log(['=> ChgPopAristos command', 'PopAristos is not an integer']);
  end;
  if Result <> 0 then Exit;
  GameSim.RegionCmns[cuRegion.UID].Pop_AristoCount := GameSim.RegionCmns[cuRegion.UID].Pop_AristoCount + STI(Params[0]);
  Log(['ChgPopAristos:', cuRegion.Name, 'new count: ', GameSim.RegionCmns[cuRegion.UID].Pop_AristoCount]);
end; 
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET POPULATION - UPPER
// in %
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetUpperPercRaw(Params : TStringList) : integer;
var Value : integer;
begin
  Result := 0;
  if not (Params.Count in [1]) then
  begin
    Inc(Result);
    Log(['=> SetPopUpperPerc command, erroneous number of params, Syntax is: PopUpperPerc = Value', Params.Count]);
  end;

  if (cuRegion = nil) then
  begin
    Inc(Result);
    Log(['=> SetPopUpperPerc command, no region selected']);
  end;
  if Params.Count = 1 then SetNumeric(Params[0], Value);

  if (Value = Low(integer)) then
  begin
    Inc(Result);
    Log(['=> SetPopUpperPerc command', 'PopUpper is not an integer']);
  end;
  if (cuRegion <> nil)and (not RegionDefs.IsLandRegion(cuRegion.UID)) then
  begin
     Inc(Result);
    Log(['=> SetPopUpperPerc command, not a land region.']);
  end;
  if (GameSim.RegionCmns[cuRegion.UID].GetPopTotalRaw = 0) then
  begin
    Inc(Result);
    Log(['SetPopUpperPerc command, Population total is not set (still at 0).']);
  end;
  if Result <> 0 then Exit;

  GameSim.RegionCmns[cuRegion.UID].Pop_UpperCount := Round(Value/100 * GameSim.RegionCmns[cuRegion.UID].GetPopTotalRaw);
  Log(['SetPopUpperPerc', cuRegion.Name, 'count set to', GameSim.RegionCmns[cuRegion.UID].Pop_UpperCount]);
end;
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// CHANGE POPULATION - UPPER
// in nb
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.ChgUpperCount(Params : TStringList) : integer;
var Value     : integer;
begin
  Result := 0;
  if not (Params.Count in [1]) then
  begin
    Inc(Result);
    Log(['=> ChgPopUpper command, erroneous number of params, Syntax is: ChgPopUpper = Value', Params.Count]);
  end;
  if (cuRegion = nil) then
  begin
    Inc(Result);
    Log(['=> ChgPopUpper command, no region selected']);
  end;
  if (cuRegion <> nil) and (not RegionDefs.IsLandRegion(cuRegion.UID)) then
  begin
    Log(['ChgPopUpper command, not a land region.']);
  end;
  if Params.Count = 1 then SetNumeric(Params[0], Value);
  if (Value = Low(integer)) then
  begin
    Inc(Result);
    Log(['=> ChgPopUpper command', 'PopUpper is not an integer']);
  end;  
  if Result <> 0 then Exit;
  GameSim.RegionCmns[cuRegion.UID].Pop_UpperCount := GameSim.RegionCmns[cuRegion.UID].Pop_UpperCount + STI(Params[0]);
  Log(['ChgPopUpper:', cuRegion.Name, 'new count: ', GameSim.RegionCmns[cuRegion.UID].Pop_UpperCount]);
end; 
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET POPULATION - MIDDLE
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetMiddlePercRaw(Params : TStringList) : integer;
var Value : integer;
begin
  Result := 0;
  if not (Params.Count in [1]) then
  begin
    Inc(Result);
    Log(['=> SetPopMiddlePerc command, erroneous number of params, Syntax is: PopMiddlePerc = Value', Params.Count]);
  end;

  if (cuRegion = nil) then
  begin
    Inc(Result);
    Log(['=> SetPopMiddlePerc command, no region selected']);
  end;
  if Params.Count = 1 then SetNumeric(Params[0], Value);

  if (Value = Low(integer)) then
  begin
    Inc(Result);
    Log(['=> SetPopMiddlePerc command', 'PopMiddle is not an integer']);
  end;
  if (cuRegion <> nil)and (not RegionDefs.IsLandRegion(cuRegion.UID)) then
  begin
    Inc(Result);
    Log(['=> SetPopMiddlePerc command, not a land region.']);
  end;
  if (GameSim.RegionCmns[cuRegion.UID].GetPopTotalRaw = 0) then
  begin
    Inc(Result);
    Log(['=> SetPopMiddlePerc command, Population total is not set (still at 0).']);
  end;
  if Result <> 0 then Exit;

  GameSim.RegionCmns[cuRegion.UID].Pop_MiddleCount := Round(Value/100 * GameSim.RegionCmns[cuRegion.UID].GetPopTotalRaw);
  Log(['SetPopMiddlePerc', cuRegion.Name, ' count set to', GameSim.RegionCmns[cuRegion.UID].Pop_MiddleCount]);
end;
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// CHANGE POPULATION - MIDDLE
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.ChgMiddleCount(Params : TStringList) : integer;
var Value     : integer;
begin
  Result := 0;
  if not (Params.Count in [1]) then
  begin
    Inc(Result);
    Log(['=> ChgPopMiddle command, erroneous number of params, Syntax is: ChgPopMiddle = Value', Params.Count]);
  end;
  if (cuRegion = nil) then
  begin
    Inc(Result);
    Log(['=> ChgPopMiddle command, no region selected']);
  end;
  if (cuRegion <> nil) and (not RegionDefs.IsLandRegion(cuRegion.UID)) then
  begin
    Log(['ChgPopMiddle command, not a land region.']);
  end;
  if Params.Count = 1 then SetNumeric(Params[0], Value);
  if (Value = Low(integer)) then
  begin
    Inc(Result);
    Log(['=> ChgPopMiddle command', 'PopMiddle is not an integer']);
  end;  
  if Result <> 0 then Exit;
  GameSim.RegionCmns[cuRegion.UID].Pop_MiddleCount := GameSim.RegionCmns[cuRegion.UID].Pop_MiddleCount + STI(Params[0]);
  Log(['ChgPopMiddle:', cuRegion.Name, 'new count: ', GameSim.RegionCmns[cuRegion.UID].Pop_MiddleCount]);
end; 
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET POPULATION - WORKERS
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetWorkersPercRaw(Params : TStringList) : integer;
var Value : integer;
begin
  Result := 0;
  if not (Params.Count in [1]) then
  begin
    Inc(Result);
    Log(['=> SetPopWorkersPerc command, erroneous number of params, Syntax is: PopWorkersPerc = Value', Params.Count]);
  end;

  if (cuRegion = nil) then
  begin
    Inc(Result);
    Log(['=> SetPopWorkersPerc command, no region selected']);
  end;
  if Params.Count = 1 then SetNumeric(Params[0], Value);

  if (Value = Low(integer)) then
  begin
    Inc(Result);
    Log(['=> SetPopWorkersPerc command', 'PopWorkers is not an integer']);
  end;
  if (cuRegion <> nil)and (not RegionDefs.IsLandRegion(cuRegion.UID)) then
  begin
    Inc(Result);
    Log(['=> SetPopWorkersPerc command, not a land region.']);
  end;
  if (GameSim.RegionCmns[cuRegion.UID].GetPopTotalRaw = 0) then
  begin
    Inc(Result);
    Log(['=> SetPopWorkersPerc command, Population total is not set (still at 0).']);
  end;
  if Result <> 0 then Exit;

  GameSim.RegionCmns[cuRegion.UID].Pop_WorkerCount := Round(Value/100 * GameSim.RegionCmns[cuRegion.UID].GetPopTotalRaw);
  Log(['SetPopWorkersPerc', cuRegion.Name, 'count set to', GameSim.RegionCmns[cuRegion.UID].Pop_WorkerCount]);
end;
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// CHANGE POPULATION - WORKERS
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.ChgWorkersCount(Params : TStringList) : integer;
var Value     : integer;
begin
  Result := 0;
  if not (Params.Count in [1]) then
  begin
    Inc(Result);
    Log(['=> ChgPopWorkers command, erroneous number of params, Syntax is: ChgPopWorkers = Value', Params.Count]);
  end;
  if (cuRegion = nil) then
  begin
    Inc(Result);
    Log(['=> ChgPopWorkers command, no region selected']);
  end;
  if (cuRegion <> nil) and (not RegionDefs.IsLandRegion(cuRegion.UID)) then
  begin
    Log(['ChgPopWorkers command, not a land region.']);
  end;
  if Params.Count = 1 then SetNumeric(Params[0], Value);
  if (Value = Low(integer)) then
  begin
    Inc(Result);
    Log(['=> ChgPopWorkers command', 'PopWorkers is not an integer']);
  end;  
  if Result <> 0 then Exit;
  GameSim.RegionCmns[cuRegion.UID].Pop_WorkerCount := GameSim.RegionCmns[cuRegion.UID].Pop_WorkerCount + STI(Params[0]);
  Log(['ChgPopWorkers:', cuRegion.Name, 'new count: ', GameSim.RegionCmns[cuRegion.UID].Pop_WorkerCount]);
end; 
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET POPULATION - PEASANTS
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetPeasantsPercRaw(Params : TStringList) : integer;
var Value : integer;
begin
  Result := 0;
  if not (Params.Count in [1]) then
  begin
    Inc(Result);
    Log(['=> SetPopPeasantsPerc command, erroneous number of params, Syntax is: PopPeasantsPerc = Value', Params.Count]);
  end;

  if (cuRegion = nil) then
  begin
    Inc(Result);
    Log(['=> SetPopPeasantsPerc command, no region selected']);
  end;
  if Params.Count = 1 then SetNumeric(Params[0], Value);

  if (Value = Low(integer)) then
  begin
    Inc(Result);
    Log(['=> SetPopPeasantsPerc command', 'PopPeasants is not an integer']);
  end;
  if (cuRegion <> nil)and (not RegionDefs.IsLandRegion(cuRegion.UID)) then
  begin
    Log(['=> SetPopPeasantsPerc command, not a land region.']);
  end;
  if (GameSim.RegionCmns[cuRegion.UID].GetPopTotalRaw = 0) then
  begin
    Log(['=> SetPopPeasantsPerc command, Population total is not set (still at 0).']);
  end;
  if Result <> 0 then Exit;

  GameSim.RegionCmns[cuRegion.UID].Pop_PeasantCount := Round(Value/100 * GameSim.RegionCmns[cuRegion.UID].GetPopTotalRaw);
  Log(['SetPopPeasantsPerc', cuRegion.Name, 'count set to', GameSim.RegionCmns[cuRegion.UID].Pop_PeasantCount]);
end;
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// CHANGE POPULATION - PEASANTS
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.ChgPeasantsCount(Params : TStringList) : integer;
var Value     : integer;
begin
  Result := 0;
  if not (Params.Count in [1]) then
  begin
    Inc(Result);
    Log(['=> ChgPopPeasants command, erroneous number of params, Syntax is: ChgPopPeasants = Value', Params.Count]);
  end;
  if (cuRegion = nil) then
  begin
    Inc(Result);
    Log(['=> ChgPopPeasants command, no region selected']);
  end;
  if (cuRegion <> nil) and (not RegionDefs.IsLandRegion(cuRegion.UID)) then
  begin
    Log(['ChgPopPeasants command, not a land region.']);
  end;
  if Params.Count = 1 then SetNumeric(Params[0], Value);
  if (Value = Low(integer)) then
  begin
    Inc(Result);
    Log(['=> ChgPopPeasants command', 'PopPeasants is not an integer']);
  end;  
  if Result <> 0 then Exit;
  GameSim.RegionCmns[cuRegion.UID].Pop_PeasantCount := GameSim.RegionCmns[cuRegion.UID].Pop_PeasantCount + STI(Params[0]);
  Log(['ChgPopPeasants:', cuRegion.Name, 'new count: ', GameSim.RegionCmns[cuRegion.UID].Pop_PeasantCount]);
end; 
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET POPULATION - SERVILES
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetServilesPercRaw(Params : TStringList) : integer;
var Value : integer;
begin
  Result := 0;
  if not (Params.Count in [1]) then
  begin
    Inc(Result);
    Log(['=> SetPopServilesPerc command, erroneous number of params, Syntax is: PopServilesPerc = Value', Params.Count]);
  end;

  if (cuRegion = nil) then
  begin
    Inc(Result);
    Log(['=> SetPopServilesPerc command, no region selected']);
  end;
  if Params.Count = 1 then SetNumeric(Params[0], Value);

  if (Value = Low(integer)) then
  begin
    Inc(Result);
    Log(['=> SetPopServilesPerc command', 'PopServiles is not an integer']);
  end;
  if (cuRegion <> nil)and (not RegionDefs.IsLandRegion(cuRegion.UID)) then
  begin
    Inc(Result);
    Log(['=> SetPopServilesPerc command, not a land region.']);
  end;
  if (GameSim.RegionCmns[cuRegion.UID].GetPopTotalRaw = 0) then
  begin
    Inc(Result);
    Log(['=> SetPopServilesPerc command, Population total is not set (still at 0).']);
  end;
  if Result <> 0 then Exit;

  GameSim.RegionCmns[cuRegion.UID].Pop_ServileCount := Round(Value/100 * GameSim.RegionCmns[cuRegion.UID].GetPopTotalRaw);
  Log(['SetPopServilesPerc', cuRegion.Name, 'count set to', GameSim.RegionCmns[cuRegion.UID].Pop_ServileCount]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// CHANGE POPULATION - SERVILES
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.ChgServilesCount(Params : TStringList) : integer;
var Value     : integer;
begin
  Result := 0;
  if not (Params.Count in [1]) then
  begin
    Inc(Result);
    Log(['=> ChgPopServiles command, erroneous number of params, Syntax is: ChgPopServiles = Value', Params.Count]);
  end;
  if (cuRegion = nil) then
  begin
    Inc(Result);
    Log(['=> ChgPopServiles command, no region selected']);
  end;
  if (cuRegion <> nil) and (not RegionDefs.IsLandRegion(cuRegion.UID)) then
  begin
    Log(['ChgPopServiles command, not a land region.']);
  end;
  if Params.Count = 1 then SetNumeric(Params[0], Value);
  if (Value = Low(integer)) then
  begin
    Inc(Result);
    Log(['=> ChgPopServiles command', 'PopServiles is not an integer']);
  end;
  if Result <> 0 then Exit;
  GameSim.RegionCmns[cuRegion.UID].Pop_ServileCount := GameSim.RegionCmns[cuRegion.UID].Pop_ServileCount + STI(Params[0]);
  Log(['ChgPopServiles:', cuRegion.Name, 'new count: ', GameSim.RegionCmns[cuRegion.UID].Pop_ServileCount]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// CHANGE POPULATION - SERVILES
// Use _noUID when don't care for Ethnic and Religion
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.ChgPopulation(Params : TStringList) : integer;
var Amount,rgn      : integer;
    Area, Ethnic, Religion : TUID;
    IL : TIntList;
    RgnCmn : TRegionCmn;
begin
  Result := 0;
  if not (Params.Count in [3, 4]) then
  begin
    Inc(Result);
    Log(['=> ChgPopulation command, erroneous number of params, Syntax is: ChgPopulation = Amount; EthnicUID; ReligionUID [;AreaUID]', Params.Count]);
  end;
  if (cuRegion = nil) and (Params.Count <> 4) then
  begin
    Inc(Result);
    Log(['=> ChgPopulation command, no region selected']);
  end;
  if (cuRegion <> nil) and (not RegionDefs.IsLandRegion(cuRegion.UID)) then
  begin
    Log(['ChgPopulation command, not a land region.']);
  end;
  if Params.Count = 4 then
    SetNumeric(Params[3], Area)
  else
    Area := _noUID;
  if (Area = Low(integer)) then
  begin
    Inc(Result);
    Log(['=> ChgPopulation command', 'AreaUID is not an integer']);
  end;
  if Result > 0 then Exit;
  SetNumeric(Params[0], Amount);
  SetNumeric(Params[1], Ethnic);
  SetNumeric(Params[2], Religion);
  if (Amount = 0)or(Amount = Low(integer))or(Ethnic = Low(integer))or(Religion = Low(integer)) then
  begin
    Inc(Result);
    Log(['=> ChgPopulation command', 'Amount or EthnicUID or ReligionUID is not an integer (wrong alias used) or Amount equals 0.']);
  end;
  if Result <> 0 then Exit;

  IL := TIntList.Create;
  if (Area <> _noUID) and (Area < Areas.Count) then
    IL.AssignFrom(Areas[Area])
  else
    IL.Add(cuRegion.UID);

  for rgn := 0 to IL.Count-1 do
  begin
    RgnCmn := GameSim.RegionCmns[il[rgn]];
    if not RegionDefs.IsLandRegion(RgnCmn.UID) then Continue; 
    if GameSim.RegionCmns[il[rgn]].Pop_Total + Amount <= 0 then
    begin
      GameSim.RegionCmns[il[rgn]].SetPopTotalToZero;
      Log(['ChgPopulation:', RgnCmn.UID, RgnCmn.Name, 'has now 0 population.']);
      Continue;
    end;

    RgnCmn.ChangePopulation(Amount, Ethnic, Religion);

    Log(['ChgPopulation:', RgnCmn.Name, 'new population: ', GameSim.RegionCmns[il[rgn]].Pop_Total]);
  end;
  IL.Free;
end;
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET POPULATION - EDUCATION
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetPopEducPerc(Params : TStringList) : integer;
var Value : integer;
begin
  Result := 0;
  if not (Params.Count in [1]) then
  begin
    Inc(Result);
    Log(['=> SetPopEducPerc command, erroneous number of params, Syntax is: PopEducationPerc = Value', Params.Count]);
  end;

  if (cuRegion = nil) then
  begin
    Inc(Result);
    Log(['=> SetPopEducPerc command, no region selected']);
  end;
  if Params.Count = 1 then SetNumeric(Params[0], Value);

  if (Value = Low(integer)) then
  begin
    Inc(Result);
    Log(['=> SetPopEducPerc command', 'PopEducation is not an integer']);
  end;
  if (cuRegion <> nil)and (not RegionDefs.IsLandRegion(cuRegion.UID)) then
  begin
    Log(['SetPopEducPerc command, not a land region.']);
  end;
  if Result <> 0 then Exit;

  GameSim.RegionCmns[cuRegion.UID].Pop_EducationPerc := Value;
  Log(['SetPopEducPerc', cuRegion.Name, 'set to', Value]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// CHANGE POPULATION - EDUCATION
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.ChgPopEducation(Params : TStringList) : integer;
var Value     : integer;
begin
  Result := 0;
  if not (Params.Count in [1]) then
  begin
    Inc(Result);
    Log(['=> ChgPopEducation command, erroneous number of params, Syntax is: ChgPopEducation = Value', Params.Count]);
  end;
  if (cuRegion = nil) then
  begin
    Inc(Result);
    Log(['=> ChgPopEducation command, no region selected']);
  end;
  if (cuRegion <> nil) and (not RegionDefs.IsLandRegion(cuRegion.UID)) then
  begin
    Log(['ChgPopEducation command, not a land region.']);
  end;
  if Params.Count = 1 then SetNumeric(Params[0], Value);
  if (Value = Low(integer)) then
  begin
    Inc(Result);
    Log(['=> ChgPopEducation command', 'PopEducation is not an integer']);
  end;
  if Result <> 0 then Exit;
  GameSim.RegionCmns[cuRegion.UID].Pop_EducationPerc := GameSim.RegionCmns[cuRegion.UID].Pop_EducationPerc + STI(Params[0]);
  Log(['ChgPopEducation:', cuRegion.Name, 'new value: ', GameSim.RegionCmns[cuRegion.UID].Pop_EducationPerc]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// CHANGE POP EDUCATION EX
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.ChgPopEducationEx (Params : TStringList) : integer;
var Value1, Value2, Value3, Value4, Value5, Value6     : integer;
    All : boolean;
    i : integer;
begin
  Result := 0;
  if not (Params.Count in [6, 7]) then
  begin
    Inc(Result);
    Log(['=> ChgPopEducationEx command, erroneous number of params, Syntax is: ChgPopEducationEx = ServileValue; PeasantValue; WorkerValue; MiddleValue; UpperValue; AristoValue [;ALL]', Params.Count]);
  end;

  All := ExtractParam('ALL', Params) <> '';

  if not All then
  begin
    if (cuRegion = nil) then
    begin
      Inc(Result);
      Log(['=> ChgPopEducationEx command, no region selected']);
    end;
    if (cuRegion <> nil) and (not RegionDefs.IsLandRegion(cuRegion.UID)) then
    begin
      Log(['ChgPopEducationEx command, not a land region.']);
    end;
  end;
  if (cuFaction = nil) then
  begin
    Inc(Result);
    Log(['=> ChgPopEducationEx command, no faction selected']);
  end;

  SetNumeric(Params[0], Value1);
  SetNumeric(Params[1], Value2);
  SetNumeric(Params[2], Value3);
  SetNumeric(Params[3], Value4);
  SetNumeric(Params[4], Value5);
  SetNumeric(Params[5], Value6);
  if (Value1 = Low(integer))or(Value2 = Low(integer))or(Value3 = Low(integer))
  or(Value4 = Low(integer))or(Value5 = Low(integer))or(Value6 = Low(integer)) then
  begin
    Inc(Result);
    Log(['=> ChgPopEducationEx command', 'One of the value is not an integer']);
  end;

  if Result <> 0 then Exit;

  if not All then
  begin
    GameSim.RegionCmns[cuRegion.UID].ChangeEducationPercEx(Value1, Value2, Value3, Value4, Value5, Value6);
    Log(['ChgPopEducationEx:', cuRegion.Name, 'new (agregated) value: ', GameSim.RegionCmns[cuRegion.UID].Pop_EducationPerc]);
  end
  else
  begin
    for i := 0 to cuFaction.ilNationalRgns.Count-1 do
    begin
      cuRegion := cuFaction.Regions.Fetch(cuFaction.ilNationalRgns[i]);
      if GameSim.RegionCmns.FetchOwner(cuRegion.UID) = cuFaction then
      begin
        GameSim.RegionCmns[cuRegion.UID].ChangeEducationPercEx(Value1, Value2, Value3, Value4, Value5, Value6);
        Log(['ChgPopEducationEx:', cuRegion.Name, 'new (agregated) value: ', GameSim.RegionCmns[cuRegion.UID].Pop_EducationPerc]);
      end;
    end;
  end;
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET POPULATION - CONTENTMENT
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetPopContentPerc(Params : TStringList) : integer;
var Value : integer;
begin
  Result := 0;
  if not (Params.Count in [1]) then
  begin
    Inc(Result);
    Log(['=> SetPopContentment command, erroneous number of params, Syntax is: SetPopContentment = Value', Params.Count]);
  end;

  if (cuRegion = nil) then
  begin
    Inc(Result);
    Log(['=> SetPopContentment command, no region selected']);
  end;
  if Params.Count = 1 then SetNumeric(Params[0], Value);

  if (Value = Low(integer)) then
  begin
    Inc(Result);
    Log(['=> SetPopContentment command', 'SetPopContentment is not an integer']);
  end;
  if (cuRegion <> nil)and (not RegionDefs.IsLandRegion(cuRegion.UID)) then
  begin
    Log(['SetPopContentment command, not a land region.']);
  end;
  if Result <> 0 then Exit;

  GameSim.RegionCmns[cuRegion.UID].Pop_ContentmentPerc := Value;
  Log(['SetPopContentment', cuRegion.Name, 'set to', Value]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// CHANGE POP CONTENTMENT
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.ChgPopContent(Params : TStringList) : integer;
var Value     : integer;
begin
  Result := 0;
  if not (Params.Count in [1]) then
  begin
    Inc(Result);
    Log(['=> ChgPopContent command, erroneous number of params, Syntax is: ChgPopContent = Value', Params.Count]);
  end;
  if (cuRegion = nil) then
  begin
    Inc(Result);
    Log(['=> ChgPopContent command, no region selected']);
  end;
  if (cuRegion <> nil) and (not RegionDefs.IsLandRegion(cuRegion.UID)) then
  begin
    Log(['ChgPopContent command, not a land region.']);
  end;
  if Params.Count = 1 then SetNumeric(Params[0], Value);
  if (Value = Low(integer)) then
  begin
    Inc(Result);
    Log(['=> ChgPopContent command', 'PopContent is not an integer']);
  end;
  if Result <> 0 then Exit;
  GameSim.RegionCmns[cuRegion.UID].Pop_ContentmentPerc := GameSim.RegionCmns[cuRegion.UID].Pop_ContentmentPerc + STI(Params[0]);
  Log(['ChgPopContent:', cuRegion.Name, 'new value: ', GameSim.RegionCmns[cuRegion.UID].Pop_ContentmentPerc]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// CHANGE POP CONTENTMENT EX
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.ChgPopContentmentEx (Params : TStringList) : integer;
var Value1, Value2, Value3, Value4, Value5, Value6     : integer;
    All : boolean;
    i : integer;
begin
  Result := 0;
  if not (Params.Count in [6, 7]) then
  begin
    Inc(Result);
    Log(['=> ChgPopContentmentEx command, erroneous number of params, Syntax is: ChgPopContentmentEx = ServileValue; PeasantValue; WorkerValue; MiddleValue; UpperValue; AristoValue [;ALL]', Params.Count]);
  end;

  All := ExtractParam('ALL', Params) <> '';

  if not All then
  begin
    if (cuRegion = nil) then
    begin
      Inc(Result);
      Log(['=> ChgPopContentmentEx command, no region selected']);
    end;
    if (cuRegion <> nil) and (not RegionDefs.IsLandRegion(cuRegion.UID)) then
    begin
      Log(['ChgPopContentmentEx command, not a land region.']);
    end;
  end;
  if (cuFaction = nil) then
  begin
    Inc(Result);
    Log(['=> ChgPopContentmentEx command, no faction selected']);
  end;

  SetNumeric(Params[0], Value1);
  SetNumeric(Params[1], Value2);
  SetNumeric(Params[2], Value3);
  SetNumeric(Params[3], Value4);
  SetNumeric(Params[4], Value5);
  SetNumeric(Params[5], Value6);
  if (Value1 = Low(integer))or(Value2 = Low(integer))or(Value3 = Low(integer))
  or(Value4 = Low(integer))or(Value5 = Low(integer))or(Value6 = Low(integer)) then
  begin
    Inc(Result);
    Log(['=> ChgPopContentmentEx command', 'One of the value is not an integer']);
  end;

  if Result <> 0 then Exit;

  if not All then
  begin
    GameSim.RegionCmns[cuRegion.UID].ChangeContentmentPercEx(Value1, Value2, Value3, Value4, Value5, Value6);
    Log(['ChgPopContentmentEx:', cuRegion.Name, 'new (agregated) value: ', GameSim.RegionCmns[cuRegion.UID].Pop_ContentmentPerc]);
  end
  else
  begin
    for i := 0 to cuFaction.ilNationalRgns.Count-1 do
    begin
      cuRegion := cuFaction.Regions.Fetch(cuFaction.ilNationalRgns[i]);
      if GameSim.RegionCmns.FetchOwner(cuRegion.UID) = cuFaction then
      begin
        GameSim.RegionCmns[cuRegion.UID].ChangeContentmentPercEx(Value1, Value2, Value3, Value4, Value5, Value6);
        Log(['ChgPopContentmentEx:', cuRegion.Name, 'new (agregated) value: ', GameSim.RegionCmns[cuRegion.UID].Pop_ContentmentPerc]);
      end;
    end;
  end;
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// CONVERT SOC CLASS
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.ConvertSocClass(Params : TStringList) : integer;
var 
    All       : boolean;
    Src       : TUID;
    Dst       : TUID;
    i         : integer;
    RgnUID    : TUID;
    Perc      : integer;
    Change    : integer;
    ilRegions : TIntList;
begin
  Result := 0;
  if not (Params.Count in [3..4]) then
  begin
    Inc(Result);
    Log(['=> ConvertSocClass command, erroneous number of params, Syntax is: ConvertSocClass = <SourceSocClass>;<DestSocClass>;<PercConverted>;[;All]', Params.Count]);
  end;

  All := ExtractParam('ALL', Params) <> '';

  if (cuRegion = nil) and (not All) then
  begin
    Inc(Result);
    Log(['=> ConvertSocClass command, no region selected']);
  end;
  if (cuFaction = nil) then
  begin
    Inc(Result);
    Log(['=> ConvertSocClass command, no faction selected']);
  end;

  Src := STI(params[0], True, -1);
  Dst := STI(params[1], True, -1);
  Perc:= STI(params[2], True, 0);

  if (Src = _noUID) or (Dst = _noUID) or (Perc <= 0) then
  begin
    Inc(Result);
    Log(['=> ConvertSocClass command, one of the alias was not converted correctly.', Src, Dst]);
  end;
  Bound(Perc, 0, 100);
  
  if Result <> 0 then Exit;

  ilRegions := TIntList.Create;

  if All then
  begin
    for i := 0 to cuFaction.ilNationalRgns.Count-1 do
    begin
      RgnUID := cuFaction.ilNationalRgns[i];
      if GameSim.RegionCmns[RgnUID].OwningFac <> cuFaction then Continue;
      ilRegions.Add(RgnUID);
    end;
  end
  else ilRegions.Add(cuRegion.UID);

  for i := 0 to ilRegions.Count-1 do
  begin
    RgnUID := ilRegions[i];

    case Src of
      _popServiles    : begin Change := RoundRandom(GameSim.RegionCmns[RgnUID].Pop_ServileCount * Perc/100); GameSim.RegionCmns[RgnUID].Pop_ServileCount := GameSim.RegionCmns[RgnUID].Pop_ServileCount - Change; end;
      _popPeasants    : begin Change := RoundRandom(GameSim.RegionCmns[RgnUID].Pop_PeasantCount * Perc/100); GameSim.RegionCmns[RgnUID].Pop_PeasantCount := GameSim.RegionCmns[RgnUID].Pop_PeasantCount - Change; end;
      _popWorkers     : begin Change := RoundRandom(GameSim.RegionCmns[RgnUID].Pop_WorkerCount * Perc/100);  GameSim.RegionCmns[RgnUID].Pop_WorkerCount := GameSim.RegionCmns[RgnUID].Pop_WorkerCount - Change; end;
      _popMiddleClass : begin Change := RoundRandom(GameSim.RegionCmns[RgnUID].Pop_MiddleCount * Perc/100);  GameSim.RegionCmns[RgnUID].Pop_MiddleCount := GameSim.RegionCmns[RgnUID].Pop_MiddleCount - Change; end;
      _popUpperClass  : begin Change := RoundRandom(GameSim.RegionCmns[RgnUID].Pop_UpperCount * Perc/100);   GameSim.RegionCmns[RgnUID].Pop_UpperCount := GameSim.RegionCmns[RgnUID].Pop_UpperCount - Change; end;
      _popAristocrats : begin Change := RoundRandom(GameSim.RegionCmns[RgnUID].Pop_AristoCount * Perc/100);  GameSim.RegionCmns[RgnUID].Pop_AristoCount := GameSim.RegionCmns[RgnUID].Pop_AristoCount - Change; end;
      else Change := 0;
    end;
    if Change > 0 then
    case Dst of
      _popServiles    : GameSim.RegionCmns[RgnUID].Pop_ServileCount := GameSim.RegionCmns[RgnUID].Pop_ServileCount + Change;
      _popPeasants    : GameSim.RegionCmns[RgnUID].Pop_PeasantCount := GameSim.RegionCmns[RgnUID].Pop_PeasantCount + Change;
      _popWorkers     : GameSim.RegionCmns[RgnUID].Pop_WorkerCount := GameSim.RegionCmns[RgnUID].Pop_WorkerCount + Change;
      _popMiddleClass : GameSim.RegionCmns[RgnUID].Pop_MiddleCount := GameSim.RegionCmns[RgnUID].Pop_MiddleCount + Change;
      _popUpperClass  : GameSim.RegionCmns[RgnUID].Pop_UpperCount := GameSim.RegionCmns[RgnUID].Pop_UpperCount + Change;
      _popAristocrats : GameSim.RegionCmns[RgnUID].Pop_AristoCount := GameSim.RegionCmns[RgnUID].Pop_AristoCount + Change;
    end;
    Log(['ConvertSocClass:', cuFaction.Name, RgnUID, GameSim.RegionCmns[RgnUID].Name, ' changed', 'Src Pop Class', Src, 'Dst Pop Class', Dst, 'Conversion Perc', Perc]);
  end;



  ilRegions.Free;
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// CHANGE POPULATION DATA BY ETHNIC/RELIGION
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.ChgPopDataByEthnic(Params : TStringList) : integer;
var Value1, Value2, Value3, TheUID     : integer;
    All : boolean;
    i : integer;
begin
  Result := 0;
  if not (Params.Count in [4]) then
  begin
    Inc(Result);
    Log(['=> ChgPopDataByEthnic command, erroneous number of params, Syntax is: ChgPopDataByEthnic = EthnicUID; ContentmentChg; EducationChg; MilitancyChg [;ALL]', Params.Count]);
  end;

  All := ExtractParam('ALL', Params) <> '';

  if not All then
  begin
    if (cuRegion = nil) then
    begin
      Inc(Result);
      Log(['=> ChgPopDataByEthnic command, no region selected']);
    end;
    if (cuRegion <> nil) and (not RegionDefs.IsLandRegion(cuRegion.UID)) then
    begin
      Log(['ChgPopDataByEthnic command, not a land region.']);
    end;
  end;
  if (cuFaction = nil) then
  begin
    Inc(Result);
    Log(['=> ChgPopDataByEthnic command, no faction selected']);
  end;

  SetNumeric(Params[0], TheUID);
  SetNumeric(Params[1], Value1);
  SetNumeric(Params[2], Value2);
  SetNumeric(Params[3], Value3);

  if (TheUID = Low(integer))or(Value1 = Low(integer))or(Value2 = Low(integer))or(Value3 = Low(integer)) then
  begin
    Inc(Result);
    Log(['=> ChgPopDataByEthnic command', 'One of the value is not an integer']);
  end;

  if Result <> 0 then Exit;

  if not All then
  begin
    GameSim.RegionCmns[cuRegion.UID].ChangePopDataByEthnic(TheUID, Value1, Value2, Value3);
    Log(['ChgPopDataByEthnic:', cuRegion.Name, 'new Contentment value: ', GameSim.RegionCmns[cuRegion.UID].Pop_ContentmentPerc]);
    Log(['ChgPopDataByEthnic:', cuRegion.Name, 'new Education value: ', GameSim.RegionCmns[cuRegion.UID].Pop_EducationPerc]);
    Log(['ChgPopDataByEthnic:', cuRegion.Name, 'new Militancy value: ', GameSim.RegionCmns[cuRegion.UID].Pop_MilitancyPerc]);
  end
  else
  begin
    for i := 0 to cuFaction.ilNationalRgns.Count-1 do
    begin
      cuRegion := cuFaction.Regions.Fetch(cuFaction.ilNationalRgns[i]);
      if GameSim.RegionCmns.FetchOwner(cuRegion.UID) = cuFaction then
      begin
        GameSim.RegionCmns[cuRegion.UID].ChangePopDataByEthnic(TheUID, Value1, Value2, Value3);
        Log(['ChgPopDataByEthnic:', cuRegion.Name, 'new Contentment value: ', GameSim.RegionCmns[cuRegion.UID].Pop_ContentmentPerc]);
        Log(['ChgPopDataByEthnic:', cuRegion.Name, 'new Education value: ', GameSim.RegionCmns[cuRegion.UID].Pop_EducationPerc]);
        Log(['ChgPopDataByEthnic:', cuRegion.Name, 'new Militancy value: ', GameSim.RegionCmns[cuRegion.UID].Pop_MilitancyPerc]);
      end;
    end;
  end;
end;

function TScriptEngine.ChgPopDataByReligion( Params : TStringList) : integer;
var Value1, Value2, Value3, TheUID     : integer;
    All : boolean;
    i : integer;
begin
  Result := 0;
  if not (Params.Count in [4]) then
  begin
    Inc(Result);
    Log(['=> ChgPopDataByReligion command, erroneous number of params, Syntax is: ChgPopDataByReligion = ReligionUID; ContentmentChg; EducationChg; MilitancyChg [;ALL]', Params.Count]);
  end;

  All := ExtractParam('ALL', Params) <> '';

  if not All then
  begin
    if (cuRegion = nil) then
    begin
      Inc(Result);
      Log(['=> ChgPopDataByReligion command, no region selected']);
    end;
    if (cuRegion <> nil) and (not RegionDefs.IsLandRegion(cuRegion.UID)) then
    begin
      Log(['ChgPopDataByReligion command, not a land region.']);
    end;
  end;
  if (cuFaction = nil) then
  begin
    Inc(Result);
    Log(['=> ChgPopDataByReligion command, no faction selected']);
  end;

  SetNumeric(Params[0], TheUID);
  SetNumeric(Params[1], Value1);
  SetNumeric(Params[2], Value2);
  SetNumeric(Params[3], Value3);

  if (TheUID = Low(integer))or(Value1 = Low(integer))or(Value2 = Low(integer))or(Value3 = Low(integer)) then
  begin
    Inc(Result);
    Log(['=> ChgPopDataByReligion command', 'One of the value is not an integer']);
  end;

  if Result <> 0 then Exit;

  if not All then
  begin
    GameSim.RegionCmns[cuRegion.UID].ChangePopDataByReligion(TheUID, Value1, Value2, Value3);
    Log(['ChgPopDataByReligion:', cuRegion.Name, 'new Contentment value: ', GameSim.RegionCmns[cuRegion.UID].Pop_ContentmentPerc]);
    Log(['ChgPopDataByReligion:', cuRegion.Name, 'new Education value: ', GameSim.RegionCmns[cuRegion.UID].Pop_EducationPerc]);
    Log(['ChgPopDataByReligion:', cuRegion.Name, 'new Militancy value: ', GameSim.RegionCmns[cuRegion.UID].Pop_MilitancyPerc]);
  end
  else
  begin
    for i := 0 to cuFaction.ilNationalRgns.Count-1 do
    begin
      cuRegion := cuFaction.Regions.Fetch(cuFaction.ilNationalRgns[i]);
      if GameSim.RegionCmns.FetchOwner(cuRegion.UID) = cuFaction then
      begin
        GameSim.RegionCmns[cuRegion.UID].ChangePopDataByReligion(TheUID, Value1, Value2, Value3);
        Log(['ChgPopDataByReligion:', cuRegion.Name, 'new Contentment value: ', GameSim.RegionCmns[cuRegion.UID].Pop_ContentmentPerc]);
        Log(['ChgPopDataByReligion:', cuRegion.Name, 'new Education value: ', GameSim.RegionCmns[cuRegion.UID].Pop_EducationPerc]);
        Log(['ChgPopDataByReligion:', cuRegion.Name, 'new Militancy value: ', GameSim.RegionCmns[cuRegion.UID].Pop_MilitancyPerc]);
      end;
    end;
  end;

end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// CHANGE SELL TRANSACTION
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.ChangeSellTransac (Params : TStringList) : integer;
var AreaUID, ResUID, Amount, Price : integer;
    ERI : TEcoResourceInfos;
begin
  Result := 0;
  if not (Params.Count in [4]) then
  begin
    Inc(Result);
    Log(['=> ChangeSellTransaction command, erroneous number of params (need 4), Syntax is: ChangeSellTransaction = <AreaUID>; <ResourceUID>; <Amount>; <Price_Or-1>', Params.Count]);
  end;

  if (cuFaction = nil) then
  begin
    Inc(Result);
    Log(['=> ChangeSellTransaction command, no faction selected']);
  end;

  AreaUID := STI(Params[0], True, _noUID);
  ResUID  := STI(Params[1], True, _noUID);
  Amount  := STI(Params[2], True, -999999);
  Price   := STI(Params[3], True, -1);

  if (AreaUID <= -1) or (AreaUID > Areas.Count-1) then
  begin
    Inc(Result);
    Log(['=> ChangeSellTransaction command', 'Area alias is not an integer (erroneous alias) or is not a valid area.']);
  end;
  if (ResUID <= -1) or (ResUID > GameSim.MerchandisesDB.Count-1) then
  begin
    Inc(Result);
    Log(['=> ChangeSellTransaction command', 'Resource alias is not an integer (erroneous alias) or is not a valid resource.']);
  end;
  if Result <> 0 then Exit;

  cuFaction.ChangeResourceSold(AreaUID, ResUID, Amount, Price, False); // no check on Sell Limit, during scen generation data are not all done
  ERI := cuFaction.GetResourceBeingSold(AreaUID, ResUID);
  if ERI = nil then
    Log(['ChangeSellTransaction', 'This resource:', GameSim.MerchandisesDB[ResUID].Name, 'is not sold anymore by', cuFaction.Name, 'in', Areas[AreaUID].UID, Areas[AreaUID].Name])
  else
    Log(['ChangeSellTransaction', 'This resource:', GameSim.MerchandisesDB[ResUID].Name, 'is sold by', cuFaction.Name, 'in', Areas[AreaUID].UID, Areas[AreaUID].Name, 'with this amount:', ERI.Amount, 'and Price:', ERI.Price])
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// CHANGE BUY TRANSACTION
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.ChangeBuyTransac  (Params : TStringList) : integer;
var AreaUID, ResUID, Amount, Price : integer;
    ERI : TEcoResourceInfos;
begin
  Result := 0;
  if not (Params.Count in [4]) then
  begin
    Inc(Result);
    Log(['=> ChangeBuyTransaction command, erroneous number of params (need 4), Syntax is: ChangeBuyTransaction = <AreaUID>; <ResourceUID>; <Amount>; <Price_Or-1>', Params.Count]);
  end;

  if (cuFaction = nil) then
  begin
    Inc(Result);
    Log(['=> ChangeBuyTransaction command, no faction selected']);
  end;

  AreaUID := STI(Params[0], True, _noUID);
  ResUID  := STI(Params[1], True, _noUID);
  Amount  := STI(Params[2], True, -999999);
  Price   := STI(Params[3], True, -1);

  if (AreaUID <= -1) or (AreaUID > Areas.Count-1) then
  begin
    Inc(Result);
    Log(['=> ChangeBuyTransaction command', 'Area alias is not an integer (erroneous alias) or is not a valid area.']);
  end;
  if (ResUID <= -1) or (ResUID > GameSim.MerchandisesDB.Count-1) then
  begin
    Inc(Result);
    Log(['=> ChangeBuyTransaction command', 'Resource alias is not an integer (erroneous alias) or is not a valid resource.']);
  end;
  if Result <> 0 then Exit;

  cuFaction.ChangeResourceBought(AreaUID, ResUID, Amount, Price, True);  // no check on Buy Limit, during scen generation data are not all done
  ERI := cuFaction.GetResourcesBoughtInfos(AreaUID, ResUID);
  if ERI = nil then
    Log(['ChangeBuyTransaction', 'This resource:', GameSim.MerchandisesDB[ResUID].Name, 'is not bought anymore by', cuFaction.Name, 'in', Areas[AreaUID].UID, Areas[AreaUID].Name])
  else
    Log(['ChangeBuyTransaction', 'This resource:', GameSim.MerchandisesDB[ResUID].Name, 'is bought by', cuFaction.Name, 'in', Areas[AreaUID].UID, Areas[AreaUID].Name, 'with this amount:', ERI.Amount, 'and Price:', ERI.Price])
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// CHANGE INTRINSIC PRODUCTION COEFFICIENT
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.ChangeIntrProd(Params : TStringList) : integer;
var UID, Amount, i : integer;
    Faction : TFaction;
begin
  Result := 0;
  if rulEconomyMode_ < _gmdAdvanced then
  begin
    Inc(Result);
    Log(['=> ChangeIntrProd command, Rule Advanced Economy is not active. Command ignored.']);
  end;
  if (cuFaction = nil)and(Params.Count <= 2) then
  begin
    Inc(Result);
    Log(['=> ChangeIntrProd command, no faction selected.']);
  end;
  if  (Params.Count < 2) then
  begin
    Inc(Result);
    Log(['=> ChangeIntrProd command, erroneous number of params, Syntax is: ChangeIntrProd = ResAlias; Coeff<;FAC1><;FAC2> ..']);
  end;
  UID := STI(Params[0], True, -1);
  Amount := STI(Params[1]);

  if (UID < 0) or (UID > GameSim.MerchandisesDB.Count-1) then
  begin
    Inc(Result);
    Log(['=> ChangeIntrProd command, invalid resource UID, found:', UID]);
  end;
  if Result > 0 then Exit;

  if Params.Count = 2 then
  begin
    cuFaction.ChangeIntrinsicProdCoeff_SE(UID, Amount);
    Log(['ChangeIntrProd command (Single Faction), Resource', GameSim.MerchandisesDB[UID].Name, ' coefficient changed to ', cuFaction.GetIntrinsicProdCoeff_SE(UID)]);
  end
  else
  if Params.Count > 2 then
  begin
    for i := 2 to Params.Count-1 do
    begin
      Faction := FactionTagToFaction(Params[i]);
      if Faction = nil then
      begin
        Log(['=> [Warning] ChangeIntrProd command, Faction not found', Params[i]]);
        Continue;
      end;
      Faction.ChangeIntrinsicProdCoeff_SE(UID, Amount);
      Log(['ChangeIntrProd command', Faction.Name, 'Resource', GameSim.MerchandisesDB[UID].Name, ' coefficient changed to ', Faction.GetIntrinsicProdCoeff_SE(UID)]);
    end;
  end;
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// CHANGE RESOURCE DISCOVERY PERCENTAGE
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.ChangeResDiscovery(Params : TStringList) : integer;
var UID, Perc, i : integer;
    Faction : TFaction;
begin
  Result := 0;
  if rulEconomyMode_ < _gmdAdvanced then
  begin
    Inc(Result);
    Log(['=> ChangeResDiscovery command, Rule Advanced Economy is not active. Command ignored.']);
  end;
  if (cuFaction = nil)and(Params.Count <= 2) then
  begin
    Inc(Result);
    Log(['=> ChangeResDiscovery command, no faction selected.']);
  end;
  if  (Params.Count < 2) then
  begin
    Inc(Result);
    Log(['=> ChangeResDiscovery command, erroneous number of params, Syntax is: ChangeResDiscovery = ResAlias; Perc<;FAC1><;FAC2> ..']);
  end;
  UID := STI(Params[0], True, -1);
  Perc := STI(Params[1]);

  if (UID < 0) or (UID > GameSim.MerchandisesDB.Count-1) then
  begin
    Inc(Result);
    Log(['=> ChangeResDiscovery command, invalid resource UID, found:', UID]);
  end;
  if Result > 0 then Exit;

  if Params.Count = 2 then
  begin
    cuFaction.ChangeResDiscoveryPerc(UID, Perc);
    Log(['ChangeResDiscovery command (Single Faction), Resource', GameSim.MerchandisesDB[UID].Name, ' % changed to ', cuFaction.GetResDiscoveryPerc(UID)]);
  end
  else
  if Params.Count > 2 then
  begin
    for i := 2 to Params.Count-1 do
    begin
      Faction := FactionTagToFaction(Params[i]);
      if Faction = nil then
      begin
        Log(['=> [Warning] ChangeResDiscovery command, Faction not found', Params[i]]);
        Continue;
      end;
      Faction.ChangeResDiscoveryPerc(UID, Perc);
      Log(['ChangeResDiscovery command', Faction.Name, 'Resource', GameSim.MerchandisesDB[UID].Name, ' % changed to ', Faction.GetResDiscoveryPerc(UID)]);
    end;
  end;
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// POPULATION - RELIGIONS
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.ChgPopReligions(Params : TStringList) : integer;
var i         : integer;
    ObjectUID : TUID;
    Value     : integer;
begin
  Result := 0;
  if not (Params.Count > 0) then
  begin
    Inc(Result);
    Log(['=> ChgPopReligions: Syntax is ChgPopReligions = [;<ReligionAlias|Value>] ..']);
  end;
  if Params.Count mod 2 <> 0 then
  begin
    Inc(Result);
    Log(['=> ChgPopReligions, Specify a list of paired values: ReligionAlias|Value']);
  end;
  if (cuRegion = nil) then
  begin
    Inc(Result);
    Log(['=> ChgPopReligions command, no region selected']);
  end;
  if (cuRegion <> nil)and (not RegionDefs.IsLandRegion(cuRegion.UID)) then
  begin
    Log(['ChgPopReligions command, not a land region.']);
  end;
  if Result <> 0 then Exit;

  for i := 1 to Params.Count div 2 do
  begin
    ObjectUID := STI(Params[(i-1)*2]);
    Value     := STI(Params[(i-1)*2 + 1]);
    GameSim.RegionCmns[cuRegion.UID].dlReligions.AlterValByKey(ObjectUID, Value, True);  // somme arithmtique & suppression si <0
    Log(['ChgPopReligions', cuRegion.Name, 'Religion of alias', ObjectUID, 'changed by:', Value, '%']);
  end;
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// POPULATION - ETHNICS
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.ChgPopEthnics(Params : TStringList) : integer;
var i         : integer;
    ObjectUID : TUID;
    Value     : integer;
begin
  Result := 0;
  if not (Params.Count > 0) then
  begin
    Inc(Result);
    Log(['=> ChgPopEthnics: Syntax is ChgPopEthnics = [;<EthnicAlias|Value>] ..']);
  end;
  if Params.Count mod 2 <> 0 then
  begin
    Inc(Result);
    Log(['=> ChgPopEthnics, Specify a list of paired values: EthnicAlias|Value']);
  end;
  if (cuRegion = nil) then
  begin
    Inc(Result);
    Log(['=> ChgPopEthnics command, no region selected']);
  end;
  if (cuRegion <> nil)and (not RegionDefs.IsLandRegion(cuRegion.UID)) then
  begin
    Log(['ChgPopEthnics command, not a land region.']);
  end;
  if Result <> 0 then Exit;

  for i := 1 to Params.Count div 2 do
  begin
    ObjectUID := STI(Params[(i-1)*2]);
    Value     := STI(Params[(i-1)*2 + 1]);
    GameSim.RegionCmns[cuRegion.UID].dlEthnics.AlterValByKey(ObjectUID, Value, True);  // somme arithmtique & suppression si <0
    Log(['ChgPopEthnics', cuRegion.Name, 'Ethnic of alias', ObjectUID, 'changed by:', Value, '%']);
  end;
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET COLONIAL PEN/REVOLT 
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//function  TScriptEngine.SetColonialPen	(Params : TStringList) : integer;
//begin
//end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET FACTION ATTRIBUTE 
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetFactionAttribute(Params : TStringList) : integer;
  var AttKind  : integer;
      AttValue : integer;

begin
  Result := 0;

  if Params.Count <> 2 then
  begin
    Inc(Result);
    Log(['=> SetFactionAttribute, need to have exactly 2 parameter: The Attribute Kind (e.g. $fatTech) and the Attribute Value (e.g. $fatTechStandart)']);
    Exit;
  end;
  if cuFaction = nil then
  begin
    Log(['=> SetFactionAttribute: No faction selected']);
    Inc(Result);
    Exit;
  end;
  if not IsNumeric(Params[0]) then
  begin
    Log(['=> SetFactionAttribute: Parameter 1 is not a number. Check if correct Alias used.']);
    Inc(Result);
    Exit;
  end;
  if not IsNumeric(Params[1]) then
  begin
    Log(['=> SetFactionAttribute: Parameter 2 is not a number. Check if correct Alias used.']);
    Inc(Result);
    Exit;
  end;

  AttKind  := STI(Params[0]);
  AttValue := STI(Params[1]);

  if AttKind = _fatTech then
  begin
    if (AttValue >= 0) and (AttValue <= _fatTechNb -1) then
    begin
      cuFaction.Attributes.Tech := AttValue;
      Log(['SetFactionAttribute: Attribute "Tech" of Faction', cuFaction.Name, 'set to', AttValue]);
    end
    else
    begin
      Log(['=> SetFactionAttribute: Attribute Value out of range, set to NoUID.']);
      Inc(Result);
      cuFaction.Attributes.Tech := _NoUID;
    end;
  end
  else
  if AttKind = _fatTrade then
  begin
    if (AttValue >= 0) and (AttValue <= _fatTradeNb -1) then
    begin
      cuFaction.Attributes.Trade := AttValue;
      Log(['SetFactionAttribute: Attribute "Trade" of Faction', cuFaction.Name, 'set to', AttValue]);
    end
    else
    begin
      Log(['=> SetFactionAttribute: Attribute Value out of range, set to NoUID.']);
      Inc(Result);
      cuFaction.Attributes.Trade := _NoUID;
    end;
  end
  else
  if AttKind = _fatGovernment then
  begin
    if (AttValue >= 0) and (AttValue <= _fatGovernmentNb -1) then
    begin
      cuFaction.Attributes.Government := AttValue;
      Log(['SetFactionAttribute: Attribute "Government" of Faction', cuFaction.Name, 'set to', AttValue]);
    end
    else
    begin
      Log(['=> SetFactionAttribute: Attribute Value out of range, set to NoUID.']);
      Inc(Result);
      cuFaction.Attributes.Government := _NoUID;
    end;
  end
  else
  if AttKind = _fatSocial then
  begin
    if (AttValue >= 0) and (AttValue <= _fatSocialNb -1) then
    begin
      cuFaction.Attributes.Social := AttValue;
      Log(['SetFactionAttribute: Attribute "Social" of Faction', cuFaction.Name, 'set to', AttValue]);
    end
    else
    begin
      Log(['=> SetFactionAttribute: Attribute Value out of range, set to NoUID.']);
      Inc(Result);
      cuFaction.Attributes.Social := _NoUID;
    end;
  end
  else
  if AttKind = _fatReligion then
  begin
    if (AttValue >= 0) and (AttValue <= _fatReligionNb -1) then
    begin
      cuFaction.Attributes.Religion := AttValue;
      Log(['SetFactionAttribute: Attribute "Religion" of Faction', cuFaction.Name, 'set to', AttValue]);
    end
    else
    begin
      Log(['=> SetFactionAttribute: Attribute Value out of range, set to NoUID.']);
      Inc(Result);
      cuFaction.Attributes.Religion := _NoUID;
    end;
  end
  else
  if AttKind = _fatEconomy then
  begin
    if (AttValue >= 0) and (AttValue <= _fatEconomyNb -1) then
    begin
      cuFaction.Attributes.Economy := AttValue;
      Log(['SetFactionAttribute: Attribute "Economy" of Faction', cuFaction.Name, 'set to', AttValue]);
    end
    else
    begin
      Log(['=> SetFactionAttribute: Attribute Value out of range, set to NoUID.']);
      Inc(Result);
      cuFaction.Attributes.Economy := _NoUID;
    end;
  end
  else
  if AttKind = _fatMilitary then
  begin
    if (AttValue >= 0) and (AttValue <= _fatMilitaryNb -1) then
    begin
      cuFaction.Attributes.Military := AttValue;
      Log(['SetFactionAttribute: Attribute "Military" of Faction', cuFaction.Name, 'set to', AttValue]);
    end
    else
    begin
      Log(['=> SetFactionAttribute: Attribute Value out of range, set to NoUID.']);
      Inc(Result);
      cuFaction.Attributes.Military := _NoUID;
    end;
  end
  else
  if AttKind = _fatProjection then
  begin
    if (AttValue >= 0) and (AttValue <= _fatProjectionNb -1) then
    begin
      cuFaction.Attributes.Projection := AttValue;
      Log(['SetFactionAttribute: Attribute "Projection" of Faction', cuFaction.Name, 'set to', AttValue]);
    end
    else
    begin
      Log(['=> SetFactionAttribute: Attribute Value out of range, set to NoUID.']);
      Inc(Result);
      cuFaction.Attributes.Projection := _NoUID;
    end;
  end
  else
  if AttKind = _fatEducation then
  begin
    if (AttValue >= 0) and (AttValue <= _fatEducationNb -1) then
    begin
      cuFaction.Attributes.Education := AttValue;
      Log(['SetFactionAttribute: Attribute "Education" of Faction', cuFaction.Name, 'set to', AttValue]);
    end
    else
    begin
      Log(['=> SetFactionAttribute: Attribute Value out of range, set to NoUID.']);
      Inc(Result);
      cuFaction.Attributes.Education := _NoUID;
    end;
  end
  else
  if AttKind = _fatNationIdent then
  begin
    if (AttValue >= 0) and (AttValue <= _fatNationIdentNb -1) then
    begin
      cuFaction.Attributes.NationIdent := AttValue;
      Log(['SetFactionAttribute: Attribute "NationIdent" of Faction', cuFaction.Name, 'set to', AttValue]);
    end
    else
    begin
      Log(['=> SetFactionAttribute: Attribute Value out of range, set to NoUID.']);
      Inc(Result);
      cuFaction.Attributes.NationIdent := _NoUID;
    end;
  end
  else
  if AttKind = _fatBureaucracy then
  begin
    if (AttValue >= 0) and (AttValue <= _fatBureaucracyNb -1) then
    begin
      cuFaction.Attributes.Bureaucracy := AttValue;
      Log(['SetFactionAttribute: Attribute "Bureaucracy" of Faction', cuFaction.Name, 'set to', AttValue]);
    end
    else
    begin
      Log(['=> SetFactionAttribute: Attribute Value out of range, set to NoUID.']);
      Inc(Result);
      cuFaction.Attributes.Bureaucracy := _NoUID;
    end;
  end
  else
  if AttKind = _fatOrganization then
  begin
    if (AttValue >= 0) and (AttValue <= _fatOrganizationNb -1) then
    begin
      cuFaction.Attributes.Organization := AttValue;
      Log(['SetFactionAttribute: Attribute "Organization" of Faction', cuFaction.Name, 'set to', AttValue]);
    end
    else
    begin
      Log(['=> SetFactionAttribute: Attribute Value out of range, set to NoUID.']);
      Inc(Result);
      cuFaction.Attributes.Organization := _NoUID;
    end;
  end
  else
  begin
    Log(['=> SetFactionAttribute: Unknown Attribute Kind.']);
    Inc(Result);
  end;

end;


// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET TECH DISCOVERY
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function  TScriptEngine.SetTechDiscovery(Params : TStringList) : integer;
  var RDidx    : integer;
      RI_UID   : integer;
      DiscPerc : integer;
      RI       : TResearchItem;

begin
  Result := 0;
  RI := nil;

  if rulResearches_ = False then
  begin
    Inc(Result);
    Log(['=> SetTechDiscovery: Game rule not activated, command ignored']);
    Exit;
  end;
  if not (Params.Count = 2) then
  begin
    Inc(Result);
    Log(['=> SetTechDiscovery: Syntax is SetTechDiscovery = <RIAlias>;<Value>']);
  end;
  if cuFaction = nil then
  begin
    Log(['=> SetTechDiscovery: No faction selected']);
    Inc(Result);
  end;
  if not IsNumeric(Params[0]) then
  begin
    Log(['=> SetTechDiscovery: Parameter 1 is not a number. Check if correct Alias used.']);
    Inc(Result);
  end;
  if not IsNumeric(Params[1]) then
  begin
    Log(['=> SetTechDiscovery: Parameter 2 is not a number. Must be Discovery Percentage: [0..100]']);
    Inc(Result);
  end;

  if Result > 0 then Exit;

  RI_UID   := STI(Params[0]);
  DiscPerc := STI(Params[1]);

  if (DiscPerc < 0) or (DiscPerc > 100) then
  begin
    Log(['=> SetTechDiscovery: Discovery Percentage =', DiscPerc, 'Out of allowed range [0..100]']);
    Inc(Result);
    Exit;
  end;
  if (RI_UID < 0) or (RI_UID > TGameSim(Owner).ResearchDB.Last.UID) then
  begin
    Log(['=> SetTechDiscovery: RI UID =', RI_UID, 'Out of range']);
    Inc(Result);
    Exit;
  end;
  if cuFaction.olResearch.IsMastered(RI_UID) then
  begin
    Log(['SetTechDiscovery: RI',RI_UID ,'already mastered, Discovery Percentage can not be changed anymore']);
    Exit;
  end;

  if cuFaction.olResearch.IsKnown(RI_UID) then
  begin
    cuFaction.olResearch.GetByUID(RI_UID).Discovery := DiscPerc;
  end
  else
  for RDidx := 0 to TGameSim(Owner).ResearchDB.Count -1 do
  begin
    if TGameSim(Owner).ResearchDB.Items[RDidx].UID = RI_UID then
    begin
      Log(['SetTechDiscovery: RI not known, will be added']);
      RI := cuFaction.olResearch.AddItem;
      RI.UID  := TGameSim(Owner).ResearchDB.Items[RDidx].UID;
      RI.Name := TGameSim(Owner).ResearchDB.Items[RDidx].Name;
      RI.Kind := TGameSim(Owner).ResearchDB.Items[RDidx].Kind;
      RI.TurnKnown := TGameSim(Owner).cuTurn;
      RI.Discovery := DiscPerc;
      if RI.Discovery < DiscPerc then  {Normaly it is not allowed to set the Discovery percentage above MaxPerc but the Script Command can do this.}
        RI.BoostDiscovery(DiscPerc-RI.Discovery);
    end;
  end;
  Log(['SetTechDiscovery: Discovery Percentage for RI',RI_UID,'set to',RI.Discovery]);
  if RI.Discovery = 100 then
    RI.ApplyEffects(True);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ADD / REM NATIONAL REGIONS
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.AddNationalRegions(Params : TStringList) : integer;
var i         : integer;
    Value     : integer;
    ZeroFound : boolean;
begin
  Result := 0;
  if not (Params.Count > 0) then
  begin
    Inc(Result);
    Log(['=> AddNationalRegions: Syntax is AddNationalRegions = [RgnUID1;RgnUID2;..]']);
  end;
  if (cuFaction = nil) then
  begin
    Inc(Result);
    Log(['=> AddNationalRegions command, no faction selected']);
  end;
  if Result <> 0 then Exit;

  ZeroFound := False;

  for i := 0 to Params.Count-1 do
  begin
    Value     := STI(Params[i]);
    if not cuFaction.ilForcedNatRgns.InList(Value) then
    begin
      if ZeroFound and (Value = 0) then
      begin
        Log(['=> AddNationalRegions', cuFaction.Name, 'Two regions have the UID 0, alias problem in:', Params.Text]);
        Inc(Result);
        Continue;
      end;
      if Value = 0 then ZeroFound := True;
      cuFaction.ilForcedNatRgns.Add(Value);
      Log(['AddNationalRegions', cuFaction.Name, 'National region added', Value, RegionDefs[Value].Name]);
    end;
  end;
end;

function TScriptEngine.RemNationalRegions(Params : TStringList) : integer;
var i         : integer;
    Value     : integer;
    ZeroFound : boolean;
begin
  Result := 0;
  if not (Params.Count > 0) then
  begin
    Inc(Result);
    Log(['=> RemNationalRegions: Syntax is RemNationalRegions = [RgnUID1;RgnUID2;..]']);
  end;
  if (cuFaction = nil) then
  begin
    Inc(Result);
    Log(['=> RemNationalRegions command, no faction selected']);
  end;
  if Result <> 0 then Exit;

  ZeroFound := False;

  for i := 1 to Params.Count-1 do
  begin
    Value     := STI(Params[i]);
    if cuFaction.ilForcedNatRgns.InList(Value) then
    begin
      if ZeroFound and (Value = 0) then
      begin
        Log(['=> RemNationalRegions', cuFaction.Name, 'Two regions have the UID 0, alias problem in:', Params.Text]);
        Inc(Result);
        Continue;
      end;
      if Value = 0 then ZeroFound := True;
      cuFaction.ilForcedNatRgns.DeleteByVal(Value);
      Log(['RemNationalRegions', cuFaction.Name, 'National region removed', Value, RegionDefs[Value].Name]);
    end;
  end;
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ADD / REM CLAIMED REGIONS
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.AddClaimedRegions(Params : TStringList) : integer;
var i         : integer;
    Value     : integer;
    ZeroFound : boolean;
begin
  Result := 0;
  if not (Params.Count > 0) then
  begin
    Inc(Result);
    Log(['=> AddClaimedRegions: Syntax is AddClaimedRegions = [RgnUID1;RgnUID2;..]']);
  end;
  if (cuFaction = nil) then
  begin
    Inc(Result);
    Log(['=> AddClaimedRegions command, no faction selected']);
  end;
  if Result <> 0 then Exit;

  ZeroFound := False;

  for i := 0 to Params.Count-1 do
  begin
    Value     := STI(Params[i]);
    if not cuFaction.ilForcedClaimRgns.InList(Value) then
    begin
      if ZeroFound and (Value = 0) then
      begin
        Log(['=> AddClaimedRegions', cuFaction.Name, 'Two regions have the UID 0, alias problem in:', Params.Text]);
        Inc(Result);
        Continue;
      end;
      if Value = 0 then ZeroFound := True;
      cuFaction.ilForcedClaimRgns.Add(Value);
      Log(['AddClaimedRegions', cuFaction.Name, 'Claimed region added', Value, RegionDefs[Value].Name]);
    end;
  end;
end;

function TScriptEngine.RemClaimedRegions(Params : TStringList) : integer;
var i         : integer;
    Value     : integer;
    ZeroFound : boolean;
begin
  Result := 0;
  if not (Params.Count > 0) then
  begin
    Inc(Result);
    Log(['=> RemClaimedRegions: Syntax is RemClaimedRegions = [RgnUID1;RgnUID2;..]']);
  end;
  if (cuFaction = nil) then
  begin
    Inc(Result);
    Log(['=> RemClaimedRegions command, no faction selected']);
  end;
  if Result <> 0 then Exit;

  ZeroFound := False;

  for i := 1 to Params.Count-1 do
  begin
    Value     := STI(Params[i]);
    if cuFaction.ilForcedClaimRgns.InList(Value) then
    begin
      if ZeroFound and (Value = 0) then
      begin
        Log(['=> RemClaimedRegions', cuFaction.Name, 'Two regions have the UID 0, alias problem in:', Params.Text]);
        Inc(Result);
        Continue;
      end;
      if Value = 0 then ZeroFound := True;
      cuFaction.ilForcedClaimRgns.DeleteByVal(Value);
      Log(['RemClaimedRegions', cuFaction.Name, 'Claimed region removed', Value, RegionDefs[Value].Name]);
    end;
  end;
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ADD / REM DE JURE REGIONS
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.AddDeJureRegions(Params : TStringList) : integer;
var i         : integer;
    Value     : integer;
    ZeroFound : boolean;
begin
  Result := 0;
  if not (Params.Count > 0) then
  begin
    Inc(Result);
    Log(['=> AddDeJureRegions: Syntax is AddDeJureRegions = [RgnUID1;RgnUID2;..]']);
  end;
  if (cuFaction = nil) then
  begin
    Inc(Result);
    Log(['=> AddDeJureRegions command, no faction selected']);
  end;
  if Result <> 0 then Exit;

  ZeroFound := False;

  for i := 0 to Params.Count-1 do
  begin
    Value     := STI(Params[i]);
    if not cuFaction.ilDeJureRgns.InList(Value) then
    begin
      if ZeroFound and (Value = 0) then
      begin
        Log(['=> AddDeJureRegions', cuFaction.Name, 'Two regions have the UID 0, alias problem in:', Params.Text]);
        Inc(Result);
        Continue;
      end;
      if Value = 0 then ZeroFound := True;
      cuFaction.ilDeJureRgns.Add(Value);
      Log(['AddDeJureRegions', cuFaction.Name, 'De Jure region added', Value, RegionDefs[Value].Name]);
    end;
  end;
end;

function TScriptEngine.RemDeJureRegions(Params : TStringList) : integer;
var i         : integer;
    Value     : integer;
    ZeroFound : boolean;
begin
  Result := 0;
  if not (Params.Count > 0) then
  begin
    Inc(Result);
    Log(['=> RemDeJureRegions: Syntax is RemDeJureRegions = [RgnUID1;RgnUID2;..]']);
  end;
  if (cuFaction = nil) then
  begin
    Inc(Result);
    Log(['=> RemDeJureRegions command, no faction selected']);
  end;
  if Result <> 0 then Exit;

  ZeroFound := False;

  for i := 1 to Params.Count-1 do
  begin
    Value     := STI(Params[i]);
    if cuFaction.ilDeJureRgns.InList(Value) then
    begin
      if ZeroFound and (Value = 0) then
      begin
        Log(['=> RemDeJureRegions', cuFaction.Name, 'Two regions have the UID 0, alias problem in:', Params.Text]);
        Inc(Result);
        Continue;
      end;
      if Value = 0 then ZeroFound := True;
      cuFaction.ilDeJureRgns.DeleteByVal(Value);
      Log(['RemDeJureRegions', cuFaction.Name, 'De Jure region removed', Value, RegionDefs[Value].Name]);
    end;
  end;
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// CHANGE RESOURCES SITES
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.ChgResources_Sites(Params : TStringList) : integer;
var i         : integer;
    ObjectUID : TUID;
    Value     : integer;
begin
  Result := 0;
  if not (Params.Count > 0) then
  begin
    Inc(Result);
    Log(['=> ChgResSites: Syntax is ChgResSites = [;<Res_SiteAlias|Value>] ..']);
  end;
  if Params.Count mod 2 <> 0 then
  begin
    Inc(Result);
    Log(['=> ChgResSites, Specify a list of paired values: Res_SiteAlias|Value']);
  end;
  if (cuRegion = nil) then
  begin
    Inc(Result);
    Log(['=> ChgResSites command, no region selected']);
  end;
  if Result <> 0 then Exit;

  for i := 1 to Params.Count div 2 do
  begin
    ObjectUID := STI(Params[(i-1)*2]);
    Value     := STI(Params[(i-1)*2 + 1]);
    GameSim.RegionCmns[cuRegion.UID].dlResources.AlterValByKey(ObjectUID, Value, True);  // somme arithmtique & suppression si <0
    Log(['ChgResSites', cuRegion.Name, 'Resource Site', GameSim.MerchandisesDB[ObjectUID].Name, 'changed by:', Value]);
  end;
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// DISCOVER RESOURCES SITES
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.DiscoverResources_Sites(Params : TStringList) : integer;
var i         : integer;
    ObjectUID : TUID;
    Value     : integer;
    MaxVal    : integer;
begin
  Result := 0;
  if not (Params.Count > 0) then
  begin
    Inc(Result);
    Log(['=> DiscoverResSites: Syntax is DiscoverResSites = [;<Res_SiteAlias|Value>] .. OR DiscoverResSites = -999/999']);
  end;
  if (Params.Count >= 2) and (Params.Count mod 2 <> 0) then
  begin
    Inc(Result);
    Log(['=> DiscoverResSites, Specify a list of paired values: Res_SiteAlias|Value']);
  end;
  if (cuRegion = nil) then
  begin
    Inc(Result);
    Log(['=> DiscoverResSites command, no region selected']);
  end;
  if Result <> 0 then Exit;

  if Params.Count = 1 then
  begin
    Value := STI(Params[0]);
    if Value = -999 then
    begin
      GameSim.RegionCmns[cuRegion.UID].dlKnownResources.Clear;
      Log(['DiscoverResSites', cuRegion.Name, 'All Resources Sites unknown.']);
    end
    else
    begin
      GameSim.RegionCmns[cuRegion.UID].dlKnownResources.Assign(GameSim.RegionCmns[cuRegion.UID].dlResources);
      Log(['DiscoverResSites', cuRegion.Name, 'All Resources Sites defined so far known.']);
    end;
    Exit;
  end;

  for i := 1 to Params.Count div 2 do
  begin
    ObjectUID := STI(Params[(i-1)*2]);
    Value     := STI(Params[(i-1)*2 + 1]);
    GameSim.RegionCmns[cuRegion.UID].dlKnownResources.AlterValByKey(ObjectUID, Value, True);  // somme arithmtique & suppression si <0
    MaxVal := GameSim.RegionCmns[cuRegion.UID].dlResources.GetValByKey(ObjectUID);
    if MaxVal < GameSim.RegionCmns[cuRegion.UID].dlKnownResources.GetValByKey(ObjectUID) then
      GameSim.RegionCmns[cuRegion.UID].dlKnownResources.AlterValByKey(ObjectUID, MaxVal); // we limit to what exists
    Log(['DiscoverResSites', cuRegion.Name, '# of known Resource Site', GameSim.MerchandisesDB[ObjectUID].Name, 'changed by:', Value]);
  end;
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET ALLIANCE
// on veut indiquer si la faction est cot insurge ou loyaliste (rgle SimpleLoyalty pour BoA)
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetAlliance(Params : TStringList) : integer;
begin
  Result := 0;
  if rulSimpleLoyalty_ then
  begin
    if not (Params.Count in [1]) then
    begin
      Inc(Result);
      Log(['=> SetAlliance command, erroneous number of params', Params.Count]);
    end;
    if (Params.Count = 1)and(cuFaction = nil) then
    begin
      Inc(Result);
      Log(['=> SetAlliance command, no faction selected']);
    end;
    if Result <> 0 then Exit;

    if Params.Count = 1 then
      cuFaction.AllianceUID := STI(Params[0]);
    Log(['SetAlliance', cuFaction.Name, 'set to alliance: ', cuFaction.AllianceUID]);
  end
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// CHANGE REMPLACEMENTS
// Indication du nb de remplacements dispo pour un sous-type de SU donn, pour la faction en cours.
// Format Remplacement = FactionTag|SousType|Nombre ou esv
// ex: Remplacement = FRA|$Cavalry|6
// ex: Remplacement = AME|$Militia|esvintVar0
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.ChangeRempls (Params : TStringList) : integer;
var Value : integer;
begin
  Result := 0;
  if not (Params.Count in [3]) then
  begin
    Inc(Result);
    Log(['=> Change Remplacements command, erroneous number of params', Params.Count]);
  end;
  if cuFaction = nil then
  begin
    Inc(Result);
    Log(['=> Change Remplacements command, no faction selected']);
  end;
  if Result <> 0 then Exit;

  if IsExposedVar(Params[2]) then
    Value := GetExposedVarValue(Params[2])
  else
    Value := STI(Params[2]);

  // bonus/malus IA
  if cuFaction.AILevel = _aiLvlHard   then Value := RoundRandom(Value * aiLvlRPLCoeffHard_ / 100)  else
  if cuFaction.AILevel = _aiLvlVHard  then Value := RoundRandom(Value * aiLvlRPLCoeffVHard_ / 100);

  cuFaction.ChangeRemplNumber(Params[0], STI(Params[1]), Value);
  Log(['Change Remplacements', cuFaction.Name, 'FactionTag:', Params[0], 'Rpl type:', Params[1], 'changed by:', Value]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET AUTO RAISE
// Format AutoRaise = $uni_AME_Reb1|3|25
// AutoRaise = UnitAlias|MaxNb|DismissPerc
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetAutoRaise (Params : TStringList) : integer;
begin
  Result := 0;
  if not (Params.Count in [3]) then
  begin
    Inc(Result);
    Log(['=> SetAutoRaise command, erroneous number of params, syntax is SetAutoRaise = UnitAlias|MaxNb|DismissPerc']);
  end;
  if cuFaction = nil then
  begin
    Inc(Result);
    Log(['=> SetAutoRaise command, no faction selected']);
  end;
  if Result <> 0 then Exit;

  cuFaction.SetAutoRaise(STI(Params[0]), STI(Params[1]), STI(Params[2]));
  Log(['SetAutoRaise', cuFaction.Name, 'Unit type:', Params[0], 'Max Number:', Params[1], 'Dismiss perc:', Params[2]]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// CHANGE AUTO RAISE
// Format ChgAutoRaise = $uni_AME_Reb1|3|25
// ChgAutoRaise = UnitAlias|MaxNb|DismissPerc
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.ChgAutoRaise (Params : TStringList) : integer;
begin
  Result := 0;
  if not (Params.Count in [3]) then
  begin
    Inc(Result);
    Log(['=> ChgAutoRaise command, erroneous number of params, syntax is ChgAutoRaise = UnitAlias|MaxNb|DismissPerc']);
  end;
  if cuFaction = nil then
  begin
    Inc(Result);
    Log(['=> ChgAutoRaise command, no faction selected']);
  end;
  if Result <> 0 then Exit;

  cuFaction.ChangeAutoRaise(STI(Params[0]), STI(Params[1]), STI(Params[2]));
  if cuFaction.GetAutoRaise(STI(Params[0])) > 0 then
    Log(['ChgAutoRaise', cuFaction.Name, 'Unit type:', Params[0], 'Max Number:',
    cuFaction.GetAutoRaise(STI(Params[0])), 'Dismiss perc:', cuFaction.GetAutoDismiss(STI(Params[0]))])
  else
    Log(['ChgAutoRaise', cuFaction.Name, 'Unit type:', Params[0], 'No more autoraise.']);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// UPGRADE UNIT
// Syntaxe : UpgradeUnit IsTechUpg [;ForceTo <ModelUID>]
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.UpgradeUnit(Params : TStringList) : integer;
var ForceTo : integer;
    Param : string;
begin
  Result := 0;
  if not (Params.Count in [1..2]) then
  begin
    Inc(Result);
    Log(['=> UpgradeUnit command, erroneous number of params, syntax is: UpgradeUnit IsTechUpg [;ForceTo <ModelUID>]']);
  end;
  if Result <> 0 then Exit;

  if cuUnit = nil then
  begin
    Log(['=> UpgradeUnit command, no unit selected']);
    Exit;
  end;

  ForceTo  := _noUID;

  Param := ExtractParam('ForceTo', Params);
  if Param <> '' then
  begin
    ExtractArgs(Param);
    ForceTo := STI(Args[0]);
  end;

  cuUnit.UpgradeSU(ExtractParam('IsTechUpg', Params) = '1', ForceTo);

  Log(['Unit upgraded', cuUnit.Name]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// AUTO RAISE ON
// Active la rgle d'autoraise pour la faction
// AutoRaise = {0, 1}
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.AutoRaiseOn(Params : TStringList) : integer;
begin
  Result := 0;

  if not (Params.Count in [1]) then
  begin
    Inc(Result);
    Log(['=> AutoRaiseOn command, erroneous number of params', Params.Count]);
  end;
  if (Params.Count = 1)and(cuFaction = nil) then
  begin
    Inc(Result);
    Log(['=> AutoRaiseOn command, no faction selected']);
  end;
  if Result <> 0 then Exit;

  if Params.Count = 1 then
    cuFaction.AutoRaiseOn := IIF(Params[0] = '1', 1, 0);
  Log(['AutoRaiseOn', cuFaction.Name, 'set to: ', cuFaction.AutoRaiseOn]);

end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET NUM CREATE
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetNumCreate(Params : TStringList) : integer;
begin
  Result := 0;

  if not (Params.Count in [1]) then
  begin
    Inc(Result);
    Log(['=> SetNumCreate command, erroneous number of params', Params.Count]);
  end;
  if Result <> 0 then Exit;

  if Params.Count = 1 then
    NumCreate := STI(Params[0]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET ENTRANCH
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetEntranch(Params : TStringList) : integer;
begin
  Result := 0;

  if not (Params.Count in [1]) then
  begin
    Inc(Result);
    Log(['=> SetEntranch command, erroneous number of params', Params.Count]);
  end;
  if Result <> 0 then Exit;

  if Params.Count = 1 then
    Entrench := STI(Params[0]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET IN CITYSCAPE
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetInCS(Params : TStringList) : integer;
begin
  Result := 0;

  if not (Params.Count in [1]) then
  begin
    Inc(Result);
    Log(['=> SetInCS command, erroneous number of params', Params.Count]);
  end;
  if Result <> 0 then Exit;

  if Params.Count = 1 then
    InCS := Params[0] = '1';
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET FIX TYPE
// Only used in group creation
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetFixType(Params : TStringList) : integer;
begin
  Result := 0;

  if not (Params.Count in [1]) then
  begin
    Inc(Result);
    Log(['=> SetFixType command, erroneous number of params', Params.Count]);
  end;
  if Result <> 0 then Exit;

  if Pos('/', Params[0]) > 0 then
    FixType := DateToTurn(Params[0]) - GameSim.cuTurn
  else
    FixType := STI(Params[0]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// CREATE FACTION
// demande un faction tag comme param
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.CreateFaction (Params : TStringList) : integer;
begin
  Result := 0;

  if not (Params.Count in [1]) then
  begin
    Inc(Result);
    Log(['=> CreateFaction command, erroneous number of params', Params.Count]);
  end;
  if Result <> 0 then Exit;

  if GameSim.Factions.GetFactionByTag(Params[0], False) <> nil then
  begin
    Log(['[WARNING] CreateFaction command, this faction already exists, exiting silently', Params[0]]);
    Exit;
  end;

  if not GameSim.CreateFaction(Params[0]) then
  begin
    Inc(Result);
    Log(['=> CreateFaction command, parameter', Params[0], 'is not a known Faction Tag or the Faction tag was already existing.']);
  end;
  if Result <> 0 then Exit;

  // if not in scen making more, then the faction will be created as AI controlled

  Log(['Faction', Params[0], 'created.']);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET MAP OFFSETS
// offsets x et y, pour une faction donne - reprsente le pt de focus sur la carte
// en arrive du main menu
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetMapOffsets (Params : TStringList) : integer;
begin
  Result := 0;

  if not (Params.Count in [2]) then
  begin
    Inc(Result);
    Log(['=> MapOffsets command, erroneous number of params', Params.Count, 'you need the x and y coordinates']);
    Exit;
  end;
  if cuFaction = nil then
  begin
    Inc(Result);
    Log(['=> MapOffsets command, no faction selected']);
    Exit;
  end;

  cuFaction.OffsetX := -Abs(STI(Params[0]));
  cuFaction.OffsetY := -Abs(STI(Params[1]));

  Log(['Faction', cuFaction.Name, 'Map offsets set to', cuFaction.OffsetX, cuFaction.OffsetY]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET MAP BOUNDS
// Spcification (au niveau gamesim) des bornes de scrolling, pour empecher le
// scroll hors d'un carr prdfini
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetMapBounds (Params : TStringList) : integer;
begin
  Result := 0;

  if not (Params.Count in [4]) then
  begin
    Inc(Result);
    Log(['=> MapBounds command, erroneous number of params', Params.Count, 'you need 4 coordinates, top left X / Y and bottom right X / Y']);
    Exit;
  end;

  GameSim.UI_TopLeftBound.x := STI(Params[0]);
  GameSim.UI_TopLeftBound.y := STI(Params[1]);
  GameSim.UI_BottomRightBound.x := STI(Params[2]);
  GameSim.UI_BottomRightBound.y := STI(Params[3]);

  if Result <> 0 then Exit;

  Log(['Map boundaries set to', GameSim.UI_TopLeftBound.x, GameSim.UI_TopLeftBound.y, GameSim.UI_BottomRightBound.x, GameSim.UI_BottomRightBound.y]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET SCEN TURNS
// Bornes de tours (tour courant, tour max)
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetScenTurns  (Params : TStringList) : integer;
begin
  Result := 0;

  if not (Params.Count in [2]) then
  begin
    Inc(Result);
    Log(['=> ScenTurns command, erroneous number of params', Params.Count, 'you need the current and max turn of the scenario']);
  end;
  if Result <> 0 then Exit;

  GameSim.cuTurn  := STI(Params[0]);
  GameSim.MaxTurn := STI(Params[1]);

  Log(['Current Turn:', GameSim.cuTurn, 'Max Turn:', GameSim.MaxTurn]);

end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET SCEN START
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetScenStart(Params : TStringList) : integer;
begin
  Result := 0;

  if not (Params.Count in [1]) then
  begin
    Inc(Result);
    Log(['=> SetScenStart command, erroneous number of params', Params.Count, 'you need the start date of the scenario']);
  end;
  if Result <> 0 then Exit;

  GameSim.cuTurn  := DateToTurn(NormalizeDate(Params[0]));

  // update the turn to use if not hard fixed
  if sysScriptForceTurn_ = '' then
    UsedTurn := TGameSim(Owner).cuTurn;

  Log(['Start scenario Turn:', GameSim.cuTurn]);

end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET SCEN END
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetScenEnd(Params : TStringList) : integer;
begin
  Result := 0;

  if not (Params.Count in [1]) then
  begin
    Inc(Result);
    Log(['=> SetScenEnd command, erroneous number of params', Params.Count, 'you need the end date of the scenario']);
  end;
  if Result <> 0 then Exit;

  GameSim.MaxTurn  := DateToTurn(NormalizeDate(Params[0]), True);

  Log(['End scenario Turn:', GameSim.cuTurn]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// CHANGE SCEN END
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.ChgScenEnd(Params : TStringList) : integer;
begin
  Result := 0;

  if not (Params.Count in [1]) then
  begin
    Inc(Result);
    Log(['=> ChgScenEnd command, erroneous number of params', Params.Count, 'you need to give a number of turns.']);
  end;
  if Result <> 0 then Exit;

  if GameSim.cuTurn < GameSim.MaxTurn then
    GameSim.MaxTurn  := Max(GameSim.cuTurn, GameSim.MaxTurn + STI(Params[0]));

  Log(['End scenario Turn:', GameSim.cuTurn]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// FORCE DATE TO
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.ForceDateTo(Params : TStringList) : integer;
begin
  Result := 0;

  if not (Params.Count in [1]) then
  begin
    Inc(Result);
    Log(['=> ForceDateTo command, erroneous number of params', Params.Count, 'you need to specify a date.']);
  end;
  if Result <> 0 then Exit;

  GameSim.cuTurn := DateToTurn(NormalizeDate(Params[0]), True);

  Log(['Current turn is now Turn:', GameSim.cuTurn, 'current date is', NormalizeDate(Params[0])]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// TURNS PER YEAR
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.TurnsPerYear(Params : TStringList) : integer;
var TPY : integer;
begin
  Result := 0;

  if not (Params.Count in [1]) then
  begin
    Inc(Result);
    Log(['=> TurnsPerYear command, erroneous number of params', Params.Count, 'you need the turns per year format of the scenario']);
  end;
  TPY := STI(Params[0]);

  if not (TPY in [4, 12, 24, 36, 48]) then
  begin
    Inc(Result);
    Log(['=> TurnsPerYear command, the accepted value is 4, 12, 24, 36, 48']);
  end;

  if Result <> 0 then Exit;

  GameSim.TurnsPerYear  := TPY;

  if GameSim.TurnsPerYear = 0 then TurnsPerYear_ := 12;
  TurnsPerYear_ := GameSim.TurnsPerYear;
  DaysPerTurn_ := 360 div Max(4, TurnsPerYear_);

  LoadTurnNames;
  
  Log(['Turns per year fixed to ', GameSim.TurnsPerYear, 'for this scenario']);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET SCEN NAME
// Nom du scnario (FileName) en cration
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetScenName   (Params : TStringList) : integer;
begin
  Result := 0;

  if not (Params.Count in [1]) then
  begin
    Inc(Result);
    Log(['=> ScenName command, erroneous number of params', Params.Count, 'you need to enter the scenario name']);
  end;
  if Result <> 0 then Exit;

  // on enlve une ventuelle extension et on mettra .scn en sauvegarde (partie Hosting)
  ChangeFileExt(Params[0], '');

  if Pos('\', Params[0]) = 1 then
    GameSim.DataFile := Params[0]
  else
    GameSim.DataFile := ScenPath_ + '\' + Params[0];


  Log(['Creating scenario, path & name:', GameSim.DataFile ]);

end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET SCEN DISPLAY NAME
// Nom du scnario (localis  la vole)
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetScnDispName(Params : TStringList) : integer;
begin
  Result := 0;

  if not (Params.Count in [1]) then
  begin
    Inc(Result);
    Log(['=> Scen Display Name command, erroneous number of params', Params.Count, 'you need to enter the scenario localizable name']);
  end;
  if Result <> 0 then Exit;

  GameSim.ScenDispName := Params[0];

  Log(['Scenario Display Name', GameSim.ScenDispName ]);

end;
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET POSTURE
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetPosture(Params : TStringList) : integer;
begin
  Result := 0;

  if not (Params.Count in [1]) then
  begin
    Inc(Result);
    Log(['=> Posture command, erroneous number of params', Params.Count, 'you need to enter posture as an integer (or alias)']);
  end;
  if Result <> 0 then Exit;

  GroupPosture := STI(Params[0]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// GENERATE (FLAVOR) MESSAGE - GENMSG command
// si le record PlayerMessage avait un UID de rgion (renseign uniquement en StartEvent), le message attach  l'evt
// a t gnr automatiquement en dbut de partie Action, autrement il faut appeler
// cette commande, pour que la rgion de focus soit bien  jour (issue d'une slection conditionnelle ou alatoire)
// on peut aussi gnrer un msg sans rgion de focus (remplacements)
//
// CSVSplitter ajoute un GenMsg en automatique sur certaines commandes

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.GenFlavorMsg  (Params : TStringList) : integer;
var i : integer;
    RgnName : string;
begin
  Result := 0;                    // LIRE COMMENTAIRE CI-DESSUS SI DOUBLE MESSAGE GENERE DANS LE JEU...

  if (cuFaction = nil)and(not PlayerMessage.IsBroadCast) then
  begin
    Log(['[WARNING] GenMsg: No Selected Faction and not a broadcast, exiting.']);
    Exit;
  end;
  if cuRegion <> nil then
  begin
    PlayerMessage.RgnUID := cuRegion.UID;
    RgnName := cuRegion.Name;
  end
  else
    PlayerMessage.RgnUID := _noUID;

  if MsgNeedRegion and (PlayerMessage.RgnUID = _noUID) then
  begin
    Log(['[WARNING] Message generation ignored, no focus region and flag MsgNeedRegion set to ON.']);
    Exit;
  end;

  with PlayerMessage do
  begin
    if (DisplayType <> _evtNoDisplay) then
    begin
      if not PlayerMessage.IsBroadCast then
      begin
        cuFaction.Messages.MakeFlavorMsg(FStr(Title, [RgnName]), Text, ImageID, DisplayType, RgnUID, SoundID);
        if RgnUID = _noUID then
          Log(['Flavor Message Generated:', Title, 'no focus region'])
        else
          Log(['Flavor Message Generated:', Title, 'Region', RgnUID, RegionDefs[RgnUID].Name])
      end
      else
      begin
        for i := 0 to GameSim.Factions.Count-1 do
          with PlayerMessage do
            GameSim.Factions[i].Messages.MakeFlavorMsg(FStr(Title, [RgnName]), Text, ImageID, DisplayType, RgnUID, SoundID);
      end;
    end;
    DisplayType := _evtNoDisplay; // pour ne le gnrer qu'une fois par evt
  end;
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// GEN TEXT MESSAGE  GENTEXTMSG command
// A appeler spcifiquement (elle a besoin de ses propres params, alors que ci dessus, GenMsg, prend ceux de la ligne StartEvent,  l'exception possible de la region, cf ci dessus)
// _evtNoDisplay = 0;
// _evtLineDisplay = 1;
// _evtFullDisplay = 2;
// GenTextMsg Title|_evtLineDisplay|NULL|NULL|NULL
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.GenTextMsg(Params : TStringList) : integer;
var RgnUID, SndUID,i : TUID;
    RgnName : string;
begin
  Result := 0;

  if (cuFaction = nil)or(cuFaction.FactionDef.FactionTag = 'CMN') then
    Log(['GenTextMsg: This is a broadcast.']);

  // pas obligatoire
  if cuRegion = nil then RgnUID := _noUID
  else
  begin
    RgnUID := cuRegion.UID;
    RgnName := cuRegion.Name;
  end;

  if Params.Count <> 5 then
  begin
    Inc(Result);
    Log(['=> GenTextMsg need 5 params: Title, DisplayType, FlavorText, ImageID, SoundID']);
    Exit;
  end;

  if (Params[4] = 'NULL') or (Params[4] = '') then
    SndUID := _noUID
  else
    SndUID := STI(Params[4]);

  if (cuFaction = nil)or(cuFaction.FactionDef.FactionTag = 'CMN') then
  begin                                            
    for i := 0 to GameSim.Factions.Count-1 do
      GameSim.Factions[i].Messages.MakeFlavorMsg(FStr(Translate(Params[0]), [RgnName]), Params[2], Params[3], STI(Params[1]), RgnUID, SndUID, False);
    Log(['GenTextMsg Message Broadcasted:', Params[0]]);
  end
  else
  begin
    cuFaction.Messages.MakeFlavorMsg(FStr(Translate(Params[0]), [RgnName]), Params[2], Params[3], STI(Params[1]), RgnUID, SndUID, False);
    Log(['GenTextMsg Message Generated:', Params[0]]);
  end;
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// GEN MULTI CHOICE EVENT
// Shows a MC event in the message panel. EventID is a string (ex: FRA_MilitaryOption2)
// referencing an active ActorPool defined previously of type $gmaMCEvent and subtype EventID.
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.GenMCEvent(Params : TStringList) : integer;
begin
  Result := 0;

  if Params.Count <> 2 then
  begin
    Inc(Result);
    Log(['=> GenMCEvent need 2 params: Title, EventID']);
    Exit;
  end;
  if cuFaction = nil then
  begin
    Log(['[Warning] GenMCEvent: No Selected Faction']);
    Inc(Result);
    Exit;
  end;
  cuFaction.Messages.MakeMCEvent(Params[0], Params[1]);
  Log(['MC Event linked to message panel', Params[0], Params[1]]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// GEN POP UP TEXT
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.GenPopUpText(Params : TStringList) : integer;
begin
  Result := 0;

  if cuFaction = nil then
  begin
    Log(['[Warning] PopUpText: No Selected Faction']);
    Inc(Result);
    Exit;
  end;

  if Params.Count <> 6 then
  begin
    Inc(Result);
    Log(['=> PopUpText need 6 params: Title, DisplayType, Verbosity, FlavorText, ImageID, SoundID']);
    Exit;
  end;
  cuFaction.Messages.MakePopUpText(Params[0], Params[1], Params[2], Params[3], Params[4], STI(Params[5]), True);
  Log(['PopUpText Message Generated:', Params[0]]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// GEN ADVANCED POP UP TEXT
// Serves no purpose and is buggy?
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.GenPopUpAdvText(Params : TStringList) : integer;
begin
  Result := 0;

  if cuFaction = nil then
  begin
    Log(['[Warning] PopUpAdvText: No Selected Faction']);
    Inc(Result);
    Exit;
  end;

  if Params.Count <> 6 then
  begin
    Inc(Result);
    Log(['=> PopUpAdvText need 6 params: Title, DisplayType, Verbosity, FlavorText, ImageID, SoundID']);
    Exit;
  end;
  cuFaction.Messages.MakePopUpAdvText(Params[0], Params[1], Params[2], Params[3], Params[4], STI(Params[5]), True);
  Log(['PopUpAdvText Message Generated:', Params[0]]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// CHANGE EVENT OCCURS
// Change par une certaine quantit le nombre d'occurences ralises d'un evt
// (ngatif autoris
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.ChgEventOccurs(Params : TStringList) : integer;
var EventName : string;
    OccursType : string;
    Amount : integer;
    ES : TEventState;
begin
  Result := 0;

  if not (Params.Count in [3]) then
  begin
    Inc(Result);
    Log(['=> Change Event Occurs command, erroneous number of params', Params.Count, 'you need 3: ChgEvtOccurs <Name>;MaxOccurs/CuOccurs;<ChangeAmount>']);
  end;
  if Result <> 0 then Exit;
  EventName := Params[0];
  OccursType := Params[1];
  Amount := STI(Params[2]);

  ES := GameSim.EventStates.GetByName(EventName);
  if ES = nil then
  begin
    Log(['[Warning] Event', EventName, 'not yet referenced']);
    Exit;
  end;

  // changement du nombre d'occurences
  if UpperCase(OccursType) = 'CUOCCURS' then
    ES.cuOccurs := ES.cuOccurs + Amount    // permettre un nb ngatif, pour dsactiver les OnAttackTrigger qui relve le nb d'occurences permises
  else {MAXOCCURS}
    ES.MaxOccurs := ES.MaxOccurs + Amount;

  Log(['Change Event Occurs command:', EventName, 'Max Occurences:', ES.MaxOccurs, 'Current Occurences:', ES.cuOccurs]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET EVENT OCCURS
// Set either the max nb of occurences allowed or the current nb of occurences
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetEventOccurs(Params : TStringList) : integer;
var EventName : string;
    OccursType : string;
    Amount : integer;
    ES : TEventState;
begin
  Result := 0;

  if not (Params.Count in [3]) then
  begin
    Inc(Result);
    Log(['=> Set Event Occurs command, erroneous number of params', Params.Count, 'you need 3: SetEvtOccurs <Name>;MaxOccurs/CuOccurs;<SetAmount>']);
  end;
  if Result <> 0 then Exit;
  EventName := Params[0];
  OccursType := Params[1];
  Amount := STI(Params[2]);

  ES := GameSim.EventStates.GetByName(EventName);
  if ES = nil then
  begin
    Log(['[Warning] Event', EventName, 'not yet referenced']);
    Exit;
  end;

  // changement du nombre d'occurences
  if UpperCase(OccursType) = 'CUOCCURS' then
    ES.cuOccurs := Amount
  else {MAXOCCURS}
    ES.MaxOccurs := Amount;

  Log(['Set Event Occurs command:', EventName, 'Max Occurences:', ES.MaxOccurs, 'Current Occurences:', ES.cuOccurs]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET WIN LEVEL
// Nb de villes strats que doit avoir la faction pour gagner - que BOA1
// SetWinLevel = x
// OR for _morMoraleWinByVp (VGN) this is the % coeff compared to 2nd to win
// example: I'm GBR with winlevel of 300, I need 3x the total of the 2nd major to win
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetWinLevel(Params : TStringList) : integer;
begin
  Result := 0;
  if (cuFaction = nil) then
  begin
    Inc(Result);
    Log(['=> SetWinLevel: No faction selected']);
  end;
  if (Params.Count <> 1) then
  begin
    Inc(Result);
    Log(['=> SetWinLevel: You need to specify the number of Strat regions required to win']);
  end;
  if Result <> 0 then Exit;

  cuFaction.VPWinLevel := STI(Params[0]);
  Log(['SetWinLevel:', cuFaction.Name, 'Level set to:', cuFaction.VPWinLevel]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// CHANGE VP COUNT
// Modif du nb de pts de victoire donns  la faction (le + haut gagne en fin de scen)
// ChangeVPCount / ChgVPCount = x
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.ChgVPCount(Params : TStringList) : integer;
begin
  Result := 0;
  if (cuFaction = nil) then
  begin
    Inc(Result);
    Log(['=> ChangeVPCount: No faction selected']);
  end;
  if (Params.Count <> 1) then
  begin
    Inc(Result);
    Log(['=> ChangeVPCount: You need to specify a number of VP']);
  end;
  if Result <> 0 then Exit;

  cuFaction.VPTotal := cuFaction.VPTotal + STI(Params[0]);
  Log(['ChangeVPCount:', cuFaction.Name, 'new value: ', cuFaction.VPTotal]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET SUDDEN DEATH
// Possibilit oui/non de victoire mort subite
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetSudDeath(Params : TStringList) : integer;
begin
  Result := 0;
  if (cuFaction = nil) then
  begin
    Inc(Result);
    Log(['=> RulSuddenDeath: No faction selected']);
  end;
  if (Params.Count <> 1) then
  begin
    Inc(Result);
    Log(['=> RulSuddenDeath: You need to specify if the rule is activated or not with 1 or 0']);
  end;
  if Result <> 0 then Exit;

  cuFaction.RulSuddenDeath := STI(Params[0]);
  Log(['RulSuddenDeath:', cuFaction.Name, ' set to:', cuFaction.RulSuddenDeath]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET PETTY WAR
// Possibilit oui/non de gain de VP par petite guerre
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetPettyWar(Params : TStringList) : integer;
begin
  Result := 0;
  if (cuFaction = nil) then
  begin
    Inc(Result);
    Log(['=> SetPettyWar: No faction selected']);
  end;
  if (Params.Count <> 1) then
  begin
    Inc(Result);
    Log(['=> SetPettyWar: You need to specify if the rule is activated or not with 1 or 0']);
  end;
  if Result <> 0 then Exit;

  cuFaction.RulPettyWar := STI(Params[0]);
  Log(['SetPettyWar:', cuFaction.Name, ' set to:', cuFaction.RulPettyWar]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET FAC PASSIVE
// Faction tjrs passive par IA (le joueur peut la jouer)
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetFacPassive (Params : TStringList) : integer;
begin
  Result := 0;
  if (cuFaction = nil) then
  begin
    Inc(Result);
    Log(['=> SetFacPassive: No faction selected']);
  end;
  if (Params.Count <> 1) then
  begin
    Inc(Result);
    Log(['=> SetFacPassive: You need to specify if the faction is passive (1) or not (0), 0 being the default']);
  end;
  if Result <> 0 then Exit;

  cuFaction.Passive := STI(Params[0]);
  if (GameSim.DataType <> _dtScen) and (cuFaction.Passive = 0) then // remove the message the AI is taking over
    cuFaction.AILevel := aiDiffLevel_;

  Log(['SetFacPassive:', cuFaction.Name, ' set to:', cuFaction.Passive]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET FAC UNSELECTABLE
// Faction pas vue en dbut de SCN ou chargement de tour (dans PickFileList)
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetFacUnselect(Params : TStringList) : integer;
begin
  Result := 0;
  if (cuFaction = nil) then
  begin
    Inc(Result);
    Log(['=> SetFacUnselect: No faction selected']);
  end;
  if (Params.Count <> 1) then
  begin
    Inc(Result);
    Log(['=> SetFacUnselect: You need to specify if the faction is not selectable (1) or can be selected (0), 0 being the default']);
  end;
  if Result <> 0 then Exit;

  cuFaction.Unselectable := STI(Params[0]);
  Log(['SetFacUnselect:', cuFaction.Name, ' set to:', cuFaction.Unselectable]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// IS TUTORIAL
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetIsTutorial (Params : TStringList) : integer;
begin
  Result := 0;
  if (Params.Count <> 1) then
  begin
    Inc(Result);
    Log(['=> SetIsTutorial: Need one param']);
  end;
  if Result <> 0 then Exit;

  GameSim.IsTutorial := Params[0] = '1';
  Log(['SetIsTutorial:', 'Set to:', GameSim.IsTutorial]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET LOYALTY RULE
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetLoyaltyRule(Params : TStringList) : integer;
begin
  Result := 0;
  if (Params.Count <> 1) then
  begin
    Inc(Result);
    Log(['=> SetLoyaltyRule: Need one param']);
  end;
  if Result <> 0 then Exit;

  GameSim.RuleLoyaltyON := Params[0] = '1';
  Log(['SetLoyaltyRule:', 'Set to:', GameSim.RuleLoyaltyON]);
end;
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET OBJECTIVES
// Liste des rgions objectifs pour que la faction gagne
// AACW+, les objectifs ont maintenant une valeur numrique associe
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetObjectives(Params : TStringList) : integer;
var i : integer;
    Key, Value : integer;
begin
  Result := 0;
  if (cuFaction = nil) then
  begin
    Inc(Result);
    Log(['=> SetObjectives: No faction selected']);
  end;
  if Params.Count mod 2 <> 0 then
  begin
    Inc(Result);
    Log(['=> SetObjectives, Specify a list of paired values: region UID (or alias) | objective value']);
    Exit;
  end;
  if Result <> 0 then Exit;

  cuFaction.ClearObjList;

  for i := 1 to Params.Count div 2 do
  begin
    Key   := STI(Params[(i-1)*2]);
    Value := STI(Params[(i-1)*2 + 1]);
    cuFaction.AlterObjective(Key, Value); // ajoute, supprime ou altre la liste
  end;

  Log(['SetObjectives:', cuFaction.Name, cuFaction.GetObjListCount, 'Objectives defined']);
end;
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// CHANGE OBJECTIVE
// Ajoute (si Value > 0) ou enlve (si  0) un objectif de la liste
// Ne fonctionne pas sur BOA
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.ChgObjective(Params : TStringList) : integer;
var 
    Key, Value : integer;
begin
  Result := 0;
  if (cuFaction = nil) then
  begin
    Inc(Result);
    Log(['=> ChgObjective: No faction selected']);
  end;
  if Params.Count <> 2 then
  begin
    Inc(Result);
    Log(['=> ChgObjective, Specify a region UID (or alias) and objective value (if 0 then it will deletes the objective)']);
    Exit;
  end;
  if Result <> 0 then Exit;
  Key   := STI(Params[0]);
  Value := STI(Params[1]);

  cuFaction.AlterObjective(Key, Value); // ajoute, supprime ou remplace l'entre selon le cas

  Log(['ChgObjective:', cuFaction.Name, cuFaction.GetObjListCount, 'Objectives defined']);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET WIN PARAMS
// Chaines dcrivant le message de victoire
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetWinParams(Params : TStringList) : integer;
var i : integer;
begin
  Result := 0;
  if (cuFaction = nil) then
  begin
    Inc(Result);
    Log(['=> SetWinParams: No faction selected']);
  end;
  if (Params.Count <> 5) then
  begin
    Inc(Result);
    Log(['=> WinParams: Need 5 parameters, Syntax is: WinParams Title|DisplayTypeUID|Text|Bitmap|Sound']);
  end;
  if Result <> 0 then Exit;

  cuFaction.WinParams := '';
  for i := 0 to Params.Count-1 do
  begin
    cuFaction.WinParams := cuFaction.WinParams + Params[i] + '|';
  end;

  Log(['SetWinParams:', cuFaction.Name, 'WinParams set to:', cuFaction.WinParams]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET WIN PARAMS VP
// Chaines dcrivant le message de victoire, si victoire par VP
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetWinParamsVP(Params : TStringList) : integer;
var i : integer;
begin
  Result := 0;
  if (cuFaction = nil) then
  begin
    Inc(Result);
    Log(['=> SetWinParamsVP: No faction selected']);
  end;
  if (Params.Count <> 5) then
  begin
    Inc(Result);
    Log(['=> SetWinParamsVP: Need 5 parameters, Syntax is: WinParamsVP Title|DisplayTypeUID|Text|Bitmap|Sound']);
  end;
  if Result <> 0 then Exit;

  cuFaction.WinParamsVP := '';
  for i := 0 to Params.Count-1 do
  begin
    cuFaction.WinParamsVP := cuFaction.WinParamsVP + Params[i] + '|';
  end;

  Log(['SetWinParamsVP:', cuFaction.Name, 'SetWinParamsVP set to:', cuFaction.WinParamsVP]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET LOSE PARAMS
// Chaines dcrivant le message de dfaite
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetLoseParams(Params : TStringList) : integer;
var i : integer;
begin
  Result := 0;
  if (cuFaction = nil) then
  begin
    Inc(Result);
    Log(['=> LoseParams: No faction selected']);
  end;
  if (Params.Count <> 5) then
  begin
    Inc(Result);
    Log(['=> LoseParams: Need 5 parameters, Syntax is: LoseParams Title|DisplayTypeUID|Text|Bitmap|Sound']);
  end;
  if Result <> 0 then Exit;

  cuFaction.LoseParams := '';
  for i := 0 to Params.Count-1 do
  begin
    cuFaction.LoseParams := cuFaction.LoseParams + Params[i] + '|';
  end;

  Log(['SetLoseParams:', cuFaction.Name, 'LoseParams set to:', cuFaction.LoseParams]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET LOSE PARAMS VP
// Chaines dcrivant le message de dfaite par VP
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetLosePrmsVP(Params : TStringList) : integer;
var i : integer;
begin
  Result := 0;
  if (cuFaction = nil) then
  begin
    Inc(Result);
    Log(['=> LoseParamsVP: No faction selected']);
  end;
  if (Params.Count <> 5) then
  begin
    Inc(Result);
    Log(['=> LoseParamsVP: Need 5 parameters, Syntax is: LoseParamsVP Title|DisplayTypeUID|Text|Bitmap|Sound']);
  end;
  if Result <> 0 then Exit;

  cuFaction.LoseParamsVP := '';
  for i := 0 to Params.Count-1 do
  begin
    cuFaction.LoseParamsVP := cuFaction.LoseParamsVP + Params[i] + '|';
  end;

  Log(['SetLoseParamsVP:', cuFaction.Name, 'LoseParamsVP set to:', cuFaction.LoseParamsVP]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET SCEN VICTORY CONDITION
// Chaines dcrivant les conditions de victoire d'une faction
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetScenVicCond(Params : TStringList) : integer;
begin
  Result := 0;
  if (cuFaction = nil) then
  begin
    Inc(Result);
    Log(['=> ScenVicCondition: No faction selected']);
  end;
  if (Params.Count <> 1) then
  begin
    Inc(Result);
    Log(['=> ScenVicCondition: Need to give the text describing the victory conditions.']);
  end;
  if Result <> 0 then Exit;

  cuFaction.VictoryCond := Params[0];

  Log(['ScenVicCondition:', cuFaction.Name, 'Textual Victory Conditions set to:', cuFaction.VictoryCond]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET SCEN TEXT
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetScenText(Params : TStringList) : integer;
begin
  Result := 0;
  if (Params.Count <> 1) then
  begin
    Inc(Result);
    Log(['=> SetScenText: You need to give the scenario description']);
  end;
  if Result <> 0 then Exit;

  GameSim.ScenText := Params[0];
  Log(['SetScenText:', 'Set to:', GameSim.ScenText]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET SIDE1
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetSide1(Params : TStringList) : integer;
var i : integer;
begin
  Result := 0;
  if (Params.Count = 0) then
  begin
    Inc(Result);
    Log(['=> SetSide1: You need to give the list of factions tag for side 1']);
  end;
  if Result <> 0 then Exit;

  GameSim.FactionTagSide1 := '';
  for i := 0 to Params.Count -1 do
    GameSim.FactionTagSide1 := GameSim.FactionTagSide1 + Params[i] + '|';
  Log(['SetSide1:', 'Set to:', GameSim.FactionTagSide1 ]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET SIDE2
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetSide2(Params : TStringList) : integer;
var i : integer;
begin
  Result := 0;
  if (Params.Count = 0) then
  begin
    Inc(Result);
    Log(['=> SetSide2: You need to give the list of factions tag for side2']);
  end;
  if Result <> 0 then Exit;

  GameSim.FactionTagSide2 := '';
  for i := 0 to Params.Count -1 do
    GameSim.FactionTagSide2 := GameSim.FactionTagSide2 + Params[i] + '|';
  Log(['SetSide2:', 'Set to:', GameSim.FactionTagSide2 ]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET BACKGROUND 1
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.ScenBGround1(Params : TStringList) : integer;
begin
  Result := 0;
  if (Params.Count <> 1) then
  begin
    Inc(Result);
    Log(['=> ScenBGround1: You need to give the scenario background (part 1)']);
  end;
  if Result <> 0 then Exit;

  GameSim.ScenBGround1 := Params[0];
  Log(['ScenBGround1:', 'Set to:', GameSim.ScenBGround1]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET BACKGROUND 2
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.ScenBGround2(Params : TStringList) : integer;
begin
  Result := 0;
  if (Params.Count <> 1) then
  begin
    Inc(Result);
    Log(['=> ScenBGround2: You need to give the scenario background (part 2)']);
  end;
  if Result <> 0 then Exit;

  GameSim.ScenBGround2 := Params[0];
  Log(['ScenBGround2:', 'Set to:', GameSim.ScenBGround2]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET DURATION
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.ScenDuration(Params : TStringList) : integer;
begin
  Result := 0;
  if (Params.Count <> 1) then
  begin
    Inc(Result);
    Log(['=> ScenDuration: You need to give the scenario textual duration ']);
  end;
  if Result <> 0 then Exit;

  GameSim.ScenDuration := Params[0];
  Log(['ScenDuration:', 'Set to:', GameSim.ScenDuration]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET FAVORED
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.ScenFavored(Params : TStringList) : integer;
begin
  Result := 0;
  if (Params.Count <> 1) then
  begin
    Inc(Result);
    Log(['=> ScenFavored: You need to give the scenario favored side & hints ']);
  end;
  if Result <> 0 then Exit;

  GameSim.ScenFavored := Params[0];
  Log(['ScenFavored:', 'Set to:', GameSim.ScenFavored]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SCEN IMAGE BOX
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.ScenImageBox(Params : TStringList) : integer;
begin
  Result := 0;
  if (Params.Count <> 4) then
  begin
    Inc(Result);
    Log(['=> ScenImageBox: You need to give 4 parameters: pressed Image (ON), highlighted image (HL), not-over image (OFF), difficulty (1 to 5)']);
  end;
  if Result <> 0 then Exit;

  GameSim.ScenImageBox := Params[0] + '|' + Params[1] + '|' + Params[2] + '|' + Params[3];
  Log(['ScenImageBox:', 'Set to:', GameSim.ScenImageBox]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// CHANGE GROUP FIXING
// Pour changer la fixation d'un groupe d'une rgion
// Group  nil autoris (pas d'erreur)
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.ChangeGroupFixing  (Params : TStringList) : integer;
var i : integer;
begin
  Result := 0;
  if not (Params.Count in [1]) then
  begin
    Log(['=> ChangeGroupFixing command, you need to specify the new Fixation type']);
    Inc(Result);
    Exit;
  end;
  if cuGroup = nil then
  begin
    Log(['[Warning] ChangeGroupFixing: No Selected Group.']);
    Exit;
  end;

  for i := 0 to cuGroup.Count-1 do
    cuGroup[i].FixType := STI(Params[0]);

  Log(['Group', cuGroup.Name, 'has now this fixation value: ', cuGroup[0].FixType, 'in region', RegionDefs[cuGroup.StartSubSpace].Name]);
  if cuGroup.IsFixed then
    cuGroup.StopMove;
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET FIXING
// Pour changer la fixation d'une unit d'une rgion
// Unit  nil autorise (pas d'erreur)
// CHANGE est obsolte mais encore support, CHANGEFIXING utilise donc cette fonction
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetFixing  (Params : TStringList) : integer;
begin
  Result := 0;
  if not (Params.Count in [1]) then
  begin
    Log(['=> SetFixing command, you need to specify the new Fixation type']);
    Inc(Result);
    Exit;
  end;
  if cuUnit = nil then
  begin
    Log(['[Warning] SetFixing: No Selected Unit.']);
    Exit;
  end;

  cuUnit.FixType := STI(Params[0]);
  Log(['Unit', cuUnit.InternalName, 'has now this fixation value: ', cuUnit.FixType, 'in region', RegionDefs[cuUnit.StartSubSpace].Name]);
  if cuUnit.FixType <> _fixUnfixed then
    TGroup(cuUnit.pGroup).StopMove;
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ADD SUB FACTION REGION
// Ajout d'une region  la liste TFaction.dlSpecialRgn
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.AddSubFacRgn(Params : TStringList) : integer;
var FD : TFactionDef;
begin
  Result := 0;
  if (cuFaction = nil) then
  begin
    Log(['[Warning] AddSubFacRgn: No faction selected']);
    Exit;
  end;
  if (Params.Count <> 2) then
  begin
    Inc(Result);
    Log(['=> AddSubFacRgn: Need to give the FactionTag & Region UID.']);
    Exit;
  end;
  if Result <> 0 then Exit;
  FD := FactionDefs.GetByFactionTag(Params[0]);
  if FD = nil then
  begin
    Inc(Result);
    Log(['=> AddSubFacRgn: Unknown FactionTag', Params[0]]);
    Exit;
  end;

  cuFaction.dlSpecialRgn.Add(STI(Params[1]), FD.UID);

  Log(['AddSubFacRgn:', cuFaction.Name, 'Region:', Params[1], 'belong to SubFaction', FD.Name]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// REM SUB FACTION REGION
// Suppression d'une region  la liste TFaction.dlSpecialRgn
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.RemSubFacRgn(Params : TStringList) : integer;
var FD : TFactionDef;
    RgnUID : TUID;
    i : integer;
begin
  Result := 0;
  if (cuFaction = nil) then
  begin
    Log(['[Warning] RemSubFacRgn: No faction selected']);
    Exit;
  end;
  if (Params.Count <> 2) then
  begin
    Inc(Result);
    Log(['=> RemSubFacRgn: Need to give the FactionTag & Region UID.']);
    Exit;
  end;
  if Result <> 0 then Exit;
  FD := FactionDefs.GetByFactionTag(Params[0]);
  if FD = nil then
  begin
    Inc(Result);
    Log(['=> RemSubFacRgn: Unknown FactionTag', Params[0]]);
    Exit;
  end;
  RgnUID := STI(Params[1]);
  for i := cuFaction.dlSpecialRgn.Count-1 downto 0 do
  begin
    if (cuFaction.dlSpecialRgn.Keys[i] = RgnUID) and (cuFaction.dlSpecialRgn.Vals[i] = FD.UID) then
      cuFaction.dlSpecialRgn.Delete(i);
  end;

  Log(['RemSubFacRgn:', cuFaction.Name, 'Region:', Params[1], 'removed from SubFaction', FD.Name]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// CHANGE LOYALTY FACTION
// Change la loyaut d'une rgion ou d'une zone par une quantit, positif pour Insurgs, ngatif pour loyalistes
// ChangeLoyaltyFac = 5 {une rgion, la courante, valeur 5}
// ChangeLoyaltyFac = $Canada|5 {toute la zone Canada, de 5}
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.ChangeLoyaltyFac (Params : TStringList) : integer;
var AreaUID : TUID;
    rgn : integer;
    Area : TArea;
    Value : integer;
begin
  Result := 0;
  if not (Params.Count in [1, 2]) then
  begin
    Inc(Result);
    Log(['=> ChangeLoyaltyFac command, Invalid number of params, Params count:', Params.Count]);
  end;

  Area := nil;
  if Params.Count = 2 then
  begin
    AreaUID := STI(Params[0]);
    if AreaUID > Areas.Count-1 then
    begin
      Inc(Result);
      Log(['=> ChangeLoyaltyFac command, if two parameters, the first is an area. If one parameter, do a SelectRegion first, the parameter being the amount to change.', Params.Count]);
      Exit;
    end;
    Area := Areas[AreaUID];
    if (cuRegion = nil)and(Area.Count > 0)and(cuFaction <> nil) then cuRegion := cuFaction.Regions.Fetch(Area[0]);
  end;
  if (cuFaction = nil) then
  begin
    Inc(Result);
    Log(['=> ChangeLoyaltyFac command, no Faction selected']);
  end;
  if (cuRegion = nil) then
  begin
    Inc(Result);
    Log(['=> ChangeLoyaltyFac command, no region selected']);
  end;
  if Result <> 0 then Exit;

  if Params.Count = 1 then Value := STI(Params[0]) else
  if Params.Count = 2 then Value := STI(Params[1]) else Value := 0;

  if Params.Count = 1 then
  begin
    GameSim.RegionCmns.ChangeLoyalty(cuFaction, cuRegion.UID, Value);
    Log(['ChangeLoyaltyFac, one region:', cuRegion.Name, 'Loyalty now at', cuRegion.Loyalty]);
  end
  else
  begin
    for rgn := 0 to Area.Count -1 do
    begin
      cuRegion := cuFaction.Regions.Fetch(Area[rgn]);
      GameSim.RegionCmns.ChangeLoyalty(cuFaction, Area[rgn], Value);
      Log(['ChangeLoyaltyFac, OnArea:', GameSim.RegionCmns[Area[rgn]].Name, 'Loyalty now at', cuRegion.Loyalty, 'for', cuFaction.Name]);
    end;
  end;
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// CHANGE LOYALTY  !! ne plus utiliser, c'est une modif absolue, convertir BOA et dtruire cette fonction !!
// Change la loyaut d'une rgion ou d'une zone par une quantit, positif pour Insurgs, ngatif pour loyalistes
// ChangeLoyalty = 5 {une rgion, la courante, valeur 5}
// ChangeLoyalty = $Canada|5 {toute la zone Canada, de 5}
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.ChangeLoyalty (Params : TStringList) : integer;
var AreaUID : TUID;
    rgn : integer;
    Area : TArea;
    Value : integer;
begin
  Result := 0;  {#INWORK BOA2 - interdire aprs patch BOA1.13}
  LogMsg(['TScriptEngine.ChangeLoyalty', 'ChangeLoyalty is deprecated, use ChangeLoyaltyFac'], _Reporting);

  if rulSimpleLoyalty_ then
  begin
    if not (Params.Count in [1, 2]) then
    begin
      Inc(Result);
      Log(['=> ChangeLoyalty command, Invalid number of params, Params count:', Params.Count]);
    end;

    Area := nil;
    if Params.Count = 2 then
    begin
      AreaUID := STI(Params[0]);
      if AreaUID > Areas.Count-1 then
      begin
        Inc(Result);
        Log(['=> ChangeLoyalty command, if two parameters, the first is an area. If one parameter, do a SelectRegion first, the parameter being the amount to change.', Params.Count]);
        Exit;
      end;
      Area := Areas[AreaUID];
      if (cuRegion = nil)and(Area.Count > 0)and(cuFaction <> nil) then cuRegion := cuFaction.Regions.Fetch(Area[0]);
    end;
    if (cuRegion = nil) then
    begin
      Inc(Result);
      Log(['=> ChangeLoyalty command, no region selected']);
    end;
    if Result <> 0 then Exit;

    if Params.Count = 1 then Value := STI(Params[0]) else
    if Params.Count = 2 then Value := STI(Params[1]) else Value := 0;

    if Params.Count = 1 then
    begin
      if GameSim.RegionCmns[cuRegion.UID].SplLoyalty <> _loyNone then
      begin
        GameSim.RegionCmns[cuRegion.UID].SplLoyalty := GameSim.RegionCmns[cuRegion.UID].SplLoyalty  + Value;
      end;
      Log(['ChangeLoyalty, one region:', cuRegion.Name, 'Loyalty now at', GameSim.RegionCmns[cuRegion.UID].SplLoyalty]);
    end
    else
    begin
      for rgn := 0 to Area.Count -1 do
      begin
        if GameSim.RegionCmns[Area[rgn]].SplLoyalty <> _loyNone then
        begin
          GameSim.RegionCmns[Area[rgn]].SplLoyalty := GameSim.RegionCmns[Area[rgn]].SplLoyalty  + Value;
          Log(['ChangeLoyalty, OnArea:', GameSim.RegionCmns[Area[rgn]].Name, 'Loyalty now at', GameSim.RegionCmns[Area[rgn]].SplLoyalty]);
        end;
      end;
    end;
  end
  else
    Log(['[WARNING] ChangeLoyalty Only works with SimpleLoyalty for now']);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ADD LOYALTY FACTION COEFF
// Ajout d'un coefficient (ou remplacement d'un ancien) dans la liste des coeff
// de loyauts de la faction
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.AddLoyFacCoeff(Params : TStringList) : integer;
var AreaUID : TUID;
    Coeff : integer;
    Idx : integer;
begin
  Result := 0;
  if (cuFaction = nil) then
  begin
    Log(['[Warning] AddLoyFacCoeff: No faction selected']);
    Exit;
  end;
  if (Params.Count <> 2) then
  begin
    Inc(Result);
    Log(['=> AddLoyFacCoeff: Need to give the Area UID & Coefficient.']);
    Exit;
  end;
  if Result <> 0 then Exit;

  AreaUID := STI(Params[0]);
  if (AreaUID < 0)or(AreaUID > Areas.Count-1)or(Areas[AreaUID].HierarchyLevel <> aeaHighestHierarchy_) then
  begin
    Inc(Result);
    Log(['=> AddLoyFacCoeff: The AreaUID is invalid, this must be a GrandArea identifier, $Canada for example']);
    Exit;
  end;
  Coeff := STI(Params[1]);

  Idx := cuFaction.dlLoyaltyCoeff.Keys.IndexOf(ITS(AreaUID));
  if Idx <> -1 then
    cuFaction.dlLoyaltyCoeff.Delete(Idx);

  cuFaction.dlLoyaltyCoeff.Add(AreaUID, Coeff);

  Log(['AddLoyFacCoeff:', cuFaction.Name, Areas[AreaUID].iName, 'Loyalty Coeff. set to:', Coeff]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// MOVE UNIT
// Dplacement instantan de l'unit en cours de slection vers une rgion donne
// MoveUnit RgnUID[;NotEnemy]
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.MoveUnit(Params : TStringList) : integer;
var RgnUID : TUID;
    NewGrp : TGroup;
    Param : string;
    Bag : TRandomBag;
    AdjReg : TRegion;
    i : integer;
begin
  Result := 0;
  if Params.Count > 2 then
  begin
    Inc(Result);
    Log(['=> MoveUnit command, no region selected']);
  end;
  if Result <> 0 then Exit;

  if cuUnit = nil then
  begin
    Log(['[WARNING] MoveUnit command, no unit selected']);
    Exit
  end;
  // Gestion sur province non ennemie unqiuement
  Param := ExtractParam('NotEnemy', Params);

  RgnUID := STI(Params[0]);
  cuRegion  := cuFaction.Regions.Fetch(RgnUID);
  cuFaction := cuUnit.pFaction;

  // gestion avec NotEnemy, et rgion ennemie: prendre adjacent, o ne pas prendre
  if (Param <> '') and (cuRegion.IsEnemyControled(True)) then
  begin
    Log(['MoveUnit', cuRegion.Name, 'is enemy controlled']);
    // NotEnemy  grer
    Bag := TRandomBag.Create;
    try
    begin
      // il faut trouver une rgion terrestre adjacente & amie
      cuRegion.CalcLinks(True);
      for i := 0 to cuRegion.pListLinks.Count -1 do
      begin
        AdjReg := TRegion(cuRegion.pListLinks[i]);
        if not RegionDefs.IsLandRegion(AdjReg.UID) then Continue;
        if AdjReg.IsEnemyControled(True) then Continue;
        // on l'ajoute
        Bag.Add(AdjReg.UID);
      end;
      // on prend une rgion au hasard
      if Bag.Count > 0 then
      begin
        cuRegion := cuFaction.Regions.Fetch(Bag.PickInt);
        Log(['MoveUnit', cuRegion.Name, 'selected']);
      end
      else
      begin
        cuRegion := nil; //invalide
        Log(['MoveUnit, no region can be selected']);
      end;
    end;
    finally
      Bag.Free;
    end;
  end;

  if cuRegion = nil then Exit;
  if cuUnit.StartSubSpace = cuRegion.UID then Exit;

  NewGrp := cuFaction.CreateGroupInRegion(cuRegion, _genUID, cuUnit.Name, cuUnit.Kind);
  NewGrp.AddUnit(cuUnit);
  Log(['MoveUnit:', cuUnit.Name, 'is now in region', cuRegion.Name]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ON ATTACK
// Indique l'vnement  dclencher si une attaque contre la faction courante
// se produit dans la liste de rgions indique
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.OnAttack(Params : TStringList) : integer;
begin
  Result := 0;
  if Params.Count <> 2 then
  begin
    Inc(Result);
    Log(['=> OnAttack: need to have the name of the area and the event to trigger as arguments (alias allowed): OnAttack <AreaName> <EventName>']);
    Exit;
  end;
  if cuFaction = nil then
  begin
    Log(['=> OnAttack command, no faction selected']);
    Inc(Result);
    Exit;
  end;

  cuFaction.slOnAttack.Add(Params[0] + '|' + Params[1]);

  Log(['OnAttack', cuFaction.Name, 'AreaUID:', Params[0], 'EventName:', Params[1]]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// CHANGE SUPPLY
// Supply avan uniquement (car stock)
// Change par +- le stock de supply d'une rgion
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.ChangeSupply(Params : TStringList) : integer;
var GenSupply, Ammo, Power : integer;
begin
  Result := 0;
  if not (Params.Count in [3]) then
  begin
    Inc(Result);
    Log(['=> Change Supply command, erroneous number of params, Syntax is: ChangeSupply = GenSupply; Ammo; Power']);
  end;

  if rulSupply_ <> _gmdAdvanced then
  begin
    Inc(Result);
    Log(['=> Change Supply command, not in advanced supply.']);
  end;

  if cuRegion = nil then
  begin
    Inc(Result);
    Log(['=> Change Supply command, no region selected']);
  end;
  SetNumeric(Params[0], GenSupply);
  SetNumeric(Params[1], Ammo);
  SetNumeric(Params[2], Power);

  if (GenSupply = Low(integer))or(Ammo = Low(integer))or(GenSupply = Low(Power)) then
  begin
    Inc(Result);
    Log(['=> Change Supply command', 'One or more amount is not an integer']);
  end;
  if Result <> 0 then Exit;


  if RegionDefs.IsLandRegion(cuRegion.UID) then
  begin
    if GenSupply <> 0 then
    begin
      GameSim.RegionCmns[cuRegion.UID].SupplyStock  := GameSim.RegionCmns[cuRegion.UID].SupplyStock + GenSupply;
      Log(['SupplyStock in ', cuRegion.Name, 'set to', GameSim.RegionCmns[cuRegion.UID].SupplyStock]);
    end;
    if Ammo <> 0 then
    begin
      GameSim.RegionCmns[cuRegion.UID].AmmoStock    := GameSim.RegionCmns[cuRegion.UID].AmmoStock   + Ammo;
      Log(['AmmoStock in ', cuRegion.Name, 'set to', GameSim.RegionCmns[cuRegion.UID].AmmoStock]);
    end;
    if Power <> 0 then
    begin
      GameSim.RegionCmns[cuRegion.UID].PowerStock   := GameSim.RegionCmns[cuRegion.UID].PowerStock  + Power;
      Log(['PowerStock in ', cuRegion.Name, 'set to', GameSim.RegionCmns[cuRegion.UID].PowerStock]);
    end;
  end
  else
    Log(['Change Supply Command ignored:', cuRegion.Name, 'is a non land region']);

end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// CHANGE RESOURCE STOCK
// For adv economy: can change assets (money ...) or merchandises in stock, at the nation level
// ChangeResStock = ResAlias; Amount
// ChangeResStock = ResAlias; Amount; FAC1; FAC2 .. (2 factions at least here)
// ChangeResStock = ResAlias; Amount; FacKind
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.ChangeResStock(Params : TStringList) : integer;
var UID, Amount, i : integer;
    Faction : TFaction;
    FacKind : TUID;
begin
  Result := 0;
{ removed for vgn battle scenario with steel & goods but no adv eco
  if rulEconomyMode_ < _gmdAdvanced then
  begin
    Inc(Result);
    Log(['=> ChangeResStock command, Rule Advanced Economy is not active. Command ignored.']);
  end;
}
  // Nil faction = CMN (by kind)
  if  (Params.Count < 2) then
  begin
    Inc(Result);
    Log(['=> ChangeResStock command, erroneous number of params, Syntax is: ChangeResStock = ResAlias; Amount<;FAC1><;FAC2> ..']);
  end;
  UID := STI(Params[0], True, -1);
  Amount := STI(Params[1]);

  if (UID < 0) or (UID > GameSim.MerchandisesDB.Count-1) then
  begin
    Inc(Result);
    Log(['=> ChangeResStock command, invalid resource UID, found:', UID]);
  end;
  if Result > 0 then Exit;

  // on faction tag specified
  if Params.Count = 2 then
  begin
    cuFaction.ChangeResource(UID, Amount);
    Log(['ChangeResStock command (Single Faction), Resource', GameSim.MerchandisesDB[UID].Name, 'changed to', cuFaction.GetResource(UID)]);
  end
  else
  // one faction kind specified
  if Params.Count = 3 then
  begin
    FacKind := STI(Params[2], True, -1);
    if FacKind = -1 then
    begin
      Log(['=> ChangeResStock command, Erroneous faction kind', Params[2]]);
      Inc(Result);
      Exit;
    end;
    for i := 0 to GameSim.Factions.Count-1 do
    begin
      Faction := GameSim.Factions[i];
      if Faction.Kind <> FacKind then Continue;
      Faction.ChangeResource(UID, Amount);
      Log(['ChangeResStock command', Faction.Name, 'Resource', GameSim.MerchandisesDB[UID].Name, 'changed to', Faction.GetResource(UID)]);
    end;
  end
  else
  // several faction tags specified
  if Params.Count > 3 then
  begin
    for i := 2 to Params.Count-1 do
    begin
      Faction := FactionTagToFaction(Params[i]);
      if Faction = nil then
      begin
        Log(['=> [Warning] ChangeResStock command, Faction not found', Params[i]]);
        Continue;
      end;
      Faction.ChangeResource(UID, Amount);
      Log(['ChangeResStock command', Faction.Name, 'Resource', GameSim.MerchandisesDB[UID].Name, 'changed to', cuFaction.GetResource(UID)]);
    end;
  end;
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// CHANGE ASSETS PRODUCTION
// Gre Supply|Ammo|Power|Money|WSU|Conscripts (et tous les champs sont obligatoires}
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.ChangeAssetsProd (Params : TStringList) : integer;
var Supply, Ammo, Power, Money, WSU, Conscript : integer;
begin
  Result := 0;
  if not (Params.Count in [6]) then
  begin
    Inc(Result);
    Log(['=> Change Assets Production command, erroneous number of params, Syntax is: ChangeAssetsProd = Supply; Ammo; Power; Money; WSU; Conscript']);
    Exit;
  end;

  if cuRegion = nil then
  begin
    Inc(Result);
    Log(['=> Change Assets Production command, no region selected']);
    Exit;
  end;
  SetNumeric(Params[0], Supply);
  SetNumeric(Params[1], Ammo);
  SetNumeric(Params[2], Power);
  SetNumeric(Params[3], Money);
  SetNumeric(Params[4], WSU);
  SetNumeric(Params[5], Conscript);

  if (Supply = Low(integer))or(Ammo = Low(integer))or(Power = Low(integer))
  or (Money = Low(integer))or(WSU = Low(integer))or(Conscript = Low(integer)) then
  begin
    Inc(Result);
    Log(['=> Change Assets Production command', 'One or more amount is not an integer']);
  end;
  if Result <> 0 then Exit;


  if RegionDefs.IsLandRegion(cuRegion.UID)and(GameSim.RegionCmns.HasStructure(cuRegion.UID)) then
  begin
    if Supply <> 0 then
    begin
      GameSim.RegionCmns[cuRegion.UID].SupplyProd  := GameSim.RegionCmns[cuRegion.UID].SupplyProd + Supply;
      Log(['(External) Supply Generation in ', cuRegion.Name, 'changed to', GameSim.RegionCmns[cuRegion.UID].SupplyProd]);
    end;
    if Ammo <> 0 then
    begin
      GameSim.RegionCmns[cuRegion.UID].AmmoProd    := GameSim.RegionCmns[cuRegion.UID].AmmoProd   + Ammo;
      Log(['(External) Ammo Production in ', cuRegion.Name, 'changed to', GameSim.RegionCmns[cuRegion.UID].AmmoProd]);
    end;
    if Power <> 0 then
    begin
      GameSim.RegionCmns[cuRegion.UID].PowerProd   := GameSim.RegionCmns[cuRegion.UID].PowerProd  + Power;
      Log(['(External) Power Points in ', cuRegion.Name, 'changed to', GameSim.RegionCmns[cuRegion.UID].PowerProd]);
    end;
    if Money <> 0 then
    begin
      GameSim.RegionCmns[cuRegion.UID].MoneyProd  := GameSim.RegionCmns[cuRegion.UID].MoneyProd + Money;
      Log(['(External) Money Generation in ', cuRegion.Name, 'changed to', GameSim.RegionCmns[cuRegion.UID].MoneyProd]);
    end;
    if WSU <> 0 then
    begin
      GameSim.RegionCmns[cuRegion.UID].WSUProd    := GameSim.RegionCmns[cuRegion.UID].WSUProd   + WSU;
      Log(['(External) War Supply Production in ', cuRegion.Name, 'changed to', GameSim.RegionCmns[cuRegion.UID].WSUProd]);
    end;
    if Conscript <> 0 then
    begin
      GameSim.RegionCmns[cuRegion.UID].ConsProd   := GameSim.RegionCmns[cuRegion.UID].ConsProd  + Conscript;
      Log(['(External) Conscripts Points in ', cuRegion.Name, 'changed to', GameSim.RegionCmns[cuRegion.UID].ConsProd]);
    end;
  end
  else
    Log(['Change Assets Production Command ignored:', cuRegion.Name, 'has no structure or is a water region']);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// CHANGE RAIL POOL
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.ChangeRailPool(Params : TStringList) : integer;
begin
  Result := 0;
  if Params.Count <> 1 then
  begin
    Inc(Result);
    Log(['=> ChangeRailPool, need to have exactly one parameter: Amount of RailPool points given (or substracted)']);
    Exit;
  end;
  if cuFaction = nil then
  begin
    Log(['=> ChangeRailPool command, no faction selected']);
    Inc(Result);
    Exit;
  end;

  cuFaction.RailPool := Max(0, cuFaction.RailPool + STI(Params[0]));

  Log(['ChangeRailPool', cuFaction.Name, 'New rail pool is', cuFaction.RailPool]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// CHANGE RIVER POOL
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.ChangeRiverPool(Params : TStringList) : integer;
begin
  Result := 0;
  if Params.Count <> 1 then
  begin
    Inc(Result);
    Log(['=> ChangeRiverPool, need to have exactly one parameter: Amount of Riverpool points given (or substracted)']);
    Exit;
  end;
  if cuFaction = nil then
  begin
    Log(['=> ChangeRiverPool command, no faction selected']);
    Inc(Result);
    Exit;
  end;

  cuFaction.RiverPool := Max(0, cuFaction.RiverPool + STI(Params[0]));

  Log(['ChangeRiverPool', cuFaction.Name, 'New river pool is', cuFaction.RiverPool]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// CHANGE MONEY POOL
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.ChangeMoneyPool (Params : TStringList) : integer;
begin
  Result := 0;
  if Params.Count <> 1 then
  begin
    Inc(Result);
    Log(['=> ChangeMoneyPool, need to have exactly one parameter: Amount of Money points given (or substracted) to the current faction']);
    Exit;
  end;
  if cuFaction = nil then
  begin
    Log(['=> ChangeMoneyPool command, no faction selected']);
    Inc(Result);
    Exit;
  end;

  cuFaction.Money := cuFaction.Money + STI(Params[0]);

  Log(['ChangeMoneyPool', cuFaction.Name, 'New money reserve is', cuFaction.Money]);
end;




// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// CHANGE WSU POOL
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.ChangeWSUPool (Params : TStringList) : integer;
begin
  Result := 0;
  if Params.Count <> 1 then
  begin
    Inc(Result);
    Log(['=> ChangeWSUPool, need to have exactly one parameter: Amount of WSU points given (or substracted) to the current faction']);
    Exit;
  end;
  if cuFaction = nil then
  begin
    Log(['=> ChangeWSUPool command, no faction selected']);
    Inc(Result);
    Exit;
  end;

  cuFaction.WarSupply := Max(0, cuFaction.WarSupply + STI(Params[0]));

  Log(['ChangeWSUPool', cuFaction.Name, 'New WSU pool is', cuFaction.WarSupply]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// CHANGE CONSCRIPTS
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.ChangeConscripts(Params : TStringList) : integer;
begin
  Result := 0;
  if Params.Count <> 1 then
  begin
    Inc(Result);
    Log(['=> ChangeConscriptPool, need to have exactly one parameter: Amount of Conscripts points given (or substracted) to the current faction']);
    Exit;
  end;
  if cuFaction = nil then
  begin
    Log(['=> ChangeConscriptPool command, no faction selected']);
    Inc(Result);
    Exit;
  end;

  cuFaction.Conscript := Max(0, cuFaction.Conscript + STI(Params[0]));

  Log(['ChangeConscriptPool', cuFaction.Name, 'New Conscripts pool is', cuFaction.Conscript]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// CHANGE PRICE INCREASE
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.ChangePriceInc (Params : TStringList) : integer;
begin
  Result := 0;
  if Params.Count <> 1 then
  begin
    Inc(Result);
    Log(['=> ChangePriceInc, need to have exactly one parameter: Amount of Price Increase points given (or substracted) to the current faction']);
    Exit;
  end;
  if cuFaction = nil then
  begin
    Log(['=> ChangePriceInc command, no faction selected']);
    Inc(Result);
    Exit;
  end;

  cuFaction.Inflation := Max(aseInflationFloor_, cuFaction.Inflation + STI(Params[0]));

  Log(['ChangePriceInc', cuFaction.Name, 'New Price Increase index is', cuFaction.Inflation]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// CHANGE MORALE
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.ChangeMorale (Params : TStringList) : integer;
begin
  Result := 0;
  if Params.Count <> 1 then
  begin
    Inc(Result);
    Log(['=> ChangeMorale, need to have exactly one parameter: Amount of National Morale points given (or substracted) to the current faction']);
    Exit;
  end;
  if cuFaction = nil then
  begin
    Log(['=> ChangeMorale command, no faction selected']);
    Inc(Result);
    Exit;
  end;

  cuFaction.Morale := Max(0, cuFaction.Morale + STI(Params[0]));

  Log(['ChangeMorale', cuFaction.Name, 'New National Morale is', cuFaction.Morale]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// CHANGE ENGAGEMENT POINTS
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.ChangeEngagementPts (Params : TStringList) : integer;
var Value : integer;

begin
  Result := 0;
  Value  := 0;

  if rulEngagementPts_ = _eptNoEngagementPts then
  begin
    Inc(Result);
    Log(['=> ChgFacEngagementPts command, game rule not activated, command ignored']);
  end;

  if (Params.Count <> 1) and (Params.Count <> 2) then
  begin
    Inc(Result);
    Log(['=> ChgFacEngagementPts: Wrong number of Parameters. Syntax: ChgFacEngagementPts = Value or ChgFacEngagementPts = esvIntVar; Modifier in per cent']);
  end;

  if cuFaction = nil then
  begin
    Log(['=> ChgFacEngagementPts command, no faction selected']);
    Inc(Result);
  end;

  if Result > 0 then Exit;

  if Params.Count = 1 then
  begin
    Value := STI(Params[0]);
  end
  else
  if Params.Count = 2 then
  begin
    if IsExposedVar(Params[0]) then
    begin
      Value := RoundRandom(GetExposedVarValue(Params[0]) * STI(Params[1]) / 100);
    end
    else
    begin
      Inc(Result);
      Log(['=> ChgFacEngagementPts: Parameter not recognized as an exposed var', Params[0]]);
      Exit;
    end;
  end;

  cuFaction.EngagementPts := Max(0, cuFaction.EngagementPts + Value);

  Log(['ChgFacEngagementPts', cuFaction.Name, 'New Nb. of Engagement Points are', cuFaction.EngagementPts]);
end;


// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// CHANGE UNIT POOL
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.ChangeUnitPool(Params : TStringList) : integer;
var UnitDefUID, Value : integer;
begin
  Result := 0;
  if Params.Count <> 2 then
  begin
    Inc(Result);
    Log(['=> ChangeUnitPool, need to have exactly two parameters: Alias of UnitDef and variation of Force Pool']);
    Exit;
  end;
  if cuFaction = nil then
  begin
    Log(['=> ChangeUnitPool command, no faction selected']);
    Inc(Result);
    Exit;
  end;
  Value      := STI(Params[1]);
  if Value = 0 then Exit; // scripters like to make unneeded commands)

  UnitDefUID := STI(Params[0], True, -1);

  if UnitDefUID = -1 then
  begin
    if IsSpecialScript then
    begin
      Log(['[Warning] ChangeUnitPool command, the current faction don''t have such unit, exiting silently.']);
      Exit;
    end;

    if cuFaction = nil then
    begin
      Log(['=> ChangeUnitPool command,', Params[0], 'is an unknown unit']);
      Inc(Result);
      Exit;
    end;
    if (UnitDefUID < 0) or (UnitDefUID > UnitDefs.Count-1) then
    begin
      Log(['=> ChangeUnitPool command, the UnitDef UID given is either negative or above the units definitions UID. ', UnitDefUID]);
      Inc(Result);
      Exit;
    end;
  end;
  cuFaction.ChangeUnitMaxForcePool(UnitDefUID, Value);
  Log(['ChangeUnitPool', cuFaction.Name, 'New Force Pool for unit (UID/Name) is', UnitDefUID, UnitDefs[UnitDefUID].Name,  cuFaction.GetUnitMaxForcePool(UnitDefUID, True)]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// CHANGE UNIT TYPE
// Convert a unit DefUID into another, both on map and in the force pool (optionnally)
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.ChangeUnitType(Params : TStringList) : integer;
var OldUID    : TUID;
    NewUID    : TUID;
    InFP      : boolean;
    OnMap     : boolean;
    AreaUID   : TUID;
    Param     : string;
    Group     : TGroup;
    aUnit     : TUnit;
    grp,unt,i : integer;
    OpDone    : boolean;
    PrevName  : string;
    lsRgnList : TList;
    Region    : TRegion;
begin
  Result := 0;
  if (Params.Count < 2)or (Params.Count > 5) then
  begin
    Inc(Result);
    Log(['=> ChangeUnitType, need to have from 2 to 6 parameters: ChangeUnitType = <OldUnitDef>;<NewUnitDef>[;InFP][;OnMap][;REGorAREA_AreaUID]']);
    Exit;
  end;
  if cuFaction = nil then
  begin
    Log(['=> ChangeUnitType command, no faction selected']);
    Inc(Result);
    Exit;
  end;

  // If nothing specified, its on map. If REG, then we expect a cuRegion, if AREA, then there is an area alias.
  Param := ExtractParam('REG', Params);
  if Param <> '' then
    AreaUID := -2
  else
    AreaUID := _noUID;
  Param := ExtractParam('AREA', Params);
  if Param <> '' then
  begin
    ExtractArgs(Param);
    AreaUID := STI(Args[0]);
  end;
  // Region Checking
  if (AreaUID = -2)and(cuRegion = nil) then
  begin
    Inc(Result);
    Log(['=> ChangeUnitType command, no region selected.']);
  end;

  OldUID      := STI(Params[0], True, Low(integer));
  NewUID      := STI(Params[1], True, Low(integer));

  InFP  := ExtractParam('InFP', Params) <> '';
  OnMap := ExtractParam('OnMap', Params) <> '';

  if (not InFP)and(not OnMap) then
  begin
    Inc(Result);
    Log(['=> ChangeUnitType command, Must be either InFp or OnMap, or both.']);
  end;

  if ((OldUID < 0) or (OldUID > UnitDefs.Count-1)) then
  begin
    Log(['[Warning] ChangeUnitType command, the current faction don''t have such (old) unit, exiting silently.']);
    Exit;
  end;
  if ((NewUID < 0) or (NewUID > UnitDefs.Count-1)) then
  begin
    Log(['[Warning] ChangeUnitType command, the current faction don''t have such (new) unit, exiting silently.']);
    Exit;
  end;

  if OnMap then
  begin
    // Creating the region list
    lsRgnList := TList.Create;
    if AreaUID = -2 then lsRgnList.add(cuRegion);
    if AreaUID = _noUID then
    begin
      for i := 0 to cuFaction.Regions.Count-1 do
        lsRgnList.add(cuFaction.Regions[i]);
    end;
    if AreaUID >= 0 then
    begin
      for i := 0 to Areas[AreaUID].Count-1 do
      begin
        if cuFaction.Regions.Fetch(i, False) <> nil then
          lsRgnList.Add(cuFaction.Regions.Fetch(i));
      end;
    end;

    for i := 0 to lsRgnList.Count-1 do
    begin
      OpDone := False;
      Region := lsRgnList[i];
      for grp := 0 to Region.pListGroup.Count-1 do
      begin
        Group := Region.pListGroup[grp];
        for unt := 0 to Group.Count-1 do
        begin
        aUnit := Group[unt];
        if aUnit.UnitType = OldUID then
        begin
          OpDone := True;
          PrevName := aUnit.Name;
          aUnit.UnitType := NewUID;
          // If not a custom name, it is updated
          if PrevName = UnitDefs[OldUID].Name then
            aUnit.Name := UnitDefs[NewUID].Name;
          Log(['ChangeUnitType command', 'In Region', Region.UID, Region.Name, 'Unit:', PrevName, 'of type', OldUID, 'converted to type', NewUID, 'Database Name:', UnitDefs[NewUID].Name]);
        end;
        end;
      end;
      if OpDone then
        GameSim.RegionCmns.Update(Region.UID);
    end;
    lsRgnList.Free;
  end;

  // Force Pool changes (as a whole)
  if InFP then
  begin
    if cuFaction.ConvertUnitInMaxForcePool(OldUID, NewUID) then
      Log(['ChangeUnitType command', 'Force Pool conversion of', OldUID, UnitDefs[OldUID].Name, 'to', NewUID, UnitDefs[NewUID].Name]);
  end;

  // Adjusting AI Pool Ratios
  // AACW (area rec): weight per family, ROP+ (region rec): weight per UnitDefUID

  // this is now done implicitely each time
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// CHANGE UNIT TYPE BY GENERATION
// Convert a unit DefUID into another, both on map and in the force pool (optionnally)
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.ChangeUnitTypeByGen(Params : TStringList) : integer;
var GenName   : string;
    TechPath  : string;
    InFP      : boolean;
    OnMap     : boolean;
    AreaUID   : TUID;
    Param     : string;
    Group     : TGroup;
    aUnit     : TUnit;
    grp,unt,i : integer;
    OpDone    : boolean;
    PrevName  : string;
    lsRgnList : TList;
    Region    : TRegion;
    StdUpg    : boolean;
    NewUID    : TUID;
    OldUID    : TUID;
begin
  Result := 0;
  if (Params.Count < 2)or (Params.Count > 5) then
  begin
    Inc(Result);
    Log(['=> ChangeUnitType, need to have from 2 to 6 parameters: ChangeUnitType = <UnitGeneration>;<TechUpgPath>[;InFP][;OnMap][;REGorAREA_AreaUID]']);
    Exit;
  end;
  if cuFaction = nil then
  begin
    Log(['=> ChangeUnitType command, no faction selected']);
    Inc(Result);
    Exit;
  end;

  // If nothing specified, its on map. If REG, then we expect a cuRegion, if AREA, then there is an area alias.
  Param := ExtractParam('REG', Params);
  if Param <> '' then
    AreaUID := -2
  else
    AreaUID := _noUID;
  Param := ExtractParam('AREA', Params);
  if Param <> '' then
  begin
    ExtractArgs(Param);
    AreaUID := STI(Args[0]);
  end;
  // Region Checking
  if (AreaUID = -2)and(cuRegion = nil) then
  begin
    Inc(Result);
    Log(['=> ChangeUnitType command, no region selected.']);
  end;

  GenName      := Params[0];
  TechPath     := Params[1];
  if UpperCase(TechPath) = 'TECHUPGALT' then
    StdUpg := False // using TechUpgAlt field
  else
    StdUpg := True; // using TechAutoUpg field

  InFP  := ExtractParam('InFP', Params) <> '';
  OnMap := ExtractParam('OnMap', Params) <> '';

  if (not InFP)and(not OnMap) then
  begin
    Inc(Result);
    Log(['=> ChangeUnitType command, Must be either InFp or OnMap, or both.']);
  end;

  if OnMap then
  begin
    // Creating the region list
    lsRgnList := TList.Create;
    if AreaUID = -2 then lsRgnList.add(cuRegion);
    if AreaUID = _noUID then
    begin
      for i := 0 to cuFaction.Regions.Count-1 do
        lsRgnList.add(cuFaction.Regions[i]);
    end;
    if AreaUID >= 0 then
    begin
      for i := 0 to Areas[AreaUID].Count-1 do
      begin
        if cuFaction.Regions.Fetch(i, False) <> nil then
          lsRgnList.Add(cuFaction.Regions.Fetch(i));
      end;
    end;

    for i := 0 to lsRgnList.Count-1 do
    begin
      OpDone := False;
      Region := lsRgnList[i];
      for grp := 0 to Region.pListGroup.Count-1 do
      begin
        Group := Region.pListGroup[grp];
        for unt := 0 to Group.Count-1 do
        begin
        aUnit := Group[unt];
        if aUnit.UnitDef.PoolGeneration = GenName then // valid unit found
        begin

          PrevName := aUnit.Name;
          if StdUpg then
            NewUID := aUnit.UnitDef.TechAutoUpg
          else
            NewUID := aUnit.UnitDef.TechUpgAlt;
          if NewUID = _noUID then Continue; // check if path is valid

          OldUID := aUnit.UnitType;
          OpDone := True; // yes, do the changes
          aUnit.UnitType := NewUID;
          // If not a custom name, it is updated
          if PrevName = UnitDefs[OldUID].Name then
            aUnit.Name := UnitDefs[NewUID].Name;
          Log(['ChangeUnitType command', 'In Region', Region.UID, Region.Name, 'Unit:', PrevName, 'of type', OldUID, 'converted to type', NewUID, 'Database Name:', UnitDefs[NewUID].Name]);
        end;
        end;
      end;
      if OpDone then
        GameSim.RegionCmns.Update(Region.UID);
    end;
    lsRgnList.Free;
  end;

  // Force Pool changes (as a whole)
  if InFP then
  begin
    if cuFaction.ConvertUnitInMaxForcePool(GenName, StdUpg) then
      Log(['ChangeUnitType command', 'Force Pool conversion, units were found in FP and converted', GenName]);
  end;

  // Adjusting AI Pool Ratios
  // AACW (area rec): weight per family, ROP+ (region rec): weight per UnitDefUID

  // this is now done implicitely each time
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// OBSOLETE MODEL GEN
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.ObsoleteModelGen(Params : TStringList) : integer;
var i : integer;
begin
  Result := 0;
  if cuFaction = nil then
  begin
    Log(['=> ObsoleteModelGen command, no faction selected']);
    Inc(Result);
    Exit;
  end;
  for i := 0 to Params.Count-1 do
    cuFaction.Tech_MakeThisModelTagObsolete(Params[i]);
end;
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// OBSOLETE STRUC GEN
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.ObsoleteStrucGen(Params : TStringList) : integer;
var i : integer;
begin
  Result := 0;
  if cuFaction = nil then
  begin
    Log(['=> ObsoleteStrucGen command, no faction selected']);
    Inc(Result);
    Exit;
  end;
  for i := 0 to Params.Count-1 do
    cuFaction.Tech_MakeThisStrucTagObsolete(Params[i]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// IS OBSOLETE MODEL GENERATION
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.IsObsoleteModelGen(Params : TStringList) : integer;
var Reverse : boolean;
begin
  Result := 0;
  if (Params.Count > 2) then
  begin
    Inc(Result);
    Log(['=> IsObsoleteModelGen, Syntax is IsObsoleteModelGen = GenerationTag [;NOT]']);
    Exit;
  end;
  if cuFaction = nil then
  begin
    Log(['=> IsObsoleteModelGen command, no faction selected']);
    Inc(Result);
    Exit;
  end;
  Reverse := ExtractParam('NOT', Params) <> '';

  if (not cuFaction.Tech_IsObsoleteModelTag(Params[0])) or (Reverse and cuFaction.Tech_IsObsoleteModelTag(Params[0])) then
      Conditions := condFalse;
  if not Reverse then
    Log(['IsObsoleteModelGen:', 'Checking if this Model Generation Tag is obsolete', cuFaction.UID, cuFaction.Name, IIF(Conditions = condEvaluating, 'True', 'False')])
  else
    Log(['IsObsoleteModelGen:', 'Checking if this Model Generation Tag is NOT obsolete', cuFaction.UID, cuFaction.Name, IIF(Conditions = condEvaluating, 'True', 'False')])
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// IS OBSOLETE STRUCTURE GENERATION
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.IsObsoleteStrucGen(Params : TStringList) : integer;
var Reverse : boolean;
begin
  Result := 0;
  if (Params.Count > 2) then
  begin
    Inc(Result);
    Log(['=> IsObsoleteStrucGen, Syntax is IsObsoleteStrucGen = GenerationTag [;NOT]']);
    Exit;
  end;
  if cuFaction = nil then
  begin
    Log(['=> IsObsoleteStrucGen command, no faction selected']);
    Inc(Result);
    Exit;
  end;
  Reverse := ExtractParam('NOT', Params) <> '';

  if (not cuFaction.Tech_IsObsoleteStrucTag(Params[0])) or (Reverse and cuFaction.Tech_IsObsoleteStrucTag(Params[0])) then
      Conditions := condFalse;
  if not Reverse then
    Log(['IsObsoleteStrucGen:', 'Checking if this Structure Generation Tag is obsolete', cuFaction.UID, cuFaction.Name, IIF(Conditions = condEvaluating, 'True', 'False')])
  else
    Log(['IsObsoleteStrucGen:', 'Checking if this Structure Generation Tag is NOT obsolete', cuFaction.UID, cuFaction.Name, IIF(Conditions = condEvaluating, 'True', 'False')])
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// EVAL PRISONERS
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.EvalPrisoners(Params : TStringList) : integer;
var cuNumber  : integer;
    FactionTag: string;
    Value     : integer;
    Operator  : string;
begin
  Result := 0;
  if (Params.Count <> 3) then
  begin
    Inc(Result);
    Log(['=> Evaluate Prisoners, syntax is: EvalPrisoners = FactionTag_or_XXX ; Operator (one of: =, <, >, <=, >=, DICE) ; Value']);
  end;
  if (cuFaction = nil) then
  begin
    Inc(Result);
    Log(['=> EvalPrisoners: No faction selected']);
  end;
  if Result > 0 then Exit;

  FactionTag := Params[0];
  Operator   := Params[1];
  Value      := STI(Params[2]);
  cuNumber   := cuFaction.GetPrisoners(FactionTag); // if XXX it does automatically GetTotalPrisoners

  if Operator = '=' then
  begin
    if not (cuNumber = Value) then Conditions := condFalse;
  end
  else
  if Operator = '<' then
  begin
    if not (cuNumber < Value) then Conditions := condFalse;
  end
  else
  if Operator = '<=' then
  begin
    if not (cuNumber <= Value) then Conditions := condFalse;
  end
  else
  if Operator = '>' then
  begin
    if not (cuNumber > Value) then Conditions := condFalse;
  end
  else
  if Operator = '>=' then
  begin
    if not (cuNumber >= Value) then Conditions := condFalse;
  end
  else
  if UpperCase(Operator) = 'DICE' then
  begin
    if (Dice(Value) > cuNumber) then Conditions := condFalse;
  end;

  Log(['EvalPrisoners:', cuFaction.Name, 'Tag (XXX means all):', factionTag, 'Current Number', cuNumber, 'against', Value, ' - Operator:', Operator, 'Result:',
  IIF(Conditions = condEvaluating, 'True', 'False')]);

end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// EVAL REGION DECISION DEF
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.EvalRgnDecisionDef(Params : TStringList) : integer;
var cuNumber  : integer;
    ObjectUID : TUID;
    Value     : integer;
    Operator  : string;
begin
  Result := 0;
  if (Params.Count <> 3) then
  begin
    Inc(Result);
    Log(['=> EvalRgnDecisionDef, syntax is: EvalRgnDecisionDef = <RgnDecisionDefUID>; <Operator (one of: =, <, >, <=, >=, DICE)> ; <Value>']);
  end;
  if (cuFaction = nil) then
  begin
    Inc(Result);
    Log(['=> EvalRgnDecisionDef: No faction selected']);
  end;
  if Result > 0 then Exit;


  ObjectUID  := STI(Params[0]);
  Operator   := Params[1];
  Value      := STI(Params[2]);

  if (ObjectUID <= 0) or (ObjectUID > GameSim.RgnDecisionsDB.Count-1) then
  begin
    Inc(Result);
    Log(['=> EvalRgnDecisionDef, Invalid decision, UID found', ObjectUID]);
  end;

  cuNumber := cuFaction.GetNumRgnDecisionDefs(ObjectUID, True);

  if Operator = '=' then
  begin
    if not (cuNumber = Value) then Conditions := condFalse;
  end
  else
  if Operator = '<' then
  begin
    if not (cuNumber < Value) then Conditions := condFalse;
  end
  else
  if Operator = '<=' then
  begin
    if not (cuNumber <= Value) then Conditions := condFalse;
  end
  else
  if Operator = '>' then
  begin
    if not (cuNumber > Value) then Conditions := condFalse;
  end
  else
  if Operator = '>=' then
  begin
    if not (cuNumber >= Value) then Conditions := condFalse;
  end
  else
  if UpperCase(Operator) = 'DICE' then
  begin
    if (Dice(Value) > cuNumber) then Conditions := condFalse;
  end;

  Log(['EvalRgnDecisionDef:', cuFaction.Name,  GameSim.RgnDecisionsDB[ObjectUID].Name, 'Current Number', cuNumber, 'against', Value, ' - Operator:', Operator, 'Result:',
  IIF(Conditions = condEvaluating, 'True', 'False')]);

end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// EVAL REGION DECISION ACT
// Check if a RD is in a region
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.EvalRgnDecisionAct(Params : TStringList) : integer;
var
    ObjectID : TUID;
    Value     : integer;
begin
  Result := 0;
  if (Params.Count <> 2) then
  begin
    Inc(Result);
    Log(['=> EvalRgnDecisionAct, syntax is: EvalRgnDecisionAct = <RgnDecisionDefUID_or_Attribute>; <Value_0_or_1>']);
  end;
  if (cuFaction = nil) then
  begin
    Inc(Result);
    Log(['=> EvalRgnDecisionAct: No faction selected']);
  end;
  if cuRegion = nil then
  begin
    Inc(Result);
    Log(['=> EvalRgnDecisionAct, No region selected']);
  end;

  if Result > 0 then Exit;

  ObjectID  := STI(Params[0]);
  Value      := STI(Params[1]);

  // Two cases, either a RGD UID or a RGD attribute
  if ObjectID <= 0 then
  begin
    if cuFaction.HasThisRgnDecisionAct(cuRegion.UID, Params[0]) and (Value = 0) then
      Conditions := condFalse
    else
    if (not cuFaction.HasThisRgnDecisionAct(cuRegion.UID, Params[0])) and (Value = 1) then
      Conditions := condFalse;
      Log(['EvalRgnDecisionAct:', cuFaction.Name,  'on attribute', Params[0], 'in region', cuRegion.UID, cuRegion.Name, IIF(Value = 0, 'NOT there', 'THERE'), 'Result:',
      IIF(Conditions = condEvaluating, 'True', 'False')]);
  end
  else // Numeric UID
  begin
    if (ObjectID <= 0) or (ObjectID > GameSim.RgnDecisionsDB.Count-1) then
    begin
      Inc(Result);
      Log(['=> EvalRgnDecisionAct, Invalid decision, UID found', ObjectID]);
    end;
    if Result > 0 then Exit;

    if cuFaction.HasThisRgnDecisionAct(cuRegion.UID, ObjectID) and (Value = 0) then
      Conditions := condFalse
    else
    if (not cuFaction.HasThisRgnDecisionAct(cuRegion.UID, ObjectID)) and (Value = 1) then
      Conditions := condFalse;

    Log(['EvalRgnDecisionAct:', cuFaction.Name,  GameSim.RgnDecisionsDB[ObjectID].Name, 'in region', cuRegion.UID, cuRegion.Name, IIF(Value = 0, 'NOT there', 'THERE'), 'Result:',
    IIF(Conditions = condEvaluating, 'True', 'False')]);
  end;
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// IS REGION DECISION ACT VALID
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.IsRgnDecisionActValid(Params : TStringList) : integer;
var
    ObjectUID : TUID;
    RDD : TRgnDecisionDef;
    TrueRgnUID : TUID;
    AreaUID : TUID;
    Area : TArea;
begin
  Result := 0;
  if (Params.Count <> 1) then
  begin
    Inc(Result);
    Log(['=> IsRgnDecisionActValid, syntax is: IsRgnDecisionActValid = <RgnDecisionDefUID>']);
  end;
  if (cuFaction = nil) then
  begin
    Inc(Result);
    Log(['=> IsRgnDecisionActValid: No faction selected']);
  end;
  if cuRegion = nil then
  begin
    Inc(Result);
    Log(['=> IsRgnDecisionActValid, No region selected']);
  end;

  if Result > 0 then Exit;

  ObjectUID  := STI(Params[0]);

  if (ObjectUID <= 0) or (ObjectUID > GameSim.RgnDecisionsDB.Count-1) then
  begin
    Inc(Result);
    Log(['=> IsRgnDecisionActValid, Invalid decision, UID found', ObjectUID]);
  end;
  if Result > 0 then Exit;

  // If the Region Decision works at Area Level, we test validity check on the area capital
  // of the area the cuRegion is in
  RDD := GameSim.RgnDecisionsDB[ObjectUID];
  if RDD.MustBeHeadOfMidHierarchy then
  begin
    AreaUID := RegionDefs[cuRegion.UID].MidAreaUID;
    if AreaUID = _noUID then
    begin
      Inc(Result);
      Log(['=> IsRgnDecisionActValid, Invalid decision, it is targeting a region that is not part of any Mid Area, and the RD works at this level, RD:', ObjectUID]);
      Exit;
    end;
    Area := Areas[AreaUID];
    TrueRgnUID := GameSim.RegionCmns.GetAreaCalcCapital(Area);
    if TrueRgnUID = _noUID then
    begin
      Log(['IsRgnDecisionActValid', 'Can''t process further the RD, it works at area level and there is no area capital existing (not an error, exiting silently), RD:', ObjectUID]);
      Exit;
    end;
  end
  else
  if RDD.MustBeHeadOfHighHierarchy then
  begin
    AreaUID := RegionDefs[cuRegion.UID].GrandAreaUID;
    if AreaUID = _noUID then
    begin
      Inc(Result);
      Log(['=> IsRgnDecisionActValid, Invalid decision, it is targeting a region that is not part of any Grand Area, and the RD works at this level, RD:', ObjectUID]);
      Exit;
    end;
    Area := Areas[AreaUID];
    TrueRgnUID := GameSim.RegionCmns.GetAreaCalcCapital(Area);
    if TrueRgnUID = _noUID then
    begin
      Log(['IsRgnDecisionActValid', 'Can''t process further the RD, it works at area level and there is no area capital existing (not an error, exiting silently), RD:', ObjectUID]);
      Exit;
    end;
  end
  else
    TrueRgnUID := cuRegion.UID; // standard case

  if (cuFaction.IsRgnDecisionActValid(TrueRgnUID, ObjectUID, GameSim.cuTurn, True) <> _rgdIsValid) then
    Conditions := condFalse;

  Log(['IsRgnDecisionActValid:', cuFaction.Name,  GameSim.RgnDecisionsDB[ObjectUID].Name, 'in region', TrueRgnUID, RegionDefs[TrueRgnUID].Name, 'Result:',
  IIF(Conditions = condEvaluating, 'True', 'False')]);

end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// EVAL RULER
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.EvalRuler(Params : TStringList) : integer;
var RulerDefUID : TUID;
    Position : integer;
    ReverseResult : boolean;
begin
  Result := 0;
  if (Params.Count < 2) or (Params.Count > 3) then
  begin
    Inc(Result);
    Log(['=> EvalRuler command, syntax is: EvalRuler = <RulerUID>  ;<Position> [;<NOT>]']);
  end;
  if (cuFaction = nil) then
  begin
    Inc(Result);
    Log(['=> EvalRuler: No faction selected']);
  end;
  if Result > 0 then Exit;

  RulerDefUID := STI(Params[0], True, -1);
  Position := STI(Params[1]);

  ReverseResult := (Params.Count = 3) and (UpperCase(Params[2]) = 'NOT');

  if GameSim.RulerDB.InList(RulerDefUID) = False then
  begin
    Inc(Result);
    Log(['=> EvalRuler: Ruler UID was not found in Ruler DB. Ruler UID:', RulerDefUID]);
    Exit;
  end;

  if Result > 0 then Exit;
  if not cuFaction.RulerList.InList(RulerDefUID) then
    Conditions := condFalse;
  if (Position <> -1) and (cuFaction.RulerList.GetRulerByPos(Position) <> nil) and
  (TRulerItem(cuFaction.RulerList.GetRulerByPos(Position)).UID = RulerDefUID) then
    DoNothing
  else
    Conditions := condFalse;

  if ReverseResult then
  begin
    if Conditions = condEvaluating then
      Conditions := condFalse
    else
      Conditions := condEvaluating;
  end;

  if not ReverseResult then
    Log(['EvalRuler:', cuFaction.Name, 'searching for the presence of Ruler/Minister', GameSim.RulerDB[RulerDefUID].Name, 'Position (-1 = all):', Position, IIF(Conditions = condEvaluating, 'True', 'False')])
  else
    Log(['EvalRuler:', cuFaction.Name, 'searching for the absence of Ruler/Minister', GameSim.RulerDB[RulerDefUID].Name, 'Position (-1 = all):', Position, IIF(Conditions = condEvaluating, 'True', 'False')])
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// EVAL COLONIAL STATUS
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.EvalColStatus(Params : TStringList) : integer;
var Value : integer;
    Operator : string;
    CA : TColonialArea;
    cuStatus : TUID;
    OthFaction : TFaction;
    RgnUID : TUID;
begin
  Result := 0;
  if Params.Count <> 3 then
  begin
    Inc(Result);
    Log(['=> EvalColStatus command, syntax is: EvalColStatus = <RgnUID>;<Operator>;<Status>']);
  end;
  if (cuFaction = nil) then
  begin
    Log(['EvalColStatus: No faction, this will be an absolute test']);
  end;

  Operator := Params[0];
  RgnUID  := STI(Params[0], False, -1);
  Value    := STI(Params[2]);

  if (RgnUID < 0) or (RgnUID > RegionDefs.Count-1) then
  begin
    Inc(Result);
    Log(['=> EvalColStatus: Region alias is not correct:', Params[0]]);
    Exit;
  end;
  CA := GameSim.ColModule.GetColAreaFromRegion(RgnUID);
  if CA = nil then
  begin
    Log(['[Warning] EvalColStatus: The region is not in a colonial area', RgnUID]);
    Exit;
  end;
  if Result > 0 then Exit;

  OthFaction := nil;
  if cuFaction = nil then // CMN faction selection
  begin
    OthFaction := Gamesim.RegionCmns.GetHighestColStatus(RgnUID);
    cuStatus   := Gamesim.RegionCmns.GetHighestColStatusVal(RgnUID);
  end
  else
    cuStatus := cuFaction.Regions.Fetch(RgnUID, True).ColonialStatus;

  if Operator = '<' then
  begin
    if (cuStatus < Value) = False then Conditions := condFalse;
  end
  else
  if Operator = '<=' then
  begin
    if (cuStatus <= Value) = False then Conditions := condFalse;
  end
  else
  if Operator = '>' then
  begin
    if (cuStatus > Value) = False then Conditions := condFalse;
  end
  else
  if Operator = '>=' then
  begin
    if (cuStatus >= Value) = False then Conditions := condFalse;
  end
  else
  if Operator = '=' then
  begin
    if (cuStatus = Value) = False then Conditions := condFalse;
  end;

  if (cuFaction = nil) and (OthFaction <> nil) then
  Log(['EvalColStatus', RgnUID, RegionDefs[RgnUID].Name, 'Highest status found in the region:',
  ColStatusToString(cuStatus), 'from faction', OthFaction.UID, OthFaction.Name,
  'Test done:', ColStatusToString(cuStatus), Operator, Operator, ColStatusToString(Value), ':',
  IIF(Conditions = condEvaluating, 'True', 'False')])
  else
  if (cuFaction = nil)  then
  Log(['EvalColStatus', RgnUID, RegionDefs[RgnUID].Name, 'Highest status found in the region:',
  ColStatusToString(cuStatus), 'no faction found, Test done:',
  ColStatusToString(cuStatus), Operator, Operator, ColStatusToString(Value), ':',
  IIF(Conditions = condEvaluating, 'True', 'False')])
  else
  Log(['EvalColStatus', RgnUID, RegionDefs[RgnUID].Name, 'for', cuFaction.Name,
  'Test done:', ColStatusToString(cuStatus), Operator, ColStatusToString(Value), ':',
  IIF(Conditions = condEvaluating, 'True', 'False')]);

end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// EVAL COLONIAL STATUS AREA
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.EvalColStatusArea(Params : TStringList) : integer;
var Value : integer;
    AreaUID : TUID;
    Operator : string;
    CA : TColonialArea;
    cuStatus : TUID;
    OthFaction : TFaction;
begin
  Result := 0;
  if Params.Count <> 3 then
  begin
    Inc(Result);
    Log(['=> EvalColStatusArea command, syntax is: EvalColStatusArea = <ColAreaUID>;<Operator>;<Status>']);
  end;
  if (cuFaction = nil) then
  begin
    Log(['EvalColStatusArea: No faction, this will be an absolute test']);
  end;

  Operator := Params[0];
  AreaUID  := STI(Params[0], False, -1);
  Value    := STI(Params[2]);

  if (AreaUID < 0) or (AreaUID > Areas.Count-1) then
  begin
    Inc(Result);
    Log(['=> EvalColStatusArea: Area alias is not correct:', Params[0]]);
  end;
  CA := GameSim.ColModule.GetColAreaFromArea(AreaUID);
  if CA = nil then
  begin
    Inc(Result);
    Log(['=> EvalColStatusArea: The Area alias is not colonial area, see the file ColonialAreas.ini', Params[1]]);
  end;
  if Result > 0 then Exit;

  OthFaction := nil;
  if cuFaction = nil then
  begin
    OthFaction := Gamesim.RegionCmns.GetHighestColStatus(CA.CapitalUID);
    cuStatus   := Gamesim.RegionCmns.GetHighestColStatusVal(CA.CapitalUID);
  end
  else
    cuStatus := cuFaction.Regions.Fetch(CA.CapitalUID, True).ColonialStatus;

  if Operator = '<' then
  begin
    if (cuStatus < Value) = False then Conditions := condFalse;
  end
  else
  if Operator = '<=' then
  begin
    if (cuStatus <= Value) = False then Conditions := condFalse;
  end
  else
  if Operator = '>' then
  begin
    if (cuStatus > Value) = False then Conditions := condFalse;
  end
  else
  if Operator = '>=' then
  begin
    if (cuStatus >= Value) = False then Conditions := condFalse;
  end
  else
  if Operator = '=' then
  begin
    if (cuStatus = Value) = False then Conditions := condFalse;
  end;

  if (cuFaction = nil) and (OthFaction <> nil) then
  Log(['EvalColStatusArea', AreaUID, Areas[AreaUID].Name, 'Highest status found in the region:',
  ColStatusToString(cuStatus), 'from faction', OthFaction.UID, OthFaction.Name,
  'Test done:', ColStatusToString(cuStatus), Operator, Operator, ColStatusToString(Value), ':',
  IIF(Conditions = condEvaluating, 'True', 'False')])
  else
  if (cuFaction = nil)  then
  Log(['EvalColStatusArea', AreaUID, Areas[AreaUID].Name, 'Highest status found in the region:',
  ColStatusToString(cuStatus), 'no faction found, Test done:',
  ColStatusToString(cuStatus), Operator, Operator, ColStatusToString(Value), ':',
  IIF(Conditions = condEvaluating, 'True', 'False')])
  else
  Log(['EvalColStatusArea', AreaUID, Areas[AreaUID].Name, 'for', cuFaction.Name,
  'Test done:', ColStatusToString(cuStatus), Operator, ColStatusToString(Value), ':',
  IIF(Conditions = condEvaluating, 'True', 'False')]);

end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// EVAL SOI
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.EvalSOI(Params : TStringList) : integer;
var Value : integer;
    AreaUID : TUID;
    Operator : string;
    CA : TColonialArea;
    cuValue : integer;
begin
  Result := 0;
  if Params.Count <> 3 then
  begin
    Inc(Result);
    Log(['=> EvalSOI command, syntax is: EvalSOI = <ColAreaUID>;<Operator>;<Value>']);
  end;
  if cuFaction = nil then
  begin
    Log(['=> EvalSOI command, no faction selected']);
    Inc(Result);
  end;
  if Result > 0 then Exit;

  AreaUID  := STI(Params[0], False, -1);
  Operator := Params[1];
  Value    := STI(Params[2]);

  if (AreaUID < 0) or (AreaUID > Areas.Count-1) then
  begin
    Inc(Result);
    Log(['=> EvalSOI: Area alias is not correct:', Params[1]]);
  end;
  CA := GameSim.ColModule.GetColAreaFromArea(AreaUID);
  if CA = nil then
  begin
    Inc(Result);
    Log(['=> EvalSOI: The Area alias is not colonial area, see the file ColonialAreas.ini', Params[1]]);
  end;
  if Result > 0 then Exit;

  cuFaction.dlSOI.GetValByKey(AreaUID, cuValue, colSOIValDefault_); // direct read, not precalc, just to be sure

  if Operator = '<' then
  begin
    if (cuValue < Value) = False then Conditions := condFalse;
  end
  else
  if Operator = '<=' then
  begin
    if (cuValue <= Value) = False then Conditions := condFalse;
  end
  else
  if Operator = '>' then
  begin
    if (cuValue > Value) = False then Conditions := condFalse;
  end
  else
  if Operator = '>=' then
  begin
    if (cuValue >= Value) = False then Conditions := condFalse;
  end
  else
  if Operator = '=' then
  begin
    if (cuValue = Value) = False then Conditions := condFalse;
  end;

  Log(['EvalSOI', AreaUID, Areas[AreaUID].Name, 'for', cuFaction.Name,
  'Test done:', cuValue, Operator, Value, ':',
  IIF(Conditions = condEvaluating, 'True', 'False')]);

end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// EVAL FACTION RELATIONSHIPS
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.EvalFacRel(Params : TStringList) : integer;
var Value : integer;
    OtherFac : TFaction;
    Operator : string;
    cuValue : integer;
begin
  Result := 0;

  if (Params.Count <> 3) then
  begin
    Inc(Result);
    Log(['=> Evaluate Faction Relationships command, syntax is: EvalFacRelationships = <Faction_tag> ; <Operator (one of: < <= = >= >) >; <Value>']);
  end;
  if (cuFaction = nil) then
  begin
    Inc(Result);
    Log(['=> EvalFacRelationships: No faction selected']);
  end;
  OtherFac := FactionTagToFaction(Params[0]);
  if OtherFac <> nil then
  begin
    Log(['EvalFacRelationships: Evaluating versus', OtherFac.Name]);
  end
  else
  begin
    Log(['EvalFacRelationships: Faction not found', Params[0]]);
    Inc(Result);
  end;
  if Result > 0 then Exit;

  Operator := Params[1];

  Value := STI(Params[2]);

  cuValue := cuFaction.GetRelationship(OtherFac);

  if Operator = '<' then
  begin
    if (cuValue < Value) = False then Conditions := condFalse;
  end
  else
  if Operator = '<=' then
  begin
    if (cuValue <= Value) = False then Conditions := condFalse;
  end
  else
  if Operator = '>' then
  begin
    if (cuValue > Value) = False then Conditions := condFalse;
  end
  else
  if Operator = '>=' then
  begin
    if (cuValue >= Value) = False then Conditions := condFalse;
  end
  else
  if Operator = '=' then
  begin
    if (cuValue = Value) = False then Conditions := condFalse;
  end;

  Log(['EvalFacRelationships:', cuFaction.Name, 'relationships with:', OtherFac.Name, 'is', cuValue,
  'Evaluating', cuValue, Operator, Value, ':',
  IIF(Conditions = condEvaluating, 'True', 'False')]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// CHECK POWER RATIO
// CheckPowerRatio = Kind|Range|Ratio|Operator|FixedPercNoPerm|<FixedPercPerm>
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.CheckPowerRatio(Params : TStringList) : integer;
var sKind : string;
    Kind : TUID;
    Range  : integer;
    Ratio : integer;
    Operator : string;
    FixedPercNoPerm : integer;
    FixedPercPerm : integer;
    ilRgnList : TIntList;
    ValSelf, ValFriend,ValFoe : integer;
    cuRatio : integer;
begin
  Result := 0;
  if (not (Params.Count in [5,6])) then
  begin
    Inc(Result);
    Log(['=> Wrong nnmber of parameters, syntax is CheckPowerRatio = Kind|Range|Ratio|Operator|FixedPercNoPerm|<>FixedPercPerm>']);
  end;
  if (cuFaction = nil) then
  begin
    Inc(Result);
    Log(['=> CheckPowerRatio: No faction selected']);
  end;
  if (cuRegion = nil) then
  begin
    Inc(Result);
    Log(['=> CheckPowerRatio: No region selected']);
  end;

  sKind := UpperCase(Params[0]);
  Range := STI(Params[1]);
  Ratio := STI(Params[2]);
  Operator := Params[3];
  FixedPercNoPerm := STI(Params[4]);
  if Params.Count = 6 then
    FixedPercPerm := STI(Params[5])
  else
    FixedPercPerm := FixedPercNoPerm;

  if (sKind <> 'LAND') and (sKind <> 'NAV')  then
  begin
    Inc(Result);
    Log(['=> CheckPowerRatio: Unknown kind specified', sKind]);
  end;
  if Result > 0 then Exit;

  Kind := IIF(sKind = 'LAND', _gkLand, _gkNav);

  ilRgnList := TIntList.Create;
  RegionDefs.GetRegionsList(cuRegion.UID, ilRgnList, Range);

  ValSelf := GameSim.RegionCmns.GetFactionPowerInRgnList(ilRgnList, cuFaction.UID, Kind,
  FixedPercNoPerm, FixedPercNoPerm, FixedPercNoPerm, FixedPercNoPerm, FixedPercPerm);
  ValFriend := GameSim.RegionCmns.GetFriendlyPowerInRgnList(ilRgnList, cuFaction.UID, Kind,
  FixedPercNoPerm, FixedPercNoPerm, FixedPercNoPerm, FixedPercNoPerm, FixedPercPerm);
  ValFoe := GameSim.RegionCmns.GetEnemyPowerInRgnList(ilRgnList, cuFaction.UID, Kind,
  FixedPercNoPerm, FixedPercNoPerm, FixedPercNoPerm, FixedPercNoPerm, FixedPercPerm);

  // FriendlyPower includes Self, so we do a very complex math operation here:
  ValFriend := ValFriend - ValSelf;

  // our end 'self' value
  ValSelf := ValSelf + ValFriend div 2; // hard coded for now, Friend is 50% of Self

  ilRgnList.Free;

  cuRatio := Round(100 * ValSelf/ValFoe);

  if Operator = '<' then
  begin
    if (cuRatio < Ratio) = False then Conditions := condFalse;
  end
  else
  if Operator = '<=' then
  begin
    if (cuRatio <= Ratio) = False then Conditions := condFalse;
  end
  else
  if Operator = '>' then
  begin
    if (cuRatio > Ratio) = False then Conditions := condFalse;
  end
  else
  if Operator = '>=' then
  begin
    if (cuRatio >= Ratio) = False then Conditions := condFalse;
  end
  else
  if Operator = '=' then
  begin
    if (cuRatio = Ratio) = False then Conditions := condFalse;
  end;

  Log(['CheckPowerRatio', 'Around', cuRegion.UID, cuRegion.Name, 'Dist = ', Range, 'for', cuFaction.Name,
  'Test done:', cuRatio, Operator, Ratio, ':',
  IIF(Conditions = condEvaluating, 'True', 'False')]);

end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// EVAL AREA POWER RATIO
// EvalAreaPowerRatio = Kind|AreaUID|Ratio|Operator|FixedPercNoPerm|<>FixedPercPerm>
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.EvalAreaPwrRatio(Params : TStringList) : integer;
var sKind : string;
    Kind : TUID;
    AreaUID  : integer;
    Area : TArea;
    Ratio : integer;
    Operator : string;
    FixedPercNoPerm : integer;
    FixedPercPerm : integer;
    ValSelf, ValFriend,ValFoe : integer;
    cuRatio : integer;
begin
  Result := 0;
  if (not (Params.Count in [5,6])) then
  begin
    Inc(Result);
    Log(['=> Wrong number of parameters, syntax is EvalAreaPowerRatio = Kind|AreaUID|Ratio|Operator|FixedPercNoPerm|<FixedPercPerm>']);
  end;
  if (cuFaction = nil) then
  begin
    Inc(Result);
    Log(['=> EvalAreaPowerRatio: No faction selected']);
  end;

  sKind := UpperCase(Params[0]);
  AreaUID := STI(Params[1], True, -1);
  if (AreaUID < 0) or (AreaUID > Areas.Count) then
  begin
    Inc(Result);
    Log(['=> EvalAreaPowerRatio: Invalid Area UID', AreaUID]);
  end;
  if Result > 0 then Exit;
  Area := Areas[AreaUID];

  Ratio := STI(Params[2]);
  Operator := Params[3];
  FixedPercNoPerm := STI(Params[4]);
  if Params.Count = 6 then
    FixedPercPerm := STI(Params[5])
  else
    FixedPercPerm := FixedPercNoPerm;

  if (sKind <> 'LAND') and (sKind <> 'NAV')  then
  begin
    Inc(Result);
    Log(['=> EvalAreaPowerRatio: Unknown kind specified', sKind]);
  end;
  if Result > 0 then Exit;

  Kind := IIF(sKind = 'LAND', _gkLand, _gkNav);

  ValSelf := GameSim.RegionCmns.GetFactionPowerInRgnList(Area, cuFaction.UID, Kind,
  FixedPercNoPerm, FixedPercNoPerm, FixedPercNoPerm, FixedPercNoPerm, FixedPercPerm);
  ValFriend := GameSim.RegionCmns.GetFriendlyPowerInRgnList(Area, cuFaction.UID, Kind,
  FixedPercNoPerm, FixedPercNoPerm, FixedPercNoPerm, FixedPercNoPerm, FixedPercPerm);
  ValFoe := GameSim.RegionCmns.GetEnemyPowerInRgnList(Area, cuFaction.UID, Kind,
  FixedPercNoPerm, FixedPercNoPerm, FixedPercNoPerm, FixedPercNoPerm, FixedPercPerm);

  // FriendlyPower includes Self, so we do a very complex math operation here:
  ValFriend := ValFriend - ValSelf;

  // our end 'self' value
  ValSelf := ValSelf + ValFriend div 2; // hard coded for now, Friend is 50% of Self

  cuRatio := Round(100 * ValSelf/ValFoe);

  if Operator = '<' then
  begin
    if (cuRatio < Ratio) = False then Conditions := condFalse;
  end
  else
  if Operator = '<=' then
  begin
    if (cuRatio <= Ratio) = False then Conditions := condFalse;
  end
  else
  if Operator = '>' then
  begin
    if (cuRatio > Ratio) = False then Conditions := condFalse;
  end
  else
  if Operator = '>=' then
  begin
    if (cuRatio >= Ratio) = False then Conditions := condFalse;
  end
  else
  if Operator = '=' then
  begin
    if (cuRatio = Ratio) = False then Conditions := condFalse;
  end;

  Log(['EvalAreaPowerRatio', 'In', Area.UID, Area.Name, 'for', cuFaction.Name,
  'Test done:', cuRatio, Operator, Ratio, ':',
  IIF(Conditions = condEvaluating, 'True', 'False')]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// EVAL NATIONAL THREATS
// A command that has a variable probability to return true, depending of the test
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.EvalNatThreats(Params : TStringList) : integer;
var
    ChanceAtWarThreat, ChanceAtWar, ChanceAtPeaceThreat, ChanceAtPeace : integer;
    IsAtWarThreat, IsAtWar, IsAtPeaceThreat, IsAtPeace : boolean;
begin
  Result := 0;
  if (not (Params.Count in [4])) then
  begin
    Inc(Result);
    Log(['=> Wrong number of parameters, syntax is EvalNatThreats = ChanceOKAtWarThreat|ChanceOKAtWar|ChanceOKAtPeaceThreat|ChanceOKAtPeace']);
  end;
  if (cuFaction = nil) then
  begin
    Inc(Result);
    Log(['=> EvalNatThreats: No faction selected']);
  end;

  IsAtWarThreat := cuFaction.IsAtWarThreatened;
  IsAtPeaceThreat := cuFaction.IsAtPeaceThreatened;
  IsAtWar := cuFaction.IsAtWarNotThreatened;
  IsAtPeace := cuFaction.IsAtPeaceNotThreatened;

  ChanceAtWarThreat   := STI(Params[0]);
  ChanceAtWar         := STI(Params[1]);
  ChanceAtPeaceThreat := STI(Params[2]);
  ChanceAtPeace       := STI(Params[3]);

  if IsAtWarThreat and (Dice(100) > ChanceAtWarThreat) then Conditions := condFalse;
  if IsAtPeaceThreat and (Dice(100) > ChanceAtPeaceThreat) then Conditions := condFalse;
  if IsAtWar and (Dice(100) > ChanceAtWar) then Conditions := condFalse;
  if IsAtPeace and (Dice(100) > ChanceAtPeace) then Conditions := condFalse;
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// EVALUATE STRUCTURE PRODUCTION
// only the structure production, excluding  WSU, money from
// transactions, markets, inate income, regional production of conscripts, supply...
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.EvalStrucProduction(Params : TStringList) : integer;
var ObjectUID : integer;
    Operator : string;
    Value : integer;
    cuValue : integer;
begin
  Result := 0;

  if (Params.Count <> 3) then
  begin
    Inc(Result);
    Log(['=> Evaluate Structure Production command, syntax is: EvalStrucProduction = <$merAlias> ; <Operator (one of: < <= = >= >) >; <Value>']);
  end;
  if (cuFaction = nil) then
  begin
    Inc(Result);
    Log(['=> EvalStrucProduction: No faction selected']);
  end;
  if Result > 0 then Exit;

  ObjectUID := STI(Params[0], True, -1);
  Operator  := Params[1];
  Value     := STI(Params[2]);

  if (ObjectUID < 0) or (ObjectUID > GameSim.MerchandisesDB.Count-1) then
  begin
    Inc(Result);
    Log(['=> EvalStrucProduction: Unknown merchandise', ObjectUID]);
  end;
  if Result > 0 then Exit;

  cuFaction.CalcStructuresProduction(cuFaction.olERIProd);
  cuValue := cuFaction.FetchResourceStrucProduction(ObjectUID);

  if Operator = '<' then
  begin
    if (cuValue < Value) = False then Conditions := condFalse;
  end
  else
  if Operator = '<=' then
  begin
    if (cuValue <= Value) = False then Conditions := condFalse;
  end
  else
  if Operator = '>' then
  begin
    if (cuValue > Value) = False then Conditions := condFalse;
  end
  else
  if Operator = '>=' then
  begin
    if (cuValue >= Value) = False then Conditions := condFalse;
  end
  else
  if Operator = '=' then
  begin
    if (cuValue = Value) = False then Conditions := condFalse;
  end;

  // Jan 2011: Registry
  if ActiveReg > 0 then
  begin
    Inc(Registry[ActiveReg], cuValue);
    Log(['EvalStrucProduction', 'Registry', ActiveReg, 'used, adding',  cuValue, 'to it, for an end value of', Registry[ActiveReg]]);
  end;

  Log(['EvalStrucProduction:', cuFaction.Name, 'production of', GameSim.MerchandisesDB[ObjectUID].Name, 'is', cuValue,
  'Evaluating', cuValue, Operator, Value, ':',
  IIF(Conditions = condEvaluating, 'True', 'False')]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// EVAL STRUCTURE COUNT
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.EvalStrucCount(Params : TStringList) : integer;
type
  TStructureSearchType = (sstAny, sstKind, sstSubKind, sstFamily, sstAttribute, sstDef);
var AreaUID   : TUID;
    Param     : string;
    i,aea     : integer;
    SST       : TStructureSearchType;
    Struc     : TStructure;
    Operator  : string;
    ValueComp : integer;
    cuValue   : integer;
    FacAll    : boolean;
    Area      : TArea;
    RgnUID    : TUID;
    AreaName  : string;
begin
  Result := 0;
  if (Params.Count <> 5) then
  begin
    Inc(Result);
    Log(['=> EvalStrucCount command, syntax is: EvalStrucCount = AreaUID_or_MAP; <FactionValue>; <StructureSearchCriteria (one of STRUCKIND, STRUCDEF, STRUCSUB, STRUCFAM, STRUCATR) value > <Operator> <ValueToCompareWith>']);
  end;

  AreaUID := _noUID;
  if Params[0] <> 'MAP' then
  begin
    if IsNumeric(Params[0]) then
      AreaUID := STI(Params[0], True, -1)
    else
    begin
      Inc(Result);
      Log(['=> EvalStrucCount: Param0 is not a known Area (check if the alias is valid):', Params[0]]);
    end;
  end;
  if Result <> 0 then Exit;

  if AreaUID <> -1 then
  begin
    Area := Areas[AreaUID];
    AreaName := Area.Name;
  end
  else
  begin
    Area := nil;
    AreaName := 'MAP';
  end;

  FacAll := Params[1] = 'FACALL';                   // Note: extractparams can not discriminate the ALL of FACALL and STRUCALL, so I use STRUCANY for the 3rd param

  if not FacAll then
    cuFaction := FactionTagToFaction(Params[1]);
  cuValue := 0;

  //  STRUCANY, STRUCKIND, STRUCDEF
  SST := sstAny;
  Param := ExtractParam('STRUCKIND', Params);
  if Param <> '' then
  begin
    SST := sstKind;
    ExtractArgs(Param);
  end;
  Param := ExtractParam('STRUCDEF', Params);
  if Param <> '' then
  begin
    SST := sstDef;
    ExtractArgs(Param);
  end;
  Param := ExtractParam('STRUCSUB', Params);
  if Param <> '' then
  begin
    SST := sstSubKind;
    ExtractArgs(Param);
  end;
  Param := ExtractParam('STRUCFAM', Params);
  if Param <> '' then
  begin
    SST := sstFamily;
    ExtractArgs(Param);
  end;
  Param := ExtractParam('STRUCATR', Params);
  if Param <> '' then
  begin
    SST := sstAttribute;
    ExtractArgs(Param);
  end;

  Operator := Params[3];
  ValueComp := STI(Params[4]);

  if Params[0] <> 'MAP' then
  begin
    if FacAll then
    begin
      for aea := 0 to Area.Count-1 do
      begin
        RgnUID := Area[aea];
        for i := 0 to GameSim.RegionCmns[RgnUID].pListStruc.Count-1 do
        begin
          Struc := GameSim.RegionCmns[RgnUID].pListStruc[i];
          if SST = sstAny then Inc(cuValue);
          if (SST = sstDef)  and (Struc.DefUID  = STI(Args[0])) then Inc(cuValue);
          if (SST = sstKind) and (Struc.GetKind = STI(Args[0])) then Inc(cuValue);
          if (SST = sstSubKind) and (Struc.StrucDef.SubKind = STI(Args[0])) then Inc(cuValue);
          if (SST = sstFamily) and (Struc.StrucDef.Family = STI(Args[0])) then Inc(cuValue);
          if (SST = sstAttribute) and (Struc.StrucDef.HasThisAttribute(Args[0])) then Inc(cuValue);
        end;
      end;
    end
    else
    begin
      for aea := 0 to Area.Count-1 do
      begin
        RgnUID := Area[aea];
        cuRegion := cuFaction.Regions.Fetch(RgnUID);
        for i := 0 to cuRegion.Structures.Count-1 do
        begin
          Struc := GameSim.RegionCmns[RgnUID].pListStruc[i];
          if SST = sstAny then Inc(cuValue);
          if (SST = sstDef)  and (Struc.DefUID  = STI(Args[0])) then Inc(cuValue);
          if (SST = sstKind) and (Struc.GetKind = STI(Args[0])) then Inc(cuValue);
          if (SST = sstSubKind) and (Struc.StrucDef.SubKind = STI(Args[0])) then Inc(cuValue);
          if (SST = sstFamily) and (Struc.StrucDef.Family = STI(Args[0])) then Inc(cuValue);
          if (SST = sstAttribute) and (Struc.StrucDef.HasThisAttribute(Args[0])) then Inc(cuValue);
        end;
      end;
    end;
  end
  else // MAP
  begin
    if FacAll then
    begin
      for aea := 0 to RegionDefs.Count-1 do
      begin
        RgnUID := RegionDefs[aea].UID;
        for i := 0 to GameSim.RegionCmns[RgnUID].pListStruc.Count-1 do
        begin
          Struc := GameSim.RegionCmns[RgnUID].pListStruc[i];
          if SST = sstAny then Inc(cuValue);
          if (SST = sstDef)  and (Struc.DefUID  = STI(Args[0])) then Inc(cuValue);
          if (SST = sstKind) and (Struc.GetKind = STI(Args[0])) then Inc(cuValue);
          if (SST = sstSubKind) and (Struc.StrucDef.SubKind = STI(Args[0])) then Inc(cuValue);
          if (SST = sstFamily) and (Struc.StrucDef.Family = STI(Args[0])) then Inc(cuValue);
          if (SST = sstAttribute) and (Struc.StrucDef.HasThisAttribute(Args[0])) then Inc(cuValue);
        end;
      end;
    end
    else
    begin
      for aea := 0 to RegionDefs.Count-1 do
      begin
        RgnUID := RegionDefs[aea].UID;
        cuRegion := cuFaction.Regions.Fetch(RgnUID);
        for i := 0 to cuRegion.Structures.Count-1 do
        begin
          Struc := GameSim.RegionCmns[RgnUID].pListStruc[i];
          if SST = sstAny then Inc(cuValue);
          if (SST = sstDef)  and (Struc.DefUID  = STI(Args[0])) then Inc(cuValue);
          if (SST = sstKind) and (Struc.GetKind = STI(Args[0])) then Inc(cuValue);
          if (SST = sstSubKind) and (Struc.StrucDef.SubKind = STI(Args[0])) then Inc(cuValue);
          if (SST = sstFamily) and (Struc.StrucDef.Family = STI(Args[0])) then Inc(cuValue);
          if (SST = sstAttribute) and (Struc.StrucDef.HasThisAttribute(Args[0])) then Inc(cuValue);
        end;
      end;
    end;
  end;

  if Operator = '<' then
  begin
    if (cuValue < ValueComp) = False then Conditions := condFalse;
  end
  else
  if Operator = '<=' then
  begin
    if (cuValue <= ValueComp) = False then Conditions := condFalse;
  end
  else
  if Operator = '>' then
  begin
    if (cuValue > ValueComp) = False then Conditions := condFalse;
  end
  else
  if Operator = '>=' then
  begin
    if (cuValue >= ValueComp) = False then Conditions := condFalse;
  end
  else
  if Operator = '=' then
  begin
    if (cuValue = ValueComp) = False then Conditions := condFalse;
  end;

  // Jan 2011: Registry
  if ActiveReg > 0 then
  begin
    Inc(Registry[ActiveReg], cuValue);
    Log(['EvalStrucCount', 'Registry', ActiveReg, 'used, adding',  cuValue, 'to it, for an end value of', Registry[ActiveReg]]);
  end;

  Log(['EvalStrucCount:', IIF(FacAll, 'FAC ALL', cuFaction.Name), 'Count of structures in area', AreaName, 'found', cuValue,
  'Evaluating', cuValue, Operator, ValueComp, ':',
  IIF(Conditions = condEvaluating, 'True', 'False')]);

end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// EVAL HAS CAPITAL
// Evaluate if the country controls his capital
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.EvalHasCapital(Params : TStringList) : integer;
var Reverse : boolean;
begin
  Result := 0;
  if Params.Count > 1 then
  begin
    Inc(Result);
    Log(['=> EvalHasCapital command, syntax is: EvalHasCapital = [NOT]']);
  end;
  if (cuFaction = nil) then
  begin
    Log(['EvalHasCapital: No faction, this will be an absolute test']);
  end;
  if Result > 0 then Exit;

  Reverse := ExtractParam('NOT', Params) <> '';
  if cuFaction.GetCapital(True, True) = nil then
  begin
    if not Reverse then Conditions := condFalse;
  end
  else
  begin
    if Reverse then Conditions := condFalse;
  end;

  // Registry
  if (Conditions = condEvaluating) and (ActiveReg > 0) then
  begin
    Inc(Registry[ActiveReg], 1);
    Log(['EvalHasCapital', 'Registry', ActiveReg, 'used, adding',  1, 'to it, for an end value of', Registry[ActiveReg]]);
  end;

  Log(['EvalHasCapital:', cuFaction.Name, IIF(Reverse, 'Has not her capital under control', 'Has her capital under control'),
  IIF(Conditions = condEvaluating, 'True', 'False')]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function  TScriptEngine.EvalInflation(Params : TStringList) : integer;
var Value : integer;
begin
  Result := 0;
  if Params.Count <> 2 then
  begin
    Inc(Result);
    Log(['=> Evaluate Inflation command, syntax is: EvalInflation = <Operator>;<Value>']);
  end;
  if (cuFaction = nil) then
  begin
    Inc(Result);
    Log(['=> EvalInflation: No faction selected']);
  end;
  if Result > 0 then Exit;

  Value := STI(Params[1]);

  if Params[0] = '<' then
  begin
    if not (cuFaction.Inflation < Value) then Conditions := condFalse;
  end
  else
  if Params[0] = '<=' then
  begin
    if not (cuFaction.Inflation <= Value) then Conditions := condFalse;
  end
  else
  if Params[0] = '>' then
  begin
    if not (cuFaction.Inflation > Value) then Conditions := condFalse;
  end
  else
  if Params[0] = '>=' then
  begin
    if not (cuFaction.Inflation >= Value) then Conditions := condFalse;
  end
  else
  if Params[0] = '=' then
  begin
    if not (cuFaction.Inflation = Value) then Conditions := condFalse;
  end
  else
  if UpperCase(Params[0]) = 'DICE' then
  begin
    if (Dice(Value) > cuFaction.Inflation) then Conditions := condFalse;
  end;

  Log(['EvalInflation:', cuFaction.Name, 'Test done:', cuFaction.Inflation, Params[0], Params[1], ':',
  IIF(Conditions = condEvaluating, 'True', 'False')]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.EvalColPen(Params : TStringList) : integer;
var Value : integer;
    Operator : string;
    CA : TColonialArea;
    cuValue : integer;
    RgnUID : TUID;
    ExcludeFaction : TFaction;
begin
  Result := 0;
  if not (Params.Count in [3,4]) then
  begin
    Inc(Result);
    Log(['=> EvalColPen command, syntax is: EvalColPen = <RgnUID>;<Operator>;<Status>[;<ExcludeTag>]']);
  end;
  if (cuFaction = nil) then
  begin
    Log(['EvalColPen: No faction, this will be a global test']);
  end;

  Operator := Params[0];
  RgnUID  := STI(Params[0], False, -1);
  Value    := STI(Params[2]);
  if Params.Count = 4 then
    ExcludeFaction := FactionTagToFaction(Params[3])
  else
    ExcludeFaction := nil;

  if (RgnUID < 0) or (RgnUID > RegionDefs.Count-1) then
  begin
    Inc(Result);
    Log(['=> EvalColPen: Region alias is not correct:', Params[0]]);
    Exit;
  end;
  CA := GameSim.ColModule.GetColAreaFromRegion(RgnUID);
  if CA = nil then
  begin
    Log(['[Warning] EvalColPen: The region is not in a colonial area', RgnUID]);
    Exit;
  end;
  if Result > 0 then Exit;

  if cuFaction = nil then // CMN faction selection - get highest col pen
  begin
    cuValue   := Gamesim.RegionCmns.GetHighestColPenVal(RgnUID, ExcludeFaction);
  end
  else
    cuValue := cuFaction.Regions.Fetch(RgnUID, True).ColonialPen;

  if Operator = '<' then
  begin
    if (cuValue < Value) = False then Conditions := condFalse;
  end
  else
  if Operator = '<=' then
  begin
    if (cuValue <= Value) = False then Conditions := condFalse;
  end
  else
  if Operator = '>' then
  begin
    if (cuValue > Value) = False then Conditions := condFalse;
  end
  else
  if Operator = '>=' then
  begin
    if (cuValue >= Value) = False then Conditions := condFalse;
  end
  else
  if Operator = '=' then
  begin
    if (cuValue = Value) = False then Conditions := condFalse;
  end;

  if (cuFaction = nil) and (ExcludeFaction <> nil) then
  Log(['EvalColPen', RgnUID, RegionDefs[RgnUID].Name, 'Highest CP in the region:',
  cuValue, 'excluding faction', ExcludeFaction.UID, ExcludeFaction.Name,
  'Test done:', cuValue, Operator, Value, ':',
  IIF(Conditions = condEvaluating, 'True', 'False')])
  else
  if (cuFaction = nil)  then
  Log(['EvalColPen', RgnUID, RegionDefs[RgnUID].Name, 'Highest CP found in the region:',
  cuValue, 'no faction excluded, Test done:',
  'Test done:', cuValue, Operator, Value, ':',
  IIF(Conditions = condEvaluating, 'True', 'False')])
  else
  Log(['EvalColPen', RgnUID, RegionDefs[RgnUID].Name, 'for', cuFaction.Name,
  'Test done:', cuValue, Operator, Value, ':',
  IIF(Conditions = condEvaluating, 'True', 'False')]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// EVAL REGION OWNER KIND
// EvalRgnOwnerKind = $fkTribalNation
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.EvalRgnOwnerKind  (Params : TStringList) : integer;
var
    Kind : TUID;
    ReverseResult : boolean;
begin
  Result := 0;
  if (Params.Count < 1) or (Params.Count > 2) then
  begin
    Inc(Result);
    Log(['=> EvalRgnOwnerKind command, syntax is: EvalRgnOwnerKind = <FactionKind> [;<NOT>]']);
  end;
  if (cuFaction = nil) then
  begin
    Inc(Result);
    Log(['=> EvalRgnOwnerKind: No faction selected']);
  end;
  if (cuRegion = nil) then
  begin
    Inc(Result);
    Log(['=> EvalRgnOwnerKind: No region selected']);
  end;

  if Result > 0 then Exit;

  Kind := STI(Params[0]);
  if not (Kind in [_fkOrganizedNation.._fkTribalNation]) then begin Inc(Result); Exit; end;

  ReverseResult := (Params.Count = 2) and (UpperCase(Params[1]) = 'NOT');


  if (GameSim.RegionCmns[cuRegion.UID].OwningFac <> nil) and (GameSim.RegionCmns[cuRegion.UID].OwningFac.Kind = Kind) then
    DoNothing
  else
    Conditions := condFalse;

  if ReverseResult then
  begin
    if Conditions = condEvaluating then
      Conditions := condFalse
    else
      Conditions := condEvaluating;
  end;

  if not ReverseResult then
    Log(['EvalRgnOwnerKind:', cuRegion.UID, cuRegion.Name, 'owned by a faction kind of value', Kind, IIF(Conditions = condEvaluating, 'True', 'False')])
  else
    Log(['EvalRgnOwnerKind:', cuRegion.UID, cuRegion.Name, 'NOT owned by a faction kind of value', Kind, IIF(Conditions = condEvaluating, 'True', 'False')])
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// CHANGE STRUCTURE POOL
// Structures are added by DefUID but the construction limits are handled by family
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.ChangeStrucPool(Params : TStringList) : integer;
var StrucDefUID, Value : integer;
begin
  Result := 0;
  if Params.Count <> 2 then
  begin
    Inc(Result);
    Log(['=> ChangeStrucPool, need to have exactly two parameters: Alias of StrucDef and variation of Force Pool']);
    Exit;
  end;
  if cuFaction = nil then
  begin
    Log(['=> ChangeStrucPool command, no faction selected']);
    Inc(Result);
    Exit;
  end;
  StrucDefUID := STI(Params[0], True, -1);

  if StrucDefUID = -1 then
  if cuFaction = nil then
  begin
    Log(['=> ChangeStrucPool command,', Params[0], 'is an unknown unit']);
    Inc(Result);
    Exit;
  end;

  Value      := STI(Params[1]);
  if (StrucDefUID < 0) or (StrucDefUID > StrucDefs.Count-1) then
  begin
    Log(['=> ChangeStrucPool command, the StrucDef UID given is either negative or above the structures definitions UID. ', StrucDefUID]);
    Inc(Result);
    Exit;
  end;
  cuFaction.ChangeStrucMaxForcePool(StrucDefUID, Value);
  Log(['ChangeStrucPool', cuFaction.Name, 'New Force Pool for structure (UID/Name) is', StrucDefUID, StrucDefs[StrucDefUID].Name,  cuFaction.GetStrucMaxForcePool(StrucDefUID)]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// CHANGE STRUCTURE TYPE
// Convert a structure DefUID into another, both on map and in the force pool (optionnally)
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.ChangeStrucType(Params : TStringList) : integer;
var OldUID    : TUID;
    NewUID    : TUID;
    FP        : boolean;
    AreaUID   : TUID;
    Param     : string;
    Struc     : TStructure;
    stc,i     : integer;
    OpDone    : boolean;
    PrevName  : string;
    lsRgnList : TList;
    Region    : TRegion;
begin
  Result := 0;
  if (Params.Count < 2)or (Params.Count > 4) then
  begin
    Inc(Result);
    Log(['=> ChangeStrucType, need to have from 2 to 4 parameters: ChangeStrucType = <OldStrucDef>;<NewStrucDef>[;FP][;REGorAREA_AreaUID]']);
    Exit;
  end;
  if cuFaction = nil then
  begin
    Log(['=> ChangeStrucType command, no faction selected']);
    Inc(Result);
    Exit;
  end;

  // If nothing specified, its on map. If REG, then we expect a cuRegion, if AREA, then there is an area alias.
  Param := ExtractParam('REG', Params);
  if Param <> '' then
    AreaUID := -2
  else
    AreaUID := _noUID;
  Param := ExtractParam('AREA', Params);
  if Param <> '' then
  begin
    ExtractArgs(Param);
    AreaUID := STI(Args[0]);
  end;
  // Region Checking
  if (AreaUID = -2)and(cuRegion = nil) then
  begin
    Inc(Result);
    Log(['=> ChangeStrucType command, no region selected.']);
  end;

  OldUID      := STI(Params[0], True, Low(integer));
  NewUID      := STI(Params[1], True, Low(integer));

  FP  := ExtractParam('FP', Params) <> '';

  if (OldUID < 0) or (OldUID > StrucDefs.Count-1) then
  begin
    Log(['=> ChangeStrucType command, the Old StrucDef UID given is either negative or above the structures definitions UID. ', OldUID]);
    Inc(Result);
    Exit;
  end;
  if (NewUID < 0) or (NewUID > StrucDefs.Count-1) then
  begin
    Log(['=> ChangeStrucType command, the New StrucDef UID given is either negative or above the structures definitions UID. ', NewUID]);
    Inc(Result);
    Exit;
  end;

  // Creating the region list
  lsRgnList := TList.Create;
  if AreaUID = -2 then lsRgnList.add(cuRegion);
  if AreaUID = _noUID then
  begin
    for i := 0 to cuFaction.Regions.Count-1 do
      lsRgnList.add(cuFaction.Regions[i]);
  end;
  if AreaUID >= 0 then
  begin
    for i := 0 to Areas[AreaUID].Count-1 do
    begin
      if cuFaction.Regions.Fetch(i, False) <> nil then
        lsRgnList.Add(cuFaction.Regions.Fetch(i));
    end;
  end;

  for i := 0 to lsRgnList.Count-1 do
  begin
    OpDone := False;
    Region := lsRgnList[i];
    for stc := 0 to Region.Structures.Count-1 do
    begin
      Struc := Region.Structures[stc];
      if Struc.DefUID = OldUID then
      begin
        OpDone := True;
        PrevName := Struc.Name;
        Struc.DefUID := NewUID;
        // If not a custom name, it is updated
        if PrevName = StrucDefs[OldUID].Name then
          Struc.Name := StrucDefs[NewUID].Name;
        Log(['ChangeStrucType command', 'In Region', Region.UID, Region.Name, 'Structure:', PrevName, 'of type', OldUID, 'converted to type', NewUID, 'Database Name:', StrucDefs[NewUID].Name]);
      end;
    end;
    if OpDone then
      GameSim.RegionCmns.Update(Region.UID);
  end;

  // Force Pool changes (as a whole)
  if FP then
  begin
    if cuFaction.ConvertStrucInMaxForcePool(OldUID, NewUID) then
      Log(['ChangeStrucType command', 'Force Pool conversion of', OldUID, StrucDefs[OldUID].Name, 'to', NewUID, StrucDefs[NewUID].Name]);
  end;

  lsRgnList.Free;
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// CHANGE STRUC LEVEL
// Whatever the faction(s) possessing it. Applied to all structures having the right criteria
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.ChangeStrucLevel(Params : TStringList) : integer;
var
    Param     : string;
    Struc     : TStructure;
    i         : integer;
    iArg      : integer;
    Arg       : string;
    ByDefUID  : boolean;
    ByAttrib  : boolean;
    ByKind    : boolean;
    LevelChg  : integer;
begin
  Result := 0;
  if (Params.Count <> 2) then
  begin
    Inc(Result);
    Log(['=> ChangeStrucLevel, need to have 2 parameters: ChangeStrucLevel = <SelectionType Value>;<LevelChange>']);
    Exit;
  end;
  if cuFaction = nil then
  begin
    Log(['=> ChangeStrucLevel command, no faction selected']);
    Inc(Result);
    Exit;
  end;
  if cuRegion = nil then
  begin
    Log(['=> ChangeStrucLevel command, no region selected']);
    Inc(Result);
    Exit;
  end;

  LevelChg := STI(Params[1], True, -100);
  if LevelChg <= -100 then
  begin
    Log(['=> ChangeStrucLevel command, Level Change is not valid.']);
    Inc(Result);
    Exit;
  end;

  // Selection Type
  iArg := 0;
  ByDefUID := False;
  Param :=  ExtractParam('ByDefUID', Params);
  if Param <> '' then
  begin
    ByDefUID := True;
    ExtractArgs(Param);
    iArg := STI(Args[0]);
  end;
  ByAttrib := False;
  Param :=  ExtractParam('ByAttrib', Params);
  if Param <> '' then
  begin
    ByAttrib := True;
    ExtractArgs(Param);
    Arg := Args[0];
  end;
  ByKind := False;
  Param :=  ExtractParam('ByKind', Params);
  if Param <> '' then
  begin
    ByKind := True;
    ExtractArgs(Param);
    iArg := STI(Args[0]);
  end;

  if (not ByDefUID) and (not ByAttrib) and (not ByKind) then
  begin
    Log(['=> ChangeStrucLevel command, no selection type selected']);
    Inc(Result);
    Exit;
  end;

  for i := 0 to GameSim.RegionCmns[cuRegion.UID].pListStruc.Count-1 do
  begin
    Struc := GameSim.RegionCmns[cuRegion.UID].pListStruc[i];
    if ByDefUID and (Struc.DefUID = iArg) then
    begin
      Struc.Level := Struc.Level + LevelChg;
      Log(['ChangeStrucLevel command, structure:', Struc.UID, Struc.Name, 'new level is:', Struc.Level]);
    end
    else
    if ByAttrib and (Struc.HasThisAttribute(Arg)) then
    begin
      Struc.Level := Struc.Level + LevelChg;
      Log(['ChangeStrucLevel command, structure:', Struc.UID, Struc.Name, 'new level is:', Struc.Level]);
    end
    else
    if ByKind and (Struc.StrucDef.Kind = iArg) then
    begin
      Struc.Level := Struc.Level + LevelChg;
      Log(['ChangeStrucLevel command, structure:', Struc.UID, Struc.Name, 'new level is:', Struc.Level]);
    end;
  end;
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// MOVE STRUCTURE
// ex: EvalRgnStruc = $Paris|FACALL|STRUCATR $attIsSewer
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.MoveStruc(Params : TStringList) : integer;
var
    Param     : string;
    i         : integer;
    SST       : TStructureSearchType;
    Struc     : TStructure;
    Found     : boolean;
    FS        : TFactionStance;
    pList     : TList;
    DstUID    : TUID;
    slSub     : TStringList;
begin
  Result := 0;
  if not (Params.Count in [3]) then
  begin
    Inc(Result);
    Log(['=> MoveStruc command, syntax is: MoveStruc = <DestArea>; <FactionStance>; <StructureSearchType (one of STRUCKIND, STRUCDEF, STRUCSUB, STRUCFAM, STRUCATR) value >']);
  end;
  if cuFaction = nil then
  begin
    Log(['=> MoveStruc command, no faction selected']);
    Inc(Result);
  end;
  if cuRegion = nil then
  begin
    Log(['=> MoveStruc command, no region selected, will be used as the source region']);
    Inc(Result);
  end;

  slSub := TStringList.Create;
  slSub.Add(Params[0] + ';Squared;OwnedCities');
  cuRegList.Clear;
  SelectRegInArea(slSub, False, True);
  slSub.Free;

  if cuRegList.Count = 0 then
  begin
    Inc(Result);
    Log(['=> MoveStruc, No region added in Current Regions List can be used as destinations regions']);
    Exit;
  end;

  pList := TList.Create;

  FS := fasSelf;
  if ExtractParam('StanceSelf', Params) <> '' then DoNothing else
  if ExtractParam('StanceFriend', Params) <> '' then FS := fasFriend else
  if ExtractParam('StanceEnemy', Params) <> '' then FS := fasFoe else
  if ExtractParam('StanceAll', Params) <> '' then FS := fasAny;

  //  STRUCANY, STRUCKIND, STRUCDEF
  SST := sstAny;
  Param := ExtractParam('STRUCKIND', Params);
  if Param <> '' then
  begin
    SST := sstKind;
    ExtractArgs(Param);
  end;
  Param := ExtractParam('STRUCDEF', Params);
  if Param <> '' then
  begin
    SST := sstDef;
    ExtractArgs(Param);
  end;
  Param := ExtractParam('STRUCSUB', Params);
  if Param <> '' then
  begin
    SST := sstSubKind;
    ExtractArgs(Param);
  end;
  Param := ExtractParam('STRUCFAM', Params);
  if Param <> '' then
  begin
    SST := sstFamily;
    ExtractArgs(Param);
  end;
  Param := ExtractParam('STRUCATR', Params);
  if Param <> '' then
  begin
    SST := sstAttribute;
    ExtractArgs(Param);
  end;

  // Find the structures to move
  for i := 0 to GameSim.RegionCmns[cuRegion.UID].pListStruc.Count-1 do
  begin
    Struc := GameSim.RegionCmns[cuRegion.UID].pListStruc[i];

    Found := False;

    if (FS = fasAny) then
      DoNothing
    else
    if (FS = fasSelf)and(Struc.pFaction <> cuFaction) then Continue
    else
    if (FS = fasFoe)and(not TFaction(Struc.pFaction).IsEnemyOf(cuFaction)) then Continue
    else
    if (FS = fasFriend)and(not TFaction(Struc.pFaction).IsFriendOf(cuFaction)) then Continue // include ourself
    else
    if (FS = fasNeutral)and(not TFaction(Struc.pFaction).IsNeutralWith(cuFaction)) then Continue;


    if SST = sstAny then Found := True else
    if (SST = sstDef)  and (Struc.DefUID  = STI(Args[0])) then Found := True else
    if (SST = sstKind) and (Struc.GetKind = STI(Args[0])) then Found := True else
    if (SST = sstSubKind) and (Struc.StrucDef.SubKind = STI(Args[0])) then Found := True else
    if (SST = sstFamily) and (Struc.StrucDef.Family = STI(Args[0])) then Found := True else
    if (SST = sstAttribute) and (Struc.StrucDef.HasThisAttribute(Args[0])) then Found := True;

    if Found then
      pList.Add(Struc);
  end;

  if pList.Count = 0 then
  begin
    Log(['[WARNING] MoveStruc, No structure selected in region', cuRegion.UID, cuRegion.Name]);
    pList.Free;
    Exit;
  end
  else
  begin

    repeat // get a region, knowing that a SelectRegion after a AddToRegList would have added the current region to the list
    // something we don't want, as we move structures 
      DstUID := cuRegList[Dice(cuRegList.Count)-1];
    until (DstUID <> cuRegion.UID) or (cuRegList.Count < 2);

    Log(['MoveStruc, structures', pList.Count, 'selected in region', cuRegion.UID, cuRegion.Name, 'moving them to', DstUID, RegionDefs[DstUID].Name]);
    for i := 0 to pList.Count-1 do
    begin
      Struc := pList[i];
      TFaction(Struc.pFaction).Regions.Fetch(Struc.RgnUID).MoveStructureTo(Struc, DstUID);
    end;

  end;


  pList.Free;
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// CHANGE FORT BREACH
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.ChangeFortBreach(Params : TStringList) : integer;
var
    Struc     : TStructure;
    i         : integer;
    Breaches  : integer;
begin
  Result := 0;
  if (Params.Count <> 1) then
  begin
    Inc(Result);
    Log(['=> ChangeFortBreach, need to have 1 parameter: ChangeFortBreach = <Amount>']);
    Exit;
  end;
  if cuFaction = nil then
  begin
    Log(['=> ChangeFortBreach command, no faction selected']);
    Inc(Result);
    Exit;
  end;
  if cuRegion = nil then
  begin
    Log(['=> ChangeFortBreach command, no region selected']);
    Inc(Result);
    Exit;
  end;

  Breaches := STI(Params[0], True, -100);
  if Breaches <= -100 then
  begin
    Log(['=> ChangeFortBreach command, Breach Change is not valid.']);
    Inc(Result);
    Exit;
  end;

  for i := 0 to GameSim.RegionCmns[cuRegion.UID].pListStruc.Count-1 do
  begin
    Struc := GameSim.RegionCmns[cuRegion.UID].pListStruc[i];
    if Struc.IsFort then
    begin
      Struc.Damages := Struc.Damages + Breaches;
      Log(['ChangeFortBreach command, structure:', Struc.UID, Struc.Name, 'breaches are now at:', Struc.Damages]);
    end;
  end;
end;


// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET STRUCTURE FAMILY CUSTOM NAMES
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetStrucCustNames (Params : TStringList) : integer;
var StrucFamily : integer;
    Value   : string;
    i       : integer;
    Faction : TFaction;
begin
  Result := 0;

  if not (Params.Count > 1) then
  begin
    Inc(Result);
    Log(['=> SetStructureCustomNames: Syntax is SetStructureCustomNames = <sfmFamily>;<Name1>;<Name2>;<Name3>;..']);
  end;

  if (cuFaction = nil) then
  begin
    Inc(Result);
    Log(['=> SetStructureCustomNames: No faction selected']);
  end;

  if Result <> 0 then Exit;

  // get the global data for the operation
  Faction := cuFaction; // the current faction
  StrucFamily := STI(Params[0]); // the structure family

  if Faction = nil then
    begin
      Log(['[Warning] SetStructureCustomNames: Unknown target faction']);
      Exit;
    end;

  for i := 1 to Params.Count-1 do
  begin
    Value  := Params[i];
    Faction.SetStructureCustomName(StrucFamily, Value); // add the value to the family
  end;

  Log([Params.Count-1 , ' Structure Custom Names Set for ', Faction.Name, ' for the family ',Params[0] ]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// REMOVE STRUCTURE
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.RemoveStructure(Params : TStringList) : integer;
var
    Param     : string;
    Struc     : TStructure;
    i         : integer;
    iArg      : integer;
    Arg       : string;
    ByDefUID  : boolean;
    ByAttrib  : boolean;
    ByKind    : boolean;
    All       : boolean;
    AnyFac    : boolean;
    Bag       : TRandomPntBag;
    List      : TList;
begin
  Result := 0;
  if (Params.Count <> 3) then
  begin
    Inc(Result);
    Log(['=> RemoveStructure, need to have 3 parameters: RemoveStructure = <SelectionType(=ByDefUID or ByAttrib or ByKind) Value>;<ALL_OR_RANDOM_KEYWORD>;<ANYFAC_OR_SINGLEFAC_KEYWORD>']);
    Exit;
  end;

  AnyFac := UpperCase(Params[2]) = 'ANYFAC';

  if (not AnyFac) and (cuFaction = nil) then
  begin
    Log(['=> RemoveStructure command, no faction selected']);
    Inc(Result);
    Exit;
  end;
  if cuRegion = nil then // cuRegion should only be used for the UID, in case it is AnyFac...
  begin
    Log(['=> RemoveStructure command, no region selected']);
    Inc(Result);
    Exit;
  end;

  // Selection Type
  iArg := 0;
  ByDefUID := False;
  Param :=  ExtractParam('ByDefUID', Params);
  if Param <> '' then
  begin
    ByDefUID := True;
    ExtractArgs(Param);
    iArg := STI(Args[0]);
  end;
  ByAttrib := False;
  Param :=  ExtractParam('ByAttrib', Params);
  if Param <> '' then
  begin
    ByAttrib := True;
    ExtractArgs(Param);
    Arg := Args[0];
  end;
  ByKind := False;
  Param :=  ExtractParam('ByKind', Params);
  if Param <> '' then
  begin
    ByKind := True;
    ExtractArgs(Param);
    iArg := STI(Args[0]);
  end;

  if (not ByDefUID) and (not ByAttrib) and (not ByKind) then
  begin
    Log(['=> RemoveStructure command, no selection type selected']);
    Inc(Result);
    Exit;
  end;

  All := UpperCase(Params[1]) = 'ALL';
  Bag := TRandomPntBag.Create;
  List := TList.Create;

  for i := 0 to GameSim.RegionCmns[cuRegion.UID].pListStruc.Count-1 do
    List.Add(GameSim.RegionCmns[cuRegion.UID].pListStruc[i]); // each remove will do an update on Region & RegionCmn... better be safe

  for i := List.Count-1 downto 0 do
  begin
    Struc := List[i];
    cuRegion := TFaction(Struc.pFaction).Regions.Fetch(cuRegion.UID); // now cuRegion is really the region owning the structure
    if ByDefUID and (Struc.DefUID = iArg) and (AnyFac  or (cuFaction = Struc.pFaction)) then
    begin
      if All then
      begin
        Log(['RemoveStructure command, structure:', Struc.UID, Struc.Name, 'owned by', TFaction(Struc.pFaction).Name, 'will be removed']);
        cuRegion.RemoveStructure(Struc)
      end
      else
        Bag.Add(Struc);
    end
    else
    if ByAttrib and (Struc.HasThisAttribute(Arg)) and (AnyFac  or (cuFaction = Struc.pFaction)) then
    begin
      if All then
      begin
        Log(['RemoveStructure command, structure:', Struc.UID, Struc.Name, 'owned by', TFaction(Struc.pFaction).Name, 'will be removed']);
        cuRegion.RemoveStructure(Struc)
      end
      else
        Bag.Add(Struc);    end
    else
    if ByKind and (Struc.StrucDef.Kind = iArg) and (AnyFac  or (cuFaction = Struc.pFaction)) then
    begin
      if All then
      begin
        Log(['RemoveStructure command, structure:', Struc.UID, Struc.Name, 'owned by', TFaction(Struc.pFaction).Name, 'will be removed']);
        cuRegion.RemoveStructure(Struc)
      end
      else
        Bag.Add(Struc);
    end;
  end;

  if Bag.Count > 0 then
  begin
    Struc := Bag.PickPointer();
    Log(['RemoveStructure command, structure:', Struc.UID, Struc.Name, 'owned by', TFaction(Struc.pFaction).Name, 'will be removed']);
    cuRegion := TFaction(Struc.pFaction).Regions.Fetch(cuRegion.UID);
    cuRegion.RemoveStructure(Struc);
  end;

  Bag.Free;
  List.Free;
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET ACTOR POOL
// Actor Pool: parameters of a complex actor (option, MC option...)
// CHANGE does not exist
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetActorPool(Params : TStringList) : integer;
var i : integer;
    Key    : string;
    Value  : integer;
    sValue : string;
    ActorUID : TUID;
    APR : TActorPoolRec;
begin
  Result := 0;
  {#UPDASSETS}

  if Params.Count mod 2 = 0 then
  begin
    Inc(Result);
    Log(['=> SetActorPool, Specify an ActorUID and then a paired data (eg: an asset identifier and a value). Valid identifiers are: MaxUse, MaxChoices, ResetFreq, Title, ImageID, MsgString, Money, Conscript, WarSupply, Morale, VP,' + ' TextChoice(1 to 4), ToolChoice(1 to 4), Param(1 to 3), SParam (1 to 3), SubType']);
    Exit;
  end;
  if cuFaction = nil then
  begin
    Log(['=> SetActorPool command, no faction selected']);
    Inc(Result);
    Exit;
  end;
  ActorUID := STI(Params[0]);
  if (ActorUID < _gmaMinUID)or(ActorUID > _gmaMaxUID) then
  begin
    Log(['=> SetActorPool command, Erroneous Actor UID', ActorUID]);
    Inc(Result);
    Exit;
  end;

  APR := InitActorPoolRec;
  for i := 1 to Params.Count div 2 do
  begin
    Key     := Trim(UpperCase(Params[(i-1)*2 +1]));
    sValue  := Params[(i-1)*2 + 2];
    Value   := STI(sValue);

    if Key = 'MAXUSE' then
      APR.MaxUse := Value else
    if Key = 'MAXCHOICES' then
      APR.MaxChoices := Value else
    if Key = 'RESETFREQ' then  // is only used by AACW
      APR.ResetFreq := Value else
    if Key = 'IMAGEID' then
      APR.ImageId := sValue else {string}
    if Key = 'TITLE' then
      APR.Title := sValue else {string}
    if Key = 'MSGSTRING' then
      APR.MsgString := sValue else {string}
    if Key = 'TOOLSTRING' then
      APR.ToolString := sValue else {string}
    if Key = 'MONEY' then
      APR.Cost.Money := Value else
    if Key = 'INFLATION' then
      APR.Cost.Inflation := Value else
    if Key = 'PRICEINC' then {AACW}
      APR.Cost.Inflation := Value else
    if Key = 'CONSCRIPT' then
      APR.Cost.Conscript := Value else
    if Key = 'WARSUPPLY' then
      APR.Cost.WarSupply := Value else
    if Key = 'WSU' then
      APR.Cost.WarSupply := Value else
    if Key = 'MORALE' then
      APR.Cost.Morale := Value else
    if Key = 'VP' then
      APR.Cost.VPTotal := Value else
    if Key = 'PRESTIGE' then
      APR.Cost.VPTotal := Value else
    if Key = 'VPTOTAL' then
      APR.Cost.VPTotal := Value else

    if Key = 'CAPITAL' then
      APR.Cost.CapitalFunds := Value else
    if Key = 'STEEL' then
      APR.Cost.Steel := Value else
    if Key = 'COAL' then
      APR.Cost.Coal := Value else
    if (Key = 'GOOD') or (Key = 'GOODS') then
      APR.Cost.Goods := Value else
    if Key = 'MINERAL' then
      APR.Cost.Mineral := Value else
    if Key = 'RARE' then           // not yet serialized
      APR.Cost.Rare := Value else
    if Key = 'OIL' then
      APR.Cost.Oil := Value else
    if Key = 'DIPLOMAT' then
      APR.Cost.Diplomat := Value else
    if Key = 'ENGAGEMENTPTS' then
      APR.Cost.EngagementPts := Value else
    if Key = 'EP' then
      APR.Cost.EngagementPts := Value else
    if Key = 'FOREIGNENTRY' then             // not yet serialized
      APR.Cost.ForeignEntry := Value else
    if Key = 'RAILPOOL' then             // not yet serialized
      APR.Cost.RailPool := Value else
    if Key = 'RIVERPOOL' then
      APR.Cost.RiverPool := Value else   // not yet serialized

    if Key = 'PARAM1' then
      APR.IntParam1 := Value else
    if Key = 'PARAM2' then
      APR.IntParam2 := Value else
    if Key = 'PARAM3' then
      APR.IntParam3 := Value else
    if Key = 'SUBTYPE' then
      APR.SubType := sValue else
    if Key = 'SPARAM1' then
      APR.StrParam1 := sValue else
    if Key = 'SPARAM2' then
      APR.StrParam2 := sValue else
    if Key = 'SPARAM3' then
      APR.StrParam3 := sValue else

    if Key = 'TEXTCHOICE1' then
      APR.TextChoice1 := sValue else
    if Key = 'TOOLCHOICE1' then
      APR.ToolChoice1 := sValue else
    if Key = 'TEXTCHOICE2' then
      APR.TextChoice2 := sValue else
    if Key = 'TOOLCHOICE2' then
      APR.ToolChoice2 := sValue else
    if Key = 'TEXTCHOICE3' then
      APR.TextChoice3 := sValue else
    if Key = 'TOOLCHOICE3' then
      APR.ToolChoice3 := sValue else
    if Key = 'TEXTCHOICE4' then
      APR.TextChoice4 := sValue else
    if Key = 'TOOLCHOICE4' then
      APR.ToolChoice4 := sValue else
  end;

  cuFaction.SetActorPool(ActorUID, APR);
  if APR.MaxUse = 0 then
    Log(['SetActorPool', cuFaction.Name, 'Actor Type:', ActorUID, 'SubType:', APR.SubType, 'has been deleted'])
  else
    Log(['SetActorPool', cuFaction.Name, 'New Actor settings for Actor Type', ActorUID, ActorUID, 'SubType:', APR.SubType, 'are',
    'MaxUse:', APR.MaxUse, 'MaxChoices:', APR.MaxChoices, 'ResetFreq:', APR.ResetFreq, 'ImageID', APR.ImageID, 'Cost:', AssetCostToString(APR.Cost)]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// PAY OPTION
// Make the faction pays the cost of an option (an actor in use and validated)
// Option costs can be indicated with SetActorPool
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.PayOption(Params : TStringList) : integer;
var ActorUID  : TUID;
    SubType   : string;
    APR : TActorPoolRec;
    ForceSpend : boolean;
begin
  Result := 0;
  if not (Params.Count in [1..3]) then
  begin
    Inc(Result);
    Log(['=> PayOption, syntax is: PayOption = <GameActorUID> [;<SubType>] [;ForceSpend]']);
  end;
  if (cuFaction = nil) then
  begin
    Inc(Result);
    Log(['=> PayOption: No faction selected']);
  end;
  ActorUID := STI(Params[0]);

  if Params.Count >= 2 then
    SubType := Params[1]
  else
    SubType  := _gmaNULL;

  ForceSpend := ExtractParam('FORCESPEND', Params) <> '';

  if (ActorUID < _gmaMinUID)or(ActorUID > _gmaMaxUID) then
  begin
    Inc(Result);
    Log(['=> PayOption: Erroneous value for the Actor UID, the correct range is', _gmaMinUID, 'to', _gmaMaxUID, 'and the parameter was:', ActorUID, 'if you get a 0, then the alias can be incorrect.']);
  end;

  APR := cuFaction.GetActorPool(ActorUID, SubType);

  if IsZeroAssetCost(APR.Cost) then
  begin
    Log(['[WARNING] PayOption: Actor Pool not found or is a zero cost actor, so suspicious.']);
    Exit;
  end;
  if cuFaction.CanPayAssets(APR.Cost) then
  begin
    cuFaction.PayAssets(APR.Cost);
    Log(['PayOption :', APR.Title, 'was paid by faction', cuFaction.UID, cuFaction.Name,
    'Faction assets are:', AssetCostToString(ResourcesListToAssetCost(cuFaction.dlResourcesStock)),
    'Option costs are:', AssetCostToString(APR.Cost)]);
  end
  else
  if ForceSpend then
  begin
    cuFaction.PayAssets(APR.Cost, 0, True);
    Log(['PayOption :', APR.Title, 'was paid partially (ForceSpent ON) by faction', cuFaction.UID, cuFaction.Name,
    'Faction assets are:', AssetCostToString(ResourcesListToAssetCost(cuFaction.dlResourcesStock)),
    'Option costs are:', AssetCostToString(APR.Cost)]);
  end
  else
    Log(['PayOption :', APR.Title, 'was NOT paid by faction', cuFaction.UID, cuFaction.Name,
    'Faction assets are:', AssetCostToString(ResourcesListToAssetCost(cuFaction.dlResourcesStock)),
    'Option costs are:', AssetCostToString(APR.Cost)]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET TRANSPORT ASSETS INFOS
// UID de l'unit def Loco asset, River asset, Supply asset, qui sont utiliss
// pour dterminer les couts d'achats dans l'UTI ou la vitesse de distribution de supply
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetTRSAssetsInfos(Params : TStringList) : integer;
var RailUID, RiverUID, SupplyUID, NavalUID : integer;
begin
  Result := 0;
  if Params.Count < 3 then
  begin
    Inc(Result);
    Log(['=> SetTRSAssetsInfos, need to have exactly 3 parameters: Alias of UnitDef of the Locomotive Asset, Riverine Asset, Supply Net Unit (adv supply only), Naval Transport asset']);
    Exit;
  end;
  if cuFaction = nil then
  begin
    Log(['=> SetTRSAssetsInfos command, no faction selected']);
    Inc(Result);
    Exit;
  end;
  RailUID := STI(Params[0]);
  if (RailUID < 0) or (RailUID > UnitDefs.Count-1) then
  begin
    Log(['=> SetTRSAssetsInfos command, the Rail Asset UnitDef UID given is either negative or above the units definitions UID. ', RailUID]);
    Inc(Result);
    Exit;
  end;
  RiverUID := STI(Params[1]);
  if (RiverUID < 0) or (RiverUID > UnitDefs.Count-1) then
  begin
    Log(['=> SetTRSAssetsInfos command, the River Asset UnitDef UID given is either negative or above the units definitions UID. ', RiverUID]);
    Inc(Result);
    Exit;
  end;
  SupplyUID := STI(Params[2]);
  if (SupplyUID <= 0) or (SupplyUID > UnitDefs.Count-1) then
  begin
    Log(['=> SetTRSAssetsInfos command, the Supply Asset UnitDef UID given is either equals to zero (suspicious), negative or above the units definitions UID. ', SupplyUID]);
    Inc(Result);
    Exit;
  end;
  if Params.Count > 3 then
  begin
    NavalUID := STI(Params[3]);
    if (NavalUID <= 0) or (NavalUID > UnitDefs.Count-1) then
    begin
      Log(['=> SetTRSAssetsInfos command, the Naval Asset UnitDef UID given is either equals to zero (suspicious), negative or above the units definitions UID. ', NavalUID]);
      Inc(Result);
      Exit;
    end;
  end
  else
    NavalUID := _noUID;


  cuFaction.SetTransportAssetsUnitDef(RailUID, RiverUID, SupplyUID, NavalUID);
  Log(['SetTRSAssetsInfos', cuFaction.Name, 'New UIDs for Rail, River, Supply assets are:', RailUID, RiverUID, SupplyUID]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET REVOLTERS INFOS
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetRevoltersInfos(Params : TStringList) : integer;
var UnitDefUID : integer;
begin
  Result := 0;
  if Params.Count <> 4 then
  begin
    Inc(Result);
    Log(['=> SetRevoltersInfos, Syntax is SetRevolterInfos = RevolterUID|CityRaiseChance|DismissChance|ConvertChance']);
    Exit;
  end;
  if cuFaction = nil then
  begin
    Log(['=> SetRevoltersInfos command, no faction selected']);
    Inc(Result);
    Exit;
  end;
  UnitDefUID := STI(Params[0]);
  if (UnitDefUID < 0) or (UnitDefUID > UnitDefs.Count-1) then
  begin
    Log(['=> SetRevoltersInfos command, the UnitDef UID given is either negative or above the units definitions UID. ', UnitDefUID]);
    Inc(Result);
    Exit;
  end;

  cuFaction.SetRevoltersInfos(UnitDefUID, STI(Params[1]), STI(Params[2]), STI(Params[3]));
  Log(['SetRevoltersInfos', cuFaction.Name, 'Revolter Unit (UID/Name) is', UnitDefUID, UnitDefs[UnitDefUID].Name]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET UPRISING INFOS
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetUprisingInfos  (Params : TStringList) : integer;
var UnitDefUID1, UnitDefUID2, UnitDefUID3, FacDefUID, PopPerUnit: integer;
    FactionTag : string;
    FD : TFactionDef;
begin
  Result := 0;
  if Params.Count <> 5 then
  begin
    Inc(Result);
    Log(['=> SetUprisingInfos, Syntax is SetUprisingInfos = UnitUID_City;UnitUID_Wild;UnitUID_Other;FactionTag;PopPerUnit']);
    Exit;
  end;
  if cuFaction = nil then
  begin
    Log(['=> SetUprisingInfos command, no faction selected']);
    Inc(Result);
    Exit;
  end;
  UnitDefUID1 := STI(Params[0]);
  UnitDefUID2 := STI(Params[1]);
  UnitDefUID3 := STI(Params[2]);
  if (UnitDefUID1 < 0) or (UnitDefUID1 > UnitDefs.Count-1)
  or (UnitDefUID2 < 0) or (UnitDefUID2 > UnitDefs.Count-1)
  or (UnitDefUID3 < 0) or (UnitDefUID3 > UnitDefs.Count-1) then
  begin
    Log(['=> SetUprisingInfos command, One of the UnitDef UID given is either negative or above the units definitions UID. ', UnitDefUID1, UnitDefUID2, UnitDefUID3]);
    Inc(Result);
    Exit;
  end;

  FactionTag := Params[3];
  FD := FactionDefs.GetByFactionTag(FactionTag);
  FacDefUID := FD.UID;
  if (FacDefUID < 0) or (FacDefUID > FactionDefs.Count-1) then
  begin
    Log(['=> SetUprisingInfos command, FactionDef UID given is either negative or above the Factions definitions UID. ', FacDefUID]);
    Inc(Result);
    Exit;
  end;

  PopPerUnit := STI(Params[4]);

  cuFaction.SetUprisingInfos(UnitDefUID1, UnitDefUID2, UnitDefUID3, FacDefUID, PopPerUnit);
  Log(['SetUprisingInfos', cuFaction.Name, 'Uprising Units (UID/Name) are', UnitDefUID1, UnitDefs[UnitDefUID1].Name,
  UnitDefUID2, UnitDefs[UnitDefUID2].Name, UnitDefUID3, UnitDefs[UnitDefUID3].Name, 'Receiving Faction', FacDefUID, FactionDefs[FacDefUID].Name,
  'Pop needed per unit', PopPerUnit]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET SCORCH EARTH INFOS
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetScorchEarthInfos(Params : TStringList) : integer;
begin
  Result := 0;
  if not (Params.Count in [0, 1, 9]) then // NULL donne 0 argument  priori (ou 1?)
  begin
    Inc(Result);
    Log(['=> SetScorchEarthInfos, Syntax is either SetScorchEarthInfos = SupplyLossPercent|PillageChance|DepotBurnChance|FortBurnChance|VillageBurnChance|PlaceHolder1|PlaceHolder2|PlaceHolder3|PlaceHolder4 OR SetScorchEarthInfos = NULL']);
    Exit;
  end;
  if cuFaction = nil then
  begin
    Log(['=> SetScorchEarthInfos command, no faction selected']);
    Inc(Result);
    Exit;
  end;
  if Params.Count = 1 then
    cuFaction.SetScorchEarthInfos(0, 0, 0, 0, 0)
  else
    cuFaction.SetScorchEarthInfos(STI(Params[0]), STI(Params[1]), STI(Params[2]), STI(Params[3]), STI(Params[4]));
  Log(['SetScorchEarthInfos', cuFaction.Name, 'SupplyLossPercent:', Params[0], 'PillageChance:', Params[1], 'DepotBurnChance:', Params[2], 'FortBurnChance:', Params[3], 'VillageBurnChance:', Params[4]]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET NAVAL POOLS
// Liste des rgions  considrer pour calculer le pool de transport naval
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetNavalPools(Params : TStringList) : integer;
var RgnUID : TUID;
    Coeff : integer;
    Idx, i : integer;
begin
  Result := 0;
  if (cuFaction = nil) then
  begin
    Log(['[Warning] SetNavalPools: No faction selected']);
    Exit;
  end;
  if Params.Count mod 2 <> 0 then
  begin
    Inc(Result);
    Log(['=> SetNavalPools: Need to give pairs of Region UID & value.']);
    Exit;
  end;
  if Result <> 0 then Exit;

  for i := 1 to Params.Count div 2 do
  begin
    RgnUID := STI(Params[(i-1)*2]);
    Coeff  := STI(Params[(i-1)*2 + 1]);

    if (RgnUID < 0)or(RgnUID > RegionDefs.Count-1) then
    begin
      Inc(Result);
      Log(['=> SetNavalPools: The Region UID is not valid.']);
      Exit;
    end;

    Idx := cuFaction.dlNavalPools.Keys.IndexOf(ITS(RgnUID));
    if Idx <> -1 then
      cuFaction.dlNavalPools.Delete(Idx);

      // permet ainsi de supprimer des entres en entrant 0 sur une entre dclare
    if Coeff <> 0 then cuFaction.dlNavalPools.Add(RgnUID, Coeff);

    Log(['SetNavalPools:', cuFaction.Name, RegionDefs[RgnUID].Name, 'Naval Pools set to:', Coeff]);
  end;
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET BLOCKADES
// Quelle sont les rgions qui permettent un blocus contre la faction
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetBlockades  (Params : TStringList) : integer;
var RgnUID : TUID;
    Coeff : integer;
    Idx, i : integer;
begin
  Result := 0;
  if (cuFaction = nil) then
  begin
    Log(['[Warning] SetBlockades: No faction selected']);
    Exit;
  end;
  if Params.Count mod 2 <> 0 then
  begin
    Inc(Result);
    Log(['=> SetBlockades: Need to give pairs of Region UID & value.']);
    Exit;
  end;
  if Result <> 0 then Exit;

  for i := 1 to Params.Count div 2 do
  begin
    RgnUID := STI(Params[(i-1)*2]);
    Coeff  := STI(Params[(i-1)*2 + 1]);

    if (RgnUID < 0)or(RgnUID > RegionDefs.Count-1) then
    begin
      Inc(Result);
      Log(['=> SetBlockades: The Region UID is not valid.']);
      Exit;
    end;

    Idx := cuFaction.dlBlockades.Keys.IndexOf(ITS(RgnUID));
    if Idx <> -1 then
      cuFaction.dlBlockades.Delete(Idx);

      // permet ainsi de supprimer des entres en entrant 0 sur une entre dclare
    if Coeff <> 0 then cuFaction.dlBlockades.Add(RgnUID, Coeff);

    Log(['SetBlockades:', cuFaction.Name, RegionDefs[RgnUID].Name, 'Naval Pools set to:', Coeff]);
  end;
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET DISTANT BLOCKADES
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetDistBlockades  (Params : TStringList) : integer;
var SrcUID : TUID;
    DstUID : TUID;
    i      : integer;
begin
  Result := 0;
  if Params.Count mod 2 <> 0 then
  begin
    Inc(Result);
    Log(['=> SetDistantBlockades: Need to give pairs of blockading and blockaded Regions UID.']);
    Exit;
  end;
  if Result <> 0 then Exit;

  for i := 1 to Params.Count div 2 do
  begin
    SrcUID  := STI(Params[(i-1)*2]);
    DstUID  := STI(Params[(i-1)*2 + 1]);

    if (SrcUID <= 0)or(SrcUID > RegionDefs.Count-1) then
    begin
      Inc(Result);
      Log(['=> SetDistantBlockades: The Source UID is not valid (0 is always erroneous, check aliases too).', Params.Text]);
      Exit;
    end;
    if (SrcUID <= 0)or(SrcUID > RegionDefs.Count-1) then
    begin
      Inc(Result);
      Log(['=> SetDistantBlockades: The Destination UID is not valid (0 is always erroneous, check aliases too).', Params.Text]);
      Exit;
    end;

    GameSim.RegionCmns.AddDistantBlockade(SrcUID, DstUID);

    Log(['SetDistantBlockades:', RegionDefs[SrcUID].UID, RegionDefs[SrcUID].Name, 'can distant blockade', RegionDefs[DstUID].UID, RegionDefs[DstUID].Name]);
  end;

end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET MORALE LEVELS
// Niveau de Moral pour dfaite et victoire
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetMoraleLevels(Params : TStringList) : integer;
begin
  Result := 0;
  if Params.Count <> 2 then
  begin
    Inc(Result);
    Log(['=> Set Morale Levels , Syntax is SetMoraleLevels = Lose_Level;Win_Level']);
    Exit;
  end;
  if (cuFaction = nil) then
  begin
    Inc(Result);
    Log(['=> Set Morale Levels: No faction selected']);
  end;
  if Result <> 0 then Exit;

  cuFaction.MoraleLose  := STI(Params[0]);
  cuFaction.MoraleWin   := STI(Params[1]);

  Log(['Set Morale Levels:', cuFaction.Name, 'Lose = ', cuFaction.MoraleLose, 'Win = ', cuFaction.MoraleWin]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// CHANGE MORALE LEVELS
// Niveau de Moral pour dfaite et victoire
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.ChgMoraleLevels(Params : TStringList) : integer;
begin
  Result := 0;
  if Params.Count <> 2 then
  begin
    Inc(Result);
    Log(['=> Change Morale Levels , Syntax is ChgMoraleLevels = Lose_Level;Win_Level']);
    Exit;
  end;
  if (cuFaction = nil) then
  begin
    Inc(Result);
    Log(['=> Change Morale Levels: No faction selected']);
  end;
  if Result <> 0 then Exit;

  cuFaction.MoraleLose := cuFaction.MoraleLose + STI(Params[0]);
  cuFaction.MoraleWin   := cuFaction.MoraleWin + STI(Params[1]);

  Log(['Change Morale Levels:', cuFaction.Name, 'Lose = ', cuFaction.MoraleLose, 'Win = ', cuFaction.MoraleWin]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET AI POOL RATIO
// AACW (area rec): weight per family, ROP+ (region rec): weight per UnitDefUID
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.AISetPoolRatio(Params : TStringList) : integer;
var FamUID  : TUID;
    Value   : integer;
    i       : integer;
begin
  Result := 0;

  if _gameGen < 3 then
  begin
    if (Params.Count = 0) then
    begin
      Inc(Result);
      Log(['=> AI.SetPoolRatio: Syntax is AI.SetPoolRatio = [;<FamilyAlias>|Value] ..']);
    end;

    if Params.Count mod 2 <> 0 then
    begin
      Inc(Result);
      Log(['=> AI.SetPoolRatio, Specify a list of paired values: Family UID (or alias) | Pool value']);
      Exit;
    end;
  end
  else
  begin
    if (Params.Count = 0) then
    begin
      Inc(Result);
      Log(['=> AI.SetPoolRatio: Syntax is AI.SetPoolRatio = [;<UnitDefAlias>|Value] ..']);
    end;

    if Params.Count mod 2 <> 0 then
    begin
      Inc(Result);
      Log(['=> AI.SetPoolRatio, Specify a list of paired values: UnitDef UID (or alias) | Pool value']);
      Exit;
    end;

  end;

  if (cuFaction = nil) then
  begin
    Inc(Result);
    Log(['=> AI.SetPoolRatio: No faction selected']);
  end;
  if Result <> 0 then Exit;


  for i := 1 to Params.Count div 2 do
  begin
    FamUID   := STI(Params[(i-1)*2]);
    Value := STI(Params[(i-1)*2 + 1]);
    cuFaction.dlAIPoolRatios.AlterValByKey(FamUID, Value); // pas de somme arithmtique (= set). Suppression si entre < 0
  end;
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// AI SET LAND/NAV BUILD
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.AISetLandBuild (Params : TStringList) : integer;
begin
  Result := 0;
  if (Params.Count <> 1) then
  begin
    Log(['=> AI.SetLandBuild: Syntax is AI.SetLandBuild = Value']);
    Inc(Result);
    Exit;
  end;
  if cuFaction = nil then
  begin
    Log(['=> AI.SetLandBuild: No Defined faction.']);
    Inc(Result);
    Exit;
  end;

  cuFaction.AIGuide.LandBuildPerc := STI(Params[0]);
end;

function TScriptEngine.AISetNavBuild  (Params : TStringList) : integer;
begin
  Result := 0;
  if (Params.Count <> 1) then
  begin
    Log(['=> AI.SetNavBuild: Syntax is AI.SetNavBuild = Value']);
    Inc(Result);
    Exit;
  end;
  if cuFaction = nil then
  begin
    Log(['=> AI.SetNavBuild: No Defined faction.']);
    Inc(Result);
    Exit;
  end;

  cuFaction.AIGuide.NavBuildPerc := STI(Params[0]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET UNIT POOL INFOS
// nb AI pool ratios is to alter the AI will to buy these units, 100 being standard.
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetUnitPoolInfos(Params : TStringList) : integer;
var UnitDefUID, Value, PR : integer;
begin
  Result := 0;
  if Params.Count < 2 then
  begin
    Inc(Result);
    Log(['=> SetUnitPoolInfos, need to have exactly 3 parameters: Alias of UnitDef, variation of Force Pool [, AI Pool Ratios]']);
    Exit;
  end;
  if cuFaction = nil then
  begin
    Log(['=> SetUnitPoolInfos command, no faction selected']);
    Inc(Result);
    Exit;
  end;

  UnitDefUID := STI(Params[0], True, -1);
  Value      := STI(Params[1]);

  if Params.Count = 3 then
    PR := STI(Params[2])
  else
    PR := 100;


  if UnitDefUID = -1 then
  if cuFaction = nil then
  begin
    Log(['=> SetUnitPoolInfos command,', Params[0], 'is an unknown unit']);
    Inc(Result);
    Exit;
  end;
  if (UnitDefUID < 0) or (UnitDefUID > UnitDefs.Count-1) then
  begin
    Log(['=> SetUnitPoolInfos command, the UnitDef UID given is either negative or above the units definitions UID. ', UnitDefUID]);
    Inc(Result);
    Exit;
  end;
  cuFaction.SetUnitMaxForcePool(UnitDefUID, Value);
  cuFaction.dlAIPoolRatios.AlterValByKey(UnitDefUID, PR);
  Log(['SetUnitPoolInfos', cuFaction.Name, 'New Force Pool for unit (UID/Name) is', UnitDefUID, UnitDefs[UnitDefUID].Name,  cuFaction.GetUnitMaxForcePool(UnitDefUID, True), 'New Pool Ratio is', PR, '%']);
end;
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// CHG UNIT POOL INFOS
// Pool ratios remain in 'SET mode' but Force Pool quantity is a change as expected
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.ChgUnitPoolInfos(Params : TStringList) : integer;
var UnitDefUID, Value, PR : integer;
begin
  Result := 0;
  if Params.Count < 2 then
  begin
    Inc(Result);
    Log(['=> ChgUnitPoolInfos, need to have 2 or 3 parameters: Alias of UnitDef, variation of Force Pool[, AI Pool Ratios]']);
    Exit;
  end;
  if cuFaction = nil then
  begin
    Log(['=> ChgUnitPoolInfos command, no faction selected']);
    Inc(Result);
    Exit;
  end;
  UnitDefUID := STI(Params[0], True, -1);
  Value      := STI(Params[1]);

  if Params.Count = 3 then
    PR := STI(Params[2])
  else
    PR := 100;


  if UnitDefUID = -1 then
  if cuFaction = nil then
  begin
    Log(['=> ChgUnitPoolInfos command,', Params[0], 'is an unknown unit']);
    Inc(Result);
    Exit;
  end;
  if (UnitDefUID < 0) or (UnitDefUID > UnitDefs.Count-1) then
  begin
    Log(['=> ChgUnitPoolInfos command, the UnitDef UID given is either negative or above the units definitions UID. ', UnitDefUID]);
    Inc(Result);
    Exit;
  end;
  cuFaction.ChangeUnitMaxForcePool(UnitDefUID, Value);
  cuFaction.dlAIPoolRatios.AlterValByKey(UnitDefUID, PR);
  Log(['ChgUnitPoolInfos', cuFaction.Name, 'New Force Pool for unit (UID/Name) is', UnitDefUID, UnitDefs[UnitDefUID].Name,  cuFaction.GetUnitMaxForcePool(UnitDefUID, True), 'New Pool Ratio is', PR, '%']);
end;
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// AI CHANGE AGGRESSIVITY
// Aggro coeff at the Theater level (110%: feeling 10% stronger that the AI is really)
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.AIChangeAggro(Params : TStringList) : integer;
var i         : integer;
    ObjectUID : TUID;
    Value     : integer;
begin
  Result := 0;
  if not (Params.Count > 0) then
  begin
    Inc(Result);
    Log(['=> Syntax is AI.ChgAggro = [;<TheaterUID|Aggro_Coeff>] ..']);
  end;
  if Params.Count mod 2 <> 0 then
  begin
    Inc(Result);
    Log(['=> AI.ChgAggro, Specify a list of paired values: TheaterUID|Aggro_Coeff']);
  end;
  if cuFaction = nil then
  begin
    Inc(Result);
    Log(['=> AI.ChgAggro, No faction selected']);
  end;
  if Result <> 0 then Exit;

  for i := 1 to Params.Count div 2 do
  begin
    ObjectUID := STI(Params[(i-1)*2]);
    if (ObjectUID < 0) or (ObjectUID > Areas.Count-1) then
    begin
      Inc(Result);
      Log(['=> AI.SetAggro, Not a valid area, skipped.']);
      Continue;
    end;
    if not Areas[ObjectUID].IsGrandArea then
    begin
      Inc(Result);
      Log(['=> AI.SetAggro, Not a valid theater, skipped.']);
      Continue;
    end;

    if Result <> 0 then Exit;
        
    Value     := STI(Params[(i-1)*2 + 1]);
    cuFaction.AIGuide.dlAggressivity.AlterValByKey(ObjectUID, Value, True);  // arithmetic sum (with old value). Entry removed if sum is <= 0
    Log(['AI.ChgAggro', cuFaction.Name, 'Aggressivity coeff in Theater', ObjectUID, 'is now at', cuFaction.AIGuide.dlAggressivity.GetValByKey(ObjectUID)]);
  end;
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// AI SET AGGRESSIVITY
// Aggro coeff at the Theater level (110%: feeling 10% stronger that the AI is really)
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.AISetAggro(Params : TStringList) : integer;
var i         : integer;
    ObjectUID : TUID;
    Value     : integer;
begin
  Result := 0;
  if not (Params.Count > 0) then
  begin
    Inc(Result);
    Log(['=> Syntax is AI.SetAggro = [;<TheaterUID|Aggro_Coeff>] ..']);
  end;
  if Params.Count mod 2 <> 0 then
  begin
    Inc(Result);
    Log(['=> AI.SetAggro, Specify a list of paired values: TheaterUID|Aggro_Coeff']);
  end;
  if cuFaction = nil then
  begin
    Inc(Result);
    Log(['=> AI.SetAggro, No faction selected']);
  end;
  if Result <> 0 then Exit;

  for i := 1 to Params.Count div 2 do
  begin
    ObjectUID := STI(Params[(i-1)*2]);
    if (ObjectUID < 0) or (ObjectUID > Areas.Count-1) then
    begin
      Inc(Result);
      Log(['=> AI.SetAggro, Not a valid area, skipped.']);
      Continue;
    end;
    if not Areas[ObjectUID].IsGrandArea then
    begin
      Inc(Result);
      Log(['=> AI.SetAggro, Not a valid theater, skipped.']);
      Continue;
    end;

    if Result <> 0 then Exit;
    
    Value     := STI(Params[(i-1)*2 + 1]);
    cuFaction.AIGuide.dlAggressivity.AlterValByKey(ObjectUID, Value, False);
    Log(['AI.SetAggro', cuFaction.Name, 'Aggressivity coeff in Theater', ObjectUID, 'is now at', cuFaction.AIGuide.dlAggressivity.GetValByKey(ObjectUID)]);
  end;
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// AI CHANGE THEATER PREVALENCE
// Prevalence coeff at the Theater level (110%: 10% more important than the calculated value)
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.AIChangeTheaPrev(Params : TStringList) : integer;
var i         : integer;
    ObjectUID : TUID;
    Value     : integer;
begin
  Result := 0;
  if not (Params.Count > 0) then
  begin
    Inc(Result);
    Log(['=> Syntax is AI.ChgTheaPrev = [;<TheaterUID|Prevalence_Coeff>] ..']);
  end;
  if Params.Count mod 2 <> 0 then
  begin
    Inc(Result);
    Log(['=> AI.ChgTheaPrev, Specify a list of paired values: TheaterUID|Prevalence_Coeff']);
  end;
  if cuFaction = nil then
  begin
    Inc(Result);
    Log(['=> AI.ChgTheaPrev, No faction selected']);
  end;
  if Result <> 0 then Exit;

  for i := 1 to Params.Count div 2 do
  begin
    ObjectUID := STI(Params[(i-1)*2]);
    Value     := STI(Params[(i-1)*2 + 1]);
    cuFaction.AIGuide.dlTheaPrevalence.AlterValByKey(ObjectUID, Value, True);  // arithmetic sum (with old value). Entry removed if sum is <= 0
    Log(['AI.ChgTheaPrev', cuFaction.Name, 'Prevalence coeff in Theater', ObjectUID, 'is now at', cuFaction.AIGuide.dlTheaPrevalence.GetValByKey(ObjectUID)]);
  end;
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// AI SET THEATER PREVALENCE
// Prevalence coeff at the Theater level (110%: 10% more important than the calculated value)
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.AISetTheaPrev(Params : TStringList) : integer;
var i         : integer;
    ObjectUID : TUID;
    Value     : integer;
begin
  Result := 0;
  if not (Params.Count > 0) then
  begin
    Inc(Result);
    Log(['=> Syntax is AI.SetTheaPrev = [;<TheaterUID|Prevalence_Coeff>] ..']);
  end;
  if Params.Count mod 2 <> 0 then
  begin
    Inc(Result);
    Log(['=> AI.SetTheaPrev, Specify a list of paired values: TheaterUID|Prevalence_Coeff']);
  end;
  if cuFaction = nil then
  begin
    Inc(Result);
    Log(['=> AI.SetTheaPrev, No faction selected']);
  end;
  if Result <> 0 then Exit;

  for i := 1 to Params.Count div 2 do
  begin
    ObjectUID := STI(Params[(i-1)*2]);
    Value     := STI(Params[(i-1)*2 + 1]);
    cuFaction.AIGuide.dlTheaPrevalence.AlterValByKey(ObjectUID, Value, False);
    Log(['AI.SetTheaPrev', cuFaction.Name, 'Prevalence coeff in Theater', ObjectUID, 'is now at', cuFaction.AIGuide.dlTheaPrevalence.GetValByKey(ObjectUID)]);
  end;
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// AI CHANGE LOCAL INTEREST
// Coeff to the interest value of a region, 25%: the region get only 1/4 of the caculated interest it has
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.AIChgeLocalIntrst(Params : TStringList) : integer;
var i         : integer;
    ObjectUID : TUID;
    Value     : integer;
    IL        : TIntList;
    WasThere  : boolean;
begin
  Result := 0;
  if not (Params.Count > 0) then
  begin
    Inc(Result);
    Log(['=> Syntax is AI.ChgLocalInterest = [;<RegionUID|Interest_Coeff>] ..']);
  end;
  if Params.Count mod 2 <> 0 then
  begin
    Inc(Result);
    Log(['=> AI.ChgLocalInterest, Specify a list of paired values: RegionUID|Interest_Coeff']);
  end;
  if cuFaction = nil then
  begin
    Inc(Result);
    Log(['=> AI.ChgLocalInterest, No faction selected']);
  end;
  if Result <> 0 then Exit;

  IL := TIntList.Create;

  for i := 1 to Params.Count div 2 do
  begin
    ObjectUID := STI(Params[(i-1)*2]);
    IL.Add(ObjectUID);
    Value     := STI(Params[(i-1)*2 + 1]);
    WasThere := cuFaction.AIGuide.dlLocalInterests.KeyExists(ObjectUID);
    cuFaction.AIGuide.dlLocalInterests.AlterValByKey(ObjectUID, Value, True);  // arithmetic sum (with old value). Entry removed if sum is <= 0
    // if the key does not exists anymore, then it means minimum interest, 1 (as no entry = 100 instead)
    if WasThere and (not cuFaction.AIGuide.dlLocalInterests.KeyExists(ObjectUID)) then
      cuFaction.AIGuide.dlLocalInterests.AlterValByKey(ObjectUID, 1, True);
  end;

  if IL.Count > 0 then
    Log(['AI.ChgLocalInterest', cuFaction.Name, 'Interests changed by this command:']);
  for i := 0 to IL.Count-1 do
    Log(['Region', IL[i], 'Interest Coeff.:', cuFaction.AIGuide.dlLocalInterests.GetValByKey(IL[i])]);

  IL.Free;
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// AI SET LOCAL INTEREST
// Coeff to the interest value of a region, 25%: the region get only 1/4 of the caculated interest it has
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.AISetLocalIntrst(Params : TStringList) : integer;
var i         : integer;
    ObjectUID : TUID;
    Value     : integer;
    IL        : TIntList;
    WasThere  : boolean;
begin
  Result := 0;
  if not (Params.Count > 0) then
  begin
    Inc(Result);
    Log(['=> Syntax is AI.SetLocalInterest = [;<RegionUID|Interest_Coeff>] ..']);
  end;
  if Params.Count mod 2 <> 0 then
  begin
    Inc(Result);
    Log(['=> AI.SetLocalInterest, Specify a list of paired values: RegionUID|Interest_Coeff']);
  end;
  if cuFaction = nil then
  begin
    Inc(Result);
    Log(['=> AI.SetLocalInterest, No faction selected']);
  end;
  if Result <> 0 then Exit;

  IL := TIntList.Create;

  for i := 1 to Params.Count div 2 do
  begin
    ObjectUID := STI(Params[(i-1)*2]);
    IL.Add(ObjectUID);
    Value     := STI(Params[(i-1)*2 + 1]);
    WasThere := cuFaction.AIGuide.dlLocalInterests.KeyExists(ObjectUID);
    cuFaction.AIGuide.dlLocalInterests.AlterValByKey(ObjectUID, Value, False);
    // if the key does not exists anymore, then it means minimum interest, 1 (as no entry = 100 instead)
    if WasThere and (not cuFaction.AIGuide.dlLocalInterests.KeyExists(ObjectUID)) then
      cuFaction.AIGuide.dlLocalInterests.AlterValByKey(ObjectUID, 1, True);
  end;

  if IL.Count > 0 then
    Log(['AI.SetLocalInterest', cuFaction.Name, 'Interests changed by this command:']);
  for i := 0 to IL.Count-1 do
    Log(['Region', IL[i], 'Interest Coeff.:', cuFaction.AIGuide.dlLocalInterests.GetValByKey(IL[i])]);

  IL.Free;
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// AI CHANGE LOCAL INTEREST
// Coeff to the interest value of a region, 25%: the region get only 1/4 of the caculated interest it has
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.AIChgeAreaLocalInt(Params : TStringList) : integer;
var i         : integer;
    Value     : integer;
    AreaUID   : TUID;
    RgnUID    : TUID;
    Area      : TArea;
    WasThere  : boolean;
begin
  Result := 0;
  if not (Params.Count = 2) then
  begin
    Inc(Result);
    Log(['=> Syntax is AI.ChgAreaLocalInterest = AreaUID|Interest_Coeff']);
  end;
  if cuFaction = nil then
  begin
    Inc(Result);
    Log(['=> AI.ChgAreaLocalInterest, No faction selected']);
  end;
  SetNumeric(Params[0], AreaUID);
  if (AreaUID < 0) or (AreaUID > Areas.Count-1) then
  begin
    Inc(Result);
    Log(['=> AI.ChgAreaLocalInterest, Erroneous AreaUID', AreaUID]);
  end;
  if Result <> 0 then Exit;

  Area := Areas[AreaUID];
  Value := STI(Params[1]);

  for i := 0 to Area.Count-1 do
  begin
    RgnUID := Area[i];
    WasThere := cuFaction.AIGuide.dlLocalInterests.KeyExists(RgnUID);
    cuFaction.AIGuide.dlLocalInterests.AlterValByKey(RgnUID, Value, True);  // arithmetic sum (with old value). Entry removed if sum is <= 0
    // if the key does not exists anymore, then it means minimum interest, 1 (as no entry = 100 instead)
    if WasThere and (not cuFaction.AIGuide.dlLocalInterests.KeyExists(RgnUID)) then
      cuFaction.AIGuide.dlLocalInterests.AlterValByKey(RgnUID, 1, True);
    Log(['Region', RgnUID, 'Interest Coeff.:', cuFaction.AIGuide.dlLocalInterests.GetValByKey(RgnUID)]);
  end;

end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// AI SET LOCAL INTEREST
// Coeff to the interest value of a region, 25%: the region get only 1/4 of the caculated interest it has
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.AISetAreaLocalInt(Params : TStringList) : integer;
var i         : integer;
    Value     : integer;
    AreaUID   : TUID;
    RgnUID    : TUID;
    Area      : TArea;
    WasThere  : boolean;
begin
  Result := 0;
  if not (Params.Count = 2) then
  begin
    Inc(Result);
    Log(['=> Syntax is AI.SetAreaLocalInterest = AreaUID|Interest_Coeff']);
  end;
  if cuFaction = nil then
  begin
    Inc(Result);
    Log(['=> AI.SetAreaLocalInterest, No faction selected']);
  end;
  SetNumeric(Params[0], AreaUID);
  if (AreaUID < 0) or (AreaUID > Areas.Count-1) then
  begin
    Inc(Result);
    Log(['=> AI.SetAreaLocalInterest, Erroneous AreaUID', AreaUID]);
  end;
  if Result <> 0 then Exit;

  Area := Areas[AreaUID];
  Value := STI(Params[1]);

  for i := 0 to Area.Count-1 do
  begin
    RgnUID := Area[i];
    WasThere := cuFaction.AIGuide.dlLocalInterests.KeyExists(RgnUID);
    cuFaction.AIGuide.dlLocalInterests.AlterValByKey(RgnUID, Value, False);  // NO arithmetic sum (with old value). Entry removed if sum is <= 0
    // if the key does not exists anymore, then it means minimum interest, 1 (as no entry = 100 instead)
    if WasThere and (not cuFaction.AIGuide.dlLocalInterests.KeyExists(RgnUID)) then
      cuFaction.AIGuide.dlLocalInterests.AlterValByKey(RgnUID, 1, True);
    Log(['Region', RgnUID, 'Interest Coeff.:', cuFaction.AIGuide.dlLocalInterests.GetValByKey(RgnUID)]);
  end;
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// AI SET PARAMS
// To redefine an AI param for a particular faction
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.AISetParams(Params : TStringList) : integer;
var i : integer;
    Key    : string;
    Value  : integer;
    sValue : string;
    NoOp : boolean;
begin
  Result := 0;

  if Params.Count = 0 then
  begin
    Inc(Result);
    Log(['=> AI.SetParams, Specify paired data (eg: an param identifier and its value). <Param1 Value1> [;<Param2 Value 2>] [;<Param3 Value 3>] ... See the wiki for a full list of correct params, example: ''aiLRV_RegionBaseVal 3''']);
    Exit;
  end;
  if cuFaction = nil then
  begin
    Log(['=> AI.SetParams command, no faction selected']);
    Inc(Result);
    Exit;
  end;

  // Each param redefinition is a pair of data
  for i := 1 to Params.Count div 2 do
  begin
    Key     := Trim(UpperCase(Params[(i-1)*2 ]));
    sValue  := Params[(i-1)*2 + 1];
    Value   := STI(sValue);
    NoOp := False;

    if Key = UpperCase('aiLRV_RegionBaseVal') then
      cuFaction.AIGuide.aiLRV_RegionBaseVal := Value else
    if Key = UpperCase('aiLRV_OwnStrucVal') then
      cuFaction.AIGuide.aiLRV_OwnStrucVal := Value else
    if Key = UpperCase('aiLRV_OppStrucVal') then
      cuFaction.AIGuide.aiLRV_OppStrucVal := Value else
    if Key = UpperCase('aiLRV_MaxLevel') then
      cuFaction.AIGuide.aiLRV_MaxLevel := Value else
    if Key = UpperCase('aiLRV_FortValCoef') then
      cuFaction.AIGuide.aiLRV_FortValCoef := Value else
    if Key = UpperCase('aiLRV_HarborValCoef') then
      cuFaction.AIGuide.aiLRV_HarborValCoef := Value else
    if Key = UpperCase('aiLRV_DepotValCoef') then
      cuFaction.AIGuide.aiLRV_DepotValCoef := Value else
    if Key = UpperCase('aiLRV_OwnPopVal') then
      cuFaction.AIGuide.aiLRV_OwnPopVal := Value else
    if Key = UpperCase('aiLRV_OppPopVal') then
      cuFaction.AIGuide.aiLRV_OppPopVal := Value else
    if Key = UpperCase('aiLRV_VPValue') then
      cuFaction.AIGuide.aiLRV_VPValue := Value else
    if Key = UpperCase('aiLRV_ValSupply') then
      cuFaction.AIGuide.aiLRV_ValSupply := Value else
    if Key = UpperCase('aiLRV_Conscript') then
      cuFaction.AIGuide.aiLRV_Conscript := Value else
    if Key = UpperCase('aiLRV_Money') then
      cuFaction.AIGuide.aiLRV_Money := Value else
    if Key = UpperCase('aiLRV_WSU') then
      cuFaction.AIGuide.aiLRV_WSU := Value else
    if Key = UpperCase('aiLRV_ObjValue') then
      cuFaction.AIGuide.aiLRV_ObjValue := Value else
    if Key = UpperCase('aiLRV_VPTurnBonusStart') then
      cuFaction.AIGuide.aiLRV_VPTurnBonusStart := Value else
    if Key = UpperCase('aiLRV_VPEndBonus') then
      cuFaction.AIGuide.aiLRV_VPEndBonus := Value else
    if Key = UpperCase('aiLRV_ControlVal') then
      cuFaction.AIGuide.aiLRV_ControlVal := Value else
    if Key = UpperCase('aiLRV_HighControl') then
      cuFaction.AIGuide.aiLRV_HighControl := Value else
    if Key = UpperCase('aiLRV_FullControl') then
      cuFaction.AIGuide.aiLRV_FullControl := Value else
    if Key = UpperCase('aiLRV_AdjBonusVal') then
      cuFaction.AIGuide.aiLRV_AdjBonusVal := Value else
    if Key = UpperCase('aiLRV_OppControlPerc') then
      cuFaction.AIGuide.aiLRV_OppControlPerc := Value else
    if Key = UpperCase('aiLRV_NoAdjPerc') then
      cuFaction.AIGuide.aiLRV_NoAdjPerc := Value else
    if Key = UpperCase('aiLRV_PrevalMaxReducPerc') then
      cuFaction.AIGuide.aiLRV_PrevalMaxReducPerc := Value else
    if Key = UpperCase('aiLRV_TrackVal') then
      cuFaction.AIGuide.aiLRV_TrackVal := Value else
    if Key = UpperCase('aiLRV_RoadVal') then
      cuFaction.AIGuide.aiLRV_RoadVal := Value else
    if Key = UpperCase('aiLRV_MRoadVal') then
      cuFaction.AIGuide.aiLRV_MRoadVal := Value else
    if Key = UpperCase('aiLRV_RailVal') then
      cuFaction.AIGuide.aiLRV_RailVal := Value else
    if Key = UpperCase('aiLRV_DestRailVal') then
      cuFaction.AIGuide.aiLRV_DestRailVal := Value else

    if Key = UpperCase('aiCLR_LocalDistance') then
      cuFaction.AIGuide.aiCLR_LocalDistance := Value else
    if Key = UpperCase('aiCLR_StdDistance') then
      cuFaction.AIGuide.aiCLR_StdDistance := Value else
    if Key = UpperCase('aiCLR_AutonomyStd') then
      cuFaction.AIGuide.aiCLR_AutonomyStd := Value else
    if Key = UpperCase('aiCLR_AutonomyHarsh') then
      cuFaction.AIGuide.aiCLR_AutonomyHarsh := Value else
    if Key = UpperCase('aiCLR_AutonomyVHarsh') then
      cuFaction.AIGuide.aiCLR_AutonomyVHarsh := Value else
    if Key = UpperCase('aiCLR_AutonomyNoGrowth') then
      cuFaction.AIGuide.aiCLR_AutonomyNoGrowth := Value else
    if Key = UpperCase('aiCLR_AutonomyNoSupply') then
      cuFaction.AIGuide.aiCLR_AutonomyNoSupply := Value else
    if Key = UpperCase('aiCLR_AutonomyNoCloseGoal') then
      cuFaction.AIGuide.aiCLR_AutonomyNoCloseGoal := Value else
    if Key = UpperCase('aiCLR_BigCity') then
      cuFaction.AIGuide.aiCLR_BigCity := Value else
    if Key = UpperCase('aiCLR_BigGroup') then
      cuFaction.AIGuide.aiCLR_BigGroup := Value else
    if Key = UpperCase('aiCLR_SupplyPerc') then
      cuFaction.AIGuide.aiCLR_SupplyPerc := Value else
    if Key = UpperCase('aiCLR_LoyMCPerc') then
      cuFaction.AIGuide.aiCLR_LoyMCPerc := Value else
    if Key = UpperCase('aiCLR_WinterQuarters') then
      cuFaction.AIGuide.aiCLR_WinterQuarters := Value else

    begin
      Log(['AI.SetParams', cuFaction.Name, 'Unknown parameter', Key, '=', Value]);
      NoOp := True;
    end;

    if NoOp = False then
      Log(['AI.SetParams', cuFaction.Name, 'A param have been redefined', Key, '=', Value]);
  end;
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET MAX ENTRENCH
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetFacMaxEntrench(Params : TStringList) : integer;
var i : integer;
begin
  Result := 0;

  if Params.Count <> 1 then
  begin
    Inc(Result);
    Log(['=> Set Faction Max Entrench Level: Syntax is SetFacMaxEntrench = Value']);
  end;

  if Result <> 0 then Exit;

  if (cuFaction = nil) then
  begin
    for i := 0 to GameSim.Factions.Count-1 do
      GameSim.Factions[i].MaxEntrench := STI(Params[0]);
    Log(['SetFacMaxEntrench:', 'All factions Max Level of entrenchments is:', Params[0]]);
  end
  else
  begin
    cuFaction.MaxEntrench := STI(Params[0]);
    Log(['SetFacMaxEntrench:', cuFaction.Name, 'Max Level of entrenchments is:', cuFaction.MaxEntrench]);
  end;
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// CHANGE FACTION RELATIONSHIPS
// ChgFacRelationships = Fac1|Val1|Fac2|Val2 ... [NOTALLY] or
// ChgFacRelationships = ALL|Val ... [NOTALLY]
// ChgFacRelationships = RGNOWNER|-50
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.ChangeFactionRel(Params : TStringList) : integer;
var FacTag  : string;
    Value   : integer;
    i       : integer;
    NotAlly : boolean;
    RgnOwner: boolean;
    All     : boolean;
    Target  : TFaction;
begin
  Result := 0;

  if not (Params.Count > 0) then
  begin
    Inc(Result);
    Log(['=> ChangeFactionRelationships: Syntax is ChgFacRelationships = [;<FactionTag*|Value>] .. [;NOTALLY] ']);
  end;


  if (cuFaction = nil) then
  begin
    Inc(Result);
    Log(['=> ChgFacRelationships: No faction selected']);
  end;
  if Result <> 0 then Exit;

  // We don't want to alter the rel of allies
  NotAlly := ExtractParam('NOTALLY', Params) <> '';
  // all countries in the world
  All := ExtractParam('ALL', Params) <> '';

  RgnOwner := ExtractParam('RGNOWNER', Params) <> '';
  if RgnOwner and (cuRegion = nil) then
  begin
    Inc(Result);
    Log(['=> ChgFacRelationships: No region selected and asked to check for an owner']);
  end;
  if Result <> 0 then Exit;

  // the syntax is special if RGNOWER is specified: ChgFacRelationships = RGNOWNER|-50 e.g
  if RgnOwner and (GameSim.RegionCmns[cuRegion.UID].OwningFac <> nil) then
  begin
    Target := GameSim.RegionCmns[cuRegion.UID].OwningFac;
    Value := STI(Params[1]);
    cuFaction.AlterRelationship(Target, Value); // <0 ok
    Log(['Change Faction Relationships:', cuFaction.Name, '<>', Target.Name, ('region owner'), cuFaction.GetRelationship(Target)]);
    Exit;
  end
  else
  if RgnOwner then
    Exit; // no owner

  if not all then
  for i := 1 to Params.Count div 2 do
  begin
    FacTag := Params[(i-1)*2];
    Target := GameSim.Factions.GetFactionByTag(FacTag, False);
    if Target = nil then Continue; // can also be NOTALLY ...

    if NotAlly and cuFaction.IsAllyOf(Target) then Continue;
    Value  := STI(Params[(i-1)*2 + 1]);
    cuFaction.AlterRelationship(Target, Value); // <0 ok
    Log(['Change Faction Relationships:', cuFaction.Name, '<>', Target.Name, cuFaction.GetRelationship(Target)]);
  end
  else
  begin
    for i := 0 to GameSim.Factions.Count-1 do
    begin
      Target := GameSim.Factions[i];
      if NotAlly and cuFaction.IsAllyOf(Target) then Continue;

      if Target = nil then Continue;
      Value  := STI(Params[1]);
      cuFaction.AlterRelationship(Target, Value); // <0 ok
      Log(['Change Faction Relationships:', cuFaction.Name, '<>', Target.Name, cuFaction.GetRelationship(Target)]);
    end;
  end;

end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET FACTION RELATIONSHIPS
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetFactionRel(Params : TStringList) : integer;
var FacTag  : string;
    Value   : integer;
    i       : integer;
    Faction : TFaction;
begin
  Result := 0;

  if not (Params.Count > 0) then
  begin
    Inc(Result);
    Log(['=> SetFactionRelationships: Syntax is SetFacRelationships = [;<FactionTag|Value>] ..']);
  end;

  if Params.Count mod 2 <> 0 then
  begin
    Inc(Result);
    Log(['=> SetFactionRelationships, Specify a list of paired values: FactionTag | Value']);
    Exit;
  end;

  if (cuFaction = nil) then
  begin
    Inc(Result);
    Log(['=> SetFacRelationships: No faction selected']);
  end;



  if Result <> 0 then Exit;


  for i := 1 to Params.Count div 2 do
  begin
    FacTag := Params[(i-1)*2];
    Faction := GameSim.Factions.GetFactionByTag(FacTag);
    if Faction = nil then
    begin
      Log(['[Warning] Set Faction Relationships:', cuFaction.Name, FacTag, 'unknown target faction']);
      Continue
    end;

    Value  := STI(Params[(i-1)*2 + 1]);
    cuFaction.AlterRelationship_SE(FacTag, Value, False); // <0 ok
  end;

  Log(['Set Faction Relationships:', cuFaction.Name, FacTag, cuFaction.GetRelationship_SE(FacTag)]);
end;


// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// CHANGE FOREIGN ENTRY
// Change le niveau de Foreign Entry (entre -100 et 100, avec -999 en valeur spciale, cf clef _FOREIGN_ dans UGlobVar)
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.ChangeForeignEntry(Params : TStringList) : integer;
var Faction : TFaction;
    Value : integer;
begin
  Result := 0;
  if Params.Count <> 1 then
  begin
    Inc(Result);
    Log(['=> ChangeForeignEntry , Syntax is ChangeForeignEntry = Value']);
    Exit;
  end;
  if (cuFaction = nil) then
  begin
    Inc(Result);
    Log(['=> ChangeForeignEntry: No faction selected']);
  end;
  if (cuFaction.FactionDef.ForeignEntryKind = _foeDisallowed) then
  begin
    Inc(Result);
    Log(['=> ChangeForeignEntry: You can only change the Foreign Entry of a faction using Foreign Entry.']);
  end;
  if Result <> 0 then Exit;

  Value := STI(Params[0]);
  if cuFaction.FactionDef.ForeignEntryKind = _foeCounterer then
  begin
    Faction := cuFaction.FindStrongestEnemy();
    Value := -Value;
    if (Faction <> nil) and (Faction.FactionDef.ForeignEntryKind <> _foeAllowed) then
    begin
      Log(['[WARNING] ChangeForeignEntry: The faction don''t have an enemy with allowed Foreign Entry, ignored.']);
      Exit;
    end;
  end
  else
    Faction := cuFaction; // _foeAllowed

  Faction.ForeignEntry := Faction.ForeignEntry + Value;

  Log(['Foreign Entry for ', Faction.Name, ' = ', Faction.ForeignEntry]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ADD HARBOR NO BLOCKADE
// Ajout d'une liste de rgions (paires de donnes) qui modifient le % de blocus par +- une valeur, si le port
// est ami et hors blocus.
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.AddHarborNoBlock(Params : TStringList) : integer;
var RgnUID : TUID;
    Coeff : integer;
    Idx, i : integer;
begin
  Result := 0;
  if (cuFaction = nil) then
  begin
    Log(['[Warning] AddHarborNoBlock: No faction selected']);
    Exit;
  end;
  if Params.Count mod 2 <> 0 then
  begin
    Inc(Result);
    Log(['=> AddHarborNoBlock: Need to give pairs of Region UID & modifier.']);
    Exit;
  end;
  if Result <> 0 then Exit;

  for i := 1 to Params.Count div 2 do
  begin
    RgnUID := STI(Params[(i-1)*2]);
    Coeff  := STI(Params[(i-1)*2 + 1]);

    if (RgnUID < 0)or(RgnUID > RegionDefs.Count-1) then
    begin
      Inc(Result);
      Log(['=> AddHarborNoBlock: The Region UID is not valid.']);
      Exit;
    end;

    Idx := cuFaction.dlHarborNoBlockade.Keys.IndexOf(ITS(RgnUID));
    if Idx <> -1 then
      cuFaction.dlHarborNoBlockade.Delete(Idx);

      // permet ainsi de supprimer des entres en entrant 0 sur une entre dclare
    if Coeff <> 0 then cuFaction.dlHarborNoBlockade.Add(RgnUID, Coeff);

    Log(['AddHarborNoBlock:', cuFaction.Name, RegionDefs[RgnUID].Name, 'Blockade modifier set to:', Coeff]);
  end;
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ADD HARBOR BLOCKADED
// Ajout d'une liste de rgions (paires de donnes) qui modifient le % de blocus par +- une valeur, si le port
// est ami et sous blocus.
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.AddHarborBlockaded(Params : TStringList) : integer;
var RgnUID : TUID;
    Coeff : integer;
    Idx, i : integer;
begin
  Result := 0;
  if (cuFaction = nil) then
  begin
    Log(['[Warning] AddHarborBlockaded: No faction selected']);
    Exit;
  end;
  if Params.Count mod 2 <> 0 then
  begin
    Inc(Result);
    Log(['=> AddHarborBlockaded: Need to give pairs of Region UID & modifier.']);
    Exit;
  end;
  if Result <> 0 then Exit;

  for i := 1 to Params.Count div 2 do
  begin
    RgnUID := STI(Params[(i-1)*2]);
    Coeff  := STI(Params[(i-1)*2 + 1]);

    if (RgnUID < 0)or(RgnUID > RegionDefs.Count-1) then
    begin
      Inc(Result);
      Log(['=> AddHarborBlockaded: The Region UID is not valid.']);
      Exit;
    end;

    Idx := cuFaction.dlHarborBlockaded.Keys.IndexOf(ITS(RgnUID));
    if Idx <> -1 then
      cuFaction.dlHarborBlockaded.Delete(Idx);

      // permet ainsi de supprimer des entres en entrant 0 sur une entre dclare
    if Coeff <> 0 then cuFaction.dlHarborBlockaded.Add(RgnUID, Coeff);

    Log(['AddHarborBlockaded:', cuFaction.Name, RegionDefs[RgnUID].Name, 'Blockade modifier set to:', Coeff]);
  end;
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET PROMOTE FLAG
// Met  1 ou 0 le flag SubUnit CanUpgrade pour le leader (en tant que SU) de
// l'unit en slection (fonctionne si leader intgr)
//
// TODO: A gnraliser en SetUpgradeFlag, quand on aura des slecteurs adquats sur SubUnits
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetPromoteFlag(Params : TStringList) : integer;
begin
  Result := 0;
  if (cuUnit = nil) then
  begin
    Log(['[Warning] SetPromoteFlag: No unit selected']);
    Exit;
  end;
  if (Params.Count <> 1) then
  begin
    Inc(Result);
    Log(['=> SetPromoteFlag: Need 1 or 0 as value.']);
    Exit;
  end;
  if Result <> 0 then Exit;
  if (cuUnit.GetLeader = nil) then
  begin
    Log(['[Warning] SetPromoteFlag: No leader in selected unit.']);
    Exit;
  end;

  cuUnit.GetLeader.CanUpgrade := STI(Params[0]);

  Log(['SetPromoteFlag:', cuUnit.GetLeader.Name, 'in unit', cuUnit.Name, 'has the promote flag set to', IIF(cuUnit.GetLeader.CanUpgrade = 1, 'True', 'False')]);

end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET CAPITALS
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetCapitals(Params : TStringList) : integer;
var RgnUID : TUID;
    Coeff : integer;
    i : integer;
begin
  Result := 0;
  if (cuFaction = nil) then
  begin
    Log(['=> SetCapitals: No faction selected']);
    Inc(Result);
    Exit;
  end;
  if Params.Count mod 2 <> 0 then
  begin
    Inc(Result);
    Log(['=> SetCapitals: Need to give pairs of Region UID & coefficient.']);
    Exit;
  end;
  if Result <> 0 then Exit;

  cuFaction.dlCapitals.Clear;

  for i := 1 to Params.Count div 2 do
  begin
    RgnUID := STI(Params[(i-1)*2]);
    Coeff  := STI(Params[(i-1)*2 + 1]);

    if (RgnUID < 0)or(RgnUID > RegionDefs.Count-1) then
    begin
      Inc(Result);
      Log(['=> SetCapitals: The Region UID is not valid.']);
      Exit;
    end;

    cuFaction.AlterCapital(RgnUID, Coeff);
    if Coeff <> 0 then
      Log(['SetCapitals:', cuFaction.Name, RegionDefs[RgnUID].Name, 'is a capital region with coefficient set to:', Coeff]);
  end;
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ADD XP LEVEL TO GROUP/UNIT
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.AddXPLevelToGroup (Params : TStringList) : integer;
var i : integer;
    EvenToLeaders, OnlyToLeaders : boolean;
begin
  Result := 0;
  if (cuGroup = nil) then
  begin
    Log(['[WARNING] AddXPToGroup: No group selected']);
    Exit;
  end;
  if not (Params.Count in [1..2]) then
  begin
    Inc(Result);
    Log(['=> AddXPToGroup: Syntax is AddXPToGroup = NbOfLevels [;EvenToLeaders] [;OnlyToLeaders]']);
    Exit;
  end;
  EvenToLeaders := ExtractParam('EvenToLeaders', Params) <> '';
  OnlyToLeaders := ExtractParam('OnlyToLeaders', Params) <> '';
  for i := 0 to cuGroup.Count-1 do
    cuGroup[i].AddExperienceLevels(STI(Params[0]), EvenToLeaders, OnlyToLeaders);
end;

function TScriptEngine.AddXPLevelToUnit(Params : TStringList) : integer;
var EvenToLeaders, OnlyToLeaders : boolean;
begin
  Result := 0;
  if (cuUnit = nil) then
  begin
    Log(['[WARNING] AddXPToUnit: No unit selected']);
    Exit;
  end;
  if not (Params.Count in [1..2]) then
  begin
    Inc(Result);
    Log(['=> AddXPToUnit: Syntax is AddXPToUnit = NbOfLevels [;EvenToLeaders] [;OnlyToLeaders]']);
    Exit;
  end;
  EvenToLeaders := ExtractParam('EvenToLeaders', Params) <> '';
  OnlyToLeaders := ExtractParam('OnlyToLeaders', Params) <> '';

  cuUnit.AddExperienceLevels(STI(Params[0]), EvenToLeaders, OnlyToLeaders);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// CHANGE GROUP/UNIT HEALTH
// Change par un % les hits des SU, sans les tuer ni dpasser leur max
// -10 = -10% du MaxHits aux hits courant
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.ChgGroupHealth(Params : TStringList) : integer;
var i : integer;
    EvenToLeaders, OnlyToLeaders : boolean;
begin
  Result := 0;
  if (cuGroup = nil) then
  begin
    Log(['[WARNING] ChgGroupHealth: No group selected']);
    Exit;
  end;
  if not (Params.Count in [1..2]) then
  begin
    Inc(Result);
    Log(['=> ChgGroupHealth: Syntax is ChgGroupHealth = HealthVariationPerc [;EvenToLeaders] [;OnlyToLeaders]']);
    Exit;
  end;
  EvenToLeaders := ExtractParam('EvenToLeaders', Params) <> '';
  OnlyToLeaders := ExtractParam('OnlyToLeaders', Params) <> '';
  for i := 0 to cuGroup.Count-1 do
    cuGroup[i].ChangeUnitHealth(STI(Params[0]), EvenToLeaders, OnlyToLeaders);

  Log(['ChgGroupHealth:', cuGroup.UID, cuGroup.Name, cuGroup.StartRegionName, 'health changed by', STI(Params[0]), '%']);
end;

function TScriptEngine.ChgUnitHealth(Params : TStringList) : integer;
var EvenToLeaders, OnlyToLeaders : boolean;
begin
  Result := 0;
  if (cuUnit = nil) then
  begin
    Log(['[WARNING] ChgUnitHealth: No unit selected']);
    Exit;
  end;
  if not (Params.Count in [1..2]) then
  begin
    Inc(Result);
    Log(['=> Change Unit Health: Syntax is ChgUnitHealth = HealthVariationPerc [;EvenToLeaders] [;OnlyToLeaders]']);
    Exit;
  end;
  EvenToLeaders := ExtractParam('EvenToLeaders', Params) <> '';
  OnlyToLeaders := ExtractParam('OnlyToLeaders', Params) <> '';

  cuUnit.ChangeUnitHealth(STI(Params[0]), EvenToLeaders, OnlyToLeaders);
  Log(['ChgUnitHealth:', cuUnit.UID, cuUnit.Name, TGroup(cuUnit.pGroup).StartRegionName, 'health changed by', STI(Params[0]), '%']);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET BUILD FLAG
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetBuildFlag(Params : TStringList) : integer;
begin
  Result := 0;
  if (cuUnit = nil) then
  begin
    Log(['[WARNING] SetBuildFlag: No unit selected']);
    Exit;
  end;
  if not (Params.Count in [1]) then
  begin
    Inc(Result);
    Log(['=> SetBuildFlag: Syntax is SetBuildFlag = 1 or 0']);
    Exit;
  end;

  cuUnit.SetBuildFlag(Params[0] = '1');
  Log(['SetBuildFlag:', cuUnit.UID, cuUnit.Name, TGroup(cuUnit.pGroup).StartRegionName, 'Build flag is now', IIF(Params[0] = '1', 'True', 'False')]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// CHANGE GROUP/UNIT COHESION
// Change par un % la cohesion des SU, en bornant entre 1 et leur max
// -10 = -10% de la Cohesion Max  la cohesion courante
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.ChgGroupCohesion(Params : TStringList) : integer;
var i : integer;
    EvenToLeaders, OnlyToLeaders : boolean;
begin
  Result := 0;
  if (cuGroup = nil) then
  begin
    Log(['[WARNING] ChgGroupCohesion: No group selected']);
    Exit;
  end;
  if not (Params.Count in [1..2]) then
  begin
    Inc(Result);
    Log(['=> ChgGroupCohesion: Syntax is ChgGroupCohesion = CohVariationPerc [;EvenToLeaders] [;OnlyToLeaders]']);
    Exit;
  end;
  EvenToLeaders := ExtractParam('EvenToLeaders', Params) <> '';
  OnlyToLeaders := ExtractParam('OnlyToLeaders', Params) <> '';
  for i := 0 to cuGroup.Count-1 do
    cuGroup[i].ChangeUnitCohesion(STI(Params[0]), EvenToLeaders, OnlyToLeaders);
  Log(['ChgGroupCohesion:', cuGroup.UID, cuGroup.Name, cuGroup.StartRegionName, 'cohesion changed by', STI(Params[0]), '%']);
end;

function TScriptEngine.ChgUnitCohesion(Params : TStringList) : integer;
var EvenToLeaders, OnlyToLeaders : boolean;
begin
  Result := 0;
  if (cuUnit = nil) then
  begin
    Log(['[WARNING] ChgUnitCohesion: No unit selected']);
    Exit;
  end;
  if not (Params.Count in [1..2]) then
  begin
    Inc(Result);
    Log(['=> ChgUnitCohesion: Syntax is ChgUnitCohesion = CohVariationPerc [;EvenToLeaders] [;OnlyToLeaders]']);
    Exit;
  end;
  EvenToLeaders := ExtractParam('EvenToLeaders', Params) <> '';
  OnlyToLeaders := ExtractParam('OnlyToLeaders', Params) <> '';

  cuUnit.ChangeUnitCohesion(STI(Params[0]), EvenToLeaders, OnlyToLeaders);
  Log(['ChgUnitCohesion:', cuUnit.UID, cuUnit.Name, TGroup(cuUnit.pGroup).StartRegionName, 'cohesion changed by', STI(Params[0]), '%']);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ALTER CURRENT SUB UNIT (S)
// Change les proprits de la cuSubUnit en slection
// _AlterCubSubUnit ApplyToList|Kill
// _AlterCubSubUnits autoris
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.AlterCuSubUnit(Params : TStringList) : integer;
var Param         : string;
    iArg,i,j, XP  : integer;
    sub           : integer;
    Probability   : integer;
    SL            : TStringList;
    Attempts      : integer;
    Bag           : TRandomPntBag;
    SkipProba     : integer;
    DoSkip        : boolean;
    TgtRegion     : TRegion;
begin
  Result := 0;

  if Params.Count = 0 then
  begin
    Inc(Result);
    Log(['=> AlterCuSubUnit: Syntax is AlterCuSubUnit = [ApplyToList] [;Probability Value] [;Attempts Value] [;Convert NewUID/PrimUpg/SecUpg [ NewName] ]']);
    Log(['=> [ChgSeniority Value] [;ChgPtsHealth ValueInHits] [;ChgPtsCohesion ValueInPts] [;ChgXPLevelValue]']);
    Log(['=> [;Kill] [;RerollActivation] [;SetFixing Value][;SetParentGroupFixing Value]']);
    Log(['=> [SetHealthPerc Value] [;SetCohesionPerc Value] [;ChgHealthPerc Value] [;ChgCohesionPerc Value] [;Move Out/In/RgnUID]']);
    Log(['=> [;LeaderActive] [;LeaderUnactive] [;SetParentGroupTrench Value]']);
    Log(['=> [SkipAttribs SkipProba Attrib1 Attrib2 ...] [;SkipUnitAttribs SkipProba Attrib1 Attrib2 ...']);
    Exit;                                                
  end;

  // on regarde si on traite la liste cuSUList ou la SU courante
  Param := ExtractParam('ApplyToList', Params);
  if Param = '' then
  begin
    if (cuSubUnit = nil) then
    begin
      Log(['[WARNING] AlterCuSubUnit: No Sub Unit selected']);
      Exit;
    end;

    cuSUList.Clear;
    cuSUList.Add(cuSubUnit)
  end;
  // autrement ne rien faire, tester cependant s'il y a un lment dans la liste (sortir si non)
  if cuSUList.Count = 0 then
  begin
    Log(['[WARNING] AlterCuSubUnit: Selection is empty (Apply To List)']);
    Exit;
  end;

  Probability := 100;
  Param := ExtractParam('Probability', Params);
  if Param <> '' then
  begin
    ExtractArgs(Param);
    if Args[0] = '' then
    begin
      Inc(Result);
      Log(['=> AlterCuSubUnit, Probability specified but missing argument']);
      Exit;
    end;
    Probability := STI(Args[0]);
    Log(['AlterCuSubUnit', 'Probability set to:', Probability]);
  end;

  Attempts := cuSUList.Count;
  Param := ExtractParam('Attempts', Params);
  if Param <> '' then
  begin
    ExtractArgs(Param);
    if Args[0] = '' then
    begin
      Inc(Result);
      Log(['=> AlterCuSubUnit, Attempts specified but missing argument']);
      Exit;
    end;
    Attempts := Max(0, STI(Args[0]));
    Log(['AlterCuSubUnit', 'Attempts set to:', Attempts]);
  end;


  SL := TStringList.Create;

  // Randomiser la liste
  Bag := TRandomPntBag.Create;
  for sub := 0 to cuSUList.Count-1 do
    Bag.Add(cuSUList[sub]);

  // Parcourir la liste, impacter les changements
  cuSubUnit := nil;
  cuUnit    := nil;
  cuGroup   := nil;
  cuRegion  := nil;

  while Bag.Count > 0 do
  begin

    if Attempts <= 0 then
    begin
      Log(['AlterCuSubUnit', 'All Attempts done, breaking loop']);
      Break;
    end
    else
      Dec(Attempts);

    if Dice(100) > Probability then Continue;

    cuSubUnit := Bag.PickPointer;
    cuUnit    := TUnit(cuSubUnit.pUnit);
    if cuUnit = nil then Continue; // abnormal
    cuGroup   := TGroup(cuSubUnit.pGroup);
    if cuGroup = nil then Continue; // abnormal
    cuRegion  := cuGroup.pStartRegion;
    cuFaction := TFaction(cuSubUnit.pFaction);

    // Skip the effect if some others SUs have some specific attributes in the parent ** group **?
    Param := ExtractParam('SkipAttribs', Params);
    DoSkip := False;
    if Param <> '' then
    begin

      ExtractArgs(Param);
      if Args[0] = '' then // Args 0 = probability to skip
      begin
        Inc(Result);
        Log(['=> AlterCuSubUnit, SkipAttribs specified but missing argument for Probability']);
        Exit;
      end;
      SkipProba := Max(0, STI(Args[0]));

      // Each args is an attribute which can allow the skip
      for i := 1 to High(Args) do
      begin
        if Args[1] = '' then // Args 1 = 1st attribute
        begin
          Inc(Result);
          Log(['=> AlterCuSubUnit, SkipAttribs specified but missing first attribute']);
          Exit;
        end;
        if Args[i] = '' then Break;
        if cuGroup.HasThisAttribute(Args[i]) and (Dice(100) <= SkipProba) then
        begin
          DoSkip := True;
          Log(['SkipAttribs in effect for', cuSubUnit.UID, cuSubUnit.Name, 'in group', cuGroup.UID, cuGroup.Name, 'SkipProba:', SkipProba, 'Triggering attribute:', Args[i]]);
          Break;
        end;
      end;
    end;
    // no effect on this SU, 'skip effect' in action
    if DoSkip then Continue;
        
    // Skip the effect if some others SUs have some specific attributes in the parent ** unit **?
    Param := ExtractParam('SkipUnitAttribs', Params);
    DoSkip := False;
    if Param <> '' then
    begin

      ExtractArgs(Param);
      if Args[0] = '' then // Args 0 = probability to skip
      begin
        Inc(Result);
        Log(['=> AlterCuSubUnit, SkipUnitAttribs specified but missing argument for Probability']);
        Exit;
      end;
      SkipProba := Max(0, STI(Args[0]));

      // Each args is an attribute which can allow the skip
      for i := 1 to High(Args) do
      begin
        if Args[1] = '' then // Args 1 = 1st attribute
        begin
          Inc(Result);
          Log(['=> AlterCuSubUnit, SkipUnitAttribs specified but missing first attribute']);
          Exit;
        end;
        if Args[i] = '' then Break;
        if cuUnit.HasThisAttribute(Args[i]) and (Dice(100) <= SkipProba) then
        begin
          DoSkip := True;
          Log(['SkipUnitAttribs in effect for', cuSubUnit.UID, cuSubUnit.Name, 'in group', cuGroup.UID, cuGroup.Name, 'Unit', cuUnit.UID, cuUnit.Name, 'SkipProba:', SkipProba, 'Triggering attribute:', Args[i]]);
          Break;
        end;
      end;
    end;

    // no effect on this SU, 'skip effect' in action
    if DoSkip then Continue;
        
    // Destruction
    Param := ExtractParam('Kill', Params);
    if (Param <> '') and (cuSubUnit <> nil) then
    begin
      Log(['AlterCuSubUnit, killed this SubUnit', cuSubUnit.UID, cuSubUnit.Name]);
      RemoveCuSubUnit(SL);
      cuSubUnit := nil;
      // no nil on others selector, done if needed in RemoveCuSubUnit
      Continue;
    end;

    // model change and upgrade 
    Param := ExtractParam('Convert', Params);
    if Param = '' then
      Param := ExtractParam('ModelUID', Params);
    if Param = '' then
      Param := ExtractParam('Model', Params);

    if Param <> '' then
    begin
      ExtractArgs(Param);
      if Args[0] = '' then
      begin
        Inc(Result);
        Log(['=> AlterCuSubUnit, ModelUID specified but erroneous arguments']);
        Exit;
      end;

      if (UpperCase(Args[0]) = 'PRIMUPG')or(UpperCase(Args[0]) = 'TECHUPG') then
        iArg := cuSubUnit.Model.TechUpg
      else
      if (UpperCase(Args[0]) = 'SECUPG')or(UpperCase(Args[0]) = 'TRAINUPG') then
        iArg := cuSubUnit.Model.TrainUpg
      else
      if (UpperCase(Args[0]) = 'SELF') then
        iArg := cuSubUnit.Model.UID // we just want to rename
      else
        iArg := STI(Args[0]);

      if (iArg <> _noUID) and (iArg <> cuSubUnit.Model.UID) then
      begin
        cuSubUnit.ChangeModelType(iArg);
        Log(['AlterCuSubUnit', cuSubUnit.UID, cuSubUnit.Name, 'changed to model', iArg]);
      end
      else
        Log(['AlterCuSubUnit', cuSubUnit.UID, cuSubUnit.Name, 'Upgrade could not be done, no upgrade path, found:', iArg]);

      // rename
      if Args[1] <> '' then
        cuSubUnit.Name := Args[1];
    end;

    // Seniority
    Param := ExtractParam('ChgSeniority', Params);
    if (cuSubUnit.IsLeader) and (Param <> '') then
    begin
      ExtractArgs(Param);
      if Args[0] = '' then
      begin
        Inc(Result);
        Log(['AlterCuSubUnit, ChgSeniority specified but missing argument']);
        Exit;
      end;
      iArg := STI(Args[0]);
      cuSubUnit.SeniorityChange := cuSubUnit.SeniorityChange + iArg;
      Log(['AlterCuSubUnit', cuSubUnit.UID, cuSubUnit.Name, 'seniority changed to ', cuSubUnit.Seniority]);
    end;

    Param := ExtractParam('RerollActivation', Params);
    if (cuSubUnit.IsLeader) and (Param <> '') and (cuUnit.Activation = 0) then
    begin
      SL.Clear;
      SL.Add('REROLL');
      SetLeaderActivity(SL); // au niveau unit
      Log(['AlterCuSubUnit', cuSubUnit.UID, cuSubUnit.Name, 'has its activation rerolled, now:', cuUnit.Activation]);
    end;
    Param := ExtractParam('LeaderActive', Params);
    if (cuSubUnit.IsLeader) and (Param <> '')  then
    begin
      SL.Clear;
      SL.Add('1');
      SetLeaderActivity(SL); // au niveau unit
      Log(['AlterCuSubUnit', cuSubUnit.UID, cuSubUnit.Name, 'is now activated:', cuUnit.Activation]);
    end;
    Param := ExtractParam('LeaderUnactive', Params);
    if (cuSubUnit.IsLeader) and (Param <> '') then
    begin
      SL.Clear;
      SL.Add('0');
      SetLeaderActivity(SL); // au niveau unit
      Log(['AlterCuSubUnit', cuSubUnit.UID, cuSubUnit.Name, 'is now unactive:', cuUnit.Activation]);
    end;

    Param := ExtractParam('ChgPtsHealth', Params);
    if Param <> '' then
    begin
      ExtractArgs(Param);
      if Args[0] = '' then
      begin
        Inc(Result);
        Log(['=> AlterCuSubUnit, ChgPtsHealth specified but missing argument']);
        Exit;
      end;
      iArg := STI(Args[0]);
      cuSubUnit.Health := cuSubUnit.Health + iArg;
      if cuSubUnit.Health = 0 then cuSubUnit.Health := 1; // le code permet 0 (pour tuer la SU aprs), ici non
      Log(['AlterCuSubUnit', cuSubUnit.UID, cuSubUnit.Name, 'health changed to ', cuSubUnit.Health]);
    end;

    Param := ExtractParam('ChgHealthPerc', Params);
    if Param <> '' then
    begin
      ExtractArgs(Param);
      if Args[0] = '' then
      begin
        Inc(Result);
        Log(['=> AlterCuSubUnit, ChgHealthPerc specified but missing argument']);
        Exit;
      end;
      iArg := STI(Args[0]);
      cuSubUnit.ChangeHealthPerc(iArg);
      if cuSubUnit.Health = 0 then cuSubUnit.Health := 1; // already done in function
      Log(['AlterCuSubUnit', cuSubUnit.UID, cuSubUnit.Name, 'health changed to ', cuSubUnit.Health]);
    end;

    Param := ExtractParam('SetHealthPerc', Params);
    if Param <> '' then
    begin
      ExtractArgs(Param);
      if Args[0] = '' then
      begin
        Inc(Result);
        Log(['=> AlterCuSubUnit, SetHealthPerc specified but missing argument']);
        Exit;
      end;
      iArg := STI(Args[0]);
      cuSubUnit.SetHealthPerc(iArg);
      if cuSubUnit.Health = 0 then cuSubUnit.Health := 1; // already done in function
      Log(['AlterCuSubUnit', cuSubUnit.UID, cuSubUnit.Name, 'health changed to ', cuSubUnit.Health]);
    end;

    Param := ExtractParam('ChgPtsCohesion', Params);
    if Param <> '' then
    begin
      ExtractArgs(Param);
      if Args[0] = '' then
      begin
        Inc(Result);
        Log(['=> AlterCuSubUnit, ChgPtsCohesion specified but missing argument']);
        Exit;
      end;
      iArg := STI(Args[0]);
      cuSubUnit.cuCohesion := cuSubUnit.cuCohesion + iArg;
      Log(['AlterCuSubUnit', cuSubUnit.UID, cuSubUnit.Name, 'cohesion changed to ', cuSubUnit.cuCohesion]);
    end;

    Param := ExtractParam('ChgCohesionPerc', Params);
    if Param <> '' then
    begin
      ExtractArgs(Param);
      if Args[0] = '' then
      begin
        Inc(Result);
        Log(['=> AlterCuSubUnit, ChgCohesionPerc specified but missing argument']);
        Exit;
      end;
      iArg := STI(Args[0]);
      cuSubUnit.ChangeCohesionPerc(iArg);
      Log(['AlterCuSubUnit', cuSubUnit.UID, cuSubUnit.Name, 'cohesion changed to ', cuSubUnit.cuCohesion]);
    end;

    Param := ExtractParam('SetCohesionPerc', Params);
    if Param <> '' then
    begin
      ExtractArgs(Param);
      if Args[0] = '' then
      begin
        Inc(Result);
        Log(['=> AlterCuSubUnit, SetCohesionPerc specified but missing argument']);
        Exit;
      end;
      iArg := STI(Args[0]);
      cuSubUnit.SetCohesionPerc(iArg);
      Log(['AlterCuSubUnit', cuSubUnit.UID, cuSubUnit.Name, 'cohesion changed to ', cuSubUnit.cuCohesion]);
    end;

    Param := ExtractParam('ChgXPLevel', Params);
    if Param <> '' then
    begin
      ExtractArgs(Param);
      if Args[0] = '' then
      begin
        Inc(Result);
        Log(['=> AlterCuSubUnit, ChgXPLevel specified but missing argument']);
        Exit;
      end;
      iArg := STI(Args[0]);
      // gain d'exprience
      for j := 1 to iArg do
      begin
        XP := cuSubUnit.GetNextLevelCost;
        if XP <> _expImpLevelCost then cuSubUnit.Exp := XP;
      end;
      Log(['AlterCuSubUnit', cuSubUnit.UID, cuSubUnit.Name, 'level changed to ', cuSubUnit.GetLevel]);
    end;

    Param := ExtractParam('SetFixing', Params);
    if Param <> '' then
    begin
      ExtractArgs(Param);
      if Args[0] = '' then
      begin
        Inc(Result);
        Log(['=> AlterCuSubUnit, SetFixing specified but missing argument']);
        Exit;
      end;
      iArg := STI(Args[0]);
      TUnit(cuSubUnit.pUnit).FixType := iArg;
      Log(['AlterCuSubUnit', 'Unit', TUnit(cuSubUnit.pUnit).Name, 'parent of:', cuSubUnit.UID, cuSubUnit.Name, 'fixation type changed to ', iArg]);
    end;

    Param := ExtractParam('SetParentGroupFixing', Params);
    if Param <> '' then
    begin
      ExtractArgs(Param);
      if Args[0] = '' then
      begin
        Inc(Result);
        Log(['=> AlterCuSubUnit, SetParentGroupFixing specified but missing argument']);
        Exit;
      end;
      iArg := STI(Args[0]);
      TGroup(cuSubUnit.pGroup).Fix(iArg);
      Log(['AlterCuSubUnit', 'Group', TGroup(cuSubUnit.pGroup).Name, 'parent of:', cuSubUnit.UID, cuSubUnit.Name, 'fixation type changed to ', iArg]);
    end;

    Param := ExtractParam('SetParentGroupTrench', Params);
    if Param <> '' then
    begin
      ExtractArgs(Param);
      if Args[0] = '' then
      begin
        Inc(Result);
        Log(['=> AlterCuSubUnit, SetParentGroupTrench specified but missing argument']);
        Exit;
      end;
      iArg := STI(Args[0]);
      TGroup(cuSubUnit.pGroup).EntrenchLevel := iArg;
      Log(['AlterCuSubUnit', 'Group', TGroup(cuSubUnit.pGroup).Name, 'parent of:', cuSubUnit.UID, cuSubUnit.Name, 'entrenchment changed to ', iArg]);
    end;

    Param := ExtractParam('Move', Params);
    if Param <> '' then
    begin
      ExtractArgs(Param);
      if Args[0] = '' then
      begin
        Inc(Result);
        Log(['=> AlterCuSubUnit, Move specified but missing argument']);
        Exit;
      end;
      if (UpperCase(Args[0]) = 'IN') and (cuRegion.Structures.GetFirstCSStruc <> nil) and (cuGroup.Kind = _gkLand) then
      begin
        cuGroup.InCityScape := 1;
        Log(['AlterCuSubUnit', 'Stack', cuGroup.Name, 'parent of:', cuSubUnit.UID, cuSubUnit.Name, 'moved inside']);
      end
      else
      if (UpperCase(Args[0]) = 'OUT') and (cuGroup.Kind = _gkLand) then
      begin
        cuGroup.InCityScape := 0;
        Log(['AlterCuSubUnit', 'Stack', cuGroup.Name, 'parent of:', cuSubUnit.UID, cuSubUnit.Name, 'moved outside']);
      end
      else
      begin
        iArg := STI(Args[0], True, Low(Integer));
        if (iArg > Low(integer))and (cuGroup.StartSubSpace <> iArg) then
        begin
          TgtRegion := cuFaction.Regions.Fetch(iArg);
          if (cuGroup.Kind = _gkLand) and RegionDefs.IsLandOrFrozenWaterRegion(iArg, GameSim.Weather.GetWeather(iArg)) then
          begin
            //j := cuGroup.StartSubSpace;
            TgtRegion.DisplaceGroup(cuGroup);
            //GameSim.RegionCmns.Update(j);
            //GameSim.RegionCmns.Update(iArg);
            Log(['AlterCuSubUnit', 'Land Stack', cuGroup.Name, 'parent of:', cuSubUnit.UID, cuSubUnit.Name, 'moved to ', RegionDefs[iArg].UID, RegionDefs[iArg].Name]);
          end
          else
          if (cuGroup.Kind = _gkNav) and
          (RegionDefs.IsNonFrozenWaterRegion(iArg, GameSim.Weather.GetWeather(iArg)) or (GameSim.RegionCmns.GetDockableHarbor(cuFaction, iArg) <> nil)) then
          begin
            //j := cuGroup.StartSubSpace;
            TgtRegion.DisplaceGroup(cuGroup);
            //GameSim.RegionCmns.Update(j);
            //GameSim.RegionCmns.Update(iArg);
            Log(['AlterCuSubUnit', 'Fleet', cuGroup.Name, 'parent of:', cuSubUnit.UID, cuSubUnit.Name, 'moved to ', RegionDefs[iArg].UID, RegionDefs[iArg].Name]);
          end
        end;
      end;

    end;
  end;

  Bag.Free;
  SL.Free;
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET LEADER ACTIVITY
// Met le flag Active d'une unit  1 ou 0 (n'est pertinent que si l'unit  un leader)
// REROLL: have the leader reroll for activity (if failed)
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetLeaderActivity(Params : TStringList) : integer;
begin
  Result := 0;

  if not (Params.Count in [1]) then
  begin
    Inc(Result);
    Log(['=> SetLeaderActivity: Syntax is SetLeaderActivity = [1 or 0 or REROLL]']);
    Exit;
  end;
  if cuUnit = nil then
  begin
    Log(['[Warning] SetLeaderActivity: No Selected unit']);
  end;
  if cuSubUnit = nil then
  begin
    Log(['[Warning] SetLeaderActivity: No Selected unit']);
    Exit;
  end;

  // We take care that the selectors are ok, this is an old BOA command acting on a unit
  // but is now used by selecting a SU
  if (cuUnit = nil)and(cuSubUnit <> nil) then
    cuUnit := TUnit(cuSubUnit.pUnit);
  if (cuUnit <> TUnit(cuSubUnit.pUnit)) then
    cuUnit := TUnit(cuSubUnit.pUnit);

  cuGroup := TGroup(cuSubUnit.pGroup);
  cuFaction := cuGroup.pFaction;

  if (Params[0] = 'REROLL') then
  begin
    cuUnit.CalcActivation(True) // no reroll if already activated
  end
  else
    cuUnit.Activation := STI(Params[0]);
end;


// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SUBUNIT COUNT
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function  TScriptEngine.SubUnitCount(Params : TStringList) : integer;
var Value : integer;
begin
  Result := 0;

  if (Params.Count <> 1) then
  begin
    Inc(Result);
    Log(['=> SubUnitCount command, syntax is: SubUnitCount = <esvIntVar>']);
  end;

  if (cuFaction = nil) then
  begin
    Inc(Result);
    Log(['=> SubUnitCount: No faction selected']);
  end;

  if not IsExposedVar(Params[0]) then
  begin
    Inc(Result);
    Log(['=> SubUnitCount: Parameter not recognized as an exposed var', Params[0]]);
  end;

  if Result > 0 then Exit;

  Value := cuSUList.Count;
  StoreToExposedVar(Params[0], Value);

  Log(['SubUnitsCount:', cuFaction.Name, 'possess: ', Value, ' SU, value stored in', Params[0]]);

end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SetSupplyMod
// Sets global modifiers for the supply production/consumption
// Only for WIA (no advanced supply)
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetSupplyMod(Params : TStringList) : integer;
begin

  Result := 0;

  if (Params.Count <> 2) then
  begin
    Inc(Result);
    Log(['=> SetSupplyMod command: Syntax is: SetSupplyMod = <Supply Production Modifier in per cent>;<Supply Consumption Modifier in per cent>']);
  end;

  if Result > 0 then Exit;

  supGlobSupplyModProd_   := STI(Params[0]);
  supGlobSupplyModConsum_ := STI(Params[1]);

  Log(['SetSupplyMod: Global supply production changed to', Params[0],'%, Global supply consumption changed to', Params[0], '%']);

end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// CREATE CRISIS
// Right now for debug and testing of the crisis module
// To add: Parameters for the mandatory agenda
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.CreateCrisis(Params : TStringList) : integer;
(*
var FacInitiator    : TFaction;
    FacReceiver     : TFaction;
    FacMediator     : TFaction;
    MandatoryCI     : TDiploItem;
    MandatoryCIDef  : TDiplomacyDef;
    CrisisIntensity : integer;
    InitiatorJC     : boolean;
    Crisis          : TCrisis;
    FacCrisisItem   : TCrisis;
    GS              : TGameSim;

*)
begin

  Result := 0;
(*
//  FacInitiator  := nil;
//  FacReceiver   := nil;
  FacMediator   := nil;

  GS := TGameSim(Owner);

  try
    try

      if (Params.Count <> 5 ) and (Params.Count <> 6) then
      begin
        Inc(Result);
        Log(['=> CreateCrisis command: Syntax is: CreateCrisis = <Fac Tag Initiator>;<Fac Tag Receiver>;[Fac Tag Mediator];<Mandatory Agenda>;<Crisis Intensity>;<Initiator Just Cause (True/False)>']);
        Exit;
      end;


      FacInitiator := FactionTagToFaction(Params[0]);
      if FacInitiator = nil then
      begin
        Inc(Result);
        Log(['=> CreateCrisis, the Faction (Initiator)', Params[0], 'is unknown.']);
      end;


      FacReceiver := FactionTagToFaction(Params[1]);
      if FacReceiver = nil then
      begin
        Inc(Result);
        Log(['=> CreateCrisis, the Faction (Receiver)', Params[1], 'is unknown.']);
      end;

      if Params.Count = 6 then
      begin
        FacMediator := FactionTagToFaction(Params[2]);
        if FacMediator = nil then
        begin
          Inc(Result);
          Log(['=> CreateCrisis, the Faction (Mediator)', Params[2], 'is unknown.']);
        end;
      end;


      MandatoryCI := TDiploItem.Create(nil);
      if Params.Count = 5 then
        MandatoryCIDef := GS.DiplomacyDB[STI(Params[2])]
      else
        MandatoryCIDef := GS.DiplomacyDB[STI(Params[3])];

      if MandatoryCIDef = nil then
      begin
        Inc(Result);
        Log(['=> CreateCrisis, the Mandatory Agenda is unknown. (Definition not found, check if UID or Alias is correct.)']);
      end
      else
      begin
        MandatoryCI.UID := MandatoryCIDef.UID;
        MandatoryCI.Name := MandatoryCIDef.Name;
        MandatoryCI.Kind := MandatoryCIDef.Kind;
      end;

      if Params.Count = 5 then
        CrisisIntensity := STI(Params[3])
      else
        CrisisIntensity := STI(Params[4]);

      if (CrisisIntensity < 0) or (CrisisIntensity > 100) then
      begin
       Inc(Result);
       Log(['=> CreateCrisis, Crisis Intensity out of valid range. (0...100)', 'CrisisIntensity:', CrisisIntensity]);
      end;

      if lowercase(Trim(Params[Params.Count -1])) = 'true' then
        InitiatorJC := True
      else
        InitiatorJC := False;

      if Result > 0 then Exit;

      // The Crisis Item is created. Every involved Faction will get a copy.
      Crisis := TCrisis.Create(nil); // The owner is not copied by the Assign method, which is used later (thus 'nil' is OK). Owner is the Crisis List of the corresponding Faction.

      Crisis.UID := GS.CrisisList.MakeCrisisUID; // Important! Every crisis get's a UID. But the CrisisItems of the participants of the same crisis (Initiator, Receiver, Med.) need the _same_ UID
      Crisis.State := _diCrisisNew;
      Crisis.FacInitiatorUID := FacInitiator.UID;
      Crisis.FacReceiverUID  := FacReceiver.UID;
      Crisis.MandatoryAgenda.Assign(MandatoryCI); // Object already created, so an assign is done.

      if FacMediator <> nil then
        Crisis.FacMediatorUID := FacMediator.UID
      else
        Crisis.FacMediatorUID := _noUID;

      Crisis.CrisisIntensity := CrisisIntensity;
      Crisis.InitiatorJC := InitiatorJC;

    // Initiator, Receiver and Mediator get each an identical copy of the crisis object
      FacCrisisItem := FacInitiator.CrisisList.AddItem;
      FacCrisisItem.Assign(Crisis);

      FacCrisisItem := FacReceiver.CrisisList.AddItem;
      FacCrisisItem.Assign(Crisis);

      if FacMediator <> nil then
      begin
        FacCrisisItem := FacMediator.CrisisList.AddItem;
        FacCrisisItem.Assign(Crisis);
      end;


    Log(['Create Crisis: New Crisis:', 'Initiator:', Crisis.FacInitiatorUID, '; Receiver:', Crisis.FacReceiverUID, '; Mediator:', Crisis.FacMediatorUID]);


    except
      on E : Exception do
        LogMsg(['TScriptEngine.CreateCrisis','Exception caught:', E.Message], IIF(FullDebug_, _Error, _Reporting));

    end;

  finally
    FreeAndNil(Crisis);
    FreeAndNil(MandatoryCI);
  end;

*)
end;


// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// FACTION MODIFIER EXISTS
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.FacModifierExists(Params : TStringList) : integer;
var ModUID : TUID;
    ReverseResult : boolean;
begin
  Result := 0;

  if not (Params.Count in [1, 2]) then
  begin
    Inc(Result);
    Log(['=> FacModifierExists command, syntax is: FacModifierExists = ModUID [;NOT]']);
  end;
  if (cuFaction = nil) then
  begin
    Inc(Result);
    Log(['=> FacModifierExists: No faction selected']);
  end;

  ModUID := STI(Params[0]);

  if ModUID = 0 then
  begin
    Inc(Result);
    Log(['=> FacModifierExists: Param0 is not a known Faction Modifier (check if the alias is valid):', Params[0]]);
  end;

  if Result <> 0 then Exit;
  ReverseResult := (Params.Count = 2) and (Params[1] = 'NOT');

  if (not ReverseResult) and (not cuFaction.FM_GetFactionModifierList.InList(ModUID)) then Conditions := condFalse;
  if (ReverseResult) and (cuFaction.FM_GetFactionModifierList.InList(ModUID)) then Conditions := condFalse;

  Log(['FacModifierExists:', 'Checking', ModUID, GameSim.FactionModsDB[ModUID].Name, IIF(ReverseResult, 'NOT', ''),
   'existing for faction', cuFaction.Name, ': result is ',
  IIF(Conditions = condEvaluating, 'True', 'False')]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ADD FACTION MODIFIERS
// AddFacModifiers = <FacMod1> [;<FacMod2>] ...
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.AddFacModifiers(Params : TStringList) : integer;
var i : integer;
    ModUID : TUID;
begin
  Result := 0;

  if cuFaction = nil then
  begin
    Inc(Result);
    Log(['=> AddFacModifiers command, no faction selected']);
    Exit;
  end;
  for i := 0 to Params.Count-1 do
  begin
    ModUID := STI(Params[i]);
    cuFaction.FM_AddFactionModifier(ModUID, TRUE); // TRUE => it will update the modifiers effects
  end;
  Log(['AddFacModifiers, list for Faction', cuFaction.Name, cuFaction.FM_GetFactionModifierList.Count, 'modifier(s)']);
end;


// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// REM FACTION MODIFIERS
// RemFacModifiers = <FacMod1> [;<FacMod2>] ...
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.RemFacModifiers(Params : TStringList) : integer;
var i : integer;
    ModUID : TUID;
begin
  Result := 0;

  if cuFaction = nil then
  begin
    Inc(Result);
    Log(['=> RemFacModifiers command, no faction selected']);
    Exit;
  end;
  for i := 0 to Params.Count-1 do
  begin
    ModUID := STI(Params[i]);
    cuFaction.FM_RemFactionModifier(ModUID, TRUE); // TRUE => it will update the modifiers effects
  end;
  Log(['RemFacModifiers, list for Faction', cuFaction.Name, cuFaction.FM_GetFactionModifierList.Count, 'modifier(s)']);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ENABLE DIPLO ITEM
// EnableDiploItems
// Sets the global variable diDIFlagComplexCond_. This command must be added to
// the Action section of the event which computes the complex conditions for
// DiploItems.
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.EnableDiploItem(Params : TStringList) : integer;
begin
  Result := 0;
  diDIFlagComplexCond_ := True;
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// EVALUATE FACTION ATTRIBUTE
// FacAttribType ;AttribVal1 [;AttribVal2] [;AttribVal3] [;AttribVal4] [;NOT]
// without the not, means faction must be in one of the attrib listed
// with not, means faction must not be in one of the attrib listed
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.EvalFacAttribute(Params : TStringList) : integer;
var AttribKind : TUID;
    AttribVals : array of integer;
    ReverseResult : boolean;
    i : integer;
    Found : boolean;
    Str : string;
begin
  Result := 0;

  if (Params.Count < 2 ) then
  begin
    Inc(Result);
    Log(['=> EvalFacAttribute command, syntax is: EvalFacAttribute = FacAttribType ;AttribVal1 [;AttribVal2] [;AttribVal3..] [;NOT]']);
  end;
  if (cuFaction = nil) then
  begin
    Inc(Result);
    Log(['=> EvalFacAttribute: No faction selected']);
  end;

  ReverseResult := UpperCase(Params[Params.Count-1]) = 'NOT';
  AttribKind := STI(Params[0]);
  AttribVals := nil;
  Str := '';
  for i := 1 to Params.Count-1 - IIF(ReverseResult, 1, 0) do // if reverseresult, then last param (Params.Count-1) is NOT
  begin
    SetLength(AttribVals, Length(AttribVals) +1);
    AttribVals[High(AttribVals)] := STI(Params[i]);
    Str := GameSim.FactionAttsDB.GetByKindAndSubType(AttribKind, STI(Params[i])).Name + ' ';
  end;

  Found := False;
  for i := 0 to High(AttribVals) do
  begin
    if cuFaction.Attributes.GetValue(AttribKind) = AttribVals[i] then
    begin
      Found := True;
      Break;
    end;
  end;

  if ReverseResult then Found := not Found;
  if not Found then Conditions := condFalse;

  Log(['EvalFacAttribute:', 'Checking', AttribKind, IIF(ReverseResult, 'not being any of these values:', 'being on of these values: '),
  Str, 'for faction', cuFaction.Name, ': result is ', IIF(Conditions = condEvaluating, 'True', 'False'), 'current value being ',
  GameSim.FactionAttsDB.GetByKindAndSubType(AttribKind, cuFaction.Attributes.GetValue(AttribKind)).Name]);

  AttribVals := nil;
end;
(*
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// EVAL DI ATTRIBUTE
//
// Checks all DI which exist between two factions for the existence of a certain
// Attribute and, optional, for how long the DI is in place.
//
// Syntax:
//   a) EvalDIAttribute = Fac2; Attribute; RelOp; Value
//   b) EvalDIAttribute = Fac2; Attribute; RelOp; AttValue; Duration; RelOp; DuValue
//   - Fac2 : Faction UID or alias. DI between selected Fac and Fac2 are checked.
//   - Attribute: The Attribute to check for. For Example *military*
//   - RelOp: Relational operator (>, <, >=, <=, =, <>)
//   - AttValue: Attribute level (*military*|2 <-- Level is 2)
//   - Duration: Keyword, simply write "Duration" (without the quotes)
//   - DuValue: Value for duration in month (integer).
//
// Norbert: To be tested
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.EvalDIAttribute(Params : TStringList) : integer;

var Fac2UID     : integer;
    GS          : TGameSim;
    DIList      : TDiploItemList;
    i         : integer;
    //DIDef       : TDiplomacyDef;
    AttValue    : integer;
    AttValueDef       : integer;
    CondFlag          : boolean;
    CheckDurationFlag : boolean;
    DuValue           : integer; // Duration value for condition
    DIDuration        : integer;

begin
  Result      := 0;

  AttValueDef := 0;
  DIDuration  := 0;
  DuValue     := 0;

  CondFlag          := False;
  CheckDurationFlag := False;


  try
    try
      GS      := TGameSim(Owner);
      DIList  := TDiploItemList.Create(GS);

      if cuFaction = nil then
      begin
        Inc(Result);
        Log(['=> EvalDIAttribute command, no faction selected']);
        Exit;
      end;

      if (Params.Count <> 4) and (Params.Count <> 7) then
      begin
        inc(Result);
        Log(['=> EvalDIAttribute: Wrong numbers of parameters (4 or 7). Syntax: EvalDIAttribute = Fac2; Attribute; RelOp; AttValue; Duration; RelOp; DuValue']);
        Exit;
      end;

      if not IsNumeric(Params[0]) then
      begin
        inc(Result);
        Log(['=> EvalDIAttribute: Fac2 UID: Not a number:', Params[0]]);
        Exit;
      end;

      Fac2UID := StrToInt(Params[0]);
      if GS.Factions.GetFactionByUID(Fac2UID, False) = nil then
      begin
        inc(Result);
        Log(['=> EvalDIAttribute: Fac2 UID: Faction not found:', Params[0]]);
        Exit;
      end;

      if (Params[2] <> '>') and (Params[2] <> '>=') and (Params[2] <> '=') and (Params[2] <> '<=') and (Params[2] <> '<') and (Params[2] <> '<>') then
      begin
        inc(Result);
        Log(['=> EvalDIAttribute: Invalid relational operator:', Params[2], ' Allowed: >, >=, =, <=, <, <>']);
        Exit;
      end;

      if not IsNumeric(Params[3]) then
      begin
        inc(Result);
        Log(['=> EvalDIAttribute: Attribute Value: Not a number:', Params[3]]);
        Exit;
      end;
      AttValue := StrToInt(Params[3]);

      if Params.Count = 7 then // Duration has be checked too
      begin
        if (Params[5] <> '>') and (Params[5] <> '>=') and (Params[5] <> '=') and (Params[5] <> '<=') and (Params[5] <> '<') and (Params[5] <> '<>') then
        begin
          inc(Result);
          Log(['=> EvalDIAttribute: Invalid relational operator (duration):', Params[5], ' Allowed: >, >=, =, <=, <, <>']);
          Exit;
        end;

        if not IsNumeric(Params[6]) then
        begin
          inc(Result);
          Log(['=> EvalDIAttribute: Duration value: Not a number:', Params[6]]);
          Exit;
        end;
        CheckDurationFlag := True;
        DuValue := StrToInt(Params[6]);
      end;


      GS.FacTreatyMatrix.GetTreatyList(cuFaction.UID, Fac2UID, DIList);
      for i := 0 to DIList.Count -1 do // Check every DI
      begin
        //DIDef := DIList.Items[i].DiploItemDef;
        {
        for j := 0 to DIDef.Attributes.Count -1 do // Check every attribute (IndexOf not sufficient, we need to check for several occurrences.)
        begin

          if UpperCase(DIDef.Attributes.Vals[j]) = UpperCase(Params[1]) then
          begin
            AttValueDef := DIDef.Attributes.Keys[j];
            if   ((Params[2] = '>' ) and (AttValueDef >  AttValue))
              or ((Params[2] = '>=') and (AttValueDef >= AttValue))
              or ((Params[2] = '=' ) and (AttValueDef =  AttValue))
              or ((Params[2] = '<=') and (AttValueDef <= AttValue))
              or ((Params[2] = '<' ) and (AttValueDef <  AttValue))
              or ((Params[2] = '<>') and (AttValueDef <> AttValue)) then
                CondFlag := True;
          end;

          if (CheckDurationFlag = True) and (CondFlag = True) then // Check additionally the duration
          begin
            DIDuration := (GS.cuTurn - DIList.Items[i].StartTurn) *  12 div GS.TurnsPerYear; // Duration of the DI in Month
            if   (not ((Params[5] = '>' ) and (DIDuration >  DuValue)))
              or (not ((Params[5] = '>=') and (DIDuration >= DuValue)))
              or (not ((Params[5] = '=' ) and (DIDuration =  DuValue)))
              or (not ((Params[5] = '<=') and (DIDuration <= DuValue)))
              or (not ((Params[5] = '<' ) and (DIDuration <  DuValue)))
              or (not ((Params[5] = '<>') and (DIDuration <> DuValue))) then
                CondFlag := False;
          end;


          if CondFlag = True then
            break;

        end; // End: for j := 0 to DIDef.Attributes.Count -1
        }
        if CondFlag = True then
          break;

      end; // End: for i := 0 to DIList.Count -1 do // Check every DI


      if CondFlag = False then
        Conditions := condFalse;

      if CheckDurationFlag = True then
        Log(['EvalDIAttribute:', 'Check for DI with Attribute', Params[1], 'Value', Params[2], AttValue, '(Found:', AttValueDef, ')', 'Duration', Params[5], DuValue, '(Found:', DIDuration, ')',  IIF(Conditions = condEvaluating, 'True', 'False')])
      else
        Log(['EvalDIAttribute:', 'Check for DI with Attribute', Params[1], 'Value', Params[2], AttValue, 'No Duration check', IIF(Conditions = condEvaluating, 'True', 'False')]);

    except
      on E : Exception do
        LogMsg(['TScriptEngine.EvalDIAttribute','Exception caught:', E.Message], _Error);
    end;

  finally
    FreeAndNil(DIList);
  end;

end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// EVAL DIPLO ITEM
// EvalDiploItem = AUS;$diAllianceTreaty;NOT
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.EvalDiploItem(Params : TStringList) : integer;
var GS : TGameSim;
    OtherFac : TFaction;
    Value : integer;
    EvalDIDef, DIDef : TDiplomacyDef;
//    DI : TDiploItem;
//    DIList      : TDiploItemList;
    ReverseResult : boolean;
    Param : string;
    SinceWhen : integer;
    i : integer;
    DIFound : boolean;
    //DIDuration : integer;
begin
  Result := 0;
  
  GS      := TGameSim(Owner);
//  DIList  := TDiploItemList.Create(GS); // list of items between the 2 factions
  DIFound := False;

  if cuFaction = nil then
  begin
    Inc(Result);
    Log(['=> EvalDiploItem command, no faction selected']);
    Exit;
  end;

  if (Params.Count < 2) then
  begin
    Inc(Result);
    Log(['=> EvalDiploItem: Wrong numbers of parameters (must be 3). Syntax: EvalDiploItem = <TargetFactionTag>; <DI_Alias> [;<NOT>] [;<Since TurnsNumber>]']);
    Exit;
  end;

  OtherFac := FactionTagToFaction(Params[0]);
  if OtherFac <> nil then Log(['EvalDiploItem: Evaluating against', OtherFac.Name])
  else
  begin
    Log(['[WARNING] EvalDiploItem command, no faction selected']);
    Exit;
  end;

  Value := STI(Params[1], True, -1);
  if (Value < 0) or (Value > GS.DiplomacyDB.Count-1) then
  begin
    Inc(Result);
    Log(['=> EvalDiploItem command, this value was not be understood', Params[1]]);
    Exit;
  end;
  EvalDIDef := GS.DiplomacyDB[Value];

  ReverseResult := ExtractParam('NOT', Params) <> '';
  SinceWhen := -1;
  Param := ExtractParam('Since', Params);
  if Param <> '' then
  begin
    ExtractArgs(Param);
    SinceWhen := STI(Args[0], True, -1);
  end;

  GS.FacTreatyMatrix.GetTreatyList(cuFaction.UID, OtherFac.UID, DIList);
  for i := 0 to DIList.Count -1 do // Check every DI
  begin
    DI := DIList.Items[i];
    DIDef := DI.DiploItemDef;
    if EvalDIDef = DIDef then
    begin
      if (SinceWhen = -1) or (GS.cuTurn - DI.StartTurn > SinceWhen) then
      begin
        DIFound := True;
        Break;
      //DIDuration := (GS.cuTurn - DI.StartTurn) *  12 div GS.TurnsPerYear; // Normalize duration of the DI in Month
      end;
    end;
  end;
  if ReverseResult then
    DIFound := not DIFound;

  if not DIFound then Conditions := condFalse;

  if ReverseResult then
  Log(['EvalDiploItem:', cuFaction.Name, ' don''t have treaty', EvalDIDef.Name, 'with',
  OtherFac.Name, '?', 'Result:', IIF(Conditions = condEvaluating, 'True', 'False')])
  else
  Log(['EvalDiploItem:', cuFaction.Name, ' has treaty', EvalDIDef.Name, 'with',
  OtherFac.Name, '?', 'Result:', IIF(Conditions = condEvaluating, 'True', 'False')]);

  FreeAndNil(DIList);
end;
*)
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// IS AT PEACE WITH / IS AT WAR WITH
// If the tag given is CMN, will be an absolute evaluation (IsAtPeace, IsAtWar)
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.EvalIsAtWarWith(Params : TStringList) : integer;
var
    OtherFac : TFaction;
begin
  Result := 0;
  if (Params.Count <> 1) then
  begin
    Inc(Result);
    Log(['=> EvalIsAtWarWith command, syntax is: EvalIsAtWarWith = <Faction_tag>']);
  end;
  if (cuFaction = nil) then
  begin
    Inc(Result);
    Log(['=> EvalIsAtWarWith: No faction selected']);
  end;
  if Result > 0 then Exit;


  if Params[0] <> 'CMN' then
  begin
    OtherFac := nil;
    if (Length(Params[0]) >= 3) and (Params[0][1] in ['A'..'Z']) then
    begin
      OtherFac := FactionTagToFaction(Params[0]);
      if OtherFac <> nil then Log(['EvalIsAtWarWith: Evaluating against', OtherFac.Name]);
    end;
    if (OtherFac = nil) then
    begin
      Inc(Result);
      Log(['=> EvalIsAtWarWith: Faction does not exist', Params[0]]);
    end;
    if Result > 0 then Exit;
    if not cuFaction.IsAtWarWith(OtherFac) then
      Conditions := condFalse;
    Log(['EvalIsAtWarWith:', cuFaction.Name, 'is at war with', OtherFac.Name, 'Result:', IIF(Conditions = condEvaluating, 'True', 'False')]);
  end
  else
  begin
    if not cuFaction.IsAtWar then
      Conditions := condFalse;
    Log(['EvalIsAtWarWith:', cuFaction.Name, 'is at war with someone?', 'Result:', IIF(Conditions = condEvaluating, 'True', 'False')]);
  end;
end;
function TScriptEngine.EvalIsAtPeaceWith(Params : TStringList) : integer;
var 
    OtherFac : TFaction;
begin
  Result := 0;
  if (Params.Count <> 1) then
  begin
    Inc(Result);
    Log(['=> EvalIsAtPeaceWith command, syntax is: EvalIsAtPeaceWith = <Faction_tag>']);
  end;
  if (cuFaction = nil) then
  begin
    Inc(Result);
    Log(['=> EvalIsAtPeaceWith: No faction selected']);
  end;
  if Result > 0 then Exit;


  if Params[0] <> 'CMN' then
  begin
    OtherFac := nil;  
    if (Length(Params[0]) >= 3) and (Params[0][1] in ['A'..'Z']) then
    begin
      OtherFac := FactionTagToFaction(Params[0]);
      if OtherFac <> nil then Log(['EvalIsAtPeaceWith: Evaluating against', OtherFac.Name]);
    end;
    if (OtherFac = nil) then
    begin
      Inc(Result);
      Log(['=> EvalIsAtPeaceWith: Faction does not exist', Params[0]]);
    end;
    if Result > 0 then Exit;

    if not cuFaction.IsAtPeaceWith(OtherFac) then
      Conditions := condFalse;

    Log(['EvalIsAtPeaceWith:', cuFaction.Name, 'is at peace with', OtherFac.Name, 'Result:', IIF(Conditions = condEvaluating, 'True', 'False')]);
  end
  else
  begin
    if not cuFaction.IsAtPeace then
      Conditions := condFalse;
    Log(['EvalIsAtPeaceWith:', cuFaction.Name, 'is at peace with everybody?', 'Result:', IIF(Conditions = condEvaluating, 'True', 'False')]);
  end;
end;
(*
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ADD DI TO ALL FAC
//
// Adds a DI to all factions. Very simple, made to set peace for all factions
// on scenario creation. It does simply add the DI and set a custom start date.
// Nothing more. Does _not_ remove the War DI if you add the Peace DI.
//
// Syntax:
// AddDIToAllFac = DIDefUID; StartDate
//
// Norbert: To be tested
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.AddDIToAllFac(Params : TStringList) : integer;
var i,j       :integer;
    GS        : TGameSim;
    NewDI     : TDiploItem;
    DIDefUID  : integer;
    DIDef     : TDiplomacyDef;
    StartTurn : integer;

begin

  Result := 0;

  try
    GS := TGameSim(Owner);

    if (Params.Count <> 1) and (Params.Count <> 2)  then
    begin
      inc(Result);
      Log(['=> AddDIToAllFac: Wrong numbers of parameters (1 or 2). Syntax: AddDIToAllFac = DIDefUID [; StartDate]']);
      Exit;
    end;

    if not IsNumeric(Params[0]) then
    begin
      inc(Result);
      Log(['=> AddDIToAllFac: DIDefUID: Not a number:', Params[0]]);
      Exit;
    end;
    DIDefUID := StrToInt(Params[0]);

    DIDef := GS.DiplomacyDB[DIDefUID];
    if DIDef = nil then
    begin
      inc(Result);
      Log(['=> AddDIToAllFac: DIDefUID: DI definition not found! UID:', Params[0]]);
      Exit;
    end;

    if (Params.Count = 2) and (Length(Params[1]) <> 10)then
    begin
      Inc(Result);
      Log(['=> AddDIToAllFac: Erroneous format for date (dd/mm/yyyy):', Params[1]]);
    end;

    if (Params.Count = 2) then
      StartTurn := DateToTurn(Params[1])
    else
      StartTurn := GS.cuTurn;


    for i := 0 to GS.Factions.Count -2 do
    begin
      for j := i+1 to GS.Factions.Count -1 do
      begin
        NewDI := TDiploItem.Create(GS, True);
        NewDI.DIDefUID := DIDefUID;
        NewDI.Name := NewDI.DiploItemDef.Name;
        NewDI.Kind := NewDI.DiploItemDef.Kind;
        NewDI.SenderFacUID := GS.Factions[i].UID;
        NewDI.RecipientFacUID := GS.Factions[j].UID;
        NewDI.FacUIDList.Add(GS.Factions[i].UID);
        NewDI.FacUIDList.Add(GS.Factions[j].UID);
        NewDI.State := _diTreatyInEffect;
        NewDI.StartTurn := StartTurn;
        NewDI.EndTurn := NewDI.StartTurn + NewDI.DiploItemDef.DIDuration * GS.TurnsPerYear div 12;
        GS.FacTreatyMatrix.AddTreaty(NewDI);
      end;
    end;


  except
    on E : Exception do
      LogMsg(['TScriptEngine.AddDIToAllFac','Exception caught:', E.Message], _Error);
  end;


end;
*)
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ADD DIPLO ITEM
// Add one DI between 2 factions (can be self)
// Handle unvalid DI with DiploObjectIsValid
// Don't trigger in game effects, use custom functions for that
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.AddDiploItem(Params : TStringList) : integer;
var 
    GS        : TGameSim;
    DIDefUID  : integer;
    DD        : TDiplomacyDef;
    StartTurn : integer;
    TgtFac    : TFaction;
    DOS       : TDiploObjectStruc;
    ReturnValue : TUID;
    WS        : integer;
    DObj      : TDiploObject;
begin

  Result := 0;

  try
    GS := TGameSim(Owner);

    if (Params.Count < 2) or (Params.Count > 4)  then
    begin
      inc(Result);
      Log(['=> AddDiploItem: Wrong numbers of parameters. Syntax: AddDiploItem = TargetFac; DIDefUID [; StartDate] [;Warscore]']);
      Exit;
    end;

    if cuFaction = nil then
    begin
      inc(Result);
      Log(['=> AddDiploItem: No Selected faction']);
      Exit;
    end;

    TgtFac := FactionTagToFaction(Params[0]);

    if TgtFac = nil then
    begin
      inc(Result);
      Log(['=> AddDiploItem: No target faction']);
      Exit;
    end;

    if not IsNumeric(Params[1]) then
    begin
      inc(Result);
      Log(['=> AddDiploItem: DIDefUID: Not a number:', Params[1]]);
      Exit;
    end;

    DIDefUID := STI(Params[1]);
    if (DIDefUID < 0) or (DIDefUID > GS.DiplomacyDB.Count-1) then
    begin
      inc(Result);
      Log(['=> AddDiploItem: DIDefUID: DI definition not found! UID:', Params[1]]);
      Exit;
    end;
    DD := GS.DiplomacyDB[DIDefUID];

    if (Params.Count >= 3) and (Length(Params[2]) = 10)then
      StartTurn := DateToTurn(Params[2])
    else
    begin
      StartTurn := GS.cuTurn;
      Log(['AddDiploItem: Using current turn.']);
    end;

    DOS := InitDiploObjectStruc;
    DOS.StartTurn := StartTurn;
    DOS.EndTurn := StartTurn + GS.DiplomacyDB[DIDefUID].DIDuration;
    DOS.DefUID := DIDefUID;
    DOS.SenderFacUID := cuFaction.UID;
    DOS.TargetFacUID := TgtFac.UID;
    DOS.State := _diTreatyInEffect;

    ReturnValue := GS.DipTreatyList.DiploObjectIsValid(DOS);
    DObj := nil;
    if ReturnValue = _dipValid then
    begin
      DObj := GS.DipTreatyList.AddNewItem(DOS); // in the GS list (list of DObj in effects)
      if GameSim.DataType <> _dtScen then
      begin
        if DObj.DiploItemDef.Unilateral then
        begin
          if DObj.DiploItemDef.HasThisAttribute(_attWar) then
            cuFaction.Messages.MakeBroadcastedLineMsg('strCMNDOWInEffect', _noUID, _msgTypeDiplomacy, cuFaction.Name, TgtFac.Name)
          else
          begin
            cuFaction.Messages.MakeBroadcastedLineMsgEx('strCMNDiploObjUnilateralInEffect', _noUID, _msgTypeDiplomacy, cuFaction, TgtFac, TgtFac.Name, DObj.DiploItemDef.Name, cuFaction.Name);
            cuFaction.Messages.MakeLineMsg('strOurDiploObjUnilateralInEffect', _noUID, _msgTypeDiplomacy, TgtFac.Name, DObj.DiploItemDef.Name);
            TgtFac.Messages.MakeLineMsg('strOtherDiploObjUnilateralInEffect', _noUID, _msgTypeDiplomacy, cuFaction.Name, DObj.DiploItemDef.Name);
          end;
        end
        else
        begin
          cuFaction.Messages.MakeBroadcastedLineMsgEx('strCMNDiploObjGenericAnswerYes', _noUID, _msgTypeDiplomacy, cuFaction, TgtFac, DObj.DiploItemDef.Name, cuFaction.Name, TgtFac.Name);
          cuFaction.Messages.MakeLineMsg('strOurDiploObjGenericAnswerYes', _noUID, _msgTypeDiplomacy, TgtFac.Name, DObj.DiploItemDef.Name);
          TgtFac.Messages.MakeLineMsg('strOtherDiploObjGenericAnswerYes', _noUID, _msgTypeDiplomacy, cuFaction.Name, DObj.DiploItemDef.Name);
        end;
      end;

      Log(['AddDiploItem: This Diplo Object', DOS.DefUID, GS.DiplomacyDB[DOS.DefUID].Name, 'has been added']);
    end
    else
      Log(['AddDiploItem: Can''t add this Diplo Object', DOS.DefUID, DD.Name, 'error code', ReturnValue]);

    if (DObj <> nil) and DD.HasThisAttribute(_attWar) and (Params.Count > 3) then
    begin
      WS := STI(Params[3]);
      DObj.WarScore := WS;
    end;

  except
    on E : Exception do
      LogMsg(['TScriptEngine.AddDiploItem','Exception caught:', E.Message], _Error);
  end;

end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// REM DIPLO ITEM
// Don't trigger in game effects, use custom functions for that
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.RemDiploItem(Params : TStringList) : integer;
var 
    GS        : TGameSim;
    DIDefUID  : integer;
    DD        : TDiplomacyDef;
    TgtFac    : TFaction;
    DObj      : TDiploObject;
begin

  Result := 0;

  try
    GS := TGameSim(Owner);

    if (Params.Count < 2) or (Params.Count > 2)  then
    begin
      inc(Result);
      Log(['=> RemDiploItem: Wrong numbers of parameters. Syntax: RemDiploItem = TargetFac; DIDefUID']);
      Exit;
    end;

    if cuFaction = nil then
    begin
      inc(Result);
      Log(['=> RemDiploItem: No Selected faction']);
      Exit;
    end;

    TgtFac := FactionTagToFaction(Params[0]);

    if TgtFac = nil then
    begin
      inc(Result);
      Log(['=> RemDiploItem: No target faction']);
      Exit;
    end;

    if not IsNumeric(Params[1]) then
    begin
      inc(Result);
      Log(['=> RemDiploItem: DIDefUID: Not a number:', Params[1]]);
      Exit;
    end;

    DIDefUID := STI(Params[1]);
    if (DIDefUID < 0) or (DIDefUID > GS.DiplomacyDB.Count-1) then
    begin
      inc(Result);
      Log(['=> RemDiploItem: DIDefUID: DI definition not found! UID:', Params[1]]);
      Exit;
    end;
    DD := GS.DiplomacyDB[DIDefUID];

    DObj := GS.DipTreatyList.FetchDiploObjectUniDir(cuFaction, TgtFac, DD.UID);
    if DObj <> nil then
    begin
      GameSim.DipTreatyList.RemoveTreaty(cuFaction, DObj, False);
      if GameSim.DataType <> _dtScen then
      begin
        cuFaction.Messages.MakeLineMsg('strOurDiploObjPastTurnAborted', _noUID, _msgTypeDiplomacy, TgtFac.Name, DObj.DiploItemDef.Name);
        TgtFac.Messages.MakeLineMsg('strOtherDiploObjPastTurnAborted', _noUID, _msgTypeDiplomacy, cuFaction.Name, DObj.DiploItemDef.Name);
      end;  
    end;
    DObj := GS.DipTreatyList.FetchDiploObjectUniDir(TgtFac, cuFaction, DD.UID);
    if DObj <> nil then
    begin
      GameSim.DipTreatyList.RemoveTreaty(TgtFac, DObj, False);
      if GameSim.DataType <> _dtScen then
      begin
        TgtFac.Messages.MakeLineMsg('strOurDiploObjPastTurnAborted', _noUID, _msgTypeDiplomacy, cuFaction.Name, DObj.DiploItemDef.Name);
        cuFaction.Messages.MakeLineMsg('strOtherDiploObjPastTurnAborted', _noUID, _msgTypeDiplomacy, TgtFac.Name, DObj.DiploItemDef.Name);
      end;  
    end;

  except
    on E : Exception do
      LogMsg(['TScriptEngine.RemDiploItem','Exception caught:', E.Message], _Error);
  end;

end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SETTLE WHITE PEACE
// High level command that test the validity of the peace item and triggers
// all in game effects if a peace can be settled. Don't support yet peace proposals
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SettleWhitePeace  (Params : TStringList) : integer;
var TargetFaction : TFaction;
    fac : integer;
    OpDone : boolean;
begin
  Result := 0;
  if Params.Count <> 1 then
  begin
    inc(Result);
    Log(['=> SettleWhitePeace: Wrong numbers of parameters . Syntax: SettleWhitePeace = <TargetFac_or_ALL>']);
    Exit;
  end;

  if cuFaction = nil then
  begin
    inc(Result);
    Log(['=> SettleWhitePeace: No Selected faction']);
    Exit;
  end;

  if Uppercase(Params[0]) <> 'ALL' then
  begin
    TargetFaction := FactionTagToFaction(Params[0]);
    if TargetFaction = nil then
    begin
      inc(Result);
      Log(['=> SettleWhitePeace: No target faction']);
      Exit;
    end;
  end
  else
    TargetFaction := nil;

  if (TargetFaction <> nil) and cuFaction.SettleWhitePeaceWith(TargetFaction) then
    Log(['SettleWhitePeace: Peace settled with', TargetFaction.Name])
  else
  if (TargetFaction <> nil) then
    Log(['SettleWhitePeace: Peace could not be settled with', TargetFaction.Name, 'a formal War must exist.']);

  if TargetFaction = nil then
  begin
    for fac := 0 to GameSim.Factions.Count-1 do
    begin
      TargetFaction := GameSim.Factions[fac];
      OpDone := cuFaction.SettleWhitePeaceWith(TargetFaction);
      if OpDone then
        Log(['SettleWhitePeace: Peace settled with', TargetFaction.Name]);
    end;

  end;

    
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// CHANGE PRISONERS
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.ChangePrisoners(Params : TStringList) : integer;
begin
  Result := 0;
  if Params.Count <> 2 then
  begin
    Inc(Result);
    Log(['=> ChgPrisoners, need to have exactly two parameters: ChgPrisoners = FactionTag (or XXX for all) and Amount of Prisoners points given (or substracted) to the current faction']);
    Exit;
  end;
  if cuFaction = nil then
  begin
    Log(['=> ChgPrisoners command, no faction selected']);
    Inc(Result);
    Exit;
  end;

  cuFaction.ChangePrisoners(Params[0], STI(Params[1]));

  Log(['ChangePrisoners', cuFaction.Name, 'New prisoners total is', cuFaction.GetTotalPrisoners]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// CHANGE OWNERSHIP
// ChangeRgnOwner = [NotFort;] FactionTag1_or_ALL [;FactionTag2] etc.
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.ChangeRgnOwner(Params : TStringList) : integer;
var Faction : TFaction;
    i       : integer;
    NotFort : boolean;
    TransfertStrucs : boolean;
begin
  Result := 0;
  if cuFaction = nil then
  begin
    Log(['=> ChangeRgnOwner command, no faction selected']);
    Inc(Result);
    Exit;
  end;
  if (cuRegion = nil) then
  begin
    Log(['=> ChangeRgnOwner: No region selected']);
    Inc(Result);
    Exit;
  end;

  NotFort := False;

  for i := 0 to Params.Count-1 do
  begin
    if Params[i] = 'NotFort' then
    begin
      NotFort := True;
      Continue;
    end;

    if UpperCase(Params[i]) = 'ALL' then
      Faction := GameSim.RegionCmns[cuRegion.UID].OwningFac
    else
      Faction := FactionTagToFaction(Params[i]);
    if Faction = nil then
    begin
      Inc(Result);
      Log(['=> ChangeRgnOwner: No faction found with tag ', Params[i]]);
      Continue;
    end;

    TransfertStrucs :=  (not NotFort) or (GameSim.RegionCmns.GetFort(cuRegion.UID) = nil);
    GameSim.RegionCmns.ChangeOwner(cuFaction, Faction, cuRegion.UID, True, TransfertStrucs);

    if TransfertStrucs and (GameSim.RegionCmns[cuRegion.UID].pListStruc.Count > 0) then
      Log(['ChangeRgnOwner: Structures were present and could have been transfered (depending of faction owner)'])
    else
      Log(['ChangeRgnOwner: No Structure transfert done.', Params[i]]);
  end;
end;
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ABSORB FACTION
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.AbsorbFaction(Params : TStringList) : integer;
var Subjected : TFaction;
    PercMatAssets, PercImmatAssets, PercLoy, PercMC, PercColPen, PercLand, PercNav : integer;
    AreaUID : TUID;
begin
  Result := 0;
  if cuFaction = nil then
  begin
    Log(['=> AbsorbFaction command, no beneficiary faction selected']);
    Inc(Result);
  end;
  Subjected := GameSim.Factions.GetFactionByTag(Params[0], False);
  if Subjected = nil then
  begin
    Log(['=> AbsorbFaction command, no Subjected faction selected']);
    Inc(Result);
  end;
  if Params.Count < 8 then
  begin
    Inc(Result);
    Log(['=> AbsorbFaction, need to have at least 8 parameters: AbsorbFaction = <SubjectedFacTag>; <PercMatAssets>; <PercImmatAssets>; <PercLoy>; <PercMC>; <PercColPen>; <PercLand>; <PercNav> [:<AreaUID>]']);
  end;
  
  if Result > 0 then Exit;
  PercMatAssets   := STI(Params[1], True, 100);
  PercImmatAssets := STI(Params[2], True, 100);
  PercLoy         := STI(Params[3], True, 100);
  PercMC          := STI(Params[4], True, 100);
  PercColPen      := STI(Params[5], True, 100);
  PercLand        := STI(Params[6], True, 100);
  PercNav         := STI(Params[7], True, 100);
  if Params.Count > 8 then
    AreaUID := STI(Params[8], True, _noUID)
  else
    AreaUID := _noUID;

  GameSim.RegionCmns.AbsorbFaction(cuFaction, Subjected, PercMatAssets, PercImmatAssets, PercLoy, PercMC, PercColPen, PercLand, PercNav, AreaUID);

  Log(['AbsorbFaction', Subjected.Name, 'has been absorbed by', cuFaction.Name, 'using these values:',
  'PercMatAssets, PercImmatAssets, PercLoy, PercMC, PercColPen, PercLand, PercNav, AreaUID',
  PercMatAssets, PercImmatAssets, PercLoy, PercMC, PercColPen, PercLand, PercNav, AreaUID]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET SPHERES OF INFLUENCE
// Two syntaxes: SetSOIVals = FacTag;Area1;Val1;Area2;Val2 (csvsplitter way)
// or SetSOIVals = Area1;Val1;Area2;Val2 and in this case we use cuFaction
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetSOIs(Params : TStringList) : integer;
var Faction : TFaction;
    AreaUID, Coeff : integer;
    i : integer;
    CA : TColonialArea;
begin
  Result := 0;

  // if even number of param, no faction tag specified
  if Params.Count / 2 = Params.Count div 2 then
  begin
    Faction := cuFaction
  end
  else
  begin
    Faction := FactionTagToFaction(Params[0]);
    if Faction = nil then
    begin
      Inc(Result);
      Log(['=> SetSOIVals: No faction found with tag ', Params[0]]);
    end;
    Params.Delete(0);
  end;
  if Result <> 0 then Exit;


  Faction.dlSOI.Clear;

  for i := 1 to Params.Count div 2 do
  begin
    AreaUID := STI(Params[(i-1)*2]);
    Coeff  := STI(Params[(i-1)*2 + 1]);

    if (AreaUID < 0)or(AreaUID > Areas.Count-1) then
    begin
      Inc(Result);
      Log(['=> SetSOIVals: The Area UID is not valid.']);
      Exit;
    end;
    CA := GameSim.ColModule.GetColAreaFromArea(AreaUID);
    if CA = nil then
    begin
      Inc(Result);
      Log(['=> SetSOIVals: The Area alias is not colonial area, see the file ColonialAreas.ini', Params[1]]);
    end;
    if Result > 0 then Exit;

    if Coeff <> colSOIValDefault_ then
    begin
      Faction.dlSOI.AlterValByKey(AreaUID, Coeff, True, False);
      Log(['SetSOIVals', Faction.Name, AreaUID, Areas[AreaUID].Name, 'SOI Val is', Coeff]);
    end;
  end;

  // recalc now look-up aray
  Faction.CalcSOIRgnWorth;
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// CHANGE SPHERES OF INFLUENCE
// Two syntaxes: ChgSOIVals = FacTag;Area1;Val1;Area2;Val2 (csvsplitter way)
// or ChgSOIVals = Area1;Val1;Area2;Val2 and in this case we use cuFaction
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.ChangeSOIs(Params : TStringList) : integer;
var Faction : TFaction;
    AreaUID, Coeff : integer;
    i : integer;
    CA : TColonialArea;
    cuValue : integer;
begin
  Result := 0;
  
  // if even number of param, no faction tag specified
  if Params.Count / 2 = Params.Count div 2 then
  begin
    Faction := cuFaction
  end
  else
  begin
    Faction := FactionTagToFaction(Params[0]);
    if Faction = nil then
    begin
      Inc(Result);
      Log(['=> ChgSOIVals: No faction found with tag ', Params[0]]);
    end;
    Params.Delete(0);
  end;
  if Result <> 0 then Exit;

  for i := 1 to Params.Count div 2 do
  begin
    AreaUID := STI(Params[(i-1)*2]);
    Coeff  := STI(Params[(i-1)*2 + 1]);

    if (AreaUID < 0)or(AreaUID > Areas.Count-1) then
    begin
      Inc(Result);
      Log(['=> ChgSOIVals: The Area UID is not valid.']);
      Exit;
    end;
    CA := GameSim.ColModule.GetColAreaFromArea(AreaUID);
    if CA = nil then
    begin
      Inc(Result);
      Log(['=> ChgSOIVals: The Area alias is not colonial area, see the file ColonialAreas.ini', Params[1]]);
    end;
    if Result > 0 then Exit;

    Faction.dlSOI.AlterValByKey(AreaUID, Coeff, True, False);
    Faction.dlSOI.GetValByKey(AreaUID, cuValue, colSOIValDefault_); // direct read, not precalc

    if cuValue <= colSOIValDefault_ then // under the default value
    begin
      Faction.dlSOI.AlterValByKey(AreaUID, -999, True); // add & delete if neg
      Log(['ChgSOIVals', Faction.Name, AreaUID, Areas[AreaUID].Name, 'SOI Val is now at default value (unreferenced from list)']);
    end
    else
    begin
      Faction.dlSOI.GetValByKey(AreaUID, cuValue, colSOIValDefault_);
      Log(['ChgSOIVals', Faction.Name, AreaUID, Areas[AreaUID].Name, 'SOI Val is now', cuValue]);
    end;
  end;

  // recalc now look-up aray
  Faction.CalcSOIRgnWorth;

end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET ADVANCED REVOLT
// SetAdvRevolt = FacTag; RevoltStrength; RegenAmount; MaxRegen; LosePerc; WinPerc
// SetAdvRevolt = RevoltStrength; RegenAmount; MaxRegen; LosePerc; WinPerc
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetAdvRevolt(Params : TStringList) : integer;
var Faction : TFaction;
    i : integer;
begin
  Result := 0;

  if not (Params.Count in [7..8]) then
  begin
    Log(['SetAdvRevolt', 'Syntax is either SetAdvRevolt = FacTag; RevoltStrength; RegenAmount; MaxRegen; LosePerc; WinPerc; CityUp; CountryUp OR Same as first syntax, without FacTag']);
    Inc(Result);
  end;
  if Result <> 0 then Exit;

  if Params.Count = 7 then
  begin
    Faction := cuFaction;
    i := 0;
  end
  else
  begin
    Faction := FactionTagToFaction(Params[0]);
    i := 1;
    if Faction = nil then
    begin
      Inc(Result);
      Log(['=> SetAdvRevolt: No faction found with tag ', Params[0]]);
    end;
  end;
  if Result <> 0 then Exit;

  Faction.AdvRevolts.InitStrength  := STI(Params[i]);
  Faction.AdvRevolts.RegenAmount   := STI(Params[i+1]);
  Faction.AdvRevolts.MaxRegenHits  := STI(Params[i+2]);
  Faction.AdvRevolts.LosePerc      := STI(Params[i+3]);
  Faction.AdvRevolts.WinPerc       := STI(Params[i+4]);
  Faction.AdvRevolts.CityUprise    := STI(Params[i+5], True, -1); // if NULL or wrong alias then -1, no revolt
  Faction.AdvRevolts.CountryUprise := STI(Params[i+6], True, -1); // if NULL or wrong alias then -1, no revolt
  Log(['SetAdvRevolt', Faction.Name, 'Advanced revolts set with', Params.Text]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET ADVANCED REVOLT FORCE POOL
// SetAdvRevoltFP = FacTag;Unit1;Nb1;Unit2;Nb2.. OR
// SetAdvRevoltFP = Unit1;Nb1;Unit2;Nb2.. and we use current faction
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetAdvRevoltFP(Params : TStringList) : integer;
var Faction : TFaction;
    UnitUID, Num : integer;
    i : integer;
begin
  Result := 0;

  // if even number of param, no faction tag specified
  if Params.Count / 2 = Params.Count div 2 then
  begin
    Faction := cuFaction
  end
  else
  begin
    Faction := FactionTagToFaction(Params[0]);
    if Faction = nil then
    begin
      Inc(Result);
      Log(['=> SetAdvRevoltFP: No faction found with tag ', Params[0]]);
    end;
    Params.Delete(0);
  end;
  if Result <> 0 then Exit;


  Faction.AdvRevolts.dlPool.Clear;

  for i := 1 to Params.Count div 2 do
  begin
    UnitUID := STI(Params[(i-1)*2], True, -1);
    Num     := STI(Params[(i-1)*2 + 1]);

    if (UnitUID < 0)or((UnitUID = 0) and sysErroneousZero_)or(UnitUID > UnitDefs.Count-1) then            
    begin
      Inc(Result);
      Log(['=> SetAdvRevoltFP: The Unit UID is not valid (check your alias).', Params[(i-1)*2]]);
      Exit;
    end;
    Faction.AdvRevolts.dlPool.AlterValByKey(UnitUID, Num, True);
    Log(['SetAdvRevoltFP', Faction.Name, UnitUID, UnitDefs[UnitUID].Name, 'in pool, quantity:', Num]);
  end;
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET ADVANCED REVOLT LOCATION
// SetAdvRevoltLoc = FacTag;Region1;Coeff1;Region2;Coeff2.. OR
// SetAdvRevoltLoc = Region1;Coeff1;Region2;Coeff2.. and we use current faction
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetAdvRevoltLoc(Params : TStringList) : integer;
var Faction : TFaction;
    RgnUID, Num : integer;
    i : integer;
begin
  Result := 0;

  // if even number of param, no faction tag specified
  if Params.Count / 2 = Params.Count div 2 then
  begin
    Faction := cuFaction
  end
  else
  begin
    Faction := FactionTagToFaction(Params[0]);
    if Faction = nil then
    begin
      Inc(Result);
      Log(['=> SetAdvRevoltLoc: No faction found with tag ', Params[0]]);
    end;
    Params.Delete(0);
  end;
  if Result <> 0 then Exit;


  Faction.AdvRevolts.dlRegions.Clear;

  for i := 1 to Params.Count div 2 do
  begin
    RgnUID := STI(Params[(i-1)*2], True, -1);
    Num     := STI(Params[(i-1)*2 + 1]);

    if (RgnUID < 0)or((RgnUID = 0) and sysErroneousZero_)or(RgnUID > RegionDefs.Count-1) then
    begin
      Inc(Result);
      Log(['=> SetAdvRevoltLoc: The Region UID is not valid (check your alias).', Params[(i-1)*2]]);
      Exit;
    end;
    Faction.AdvRevolts.dlRegions.AlterValByKey(RgnUID, Num, True);
    Log(['SetAdvRevoltLoc', Faction.Name, RgnUID, RegionDefs[RgnUID].Name, 'is a revolt region with a coefficient of ', Num]);
  end;
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// TRIGGER ADVANCED REVOLT
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.TriggerAdvRev(Params : TStringList) : integer;
var Value : integer;
begin
  Result := 0;
  if not (Params.Count <> 1) then
  begin
    Inc(Result);
    Log(['=> Syntax is TriggerAdvRev = <RevoltTypeUID>']);
  end;
  if cuFaction = nil then
  begin
    Inc(Result);
    Log(['=> TriggerAdvRev, No faction selected']);
  end;
  if cuRegion = nil then
  begin
    Log(['=> TriggerAdvRev, No region center selected']);
    Exit;
  end;
  if cuFaction.AdvRevolts.NotParametized then
  begin
    Log(['[WARNING] TriggerAdvRev, The faction don''t have a revolt defined']);
    Exit;
  end;

  Value := STI(Params[0], False, -1);
  if not (Value in [_revPrimary, _revSecondary, _revTertiary]) then
  begin
    Inc(Result);
    Log(['=> TriggerAdvRev, Alias unknown', Params[0]]);
  end;
  if Result > 0 then Exit;

  cuFaction.TriggerAdvRevolt(cuFaction.AdvRevolts, cuRegion.UID, Value);
  Log(['TriggerAdvRev', cuFaction.UID, cuFaction.Name, 'revolt triggered, of type', Value, 'around', cuRegion.Name]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// CHANGE REGION DECISION FORCE POOL
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.ChgRgnDecisionFP  (Params : TStringList) : integer;
var i         : integer;
    ObjectUID : TUID;
    Value     : integer;
begin
  Result := 0;
  if not (Params.Count > 0) then
  begin
    Inc(Result);
    Log(['=> Syntax is ChgRgnDecisionFP = [;<RgnDecisionUID|Variation>] ..']);
  end;
  if Params.Count mod 2 <> 0 then
  begin
    Inc(Result);
    Log(['=> ChgRgnDecisionFP, Specify a list of paired values: RgnDecisionUID|Variation']);
  end;
  if cuFaction = nil then
  begin
    Inc(Result);
    Log(['=> ChgRgnDecisionFP, No faction selected']);
  end;
  if Result <> 0 then Exit;

  for i := 1 to Params.Count div 2 do
  begin
    ObjectUID := STI(Params[(i-1)*2]);
    Value     := STI(Params[(i-1)*2 + 1]);
    cuFaction.ChgRgnDecisionDefs(ObjectUID, Value);
    Log(['Current Region Decisions allowed', GameSim.RgnDecisionsDB[ObjectUID].Name, cuFaction.GetNumRgnDecisionDefs(ObjectUID)]);
  end;
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ADD/REM REGION DECISION ACT
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.AddRgnDecisionAct (Params : TStringList) : integer;
var 
    ObjectUID : TUID;
    Value     : integer;
begin
  Result := 0;
  if not (Params.Count > 0) then
  begin
    Inc(Result);
    Log(['=> Syntax is AddRgnDecisionAct = <RgnDecisionUID>;<yyyy/mm/dd_or_CURRENT>']);
  end;
  if Result <> 0 then Exit;

  if cuFaction = nil then
  begin
    Inc(Result);
    Log(['=> AddRgnDecisionAct, No faction selected']);
  end;

  if cuRegion = nil then
  begin
    Inc(Result);
    Log(['=> AddRgnDecisionAct, No region selected']);
  end;

  if UpperCase(Params[1]) = 'CURRENT' then
    Value := GameSim.cuTurn
  else
  if Length(Params[2]) = 10 then
    Value := DateToTurn(Params[2])
  else
    Value := -1;
  if (Value <= 0)  then
  begin
    Inc(Result);
    Log(['=> AddRgnDecisionAct, Invalid turn added, erroneous date, turn added', Value]);
  end;

  ObjectUID := STI(Params[0]);
  if (ObjectUID <= 0) or (ObjectUID > GameSim.RgnDecisionsDB.Count-1) then
  begin
    Inc(Result);
    Log(['=> AddRgnDecisionAct, Invalid decision, UID found', ObjectUID]);
  end;

  if Result <> 0 then Exit;

  if Value > GameSim.cuTurn then
  begin
    Log(['[WARNING] AddRgnDecisionAct, the decision can''t be added in the future', GameSim.RgnDecisionsDB[ObjectUID].Name, 'Turn Added', Value, 'Current Turn', GameSim.cuTurn]);
    Exit;
  end;
  if GameSim.RgnDecisionsDB[ObjectUID].Duration + Value <= GameSim.cuTurn then
  begin
    Log(['[WARNING] AddRgnDecisionAct, the decision is already too old, compared to its duration', GameSim.RgnDecisionsDB[ObjectUID].Name, 'Turn Added', Value, 'Current Turn', GameSim.cuTurn, 'Duration', GameSim.RgnDecisionsDB[ObjectUID].Duration]);
    Exit;
  end;

  if cuFaction.IsRgnDecisionActValid(cuRegion.UID, ObjectUID, GameSim.cuTurn, True) <> _rgdIsValid then
  begin
    Log(['[WARNING] AddRgnDecisionAct, the decision can''t be added, some conditions are not met', GameSim.RgnDecisionsDB[ObjectUID].Name]);
    Exit;
  end;

  if cuFaction.AddRgnDecisionAct(cuRegion.UID, ObjectUID, Value) then
    Log(['AddRgnDecisionAct', 'The decision', GameSim.RgnDecisionsDB[ObjectUID].Name, 'is now active in', cuRegion.UID, cuRegion.Name, 'with a start turn of', Value, '(current turn is:', GameSim.cuTurn, ')'])
  else
    Log(['[WARNING] AddRgnDecisionAct', 'the decision was not added, an identical decision was already there', GameSim.RgnDecisionsDB[ObjectUID].Name]);

end;
function TScriptEngine.RemRgnDecisionAct (Params : TStringList) : integer;
var
    ObjectID : TUID;
begin
  Result := 0;
  if not (Params.Count > 0) then
  begin
    Inc(Result);
    Log(['=> Syntax is RemRgnDecisionAct = <RgnDecisionDefUID_or_Attribute>']);
  end;
  if Result <> 0 then Exit;

  if cuFaction = nil then
  begin
    Inc(Result);
    Log(['=> RemRgnDecisionAct, No faction selected']);
  end;

  if cuRegion = nil then
  begin
    Inc(Result);
    Log(['=> RemRgnDecisionAct, No region selected']);
  end;

  ObjectID := STI(Params[0]);

  // Two cases, either a RGD UID or a RGD attribute
  if ObjectID <= 0 then
  begin
    if cuFaction.RemRgnDecisionAct(cuRegion.UID, Params[0]) then
      Log(['RemRgnDecisionAct', 'The decision with attribute', Params[0], 'has been removed in', cuRegion.UID, cuRegion.Name])
    else
      Log(['[WARNING] RemRgnDecisionAct', 'The decision with attribute', Params[0], 'has been removed in', cuRegion.UID, cuRegion.Name]);
  end
  else // Numeric UID
  begin
    if (ObjectID <= 0) or (ObjectID > GameSim.RgnDecisionsDB.Count-1) then
    begin
      Inc(Result);
      Log(['=> RemRgnDecisionAct, Invalid decision, UID found', ObjectID]);
    end;

    if Result <> 0 then Exit;

    if cuFaction.RemRgnDecisionAct(cuRegion.UID, ObjectID) then
      Log(['RemRgnDecisionAct', 'The decision', GameSim.RgnDecisionsDB[ObjectID].Name, 'has been removed in', GameSim.RgnDecisionsDB[ObjectID].Name, cuRegion.UID, cuRegion.Name])
    else
      Log(['[WARNING] RemRgnDecisionAct', 'No such decision to remove in the current region', GameSim.RgnDecisionsDB[ObjectID].Name, cuRegion.UID, cuRegion.Name]);
  end;
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ADD / REM REGION DECISION EXCLUSION
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.AddRgnDecisionExcl(Params : TStringList) : integer;
var
    ObjectUID : TUID;
begin
  Result := 0;
  if (Params.Count <> 1) then
  begin
    Inc(Result);
    Log(['=> Syntax is AddRgnDecisionExcl = <AreaUID>']);
  end;
  if Result <> 0 then Exit;

  if cuFaction = nil then
  begin
    Inc(Result);
    Log(['=> AddRgnDecisionExcl, No faction selected']);
  end;

  ObjectUID := STI(Params[0], True, -1);
  if (ObjectUID < 0) or (ObjectUID > Areas.Count-1) then
  begin
    Inc(Result);
    Log(['=> AddRgnDecisionExcl, Invalid Area, UID found', ObjectUID]);
  end;

  if Result <> 0 then Exit;

  if cuFaction.ilRgnDecisionExcl.IndexOf(ITS(ObjectUID)) = -1 then
    cuFaction.ilRgnDecisionExcl.Add(ObjectUID);
  Log(['AddRgnDecisionExcl, Area', Areas[ObjectUID].UID, Areas[ObjectUID].Name, 'added']);

end;

function TScriptEngine.RemRgnDecisionExcl(Params : TStringList) : integer;
var
    ObjectUID : TUID;
begin
  Result := 0;
  if (Params.Count <> 1) then
  begin
    Inc(Result);
    Log(['=> Syntax is RemRgnDecisionExcl = <AreaUID>']);
  end;
  if Result <> 0 then Exit;

  if cuFaction = nil then
  begin
    Inc(Result);
    Log(['=> RemRgnDecisionExcl, No faction selected']);
  end;

  ObjectUID := STI(Params[0]);
  if (ObjectUID <= 0) or (ObjectUID > Areas.Count-1) then
  begin
    Inc(Result);
    Log(['=> RemRgnDecisionExcl, Invalid Area, UID found', ObjectUID]);
  end;

  if Result <> 0 then Exit;

  if cuFaction.ilRgnDecisionExcl.IndexOf(ITS(ObjectUID)) <> -1 then
    cuFaction.ilRgnDecisionExcl.DeleteByVal(ObjectUID);
  Log(['RemRgnDecisionExcl, Area', Areas[ObjectUID].UID, Areas[ObjectUID].Name, 'deleted']);

end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET COLONIAL STATUS AREA
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetColonialStatusArea(Params : TStringList) : integer;
var Value : integer;
    AreaUID : TUID;
    CA : TColonialArea;
    Region : TRegion;
begin
  Result := 0;
  if Params.Count <> 2 then
  begin
    Inc(Result);
    Log(['=> SetColonialStatusArea command, syntax is: SetColonialStatusArea = <ColAreaUID>;<Status>']);
  end;
  if cuFaction = nil then
  begin
    Inc(Result);
    Log(['=> SetColonialStatusArea, No faction selected']);
  end;

  AreaUID  := STI(Params[0], False, -1);
  Value    := STI(Params[1]);

  if (AreaUID < 0) or (AreaUID > Areas.Count-1) then
  begin
    Inc(Result);
    Log(['=> SetColonialStatusArea: Area alias is not correct:', Params[0]]);
  end;
  {
  if Value > _colTypeDominion then
  begin
    Inc(Result);
    Log(['=> SetColonialStatusArea: Value is not a colonial status', Params[1]]);
  end;
  }
  CA := GameSim.ColModule.GetColAreaFromArea(AreaUID);
  if CA = nil then
  begin
    Inc(Result);
    Log(['=> SetColonialStatusArea: The Area alias is not colonial area, see the file ColonialAreas.ini', Params[1]]);
  end;
  if Result > 0 then Exit;

  Region := cuFaction.Regions.Fetch(CA.CapitalUID, True);
  Region.ColonialStatus := Value;

  GameSim.ColModule.SynchColonialStatus(AreaUID);


  Log(['SetColonialStatusArea', AreaUID, Areas[AreaUID].Name, 'Colonial status set to', ColStatusToString(Value)]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET PILLAGE
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetPillage(Params : TStringList) : integer;
begin
  Result := 0;
  if Params.Count <> 1 then
  begin
    Inc(Result);
    Log(['=> SetPillage command, syntax is: SetPillage = <0_or_1>']);
  end;
  if cuRegion = nil then
  begin
    Inc(Result);
    Log(['=> SetPillage command, no region in selection']);
  end;
  if Result <> 0 then Exit;

  GameSim.RegionCmns[cuRegion.UID].ForageLvl := STI(Params[0]);
  Log(['SetPillage', cuRegion.UID, cuRegion.Name, 'Pillage is now at', STI(Params[0])]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// MOBILIZE
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.Mobilize(Params : TStringList) : integer;
var Attributes : string;
begin

  Result := 0;

  if (Params.Count > 2) then
  begin
    Inc(Result);
    Log(['=> Mobilize command: Syntax is: Mobilize = <MobilizePerc> [;<Attributes Needed>]']);
  end;
  if (cuFaction = nil) then
  begin
    Inc(Result);
    Log(['=> Mobilize command: No faction in selection']);
  end;
  if Result > 0 then Exit;

  Attributes := '';
  if Params.Count = 2 then
    Attributes := Params[1];
  cuFaction.Mobilize(STI(Params[0]), Attributes);

  Log(['Mobilize', 'Triggering mobilization for', cuFaction.Name, '%', Params[0], 'Attributes used:', IIF(Attributes = '', 'None', Attributes)]);

end;

function TScriptEngine.Demobilize (Params : TStringList) : integer;
var Attributes : string;
begin

  Result := 0;

  if (Params.Count > 2) or (Params.Count = 0) then
  begin
    Inc(Result);
    Log(['=> Demobilize command: Syntax is: Demobilize = <DemobilizePerc> [;<Attributes Needed>]']);
  end;
  if (cuFaction = nil) then
  begin
    Inc(Result);
    Log(['=> Demobilize command: No faction in selection']);
  end;
  if Result > 0 then Exit;

  Attributes := '';
  if Params.Count = 2 then
    Attributes := Params[1];
  cuFaction.Demobilize(STI(Params[0]), Attributes);

  Log(['Demobilize', 'Triggering demobilization for', cuFaction.Name, '%', Params[0], 'Attributes used:', IIF(Attributes = '', 'None', Attributes)]);


end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET NEXT ELECTION
// SET THE PERIOD FOR THE NEXT ELECTION
// SetNextElection = Position;Period
// Position: 0 for ruler / 1 for cabinet
// Period: positive integer for turns
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetNextElection (Params : TStringList) : integer;
var
    Value   : integer;
    Position: integer;
    Faction : TFaction;
begin
  Result := 0;

  if not (Params.Count = 2) then // the command accepts only two parameters
  begin
    Inc(Result);
    Log(['=> SetNextElection: Syntax is SetNextElection = <Position>;<Period>']);
  end;

  if (cuFaction = nil) then
  begin
    Inc(Result);
    Log(['=> SetNextElection: No faction selected']);
  end;

  if Result <> 0 then Exit;

  // get the global data for the operation
  Faction := cuFaction; // the current faction

  if Faction = nil then
  begin
    Log(['[Warning] SetNextElection: Unknown target faction']);
    Exit;
  end;

  // get the values from the string
  Position  := STI(Params[0], False, -1);
  Value  := STI(Params[1], False, -1);

  // command accepts only position 0 or 1
  if (Position < 0) or (Position > 1) then
  begin
    Log(['=> SetNextElection: Invalid Position']);
    Exit;
  end;

  // command accepts only positive values for the period
  if (Value < 1) then
  begin
    Log(['=> SetNextElection: The Period must be an integer value higher than zero']);
    Exit;
  end;

  Faction.SetNextElection(Position, Value); // add the election time

  Log(['SetNextElection: ', Faction.Name, 'new election for position',Params[0],'after',Params[1],'turns']);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET RULERITEM ACTIVE STATE
// SetRulerItemActiveState = <RulerUID>;<ActiveState>
// ActiveState : 1=active // 0=inactive
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetRulerItemActiveState (Params : TStringList) : integer;
var
    GS        : TGameSim;
    RuDefUID  : integer;
    ActiveState  : integer;
begin
  Result := 0;
  GS := TGameSim(Owner);

  if (Params.Count < 2) then
  begin
    Inc(Result);
    Log(['=> SetRulerItemActiveState: Wrong number of parameters. Syntax: SetRulerItemActiveState = <RulerUID>;<ActiveState>']);
    Exit;
  end;

  // determines to which faction list this ruler is added...
  if cuFaction = nil then
  begin
    Inc(Result);
    Log(['=> SetRulerItemActiveState: No faction selected']);
    Exit;
  end;

  if SetNumeric(Params[0], RuDefUID) = False then
  begin
    Inc(Result);
    Log(['=> SetRulerItemActiveState: Ruler UID is no integer (check alias). Ruler UID:', Params[0]]);
    Exit;
  end;

  if SetNumeric(Params[1], ActiveState) = False then
  begin
    Inc(Result);
    Log(['=> SetRulerItemActiveState: Ruler ActiveState is no integer (0 or 1). ActiveState entered:', Params[1]]);
    Exit;
  end;

  if ((ActiveState < 0) or (ActiveState > 1)) then
  begin
    Inc(Result);
    Log(['=> SetRulerItemActiveState: Ruler ActiveState is not valid, should be 0 or 1. ActiveState entered:', Params[1]]);
    Exit;
  end;

  if GS.RulerDB.InList(RuDefUID) = False then
  begin
    Inc(Result);
    Log(['=> SetRulerItemActiveState: Ruler UID was not found in Ruler DB. Ruler UID:', RuDefUID]);
    Exit;
  end;

  Log(['SetRulerItemActiveState: Adding new ActiveState to Ruler. RulerItem:', GS.RulerDB[RuDefUID].Name, 'Faction:', cuFaction.Name]);
  // we set the new state locally to the faction...
  cuFaction.SetRIActiveState(RuDefUID, ActiveState);

end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET FA OVERRIDE
// Sets a FAOID to refference from the DB to be able to overwrite
// what comes form other sources, like the Ruler FA
// To be used on Events for example...
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetFAOverride (Params : TStringList) : integer;
var
    FAOID    : integer;
begin
  Result := 0;

  if (Params.Count < 1) then
  begin
    Inc(Result);
    Log(['=> SetFAOverride: Wrong number of parameters. Syntax: SetFAOverride = <FAOID>']);
    Exit;
  end;

  // determines to which faction the FA will be added...
  if cuFaction = nil then
  begin
    Inc(Result);
    Log(['=> SetFAOverride: No faction selected']);
    Exit;
  end;

  if SetNumeric(Params[0], FAOID) = False then
  begin
    Inc(Result);
    Log(['=> SetFAOverride: FAOID is no integer (check alias). FAOID entered:', Params[0]]);
    Exit;
  end;

  cuFaction.SetFAOverride(FAOID);

end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// REM FA OVERRIDE
// Removes a FAOID refference from the faction
// To be used on Events for example...
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.RemFAOverride (Params : TStringList) : integer;
var
    FAOID    : integer;
begin
  Result := 0;

  if (Params.Count < 1) then
  begin
    Inc(Result);
    Log(['=> RemFAOverride: Wrong number of parameters. Syntax: RemFAOverride = <FAOID>']);
    Exit;
  end;

  // determines to which faction the FA will be removed...
  if cuFaction = nil then
  begin
    Inc(Result);
    Log(['=> RemFAOverride: No faction selected']);
    Exit;
  end;

  if SetNumeric(Params[0], FAOID) = False then
  begin
    Inc(Result);
    Log(['=> RemFAOverride: FAOID is no integer (check alias). FAOID entered:', Params[0]]);
    Exit;
  end;

  cuFaction.RemFAOverride(FAOID);

end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET / CHANGE DIPLOMATIC RELATIONSHIPS BIAS
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.AISetDiploRelBias   (Params : TStringList) : integer;
var i         : integer;
    Value     : string;
    FacTag    : string;
    Faction   : TFaction;
begin
  Result := 0;
  if Params.Count < 2 then
  begin
    Inc(Result);
    Log(['=> Syntax is AI.SetDiploRelBias = FacTag;Value <FacTag;Value> ..']);
  end;
  if cuFaction = nil then
  begin
    Inc(Result);
    Log(['=> AI.SetDiploRelBias, No faction selected']);
  end;
  if Result <> 0 then Exit;

  for i := 1 to Params.Count div 2 do
  begin
    FacTag := Params[(i-1)*2];
    Faction := GameSim.Factions.GetFactionByTag(FacTag, False);
    if Faction = nil then
    begin
      Log(['[WARNING] AI.SetDiploRelBias,', 'No target faction selected:', FacTag, 'not understood.']);
      Continue;
    end;

    Value  := Params[(i-1)*2 + 1];

    cuFaction.AIGuide.SetDiploRelBias(Faction, STI(Value, True, 100));
    Log(['AI.SetDiploRelBias', cuFaction.Name, 'bias in relationships toward', Faction.Name, 'is', Value]);
  end;
end;

function TScriptEngine.AIChangeDiploRelBias(Params : TStringList) : integer;
var i         : integer;
    Value     : string;
    FacTag    : string;
    Faction   : TFaction;
begin
  Result := 0;
  if Params.Count < 2 then
  begin
    Inc(Result);
    Log(['=> Syntax is AI.ChgDiploRelBias = FacTag;Value <FacTag;Value> ..']);
  end;
  if cuFaction = nil then
  begin
    Inc(Result);
    Log(['=> AI.ChgDiploRelBias, No faction selected']);
  end;
  if Result <> 0 then Exit;

  for i := 1 to Params.Count div 2 do
  begin
    FacTag := Params[(i-1)*2];
    Faction := GameSim.Factions.GetFactionByTag(FacTag, False);
    if Faction = nil then
    begin
      Log(['[WARNING] AI.ChgDiploRelBias,', 'No target faction selected:', FacTag, 'not understood.']);
      Continue;
    end;

    Value  := Params[(i-1)*2 + 1];

    cuFaction.AIGuide.ChangeDiploRelBias(Faction, STI(Value, True, 100));
    Log(['AI.ChgDiploRelBias', cuFaction.Name, 'bias in relationships toward', Faction.Name, 'changed by', Value, 'end value being', cuFaction.AIGuide.GetDiploRelBias(Faction)]);
  end;

end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET DIPLOMATIC ITEM BIAS
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.AISetDiploItemBias  (Params : TStringList) : integer;
var i         : integer;
    Value     : string;
    FacTag    : string;
    Attrib    : string;
    Faction   : TFaction;
begin
  Result := 0;
  if Params.Count < 3 then
  begin
    Inc(Result);
    Log(['=> Syntax is AI.SetDiploItemBias = FacTag;TreatyAttrib;Value <TreatyAttrib;Value> ..']);
  end;
  if cuFaction = nil then
  begin
    Inc(Result);
    Log(['=> AI.SetDiploItemBias, No faction selected']);
  end;
  if Result <> 0 then Exit;

  FacTag := Params[0];
  Faction := GameSim.Factions.GetFactionByTag(FacTag, False);
  if Faction = nil then
  begin
    Log(['[WARNING] AI.SetDiploItemBias,', 'No target faction selected:', FacTag, 'not understood.']);
    Exit;
  end;

  Params.Delete(0);

  for i := 1 to Params.Count div 2 do
  begin
    Attrib := Params[(i-1)*2];
    Value  := Params[(i-1)*2 + 1];

    // remove erroneously typed number, like *100*
    if Pos('*', Value) = 1 then Value := Copy(Value, 2, 255);
    if Pos('*', Value) = Length(Value) then Value := Copy(Value, 1, Length(Value)-1);

    cuFaction.AIGuide.SetDiploItemBias(Faction, Attrib, STI(Value, True, 100));
    Log(['AI.SetDiploItemBias', cuFaction.Name, 'bias for Diplo Item Attribute', Attrib, 'toward', Faction.Name, 'is', Value]);
  end;
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ADD/REM MAP ALEA
// AreaUID: negative allowed, it means 'the whole map'
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.AddMapAlea (Params : TStringList) : integer;
var
    AreaUID   : TUID;
    DefUID : TUID;
    Active : boolean;
    MA : TMapAlea;
begin
  Result := 0;
  if Params.Count <> 3 then
  begin
    Inc(Result);
    Log(['=> Syntax needs 3 parameters: AddMapAlea = <AreaUID_or_neg_number>; <Type of Alea>; <Already Active>']);
  end;

  if Result > 0 then Exit;

  AreaUID   := STI(Params[0]);
  DefUID := STI(Params[1]);
  Active := Params[2] = '1';

  if GameSim.MapAleas.Exists(AreaUID, DefUID) then
  begin
    Log(['[WARNING] Can''t AddMapAlea', 'AreaUID/Type of Alea', AreaUID, DefUID, 'as an identical entry already exists.']);
    Exit;
  end;

  MA := TMapAlea.Create(GameSim.MapAleas, DefUID, AreaUID, Active);

  GameSim.MapAleas.Add(MA);
  Log(['AddMapAlea', 'Added an entry, AreaUID/DefUID', AreaUID, DefUID]);
end;

function TScriptEngine.RemMapAlea (Params : TStringList) : integer;
var
    AreaUID : TUID;
    TypeUID : TUID;
begin
  Result := 0;
  if Params.Count <> 2 then
  begin
    Inc(Result);
    Log(['=> Syntax needs 2 parameters: RemMapAlea =', '<MapArea>; <Type of Alea>']);
  end;
  if Result > 0 then Exit;

  AreaUID   := STI(Params[0]);
  TypeUID := STI(Params[1]);
  if not GameSim.MapAleas.Exists(AreaUID, TypeUID) then
  begin
    Log(['[WARNING] Can''t RemMapAlea', 'AreaUID/TypeUID', AreaUID, TypeUID, 'as no identical entry already exists.']);
    Exit;
  end;
  
  GameSim.MapAleas.RemoveEntry(AreaUID, TypeUID);
  Log(['RemMapAlea', 'Removed entry, AreaUID/TypeUID', AreaUID, TypeUID]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// AI ADD LEADER AGENT
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.AIAddLeaderAgent(Params : TStringList) : integer;
var
    Param       : string;
    UIDList     : TStringList;
    i           : integer;
    AgentUID    : TUID;
    LeaderUID   : TUID;
    SU          : TSubUnit;
    Model       : TModel;
    AIA         : TAIAgent;
begin
  Result := 0;
  if (Params.Count < 3) then
  begin
    Log(['=> AI.AddLeaderAgent: Syntax is AI.AddLeaderAgent = <AgentUID>; <LeaderUID_or_cuSubUnit>; <Regions UID1 UID2 UID3 ...>']);
    Inc(Result);
    Exit;
  end;
  if cuFaction = nil then
  begin
    Log(['=> AI.AddLeaderAgent: No Defined faction.']);
    Inc(Result);
    Exit;
  end;

  // IMPORTANT: troubles when parsing aliases within same parameter, for example
  // Regions $Reg1 $Reg2
  // so we translate manually the aliases, one by one
  // this command do all alias PARSING here, not in the main TScriptEngine.Parse

  UIDList := TStringList.Create;
  Param :=  ExtractParam('Regions', Params);
  if Param <> '' then
  begin
    ExtractArgs(Param);
    i := 0;
    while (i <= High(Args) +1)and(Args[i] <> '') do
    begin
      if i = High(Args) +1 then
      begin
        Log(['=> AI.AddLeaderAgent: Too many regions listed, max is', High(Args)]);
        Inc(Result);
        Exit;
      end;

      TFileParser.ParseForAliases(Args[i]);
      if IsNumeric(Args[i]) then
        UIDList.Add(Args[i])
      else
      begin
        Log(['=> AI.AddLeaderAgent', Args[i], 'is not a valid region alias']);
        Inc(Result);
      end;
      Inc(i);
    end;
  end
  else
  begin
    Log(['=> AI.AddLeaderAgent: Regions list is mandatory']);
    Inc(Result);
    Exit;
  end;


  AgentUID := STI( Translate(Params[0]));
  if AgentUID > GameSim.AIAgentsDB.Count-1 then
  begin
    Log(['=> AI.AddLeaderAgent: Agent UID is too high, value', AgentUID, 'Max Value', GameSim.AIAgentsDB.Count-1]);
    Inc(Result);
    Exit;
  end;

  // Select subunit as leader or specified as model UID?
  if (LowerCase(Params[1]) = 'cusubunit') and (cuSubUnit <> nil)
  and cuSubUnit.IsLeader and (cuFaction <> nil) and (cuSubUnit.FactionUID = cuFaction.UID) then
    SU := cuSubUnit
  else
  begin
    LeaderUID := STI( Translate(Params[1]));
    if ((LeaderUID <= 0)and sysErroneousZero_) or(LeaderUID <= _noUID) then
    begin
      Log(['=> AI.AddLeaderAgent: Leader UID is too low, value', LeaderUID]);
      Inc(Result);
      Exit;
    end;

    Model := Models[LeaderUID];
    SU := cuFaction.GetSubUnitByName(Model.Name, True); // this will search whatever the rank...
    if SU = nil then
    begin
      Log(['[WARNING] AI.AddLeaderAgent: Leader not found, name searched', Model.Name, 'of Def UID', Model.UID, 'higher ranks would have been found too']);
      Exit;
    end;
  end;

  AIA := cuFaction.AIGuide.AddSUAgent(SU, AgentUID);
  if AIA <> nil  then
  begin
    AIA.Priority := 500;
    AIA.ilRegionsList.Text := UIDList.Text;
    Log(['AI.AddLeaderAgent', 'Added successfully Agent', AgentUID, GameSim.AIAgentsDB[AgentUID].Name, 'on leader', SU.UID, SU.Name, 'with', UIDList.Count, 'regions in objectives list.']);
  end
  else
    Log(['AI.AddLeaderAgent', 'Failed to add Agent', AgentUID, GameSim.AIAgentsDB[AgentUID].Name, 'on leader', SU.UID, SU.Name]);

  UIDList.Free;
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// TAKE OVER UNITS
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.TakeOverUnits(Params : TStringList) : integer;
var sub       : integer;
    SU        : TSubUnit;
    pList     : TList;
    StackLdr  : boolean;
    Group     : TGroup;
    NewGrp    : TGroup;
begin
  Result := 0;
  if (Params.Count < 1) or (Params.Count > 1) then
  begin
    Inc(Result);
    Log(['=> Syntax needs 1 parameters TakeOverUnits = <ByStackLeader_or_PerUnit>']);
  end;
  if (cuFaction = nil) then
  begin
    Inc(Result);
    Log(['=> TakeOverUnits: No faction selected']);
  end;
  if (cuRegion = nil) then
  begin
    Inc(Result);
    Log(['=> TakeOverUnits: No region selected']);
  end;
  if Result > 0 then Exit;

  if cuSUList.Count = 0 then
  begin
    Log(['[WARNING] No subunits in selection, exiting.']);
    Exit;
  end;
  pList := TList.Create;
  StackLdr := ExtractParam('ByStackLeader', Params) <> '';

  // find candidates
  for sub := 0 to cuSUList.Count-1 do
  begin
    SU := cuSUList[sub];
    if StackLdr and (TGroup(SU.pGroup).GetLeader <> SU) then Continue; // must be  the stack leader
    if (not StackLdr) and (TUnit(SU.pUnit)[0] <> SU) then Continue;    // must be the first SU of the unit
    pList.Add(SU);
  end;

  if plist.Count > 0 then
    NewGrp := cuFaction.CreateGroupInRegion(cuRegion, _genUID, _genName, _gkLand, 0, False, _fixUnfixed, _psDefensive)
  else
  begin
    Log(['[WARNING]', cuSUList.Count, 'SubUnits in initial selection, but 0 remaining after filtering, exiting.']);
    Exit;
  end;

  for sub := 0 to pList.Count-1 do
  begin
    SU := pList[sub];
    Group := TGroup(SU.pGroup);
    if StackLdr then NewGrp.AddGroup(Group, True, True);
    if (not StackLdr) then
    begin
      NewGrp.BorrowUnit(TUnit(SU.pUnit), True, True);
    end;
  end;

  if NewGrp.Count > 0 then
  begin
    NewGrp.Update();
  Log(['TakeOverUnits', NewGrp.Name, 'in', cuRegion.Name, 'created with', NewGrp.Count, 'units taken away']);
  end
  else
   cuFaction.KillGroup(NewGrp);

  GameSim.RegionCmns.Update(cuRegion.UID);


  pList.Free;
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// MAKE TECH KNOWN
// MakeTechKnown = <RIUID>; <Change/Ignore/Replace> [;DiscoveryPerc] [;RunScript]
// RIUID : UID of the Research Item
// Replace: Has only an effect then the RI is already known: If the string
//          parameter "Replace" is given, the Discovery Percentage of the RI
//          will be SET to DiscoveryPerc.
// Change: Has only an effect then the RI is already known: If the string
//         parameter "Change" is given, the Discovery Percentage of the RI
//         will be CHANGED by DiscoveryPerc.
// Ignore: Has only an effect then the RI is already known: If the string
//         parameter "Ignore" is given, the script command will be aborted if
//         the RI is already known.
// DiscoveryPerc : The RI is made known and the Discovery Percentage set go the
//         given value. Otherwise the Base Precentage Value from the RI Definition
//         will be used.
// RunScript: If the keyword 'RunScript' is added the script (if any) will be
//         run if the RI is made _mastered_ (by setting the Discovery
//         Percentage to 100). The Faction Modifiers will be applied in all
//         cases
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.MakeTechKnown(Params : TStringList) : integer;
var RIUID     : integer;
    DiscPerc  : integer;
    RunScript : boolean;
    //RI        : TResearchItem;
    RD        : TResearchDef;
    GS        : TGameSim;
    Str       : string;

begin

  Result := 0;
  RunScript := False;
  GS := TGameSim(Owner);
  DiscPerc := low(integer); // low(integer) means: Optional parameter DiscoveryPerc not present.

  Log(['MakeTechKnown: Begin']);

  if (Params.Count < 2) or (Params.Count > 4) then
  begin
    Inc(Result);
    Log(['=> MakeTechKnown: Wrong number of parameters. Syntax: MakeTechKnown = <RIUID>; <Change/Ignore/Replace> [;DiscoveryPerc] [;RunScript]']);
    Exit;
  end;

  if cuFaction = nil then
  begin
    Inc(Result);
    Log(['=> MakeTechKnown: No faction selected']);
    Exit;
  end;

  if not SetNumeric(Params[0], RIUID) then
  begin
    Inc(Result);
    Log(['=> MakeTechKnown: Research Item UID is no integer. Check if correct alias used. RIUID:', Params[0] ]);
    Exit;
  end;

  if GS.ResearchDB.GetByUID(RIUID) = nil then
  begin
    Inc(Result);
    Log(['=> MakeTechKnown: Research Item UID was not found in database. Check if correct alias used. RIUID:', Params[0] ]);
    Exit;
  end;


  Str := lowercase(Trim(Params[1]));
  if (Str <> 'change') and (Str <> 'replace') and (Str <> 'ignore') then
  begin
    Inc(Result);
    Log(['=> MakeTechKnown: Incorrect parameter (2). Must be "Ignore", "Change" or "Replace". Parameter:', Str ]);
    Exit;
  end;



  // 3 Parameter present
  if Params.Count = 3 then // Parameter could be the Discovery Perc. or 'RunScript'
  begin
    // Check if RunScript parameter present
    if Trim(lowercase(Params[2])) = lowercase('RunScript') then
    begin
      RunScript := True;
      DiscPerc := low(integer);  // low(integer) means: Optional parameter DiscoveryPerc not present.
    end
    else
    // Set DiscPerc
    begin
      RunScript := False;
      if not SetNumeric(Params[2], DiscPerc) then
      begin
        Inc(Result);
        Log(['=> MakeTechKnown: Incorrect Parameter (3). Not a valid DiscoveryPerc (no integer) nor "RunScript". Parameter:', Params[1] ]);
        Exit;
      end;
    end;
  end;


  // 4 Parameter present
  if Params.Count = 4 then
  begin
    // Set DiscPerc
    if not SetNumeric(Params[2], DiscPerc) then
    begin
      Inc(Result);
      Log(['=> MakeTechKnown: Discovery Percentage is no integer. Check if correct alias used. DiscoveryPerc:', Params[2] ]);
      Exit;
    end;
    // Check RunScript parameter
    if Trim(lowercase(Params[3])) = lowercase('RunScript') then
      RunScript := True
    else
    begin
      Inc(Result);
      Log(['=> MakeTechKnown: Invalid optional parameter (4). Must be "RunScript". Parameter:', Params[3] ]);
      Exit;
    end;
  end;



  RD := GS.ResearchDB.GetByUID(RIUID); // Get the Research Def.
  if DiscPerc = low(integer) then
    DiscPerc := RD.BasePerc; //  We give the base perc if no value for DiscoveryPerc is given as parameter.

  //RI := cuFaction.olResearch.GetByUID(RIUID); // Get the RI if already present in faction


  cuFaction.Tech_Discover(RIUID, Trim(lowercase(Params[1])) = 'change', DiscPerc, RunScript);

  {
  // RI is unknown, will be added.
  if RI = nil then
  begin // If something is changed here check THost.CheckResearch too, same code there. Should maybe be put to a new method: TFaction.MakeRIKnown.
    Log(['MakeTechKnown: RI was unknown to the faction and is now added. RI:', RD.UID, RD.Name]);

    RI := cuFaction.olResearch.AddItem;
    RI.UID  := RD.UID;
    RI.Name := RD.Name;
    RI.Kind := RD.Kind;

    RI.Discovery := DiscPerc;
    Log(['MakeTechKnown: Progress set to', RI.Discovery]);

    if GameSim.DataType <> _dtScen then
      cuFaction.Messages.MakeTechKnown(RI.Name);
  end
  else // RI was already known before
  begin
    if RI.IsMastered = True then
    begin
      Log(['MakeTechKnown: RI already known and mastered. Doing nothing.']);
      Exit;
    end;
    if Trim(lowercase(Params[1])) = 'change'  then
    begin
      RI.Discovery := RI.Discovery + DiscPerc;
      Log(['MakeTechKnown: Progress changed by:', DiscPerc, 'to:', RI.Discovery]);
    end
    else
    if Trim(lowercase(Params[1])) = 'replace' then
    begin
      RI.Discovery := DiscPerc;
      Log(['MakeTechKnown: Progress set to', RI.Discovery]);
    end
    else
    begin
      Log(['MakeTechKnown: RI already known and ignored.']);
      Exit;
    end;
  end;

  // Tech is now mastered, apply the effects.
  if RI.IsMastered then
  begin
    if GameSim.DataType <> _dtScen then
      cuFaction.Messages.MakeTechMastered(RI.Name);
    if RunScript = True then
    begin
      RI.ApplyEffects(True);
      Log(['MakeTechKnown: Tech mastered. Effects applied. Script run.', RI.Discovery]);
    end
    else
    begin
      RI.ApplyEffects(False);
      Log(['MakeTechKnown: Tech mastered. Effects applied. Script not run.', RI.Discovery]);
    end;
  end;
  }
end;



// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ADD RULER TO FACTION
// AddRulerToFac = <RulerUID>;<Position> [;NoReplace]
// Position  : (integer/alias) The position of the Ruler in the Faction
// NoReplace : (string) If this parameter is added, the Ruler is not replaced if
//             the Position is already taken.
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.AddRulerToFac(Params : TStringList) : integer;
var NoReplace : boolean;
    GS        : TGameSim;
    RuDefUID  : integer;
    RuI       : TRulerItem;
    RulerPos  : integer;

begin

  Result := 0;
  NoReplace := False;
  GS := TGameSim(Owner);

  if (Params.Count < 2) or (Params.Count > 3) then
  begin
    Inc(Result);
    Log(['=> AddRulerToFac: Wrong number of parameters. Syntax: AddRulerToFac = <RulerUID>;<Position> [;NoReplace]']);
    Exit;
  end;

  if cuFaction = nil then
  begin
    Inc(Result);
    Log(['=> AddRulerToFac: No faction selected']);
    Exit;
  end;

  if SetNumeric(Params[0], RuDefUID) = False then
  begin
    Inc(Result);
    Log(['=> AddRulerToFac: Ruler UID is no integer (check alias). Ruler UID:', Params[0]]);
    Exit;
  end;

  if SetNumeric(Params[1], RulerPos) = False then
  begin
    Inc(Result);
    Log(['=> AddRulerToFac: Ruler Position is no integer (check alias). Ruler UID:', Params[1]]);
    Exit;
  end;

  if Params.Count = 3 then
    if Trim(lowercase(Params[3])) = 'noreplace' then
      NoReplace := true;


  if GS.RulerDB.InList(RuDefUID) = False then
  begin
    Inc(Result);
    Log(['=> AddRulerToFac: Ruler UID was not found in Ruler DB. Ruler UID:', RuDefUID]);
    Exit;
  end;


  RuI := cuFaction.RulerList.GetRulerByPos(RulerPos);
  if RuI <> nil then
  begin
    if (NoReplace = true) then
    begin
      Log(['AddRulerToFac: Ruler Position already taken and parameter ''NoReplace'' set: Nothing will be done.']);
      Exit;
    end
    else
    begin
      Log(['AddRulerToFac: Ruler Position already taken: Will be replaced.']);
      cuFaction.RulerList.Remove(RuI);
    end;
  end;

  RuI := cuFaction.RulerList.AddItem(RuDefUID, RulerPos);

  Log(['AddRulerToFac: New Ruler added to Faction. Ruler UID:', RuI.UID, 'Faction:', cuFaction.FactionDef.FactionTag]);

end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// CHANGE LEADERS POOL
// Change the allowance in leaders (by rank), for the provisional leader system
// 1 to 3 for land leader rank 1 to 3, 11 to 13 for nav leader rank 1 to 3
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.ChgLeadersPool(Params : TStringList) : integer;
var i         : integer;
    ObjectUID : TUID;
    Value     : integer;
begin
  Result := 0;
  if not (Params.Count > 0) then
  begin
    Inc(Result);
    Log(['=> Syntax is ChangeLeadersPool = [;<LeaderRankAlias|Variation>] ..']);
  end;
  if Params.Count mod 2 <> 0 then
  begin
    Inc(Result);
    Log(['=> ChangeLeadersPool, Specify a list of paired values: LeaderRankAlias|Variation']);
  end;
  if cuFaction = nil then
  begin
    Inc(Result);
    Log(['=> ChangeLeadersPool, No faction selected']);
  end;
  if Result <> 0 then Exit;

  for i := 1 to Params.Count div 2 do
  begin
    ObjectUID := STI(Params[(i-1)*2]);
    Value     := STI(Params[(i-1)*2 + 1]);
    cuFaction.dlLeadersPoolNb.AlterValByKey(ObjectUID, Value, True);  // arithmetic sum (with old value). Entry removed if sum is <= 0
    Log(['Number allowed for rank', ObjectUID,':', cuFaction.dlLeadersPoolNb.GetValByKey(ObjectUID)]);
  end;
end;
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ADD/REMOVE PROVISIONAL LEADERS
// The list of Provisional leaders (UnitDefs)
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.AddProvLeaders(Params : TStringList) : integer;
var i         : integer;
    Value     : integer;
begin
  Result := 0;
  if not (Params.Count > 0) then
  begin
    Inc(Result);
    Log(['=> AddProvLeaders: Syntax is AddProvLeaders = [LdrUID1;LdrUID2;..]']);
  end;
  if (cuFaction = nil) then
  begin
    Inc(Result);
    Log(['=> AddProvLeaders command, no faction selected']);
  end;
  if Result <> 0 then Exit;

  for i := 0 to Params.Count-1 do
  begin
    Value     := STI(Params[i]);
    cuFaction.ilProvLeaders.Add(Value);
    Log(['AddProvLeaders', cuFaction.Name, 'Leader', Value, UnitDefs[Value].Name, 'added.']);
  end;
end;

function TScriptEngine.RemProvLeaders(Params : TStringList) : integer;
var i         : integer;
    Value     : integer;
begin
  Result := 0;
  if not (Params.Count > 0) then
  begin
    Inc(Result);
    Log(['=> RemProvLeaders: Syntax is RemProvLeaders = [LdrUID1;LdrUID2;..]']);
  end;
  if (cuFaction = nil) then
  begin
    Inc(Result);
    Log(['=> RemProvLeaders command, no faction selected']);
  end;
  if Result <> 0 then Exit;

  for i := 0 to Params.Count-1 do
  begin
    Value     := STI(Params[i]);
    cuFaction.ilProvLeaders.DeleteByVal(Value);
    Log(['RemProvLeaders', cuFaction.Name, 'Leader', Value, UnitDefs[Value].Name, 'removed.']);
  end;
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET ALL FAC TAXES
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetAllFacTaxes(Params : TStringList) : integer;
var i, Key, Value, fac : integer;
    sValue : string;
    Faction : TFaction;
begin
  Result := 0;
  if Params.Count mod 2 = 0 then
  begin
    Inc(Result);
    Log(['=> SetAllFacTaxes, Specify a list of paired data: Tax alias (UID) and then value']);
    Exit;
  end;
  for i := 1 to Params.Count div 2 do
  begin
    Key     := STI(UpperCase(Params[(i-1)*2 +1]));
    sValue  := Params[(i-1)*2 + 2];
    Value   := STI(sValue);
    Log(['SetAllFacTaxes, Tax', Key, 'set to', Value, 'for all factions.']);
    for fac := 0 to GameSim.Factions.Count-1 do
    begin
      Faction := GameSim.Factions[fac];
      case Key of
        _taxCensus        : Faction.Tax_Census := Value;
        _taxCorporate     : Faction.Tax_Corporate := Value;
        _taxExcise        : Faction.Tax_Excise := Value;
        _taxTariff        : Faction.Tax_Tariff := Value;
        _taxIncome        : Faction.Tax_Income := Value;
        _taxMaritime      : Faction.Tax_Maritime := Value;
        else Log(['[WARNING] SetAllFacTaxes command, Erroneous Tax identifier:', Key]);
      end;
    end;
  end;
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET TAXES
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetTaxes(Params : TStringList) : integer;
var i, Key, Value : integer;
    sValue : string;
begin
  Result := 0;
  if Params.Count mod 2 = 0 then
  begin
    Inc(Result);
    Log(['=> SetTaxes, Specify a list of paired data: Tax alias (UID) and then value']);
    Exit;
  end;
  if cuFaction = nil then
  begin
    Log(['=> SetTaxes command, no faction selected']);
    Inc(Result);
    Exit;
  end;
  for i := 1 to Params.Count div 2 do
  begin
    Key     := STI(UpperCase(Params[(i-1)*2 +1]));
    sValue  := Params[(i-1)*2 + 2];
    Value   := STI(sValue);
    Log(['SetTaxes, Tax', Key, 'set to', Value, 'for all factions.']);
    case Key of
      _taxCensus        : cuFaction.Tax_Census := Value;
      _taxCorporate     : cuFaction.Tax_Corporate := Value;
      _taxExcise        : cuFaction.Tax_Excise := Value;
      _taxTariff        : cuFaction.Tax_Tariff := Value;
      _taxIncome        : cuFaction.Tax_Income := Value;
      _taxMaritime      : cuFaction.Tax_Maritime := Value;
      else Log(['[WARNING] SetTaxes command, Erroneous Tax identifier:', Key]);
    end;
  end;
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ADD DYNAMIC GRAPHIC
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.AddDynamicGraphic (Params : TStringList) : integer;
var RgnUID : TUID;
    TerrainUID : TUID;
    Layer : integer;
begin
  Result := 0;
  if not (Params.Count in [7..8]) then
  begin
    Inc(Result);
    Log(['=> AddDynamicGraphic, Wrong number of parameters, syntax is: AddDynGraphic = <RgnUID>;<ImageName>;<AltWeatherImgName>;<TerrainUID>;<CooX>;<CooY>;<Tooltip>']);
    Exit;
  end;
  RgnUID := STI(Params[0]);

  if Params.Count = 7 then
    Layer := 1
  else
    Layer := STI(Params[7]);

  if Params[3] = 'NULL' then
    TerrainUID := _noUID
  else
    TerrainUID := STI(Params[3]);
  GameSim.RegionCmns[RgnUID].AddDynamicGraphic(Params[1], Params[2], TerrainUID, STI(Params[4]), STI(Params[5]), Params[6], Layer);
  Log(['AddDynamicGraphic', RgnUID, GameSim.RegionCmns[RgnUID].Name, 'Dynamic Graphic is now:', Params[1] + ', ' + Params[2]+ ', ' + ITS(TerrainUID) + ', ' + Params[4] + ', ' + Params[5]+ ', ' + Params[6] + 'Layer ' + ITS(Layer)]);

end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// REM DYNAMIC GRAPHIC
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.RemDynamicGraphic (Params : TStringList) : integer;
var
    RgnUID : TUID;
begin
  Result := 0;
  if Params.Count <> 2 then
  begin
    Inc(Result);
    Log(['=> RemDynamicGraphic, Wrong number of parameters, syntax is: RemDynGraphic = <RgnUID>;<Image>']);
    Exit;
  end;
  RgnUID := STI(Params[0]);
  GameSim.RegionCmns[RgnUID].DelDynamicGraphic(Params[1]);

  Log(['RemDynamicGraphic', RgnUID, GameSim.RegionCmns[RgnUID].Name, 'All Dynamic Graphics using this image:', Params[1], 'in the region deleted.']);

end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// CHANGE FACTION DISPLAY TAG
// Change the display tag of a faction (allowing the dynamic change of many graphic
// files, if option inoFactionAutocalcImgName is activated). Must be a 3-letters tag.
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.ChangeFacDisplayTag  (Params : TStringList) : integer;
begin
  Result := 0;
  if Params.Count <> 1 then
  begin
    Inc(Result);
    Log(['=> ChgFacDisplayTag, Wrong number of parameters, syntax is: ChgFacDisplayTag = <NewTag>']);
    Exit;
  end;
  if cuFaction = nil then
  begin
    Log(['=> ChgFacDisplayTag command, no faction selected']);
    Inc(Result);
    Exit;
  end;
  if Length(Params[0]) <> 3 then
  begin
    Log(['=> ChgFacDisplayTag command, Display Tag is not a 3 characters string.']);
    Inc(Result);
    Exit;
  end;

  cuFaction.DisplayTag := Params[0];

  Log(['ChgFacDisplayTag', cuFaction.Name, 'Display Tag is now', Params[0]]);

end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET FACTION DISPLAY NAMES
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetFacDisplayNames(Params : TStringList) : integer;
var
    i : integer;
begin
  Result := 0;
  if (cuFaction = nil) then
  begin
    Log(['=> SetFacDisplayNames: No faction selected']);
    Inc(Result);
    Exit;
  end;
  if Params.Count<> 6 then
  begin
    Inc(Result);
    Log(['=> SetFacDisplayNames: The correct syntax is SetFacDisplayNames = Name|ShortName|Text|AdjectiveM|AdjectiveF|PopName. Also you should not use aliases (no $).']);
    Exit;
  end;
  if Result <> 0 then Exit;

  cuFaction.SerialDisplayNames := '';

  for i := 0 to Params.Count-1  do
    cuFaction.SerialDisplayNames := cuFaction.SerialDisplayNames + Params[i] + '|';

  Log(['SetFacDisplayNames:', cuFaction.Name, 'has now these data for its 6 DisplayNames', cuFaction.SerialDisplayNames]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET FACTION DISPLAY FORCE NAMES
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetFacDisplayFrcNames(Params : TStringList) : integer;
var
    i : integer;
begin
  Result := 0;
  if (cuFaction = nil) then
  begin
    Log(['=> SetFacDisplayFrcNames: No faction selected']);
    Inc(Result);
    Exit;
  end;
  if Params.Count<> 15 then
  begin
    Inc(Result);
    Log(['=> SetFacDisplayFrcNames: The correct syntax is SetFacDisplayFrcNames = LandHQSName|LandFrcName|LandDetName|NavFrcName|NavLdrName|Leader3Force|Leader2Force|Leader1Force|']);
    Log(['=> Leader3ColoForce|Leader2ColoForce|Leader1ColoForce|ColoName|SmallCombiName|MedCombiName|LargeCombiName. Also you should not use aliases (no $).']);
    Exit;
  end;
  if Result <> 0 then Exit;

  cuFaction.SerialDisplayFrcNames := '';

  for i := 0 to Params.Count-1  do
    cuFaction.SerialDisplayFrcNames := cuFaction.SerialDisplayFrcNames + Params[i] + '|';

  Log(['SetFacDisplayFrcNames:', cuFaction.Name, 'has now these data for its 15 DisplayForceNames', cuFaction.SerialDisplayFrcNames]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET FACTION DISPLAY COLORS
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetFacDisplayColors(Params : TStringList) : integer;
var
    i : integer;
begin
  Result := 0;
  if (cuFaction = nil) then
  begin
    Log(['=> SetFacDisplayColors: No faction selected']);
    Inc(Result);
    Exit;
  end;
  if Params.Count<> 3 then
  begin
    Inc(Result);
    Log(['=> SetFacDisplayColors: The correct syntax is SetFacDisplayColors = Color|SecColor|MinimapColor. Also you should use unparsable aliases (no $) and no data as in 150|25|175|255 to indicate a color.']);
    Exit;
  end;
  if Result <> 0 then Exit;

  cuFaction.SerialDisplayColors := '';

  for i := 0 to Params.Count-1  do
    cuFaction.SerialDisplayColors := cuFaction.SerialDisplayColors + Params[i] + '|';

  Log(['SetFacDisplayColors:', cuFaction.Name, 'has now these data for its 15 DisplayForceNames', cuFaction.SerialDisplayColors]);
end;

// PREVIOUS CODE, when changing FACTION TAG ITSELF WAS CONSIDERED:

(*
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// CHANGE FACTION TAG
// Change the faction def of a faction (France pointing to F01 and not FRA e.g)
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.ChangeFactionTag  (Params : TStringList) : integer;
var FD : TFactionDef;
begin
  Result := 0;
  if Params.Count <> 1 then
  begin
    Inc(Result);
    Log(['=> ChgFactionTag, Wrong number of parameters, syntax is: ChgFactionTag = <NewTag>']);
    Exit;
  end;
  if cuFaction = nil then
  begin
    Log(['=> ChgFactionTag command, no faction selected']);
    Inc(Result);
    Exit;
  end;

  FD := FactionDefs.GetByFactionTag(Params[0]);

  if FD = nil then
  begin
    Log(['=> ChgFactionTag command, The new faction tag is unknown in the database.']);
    Inc(Result);
    Exit;
  end;

  if FD.FactionFileID <> cuFaction.FactionDef.FactionFileID then
  begin
    Log(['[Warning] ChgFactionTag command, both factions must share the same FactionFileID.']);
    Exit;
  end;

  cuFaction.DefType := FD.UID;

  Log(['ChgFactionTag', cuFaction.Name, 'faction is now pointing to the definitions of this faction tag:', Params[0]]);

end;
*)

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// CHANGE GAME RULES
// Change the rules of the game. USE WITH CAUTION!
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.ChangeGameRules   (Params : TStringList) : integer;
var i : integer;
    Key    : string;
    Value  : integer;
    sValue : string;
begin
  Result := 0;

  if Params.Count mod 2 <> 0 then
  begin
    Inc(Result);
    Log(['=> ChangeGameRules, Specify a game rule (as a string, e.g Cohesion) and then a numeric data (use 1 for True and 0 for False).']);
    Log(['=> Valid Identifiers are Morale, ResearchMode, EconomyMode, DiplimacyMode, CommerceMode, PoliticalMode, SupplyDecay.']);
    Exit;
  end;

  for i := 1 to Params.Count div 2 do
  begin
    Key     := Trim(UpperCase(Params[(i-1)*2]));
    sValue  := Trim(Params[(i-1)*2 + 1]);
    Value   := STI(sValue);

    if Key = 'MORALE' then
      rulNationalMorale_ := Value else
    if Key = 'RESEARCHMODE' then
      rulResearches_ := Value = 1 else
    if Key = 'ECONOMYMODE' then
      rulEconomyMode_ := Value else
    if Key = 'DIPLOMACYMODE' then
      rulDiplomacyMode_ := Value else
    if Key = 'COMMERCEMODE' then
      rulCommerceMode_ := Value else
    if Key = 'POLITICALMODE' then
      rulPoliticalMode_ := Value else
    if Key = 'SUPPLYDECAY' then
      rulSupplyDecay_ := Value = 1 else
    if Key = 'SUPPLYFREQUENCY' then
      rulSupplyFrequency_ := Value else
    DoNothing;

    Log(['ChangeGameRules', 'Rule: ', Key, 'has been changed to this value:', sValue])
  end;



end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET (FLAVOR) MESSAGE GENERATION
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetMsgGeneration(Params : TStringList) : integer;
begin
  Result := 0;

  if not (Params.Count in [1]) then
  begin
    Inc(Result);
    Log(['=> MsgNeedRegion: Syntax is MsgNeedRegion = [1 or 0]']);
    Exit;
  end;

  MsgNeedRegion := Params[0] = '1';
  Log(['MsgNeedRegion set to ', IIF(MsgNeedRegion, 'True', 'False')]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET COMBI UNITS
// Nb d'units combines permises par faction
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetCombiUnits(Params : TStringList) : integer;
begin
  Result := 0;
  if cuFaction = nil then
  begin
    Inc(Result);
    Log(['=> SetCombiUnits command, no faction selected']);
    Exit;
  end;
  if Params.Count <> 3 then
  begin
    Inc(Result);
    Log(['=> SetCombiUnits: Syntax is SetCombiUnits = NbSmall|NbMedium|NbLarge']);
    Exit;
  end;

  cuFaction.CombiUnitsNb := Params[0] + '|' + Params[1] + '|' + Params[2];

  Log(['Combined Units Max number for faction', cuFaction.Name, ' set to ', cuFaction.CombiUnitsNb]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET FORMED COMMAND MAXIMUM
// Max nb of GHQ/HQ stacks, if not set, 9999
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.SetFormedCmdMax(Params : TStringList) : integer;
begin
  Result := 0;
  if cuFaction = nil then
  begin
    Inc(Result);
    Log(['=> SetFormedCmdMax command, no faction selected']);
    Exit;
  end;
  if Params.Count <> 2 then
  begin
    Inc(Result);
    Log(['=> SetFormedCmdMax: Syntax is SetFormedCmdMax = NbCorpsStacks|NbArmiesStacks']);
    Exit;
  end;

  cuFaction.FormedCmdInfos := Params[0] + '|' + Params[1];

  Log(['SetFormedCmdMax for faction', cuFaction.Name, ' set to ', cuFaction.FormedCmdInfos]);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// APPLY
// Execute la commande en cours
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.Apply : integer;
begin
  Result := 0;
  if WorkingGME = '' then
  begin
    Result := 1;
    Log(['=> Apply command, but nothing to do']);
    Exit;
  end;
  if WorkingGME = _geGroup then
    Result := ApplyCreateGroup
  else
  if WorkingGME = _geUnit then
    Result := ApplyCreateUnit
  else
  if WorkingGME = _geSubU then
    Result := ApplyCreateSubUnit
  else
  if WorkingGME = _geStruc then
    Result := ApplyCreateStruc;

  WorkingGME := '';
  Name := '';
  ShortName := '';
  FlavorName := '';
  SUFlavorName := '';
  Text := '';
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// APPLY CREATE GROUP
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.ApplyCreateGroup : integer;
var UID : TUID;
begin
  Result := 0;
  if cuFaction = nil then
  begin
    Inc(Result);
    Log(['=> ApplyCreateGroup command, no faction selected']);
  end;
  if cuRegion = nil then {slection de rgion vide, pas une erreur, la condition n'est tout simplement pas rempli, l'evt ne s'applique plus (mais dclench!) }
  begin
    Log(['[REPORTING] ApplyCreateGroup command, no region selected']);
    Exit;
  end;
  if Result > 0 then
  begin
    Abort;
    Exit;
  end;

  if WorkingVars.Count = 0 then
    UID := MakeGmUID(_genUID)
  else
  begin
    UID := LabelToUID(WorkingVars[0]);
    WorkingVars.Delete(0);
  end;

  if Kind = _gkUnknown then
    Kind := _gkLand;
  if Name = '' then Name := _genName;
  cuGroup := cuFaction.CreateGroupInRegion(cuRegion, UID, Name, Kind, Entrench, InCS, FixType, GroupPosture);

  cuGroup.NeedUpdate := True; // postponed update, done once all events are processed
    
  Entrench      := 0;
  InCS          := False;
  GroupPosture  := _psDefensive;                       
  if cuGroup = nil then                                    
  begin
    Inc(Result);
    Log(['=> ApplyCreateGroup command, failed to create group in region', cuRegion.Name, 'for faction', cuFaction.Name]);
    Exit;
  end
  else
    Log([cuFaction.Name, cuGroup.Name, 'created in region', cuRegion.Name]);
  Level := 0;

end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// APPLY CREATE UNIT
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.ApplyCreateUnit : integer;
var UID : TUID;
    i : integer;
    lsCombo : TList;
    aUnit   : TUnit;
    ToMerge : TUnit;
    Nb : integer;
    Total : integer;
    FD : TFactionDef;
    ErrorNb : integer;
begin
  Result := 0;
  if cuGroup = nil then
  begin
    Log(['[Warning] ApplyCreateUnit command, no group selected']);
    Exit;
  end;
  if ObjType = _noUID then
  begin
    Inc(Result);
    Log(['=> ApplyCreateUnit command, Unit creation aborted, invalid UnitDef type (check if the alias was correct).']);
    Exit;
  end;

  if WorkingVars.Count = 0 then
    UID := MakeGmUID(_genUID)
  else
  begin
    UID := LabelToUID(WorkingVars[0]);
    WorkingVars.Delete(0);
  end;

  if Name = '' then Name := _genName;

  cuUnit := cuGroup.CreateUnit(ObjType, UID, Name, ukNone); {pas d'update}
  cuUnit.FixType := FixType;
  if ShortName <> '' then cuUnit.ShortName := ShortName;
  if FlavorName <> '' then cuUnit.FlavorName := FlavorName;
  if GameSim.IsTutorial then cuUnit.Activation := 1;

  if sysErroneousZero_ and (cuUnit.UnitType = 0) then
  begin
    Inc(Result);
    Log(['=> ApplyCreateUnit command', cuUnit.Name, 'is using Erroneous UnitDef of UID 0 at line', cuLine, '(see ScriptPostParse.txt to identify this line)']);
    // no exit...
  end;

  {Name, ShortName, Text sont RZ aprs chaque apply}

  if cuUnit = nil then
  begin
    Inc(Result);
    Log(['=> ApplyCreateUnit command, failed to create unit in group', cuGroup.Name]);
    Exit;
  end
  else
    Log([cuUnit.Name, 'created in group', cuGroup.Name]);

  // si on n'est pas en Custom unit, on enchaine directement sur la cration des SU
  // en fonction du template de l'unit
  if not UseCustom then
  begin
    if not cuUnit.CreateSUFromTemplate(NumCreate, HealthPerc, CohesionPerc, Level) then
    begin
      Inc(Result);
      Log(['=> ApplyCreateUnit command, failed to create all SubUnits from template']);
      Exit;
    end;
    // Nom Flavor SU
    for i := 1 to Min(cuUnit.Count, RxEW_GetWordCount(SUFlavorName)) do
      cuUnit[i-1].Name := ExString(SUFlavorName, i);
  end
  else
  begin
    if NumCreate = -1 then NumCreate := 1;
    Log([NumCreate, 'SubUnit(s) created from template in unit', cuUnit.Name]);
  end;

  if cuUnit.Count = 0 then
  begin
    Inc(Result);
    Log(['=> ApplyCreateUnit command, the unit', cuUnit.name, 'has not subunits']);
    Exit;
  end
  else
  if sysErroneousZero_  then
  begin
    for i := 0 to cuUnit.Count-1 do
      if cuUnit[i].Model.UID = 0 then
      begin
        Inc(Result);
        Log(['=> ApplyCreateUnit command', cuUnit[i].Name, 'in', cuUnit.Name, 'is using Erroneous Model of UID 0 at line', cuLine, '(see ScriptPostParse.txt to identify this line)']);
      end;
  end;

  NumCreate := -1; // reset sur nb max
  Level := 0;

  cuGroup.CalcLeader; // CalcLeader can be needed, but no UPDATE, as the sorting will scramble the units combination!
  
  // Cration d'un GHQGroup (unit deja cre  ce stade)
  if SpecialCommit = _comGHQGroup then
  begin
    SpecialCommit := _comNone;
    if (cuGroup.Count < 2)and(rulCommandChain_ = _gmdComplex) then
    begin
      Inc(Result);
      Log(['=> ApplyCreateUnit command, not enough units declared for GHQGroup subcommand']);
      Exit;
    end;
    if (rulCommandChain_ = _gmdComplex) and
      (cuGroup[0].GetNbSUOfThisFamily(_famGHQ) = 0)and(cuGroup[1].GetNbSUOfThisFamily(_famGHQ) = 0) then
    begin
      Inc(Result);
      Log(['=> ApplyCreateUnit command, First or second unit must have a GHQ, for GHQGroup subcommand']);
      Exit;
    end;
    if (cuGroup[0].GetNbSUOfThisFamily(_famLeader) = 0) and (cuGroup[1].GetNbSUOfThisFamily(_famLeader) = 0) then
    begin
      Inc(Result);
      Log(['=> ApplyCreateUnit command, First or Second unit must be a Leader, for GHQGroup subcommand']);
      Exit;
    end;
    if cuGroup.GetLeader.LeaderRank < ldrMinRankForGHQStack_ then
    begin
      Inc(Result);
      Log(['=> ApplyCreateUnit command, leader is of too low rank, for GHQGroup subcommand']);
      Exit;
    end;
    FD := TFactionDef(cuGroup.GetFactionDefForGHQ);
    if FD = nil then
    begin
      Inc(Result);
      Log(['=> ApplyCreateUnit command, Unreferenced FactionDef, for GHQGroup subcommand']);
      Exit;
    end;

    // in all cases
    cuFaction.CalcFreeGHQList;

    // what is the index of the GHQ
    i := FD.GetIndexGHQByName(GMEParam);
    if i = -1 then
    begin
      Log(['ApplyCreateUnit command, Fetching closest valid free GHQ info.']);

      if cuFaction.ilFreeGHQ.Count = 0 then
      begin
        Inc(Result);
        Log(['=> ApplyCreateUnit command, No free GHQ for unit', cuUnit.Name, 'for GHQGroup subcommand']);
        Exit;
      end;
      if (cuRegion <> nil) then
        i := cuFaction.GetClosestFreeGHQIndex(cuRegion.UID, cuUnit[0].FactionTag);
      if i = -1 then
      begin
        Inc(Result);
        Log(['=> ApplyCreateUnit command, Search failed, No valid free GHQ for unit', cuUnit.Name, 'for GHQGroup subcommand']);
        Exit;
      end;
    end;

    // Trt de la commande
    if not cuGroup.CreateGHQStack(i) then
    begin
      Inc(Result);
      Log(['=> ApplyCreateUnit command, One or more error(s) occured when creating the GHQ, for GHQGroup subcommand']);
      Exit;
    end
    else
    begin
      // FreeGHQ update already done
      Log(['ApplyCreateUnit command, GHQStack Status', cuGroup.UID, cuGroup.Name, cuGroup.StartRegionName]);
      LastGHQGroup := cuGroup;
    end;
  end
  else
  // Cration d'un HQGroup
  if SpecialCommit = _comHQGroup then
  begin
    SpecialCommit := _comNone;
    if cuGroup.Count = 0 then
    begin
      Inc(Result);
      Log(['=> ApplyCreateUnit command, not enough units declared for HQGroup subcommand']);
      Exit;
    end;
    if cuGroup.GetLeader = nil then
    begin
      Inc(Result);
      Log(['=> ApplyCreateUnit command, no leader declared, for HQGroup subcommand']);
      Exit;
    end;
    if cuGroup.GetLeader.LeaderRank < ldrMinRankForHQStack_ then
    begin
      Inc(Result);
      Log(['=> ApplyCreateUnit command, leader is of too low rank, for HQGroup subcommand']);
      Exit;
    end;
    // Trt de la commande
    ErrorNb := cuGroup.CreateHQStack(LastGHQGroup, False, False);
    if ErrorNb <> 0 then
    begin
      Inc(Result);
      case ErrorNb of
        1: Log(['=> ApplyCreateUnit command, Error when creating the HQ, for HQGroup subcommand: Can''t be HQ Stack']);
        2: Log(['=> ApplyCreateUnit command, Error when creating the HQ, for HQGroup subcommand: No GHQ found']);
        3: Log(['=> ApplyCreateUnit command, Error when creating the HQ, for HQGroup subcommand: Update Error']);
      end;
      Exit;
    end
    else
      Log(['ApplyCreateUnit command, HQStack Status', cuGroup.UID, cuGroup.Name, cuGroup.StartRegionName, 'subordinated to', cuGroup.GetGHQStackName]);
  end
  else
  // Unit combo, il faut combiner les units en une seule
  if SpecialCommit = _comCombine  then
  begin
    SpecialCommit := _comNone;
    Nb := Min(STI(GMEParam), cuGroup.Count);
    if Nb < 2 then
    begin
      Inc(Result);
      Log(['=> ApplyCreateUnit command, not enough units declared or possible for the recombination:', Nb]);
      Exit;
    end;
    GMEParam := ITS(Low(Integer));
    lsCombo := TList.Create;
    try
      aUnit := cuGroup[cuGroup.Count - Nb];
      for i := Nb-1 downto 1 do
        lsCombo.Add(cuGroup[cuGroup.Count - i]);
      Total := aUnit.Count;
      for i := 0 to lsCombo.Count-1 do
      begin
        ToMerge := TUnit(lsCombo[i]);
        if Total + ToMerge.Count > subMaxNbPerUnit_ then
        begin
          Inc(Result);
          Log(['=> ApplyCreateUnit command, the unit', cuUnit.name, 'can''t be merged with unit', ToMerge.Name, 'the total SU exceeds the max number allowed.']);
          Continue;
        end;
        Inc(Total, ToMerge.Count);
      end;
      // on accepte toute combinaison dans le script
      aUnit.Recombine(lsCombo, True, True);
      if rulCombi_ then
      begin
        if aUnit.HasLeader and (Nb > 2) then
        begin
          aUnit.CombiKind := _combPrivilegedLarge;
          aUnit.PrevCombiKind := 1;
          if aUnit.FlavorName <> '' then
            aUnit.CombiName := aUnit.FlavorName;
        end;
      end;

    finally
      lsCombo.Free;
    end;
  end;

  cuGroup.CalcKindAndTypes;
  if IsSpecialScript then
    cuGroup.Update(); // because in case of special script, the FinalizeParsing is not done and group are thus not updated except here
  cuGroup.Name := cuGroup.Name; // will trigger the update of SetupNameInfo, see TGroup.SetName;
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// APPLY CREATE SUB UNIT
// utilis quand on ne crer pas directement via le template d'unit (obsolte +- mais encore possible)
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.ApplyCreateSubUnit : integer;
var i : integer;
begin
  Result := 0;
  if cuUnit = nil then
  begin
    Inc(Result);
    Log(['=> ApplyCreateSubUnit command, no unit selected']);
  end;
  if Result > 0 then
  begin
    Abort;
    Exit;
  end;
  if ObjType = _noUID then
  begin
    Inc(Result);
    Log(['=> ApplyCreateSubUnit command, SubUnit creation aborted, invalid Model type (check if the alias was correct).']);
    Exit;
  end;

  // pas de WorkingVars pour les SubUnits

  if Name = '' then Name := _genName;

  if cuUnit.CreateSubUnits(ObjType, NumCreate, HealthPerc, CohesionPerc, Name) = nil then
  begin
    Inc(Result);
    Log(['=> ApplyCreatSubUnit command, failed to create all SubUnits']);
    Exit;
  end
  else
    Log([NumCreate, 'SubUnit(s) created in unit', cuUnit.Name]);
  Level := 0;

  if sysErroneousZero_  then
  begin
    for i := 0 to cuUnit.Count-1 do
      if cuUnit[i].Model.UID = 0 then
      begin
        Inc(Result);
        Log(['=> ApplyCreateSubUnit command', cuUnit[i].Name, 'in', cuUnit.Name, 'is using Erroneous Model of UID 0 at line', cuLine, '(see ScriptPostParse.txt to identify this line)']);
      end;
  end;
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// APPLY CREATE STRUCTURE
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.ApplyCreateStruc : integer;
var UID         : TUID;
    ST          : TStructureStruc;
    pStructures : TStructures;
    i           : integer;
    Structure   : TStructure;
    UIDExclu    : TUID;
    Owner       : TFaction;
    Region      : TRegion;
begin
  Result := 0;
  if cuFaction = nil then
  begin
    Log(['=> ApplyCreateStruc command, no faction selected']);
    Inc(Result);
  end;
  if cuRegion = nil then
  begin
    Log(['ApplyCreateStruc command, no region selected']);
    Exit;
  end;
  if Result > 0 then
  begin
    Abort;
    Exit;
  end;

  if WorkingVars.Count = 0 then
    UID := MakeGmUID(_genUID)
  else
  begin
    UID := LabelToUID(WorkingVars[0]);
    WorkingVars.Delete(0);
  end;

  if (ObjType < 0)or(ObjType > StrucDefs.Count -1) then
  begin
    Inc(Result);
    Log(['=> ApplyCreateStruc command', ' This type:', ObjType, 'don''t exists as a structure']);
    Exit;
  end;

  // pas rgion terrestre
  if not RegionDefs.IsLandRegion(cuRegion.UID) then
  begin
    Log(['[WARNING] ApplyCreateStruc command', StrPhysicalNoStrucAtSea_]);
    Exit;
  end;

  // exclusive & already there
  pStructures := GameSim.RegionCmns.Fetch(cuRegion.UID).pListStruc;

  // in case of harbor we fix the problem ourself
  for i := pStructures.Count -1 downto 0 do
  begin
    Structure := pStructures[i];
    if (Structure.StrucDef.IsHarbor)and(StrucDefs[ObjType].IsHarbor) then
    begin
      Owner :=GameSim.Factions.GetFactionByUID(Structure.FactionUID);
      Region := Owner.Regions.Fetch(cuRegion.UID);
      Log(['[WARNING] ApplyCreateStruc command', 'An harbor is already there, owner:', Owner.Name, 'keeping harbor for faction', Structure.FactionUID, 'fixing issue.']);
      //ShowMessage('[WARNING] ApplyCreateStruc command ' + cuRegion.Name + '. An harbor is already there, owner: ' + Owner.Name + ', removing it and adding new harbor for faction ' + cuFaction.Name);
      Region.Structures.Remove(Structure);
      GameSim.RegionCmns.Update(cuRegion.UID);
    end;
  end;

  for i := 0 to pStructures.Count -1 do
  begin
    Structure := pStructures[i];
    // check the structure has no duplicate
    if (not Structure.StrucDef.CanDuplicate) then
    begin
      if GameSim.RegionCmns.StrucHasDuplicate(cuRegion.UID, Structure) then // in same faction if not exclusive, for all region if exclusive
      begin
        Inc(Result);
        Log(['=> ApplyCreateStruc command', 'Duplicate issue when adding', ObjType, StrucDefs[ObjType].Name ,
        'while testing', Structure.Name, Structure.StrucDef.Name, 'Check Kind, Family, IsDuplicate']);
        Exit;
      end;
    end;
  end;

  if Level < 1 then Level := 1;

  ST := MakeStructureStruc(ObjType, cuFaction.UID);
  ST.UID      :=  UID;
  ST.Level    := Level;
  if HealthPerc <> 100 then
    ST.PercDone := HealthPerc;
  // Name est remis  '' sur chaque Apply
  if Name = '' then ST.Name := _genName else ST.Name := Name;
  Structure := TStructure.Create(cuRegion.Structures, ST);
  Structure.Active := HealthPerc = 100;
  cuRegion.Structures.AddStruc(Structure);

  Log([cuFaction.Name, ',', StrucDefs[ObjType].Name, Name, 'created in region', cuRegion.Name]);

  // Test si les structures exclusives sont bien  la mme faction
  if pStructures.Count > 1 then
  begin
    UIDExclu := _noUID;
    for i := 0 to pStructures.Count -1 do
    begin
      if pStructures[i].IsExclusive and (UIDExclu = _noUID) then
        UIDExclu := pStructures[i].FactionUID
      else
      if pStructures[i].StrucDef.IsExclusive and (UIDExclu <> _noUID) and (UIDExclu <> pStructures[i].FactionUID) then
      begin
        Inc(Result);
        Log(['=> ApplyCreateStruc command', '2 factions have exclusive structures', 'Latest addition', pStructures[i].Name, 'UID of Owner',
        pStructures[i].FactionUID, 'Conflicting with (UID of Owner)', UIDExclu, 'Check also between the Setup tab and the primary controler in the Cities Tab']);
        Exit;
      end;
    end;
  end;

  Level := 0;
end;
{
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// APPLY CREATE POP (BRICK)
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptEngine.ApplyCreatePop : integer;
var PB : TPopBrick;
    TmpSocial, TmpEthnic, TmpRichness : integer;
begin
  Result := 0;
  if cuFaction = nil then
  begin
    Inc(Result);
    Log(['=> ApplyCreatePop command, no faction selected']);
  end;
  if cuRegion = nil then
  begin
    Inc(Result);
    Log(['=> ApplyCreatePop command, no region selected']);
  end;
  if Result > 0 then
  begin
    Abort;
    Exit;
  end;
  // pas rgion terrestre
  if not RegionDefs.IsLandRegion(cuRegion.UID) then
  begin
    Inc(Result);
    Log(['=> ApplyCreatePop command, not a land region']);
    Exit;
  end;
  TmpSocial   := SocialClass;
  TmpEthnic   := EthnicGroup;
  TmpRichness := Richness;

  if (SocialClass < 0) or (SocialClass > _PopNb) then
    TmpSocial := Random(_PopNb +1);
  if EthnicGroup = _noUID then
    TmpEthnic := cuFaction.UID;
  if Richness < 0 then
    TmpRichness := SocialClass * 20 + Random(30);

  PB := cuRegion.CreatePop(TmpSocial, TmpEthnic, Religion);
  PB.LoyaltyFac := LoyaltyFac;
  PB.LoyaltyEth := LoyaltyEth;
  PB.GrowthRate := GrowthRate;
  PB.Richness   := TmpRichness;
  Log([PB.Name, 'created in region', cuRegion.Name]);
end;
}
// ACCESSEURS
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SET CURRENT REGION
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
procedure TScriptEngine.SetCuRegion(Value : TRegion);
begin
  if Value <> fcuRegion then
    fcuRegion := Value;
end;

// TEVENT STATE
// Etat courant d'un evt
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SERIALIZE
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TEventState.Serialize  (out SL : TStringList) : boolean;
begin
  Result := False;
  if not ClearStringList(SL) then
    begin LogMsg(['TEventState.Serialize', 'Uninialied container data for event', UName], _Critical); Exit; end;
  SL.Add(_KeyStartEventSt);
  SL.Add(_KeyName + '=' + UName);
  SL.Add(_KeycuOccurs + '=' + ITS(cuOccurs));
  SL.Add(_KeyMaxOccurs + '=' + ITS(MaxOccurs));
  SL.Add(_KeyEndEventSt);
  Result := True;
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// UN SERIALIZE
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TEventState.UnSerialize(SL : TStringList)     : boolean;
begin
  Result := False;
  if not CheckStringList(SL) then
    begin LogMsg(['TEventState.UnSerialize', 'Uninialized container data'], _Critical); Exit; end;
  if SL.Count < 2 then
    begin LogMsg(['TEventState.UnSerialize', 'Too few entries', SL.Count], _Warning); Exit; end;
  if Trim(SL[0]) <> _KeyStartEventSt then
    begin LogMsg(['TEventState.UnSerialize', 'Wrong start tag, found:', SL[0], 'expected:', _KeyStartEventSt], _Warning); Exit; end;
  if Trim(SL[SL.Count-1]) <> _KeyEndEventSt then
    begin LogMsg(['TEventState.UnSerialize', 'Wrong end tag, found:', SL[SL.Count-1], 'expected:', _KeyEndEventSt], _Warning); Exit; end;

  UName     := TFileParser.ReadKeyS(_KeyName,      SL[1]);
  cuOccurs  := TFileParser.ReadKeyI(_KeycuOccurs,  SL[2]);
  MaxOccurs := TFileParser.ReadKeyI(_KeyMaxOccurs, SL[3]);

  Result := True;
end;

// TEVENT STATES

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// CREATE
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
constructor TEventStates.Create;
begin
  inherited;
  OwnsObjects := True;
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// GET BY NAME
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TEventStates.GetByName(aUName : string) : TEventState;
var i : integer;
begin
  Result := nil;
  for i := 0 to Count -1 do
  begin
    if Trim(UpperCase(Items[i].UName)) = Trim(UpperCase(aUName)) then
    begin
      Result := Items[i];
      Exit;
    end;
  end;
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ADD
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TEventStates.Add(Item: TEventState): Integer;
begin
  Result := Inherited Add(Item);
end;
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// PROTECTED
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// GET
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TEventStates.Get(Index: Integer): TEventState;
begin
  Result := TEventState(inherited Get(Index));
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// PUT
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
procedure TEventStates.Put(Index: Integer; Value: TEventState);
begin
  inherited Put(Index, Value);
end;

// *** SCRIPT VAR ***

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SERIALIZE
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptVar.Serialize  (out SL : TStringList) : boolean;
begin
  Result := False;
  if not ClearStringList(SL) then
    begin LogMsg(['TScriptVar.Serialize', 'Uninialized container data for event', Name], _Critical); Exit; end;
  SL.Add(_KeyStartScriptVar);
  SL.Add(_KeyName + '=' + Name);
  SL.Add(_KeyVarType + '=' + ITS(VarType));
  SL.Add(_KeyValue + '=' + Value);
  SL.Add(_KeyEndScriptVar);
  Result := True;
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// UNSERIALIZE
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function TScriptVar.UnSerialize(SL : TStringList)     : boolean;
begin
  Result := False;
  if not CheckStringList(SL) then
    begin LogMsg(['TScriptVar.UnSerialize', 'Uninialized container data'], _Critical); Exit; end;
  if SL.Count < 2 then
    begin LogMsg(['TScriptVar.UnSerialize', 'Too few entries', SL.Count], _Warning); Exit; end;
  if Trim(SL[0]) <> _KeyStartScriptVar then
    begin LogMsg(['TScriptVar.UnSerialize', 'Wrong start tag, found:', SL[0], 'expected:', _KeyStartScriptVar], _Warning); Exit; end;
  if Trim(SL[SL.Count-1]) <> _KeyEndScriptVar then
    begin LogMsg(['TScriptVar.UnSerialize', 'Wrong end tag, found:', SL[SL.Count-1], 'expected:', _KeyEndScriptVar], _Warning); Exit; end;

  Name     := TFileParser.ReadKeyS(_KeyName,    SL[1]);
  VarType  := TFileParser.ReadKeyI(_KeyVarType, SL[2]);
  Value    := TFileParser.ReadKeyS(_KeyValue,   SL[3]);

  Result := True;
end;


end.
